/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: advanced.c 3 2004-07-13 11:26:13Z picard $
 *
 * BetaPlayer Core
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "stdafx.h"

typedef struct advanced
{
	node VMT;

	bool_t OldShell;
	bool_t AllKeys;
	bool_t SlowVideo;
	bool_t ColorLookup;
	bool_t NoBackLight;
	bool_t HomeScreen;
	bool_t NoWMMX;
	bool_t IDCTSwap;
	bool_t BenchFromPos;
	bool_t PixelAspect;
	tick_t DropTolerance;
	tick_t SkipTolerance;
	tick_t AVOffset;

} advanced;

static const datadef AdvancedParams[] = 
{
	{ ADVPLATFORM_ID,			TYPE_CLASS },
	{ ADVPLATFORM_NOBACKLIGHT, TYPE_BOOL, DF_SETUP|DF_CHECKLIST },
	{ ADVPLATFORM_HOMESCREEN, TYPE_BOOL, DF_SETUP|DF_CHECKLIST },
	{ ADVPLATFORM_OLDSHELL,	TYPE_BOOL, DF_SETUP|DF_RESTART|DF_CHECKLIST },
#if defined(ARM) 
	{ ADVPLATFORM_NOWMMX,		TYPE_BOOL, DF_SETUP|DF_CHECKLIST|DF_RESTART },
#endif
#if !defined(SH3) && !defined(MIPS)
	{ ADVPLATFORM_SLOW_VIDEO,	TYPE_BOOL, DF_SETUP|DF_CHECKLIST },
#endif
#if defined(ARM) || defined(_M_IX86)
	{ ADVPLATFORM_IDCTSWAP,	TYPE_BOOL, DF_SETUP|DF_CHECKLIST|DF_RESTART },
#endif
#if !defined(SH3) 
	{ ADVPLATFORM_COLOR_LOOKUP,TYPE_BOOL, DF_SETUP|DF_CHECKLIST },
#endif
#if defined(_WIN32_WCE)
	{ ADVPLATFORM_ALLKEYS,	TYPE_BOOL, DF_SETUP|DF_CHECKLIST },
#endif
	{ ADVPLATFORM_BENCHFROMPOS,TYPE_BOOL, DF_SETUP|DF_CHECKLIST },
	{ ADVPLATFORM_PIXELASPECT, TYPE_BOOL, DF_HIDDEN },//DF_SETUP|DF_CHECKLIST|DF_RESTART|DF_HIDDEN },

	{ ADVPLATFORM_DROPTOL,		TYPE_TICK, DF_SETUP|DF_MINMAX|DF_MSEC|DF_GAP, 0, TICKSPERSEC },
	{ ADVPLATFORM_SKIPTOL,		TYPE_TICK, DF_SETUP|DF_MINMAX|DF_MSEC, 0, 5*TICKSPERSEC },
	{ ADVPLATFORM_AVOFFSET,		TYPE_TICK, DF_SETUP|DF_MINMAX|DF_MSEC|DF_NEG|DF_RESYNC, -2*TICKSPERSEC, 2*TICKSPERSEC },
	DATADEF_END
};

static int Enum( advanced* p, int No, datadef* Param )
{
	int Result = NodeEnumType(&No,Param,NodeParams,AdvancedParams,NULL);
	if (Result == ERR_NONE && Param->No == ADVPLATFORM_OLDSHELL && 
		((QueryPlatform(PLATFORM_CAPS) & CAPS_OLDSHELL) || QueryPlatform(PLATFORM_TYPENO) == TYPE_SMARTPHONE || (QueryPlatform(PLATFORM_TYPENO) == TYPE_POCKETPC && QueryPlatform(PLATFORM_VER) >= 421)))
		Param->Flags |= DF_HIDDEN;
	if (Result == ERR_NONE && Param->No == ADVPLATFORM_NOWMMX && !(QueryPlatform(PLATFORM_CAPS) & CAPS_ARM_WMMX))
		Param->Flags |= DF_HIDDEN;
	if (Result == ERR_NONE && Param->No == ADVPLATFORM_ALLKEYS && QueryPlatform(PLATFORM_TYPENO) == TYPE_SMARTPHONE)
		Param->Flags |= DF_HIDDEN;
	if (Result == ERR_NONE && Param->No == ADVPLATFORM_HOMESCREEN && QueryPlatform(PLATFORM_TYPENO) != TYPE_SMARTPHONE)
		Param->Flags |= DF_HIDDEN;
	return Result;
}

static int Get( advanced* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(ADVPLATFORM_ID,int); break;
	case ADVPLATFORM_OLDSHELL: GETVALUE(p->OldShell,bool_t); break;
	case ADVPLATFORM_HOMESCREEN: GETVALUE(p->HomeScreen,bool_t); break;
	case ADVPLATFORM_NOBACKLIGHT: GETVALUE(p->NoBackLight,bool_t); break;
	case ADVPLATFORM_IDCTSWAP: GETVALUE(p->IDCTSwap,bool_t); break;
	case ADVPLATFORM_NOWMMX: GETVALUE(p->NoWMMX,bool_t); break;
	case ADVPLATFORM_SLOW_VIDEO: GETVALUE(p->SlowVideo,bool_t); break;
	case ADVPLATFORM_COLOR_LOOKUP: GETVALUE(p->ColorLookup,bool_t); break;
	case ADVPLATFORM_DROPTOL: GETVALUE(p->DropTolerance,tick_t); break;
	case ADVPLATFORM_SKIPTOL: GETVALUE(p->SkipTolerance,tick_t); break;
	case ADVPLATFORM_AVOFFSET: GETVALUE(p->AVOffset,tick_t); break;
	case ADVPLATFORM_ALLKEYS: GETVALUE(p->AllKeys,bool_t); break;
	case ADVPLATFORM_BENCHFROMPOS: GETVALUE(p->BenchFromPos,bool_t); break;
	case ADVPLATFORM_PIXELASPECT: GETVALUE(p->PixelAspect,bool_t); break;
	}
	return Result;
}

static int Set( advanced* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case ADVPLATFORM_OLDSHELL: SETVALUE(p->OldShell,bool_t,ERR_NONE); break;
	case ADVPLATFORM_HOMESCREEN: SETVALUE(p->HomeScreen,bool_t,ERR_NONE); break;
	case ADVPLATFORM_NOBACKLIGHT: SETVALUE(p->NoBackLight,bool_t,ERR_NONE); break;
	case ADVPLATFORM_IDCTSWAP: SETVALUE(p->IDCTSwap,bool_t,ERR_NONE); break;
	case ADVPLATFORM_NOWMMX: SETVALUE(p->NoWMMX,bool_t,ERR_NONE); break;
	case ADVPLATFORM_SLOW_VIDEO: SETVALUE(p->SlowVideo,bool_t,ERR_NONE); break;
	case ADVPLATFORM_COLOR_LOOKUP: SETVALUE(p->ColorLookup,bool_t,ERR_NONE); break;
	case ADVPLATFORM_DROPTOL: SETVALUE(p->DropTolerance,tick_t,NodeRefresh()); break;
	case ADVPLATFORM_SKIPTOL: SETVALUE(p->SkipTolerance,tick_t,NodeRefresh()); break;
	case ADVPLATFORM_AVOFFSET: SETVALUE(p->AVOffset,tick_t,NodeRefresh()); break;
	case ADVPLATFORM_ALLKEYS: SETVALUE(p->AllKeys,bool_t,ERR_NONE); break;
	case ADVPLATFORM_BENCHFROMPOS: SETVALUE(p->BenchFromPos,bool_t,ERR_NONE); break;
	case ADVPLATFORM_PIXELASPECT: SETVALUE(p->PixelAspect,bool_t,ERR_NONE); break;
	}
	return Result;
}

static advanced Advanced;

int QueryAdvPlatform( int Param )
{
	int Value=0;
	Advanced.VMT.Get(&Advanced,Param,&Value,sizeof(Value));
	return Value;
}

void Advanced_Init()
{
	int Caps = QueryPlatform(PLATFORM_CAPS);
	advanced* p = &Advanced;
	NodeFill(&p->VMT,sizeof(advanced),(nodeenum)Enum,(nodeget)Get,(nodeset)Set);
	p->SlowVideo = 0;
	p->BenchFromPos = 0;
	p->PixelAspect = 1;
	p->ColorLookup = !(Caps & CAPS_MIPS_VR41XX) && !(Caps & CAPS_ARM_XSCALE);
	p->OldShell = (Caps & CAPS_OLDSHELL) != 0;
	p->DropTolerance = (TICKSPERSEC*55)/1000;
	p->SkipTolerance = (TICKSPERSEC*700)/1000;
	p->AVOffset = 0;

	NodeRegister(&p->VMT,PRI_MAXIMUM+50);
}

void Advanced_Done()
{
	NodeUnRegister(&Advanced.VMT);
}

