/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: format.h 144 2004-12-12 15:28:39Z picard $
 *
 * BetaPlayer Core
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#ifndef __FORMAT_H
#define __FORMAT_H

#define FTYPE_PLAYLIST		'P'
#define FTYPE_VIDEO			'V'
#define FTYPE_AUDIO			'A'

//---------------------------------------------------------------
// format params

#define FORMAT_CLASS			FOURCC('F','M','T','_')
#define FORMAT_MEDIA_CLASS		FOURCC('F','M','T','M')
#define FORMAT_PLAYLIST_CLASS	FOURCC('F','M','T','L')

// supported file extension and ype sperated by semicolon (ext:type[;ext:type])
#define FORMAT_EXTS				0x31
// input stream (stream*)
#define FORMAT_INPUT			0x32
// input stream (stream*)
#define FORMAT_OUTPUT			0x3B
// duration (tick_t)
#define FORMAT_DURATION			0x33
// approx file position (int)
#define FORMAT_FILEPOS			0x36
// file size in KB if available (int)
#define FORMAT_FILESIZE			0x44
// buffer size (int)
#define FORMAT_BUFFERSIZE		0x3A
// update streams notify
#define FORMAT_UPDATESTREAMS	0x41
// manual data feed (for probe data)
#define FORMAT_DATAFEED			0x42

// output data streams (pin) 0x1000,0x1001,0x1002...
#define FORMAT_STREAM			0x1000
// comments by streams (pin) (optional)
#define FORMAT_COMMENT			0x2000

extern DLL const datadef FormatParams[];
extern DLL const datadef FormatMediaParams[];
extern DLL const datadef FormatPlayListParams[];

// probe format (must work standalone. called before setting the input stream)
typedef int (*fmtprobe)(void* This,const void* Data,int Length,const tchar_t* URL, const tchar_t* ContentType);
// sync or seek by time or by file position (-1 for no seek)
typedef int (*fmtsync)(void* This,tick_t Time,int FilePos);
// read input (only one chunk) to buffer
typedef int (*fmtread)(void* This,int BufferPercent,int* BufferUsed);
// read packets and send them time syncorinzed to codecs 
// time is updated after syncronization (return value ERR_SYNCED)
typedef int (*fmtprocess)(void* This,tick_t* Time,bool_t* WakeInput,bool_t Fill);
// read playlist item
typedef int (*fmtreadlist)(void* This,tchar_t* Path,int PathLen,tchar_t* DispName,int DispNameLen,tick_t* Length);
// write playlist item
typedef int (*fmtwritelist)(void* This,const tchar_t* Path,const tchar_t* DispName,tick_t Length);

typedef struct format
{
	nodeenum		Enum;
	nodeget			Get;
	nodeset			Set;

	fmtprobe		Probe;
	fmtsync			Sync;
	fmtread			Read;
	fmtprocess		Process;

	fmtreadlist		ReadList;
	fmtwritelist	WriteList;

} format;

#endif
