/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: format_base.c 135 2004-12-08 07:37:19Z picard $
 *
 * BetaPlayer Core
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "stdafx.h"

#define ENDCHUNKS		8
#define SEEKTRIES		4
#define SKIPDISTANCE	64*1024

int Format_Enum(format_base* p,int No,datadef* Param)
{
	if (NodeEnumType(&No,Param,NodeParams,FormatParams,FormatMediaParams,NULL) == ERR_NONE)
		return ERR_NONE;

	if (No < p->StreamCount)
	{
		memset(Param,0,sizeof(datadef));
		Param->Class = FORMAT_CLASS;
		Param->Size = sizeof(pin);
		Param->No = FORMAT_STREAM + No;
		Param->Type = TYPE_PIN;
		Param->Format1 = p->Streams[No]->FormatType;
		Param->Flags = DF_OUTPUT;
		return ERR_NONE;
	}
	No -= p->StreamCount;

	if (No < p->StreamCount)
	{
		memset(Param,0,sizeof(datadef));
		Param->Class = FORMAT_CLASS;
		Param->Size = sizeof(pin);
		Param->No = FORMAT_COMMENT + No;
		Param->Type = TYPE_PIN;
		Param->Format1 = PACKET_COMMENT;
		Param->Flags = DF_OUTPUT;
		return ERR_NONE;
	}

	return ERR_INVALID_PARAM;
}

static INLINE void Format_BufferInsert(format_base* p,format_buffer* Buffer)
{
	Buffer->Next = NULL;
	if (p->BufferLast)
		p->BufferLast->Next = Buffer;
	else
		p->BufferFirst = Buffer;
	p->BufferLast = Buffer;
	++p->BufferAvailable;
}

format_buffer* Format_BufferAlloc(format_base*,int Mode); //0:optional 1:drop 2:new alloc
format_packet* Format_PacketAlloc( format_base* );
void Format_PacketRelease( format_base*, format_packet*);
bool_t Format_ReadBuffer( format_base* p, bool_t ToRead );

int Format_FileSeek(format_base* p,int Pos,int Origin)
{
	int RelPos = Pos;
	if (Origin == SEEK_SET)
		RelPos -= p->FilePos;
	if (Origin != SEEK_END)
	{
		// check if we can use Format_Skip
		if (RelPos >= 0 && RelPos < SKIPDISTANCE)
		{
			Format_Skip(p,RelPos);
			return ERR_NONE;
		}

		// check if we are in the same ReadBuffer
		if (RelPos < 0 && p->ReadBuffer && p->ReadPos + RelPos >= 0)
		{
			p->FilePos += RelPos;
			p->ReadPos += RelPos;
			return ERR_NONE;
		}
	}

	Pos = p->Input->Seek(p->Input,Pos,Origin);
	
	if (Pos < 0)
		return ERR_NOT_SUPPORTED;

	p->FilePos = Pos;
	Format_ReleaseBuffers(p);
	return ERR_NONE;
}

int Format_Seek(format_base* p,int Pos,int Origin)
{
	int Result = Format_FileSeek(p,Pos,Origin);
	if (Result == ERR_NONE)
		Format_AfterSeek(p);
	return Result;
}

int Format_SeekByPacket(format_base* p,tick_t Time,int FilePos)
{
	int Try;
	int LastTime;
	int LastPos;
	format_stream* Stream = NULL;

	if (FilePos < 0)
	{
		Time -= SHOWAHEAD; // we should go a little earlier
		if (Time > 0)
		{
			if (p->FileSize<0 || p->Duration<0)
				return ERR_NOT_SUPPORTED;

			FilePos = Scale(p->FileSize,Time,p->Duration);
		}
		else
			FilePos = 0;
	}

	if (Time>0)
		Stream = Format_DefSyncStream(p);

	p->InSeek = 1;

	LastPos = -1;
	LastTime = -1;

	for (Try=0;Try<SEEKTRIES;++Try) // maximum iterations
	{
		if (FilePos < p->SeekByPacket_DataStart)
			FilePos = p->SeekByPacket_DataStart;
		if (p->SeekByPacket_BlockAlign > 1)
			FilePos -= (FilePos - p->SeekByPacket_DataStart) % p->SeekByPacket_BlockAlign;

		if (Format_Seek(p,FilePos,SEEK_SET) != ERR_NONE)
		{
			p->InSeek = 0;
			return ERR_NOT_SUPPORTED;
		}

		if (Stream)
		{
			// check if really seeked to the right time (and adjust if needed)

			int SeekRead = FilePos + MINBUFFER;

			while (p->FilePos<SeekRead)
			{
				Format_ReadBuffer(p,0); // simulate input thread

				if (p->FillQueue(p) == ERR_END_OF_FILE)
					break;

				if (Stream->LastTime >= 0)
				{
					DEBUG_MSG3(DEBUG_FORMAT,T("SeekByPacket to:%d current:%d filepos:%d"),Time,Stream->LastTime,FilePos);

					if (Time>=Stream->LastTime && (Time-Stream->LastTime) <= (TICKSPERSEC/2))
						FilePos = -1; // accept
					else
					{
						int Move;
						
						if (LastPos >= 0)
 							Move = Scale(FilePos-LastPos,Time-Stream->LastTime,Stream->LastTime-LastTime);
						else
 							Move = Scale(p->FileSize,Time-Stream->LastTime,p->Duration);

						if (Move < 0)
							Move = (Move*9)/8;
						else
							Move = (Move*15)/16;

						if (Move > 64*1024 || Move < -64*1024)
						{
							LastTime = Stream->LastTime;
							LastPos = FilePos;
						}
						else
							LastPos = -1;

						FilePos += Move;
					}

					if (FilePos < 0 || FilePos >= p->FileSize)
						SeekRead = -1; // accept
					break;
				}

				if (!p->BufferAvailable && !p->ReadBuffer && p->NoMoreInput)
					break; // end of file
			}

			if (p->FilePos<SeekRead)
				continue; // try again
		}
		break;
	}

	p->InSeek = 0;

	return ERR_NONE;
}

bool_t DropBuffer( format_base* p )
{
	bool_t Dropped = 0;
	int No=0;

	for (No=0;No<p->StreamCount;++No)
	{
		format_stream* Stream = p->Streams[No];
		if (Stream->PacketFirst)
		{
			format_packet* Packet = Stream->PacketFirst;
			Stream->PacketFirst = Packet->Next;
			if (Stream->PacketLast == Packet)
				Stream->PacketLast = NULL;

			Format_PacketRelease(p,Packet);

			Stream->Packet.DropLevel = 2;
			Dropped = 1;
		}
	}

	return Dropped;
}

format_buffer* Format_BufferAlloc(format_base* p,int Mode)
{
	format_buffer* Buffer;

	if (Mode!=2)
		while (!p->FreeBuffers && (p->BufferUsed >= p->BufferSize || AvailMemory() < MEMORYLIMIT))
			if (Mode==0 || !DropBuffer(p))
				return NULL;

	Buffer = p->FreeBuffers;
	if (!Buffer)
	{
		Buffer = (format_buffer*) Alloc(sizeof(format_buffer));
		if (!Buffer)
			return NULL;
	}
	else
	{
		assert(Buffer->RefCount==0);
		p->FreeBuffers = Buffer->Next;
	}

	Buffer->Id = p->BufferId++;
	Buffer->RefCount = 1;
	Buffer->Next = NULL;
	++p->BufferUsed;

	return Buffer;
}

void CalcDuration(format_base* p)
{
	if (p->Duration<0 && p->FileSize>=0)
	{
		if (p->TimeStamps)
		{
			// try to read the last packets and figure out Duration
			int SavePos;

			LockEnter(p->InputLock);
			LockEnter(p->BufferLock);

			SavePos = p->Input->Seek(p->Input,0,SEEK_CUR);
			if (SavePos >= 0)
			{
				format_buffer* Buffer[ENDCHUNKS];
				int No,i;
				tick_t Max=-1;

				bool_t BackupNoMore = p->NoMoreInput;
				format_buffer* BackupLast = p->BufferLast;
				format_buffer* BackupFirst = p->BufferFirst;
				format_buffer* BackupRead = p->ReadBuffer;
				int BackupFilePos = p->FilePos;
				int BackupReadPos = p->ReadPos;
				int BackupAvailable = p->BufferAvailable;

				memset(Buffer,0,sizeof(Buffer));
		
				p->NoMoreInput = 1; //disable reading during ReadPacket
		
				for (No=0;No<ENDCHUNKS;++No)
				{
					int Pos = p->FileSize - (No+1)*CHUNKSIZE;
					if (Pos<0)
						break;

					Buffer[No] = Format_BufferAlloc(p,2);
					if (!Buffer[No])
						break;

					if (p->Input->Seek(p->Input,Pos,SEEK_SET)<0)
						break;

					Buffer[No]->Length = p->Input->Read(p->Input,Buffer[No]->Buffer,CHUNKSIZE);
					if (Buffer[No]->Length < 0)
						break;

					if (No)
						Buffer[No]->Next = Buffer[No-1];

					// setup temporary buffer
					p->BufferLast = Buffer[0];
					p->BufferFirst = Buffer[No];
					p->BufferAvailable = No+1;
					p->ReadPos = 0;
					p->ReadBuffer = NULL;
					p->FilePos = Pos;
					for (i=0;i<=No;++i)
						Buffer[i]->RefCount++;

					if (p->SeekByPacket_BlockAlign > 1)
					{
						int Rem = (Pos - p->SeekByPacket_DataStart) % p->SeekByPacket_BlockAlign;
						if (Rem)
							Format_Skip(p,p->SeekByPacket_BlockAlign - Rem);
					}
					if (p->BackupPacketState)
						p->BackupPacketState(p,1);

					while (p->BufferLast || p->ReadBuffer)
					{
						int Result;
						format_packet* Packet = Format_PacketAlloc(p);
						if (!Packet)
							break;

						Result = p->ReadPacket(p,Packet);

						if (Result != ERR_NONE && Result != ERR_DATA_NOT_FOUND)
						{
							p->BufferLast = NULL;
							p->ReadBuffer = NULL;
						}
						else
						if (Result == ERR_NONE && Packet->Stream && Packet->RefTime>Max)
							Max = Packet->RefTime;

						Format_PacketRelease(p,Packet);
					}

					if (p->BackupPacketState)
						p->BackupPacketState(p,0);

					for (i=0;i<=No;++i)
						Buffer[i]->RefCount=1; // ugly

					if (Max >= 0)
					{
						p->Duration = Max;
						break;
					}
				}

				for (No=0;No<ENDCHUNKS;++No)
					if (Buffer[No])
						Format_BufferRelease(p,Buffer[No]);

				p->Input->Seek(p->Input,SavePos,SEEK_SET);

				p->BufferLast = BackupLast;
				p->BufferFirst = BackupFirst;
				p->ReadBuffer = BackupRead;
				p->FilePos = BackupFilePos;
				p->ReadPos = BackupReadPos;
				p->BufferAvailable = BackupAvailable;
				p->NoMoreInput = BackupNoMore;
			}

			LockLeave(p->BufferLock);
			LockLeave(p->InputLock);
		}

		if (p->Duration<0 && p->StreamCount)
		{
			// try to sum stream byterates

			int No;
			int SumByteRate = 0;
			format_stream* Stream;

			for (No=0;No<p->StreamCount;++No)
			{
				int ByteRate = 0;
				Stream = p->Streams[No];

				switch (Stream->FormatType)
				{
				case PACKET_VIDEO:
					ByteRate = ((video*)Stream->Format)->ByteRate;
					if (!ByteRate && Stream->Pin.Node)
					{
						// maybe codec has some information about byterate
						Stream->Pin.Node->Get(Stream->Pin.Node,Stream->Pin.No|PACKET_FORMAT,Stream->Format,Stream->FormatLength);
						ByteRate = ((video*)Stream->Format)->ByteRate;
					}
					break;

				case PACKET_AUDIO:
					ByteRate = ((audio*)Stream->Format)->ByteRate;
					if (!ByteRate && Stream->Pin.Node)
					{
						// maybe codec has some information about byterate
						Stream->Pin.Node->Get(Stream->Pin.Node,Stream->Pin.No|PACKET_FORMAT,Stream->Format,Stream->FormatLength);
						ByteRate = ((audio*)Stream->Format)->ByteRate;
					}
					break;
				}

				if (ByteRate>0)
					SumByteRate += ByteRate;
				else
					break;
			}

			if (No == p->StreamCount)
				p->Duration = Scale(p->FileSize,TICKSPERSEC,SumByteRate);
		}
	}
}

void Format_BufferRelease( format_base* p, format_buffer* Buffer )
{
	//DEBUG_MSG4(DEBUG_FORMAT,"Buffer:%d release %d->%d (%d)",Buffer->Id,Buffer->RefCount,Buffer->RefCount-1,p->BufferUsed);

	LockEnter(p->BufferLock);

	assert(Buffer->RefCount>0 && Buffer->RefCount<100);

	if (--Buffer->RefCount <= 0)
	{
		assert(Buffer->RefCount==0);

		--p->BufferUsed;
		if (p->BufferUsed <= p->BufferSize)
		{
			Buffer->Next = p->FreeBuffers;
			p->FreeBuffers = Buffer;
		}
		else
			Free(Buffer);
	}

	LockLeave(p->BufferLock);
}

format_buffer* Format_BufferRemove( format_base* p )
{
	format_buffer* Buffer;
	LockEnter(p->BufferLock);
	Buffer = p->BufferFirst;
	if (Buffer)
	{
		p->BufferFirst = Buffer->Next;
		if (p->BufferLast == Buffer)
			p->BufferLast = NULL;
		--p->BufferAvailable;
		assert(p->BufferAvailable >= 0);
	}
	LockLeave(p->BufferLock);
	return Buffer;
}

void Format_SingleRefRelease( format_base* p, format_ref* Ref )
{
	if (Ref->Buffer)
		Format_BufferRelease(p,Ref->Buffer);
	Ref->Next = p->FreeRefs;
	p->FreeRefs = Ref;
}

void Format_ReleaseRef( format_base* p, format_ref* Ref )
{
	while (Ref)
	{
		format_ref* Next = Ref->Next;
		Format_SingleRefRelease(p,Ref);
		Ref = Next;
	}
}

void Format_PacketRelease( format_base* p, format_packet* Packet )
{
	Format_ReleaseRef(p,Packet->Data);
	Packet->Data = NULL;
	Packet->Next = p->FreePackets;
	p->FreePackets = Packet;
}

void Format_ReleaseStream(format_base* p,int No)
{
	format_stream* Stream = p->Streams[No];
	format_packet* Packet;

	Stream->Packet.DropLevel = 0;
	Stream->Pending = 0;
	Stream->LastTime = -1;

	while (Stream->PacketFirst)
	{
		Packet = Stream->PacketFirst;
		Stream->PacketFirst = Packet->Next;
		if (Stream->PacketLast == Packet)
			Stream->PacketLast = NULL;

		Format_PacketRelease(p,Packet);
	}

	if (p->ReleaseStream)
		p->ReleaseStream(p,Stream);
}

void Format_AfterSeek(format_base* p)
{
	int No;

	// release stream packets
	for (No=0;No<p->StreamCount;++No)
		Format_ReleaseStream(p,No);

	if (p->AfterSeek)
		p->AfterSeek(p);
}

void Format_ReleaseBuffers(format_base* p)
{
	format_buffer* Buffer;

	// release read buffers
	p->NoMoreInput = 0;

	p->ReadPos = 0;
	if (p->ReadBuffer)
	{
		Format_BufferRelease(p,p->ReadBuffer);
		p->ReadBuffer = NULL;
	}

	while ((Buffer = Format_BufferRemove(p))!=NULL)
		Format_BufferRelease(p,Buffer);
}

int Format_Reset(format_base* p)
{
	int Result = ERR_NONE;
	format_stream* Stream;
	format_packet* Packet;
	format_buffer* Buffer;
	format_ref*	Ref;
	int No;

	if (p->BufferLock) // only if it was inited
	{
		Format_AfterSeek(p);
		Format_ReleaseBuffers(p);
	}

	if (p->Done)
		p->Done(p);

	while (p->FreePackets)
	{
		Packet = p->FreePackets;
		p->FreePackets = Packet->Next;
		Free(Packet);
	}

	while (p->FreeBuffers)
	{
		Buffer = p->FreeBuffers;
		p->FreeBuffers = Buffer->Next;
		Free(Buffer);
	}

	while (p->FreeRefs)
	{
		Ref = p->FreeRefs;
		p->FreeRefs = Ref->Next;
		Free(Ref);
	}

	for (No=0;No<p->StreamCount;++No)
	{
		if (p->FreeStream)
			p->FreeStream(p,p->Streams[No]);
		Stream = p->Streams[No];
		Free(Stream->Format);
		if (Stream->ExtraLength > 0)
			Free(Stream->Extra);
		MergeDone(&Stream->Merge);
		Free(Stream);
	}

	p->StreamCount = 0;
	p->ActiveStreams = 0;

	LockDelete(p->BufferLock);
	LockDelete(p->InputLock);
	p->BufferLock = NULL;
	p->InputLock = NULL;

	if (p->Input)
	{
		if (!NodeIsClass((node*)p->Input,STREAM_CLASS))
			return ERR_INVALID_PARAM;

		if (p->Input->Get(p->Input,STREAM_LENGTH,&p->FileSize,sizeof(p->FileSize)) != ERR_NONE)
			p->FileSize = -1;

		p->InputLock = LockCreate();
		p->BufferLock = LockCreate();
		p->Duration = -1;
		p->NeedDuration = 1;
		p->NoMoreInput = 0;
		p->SyncMode = 1;			// start with syncmode
		p->SyncStream = NULL;
		p->HeaderLoaded = 0;
		p->FilePos = 0;

		p->SeekByPacket_DataStart = 0;
		p->SeekByPacket_BlockAlign = 1;

		if (p->Init)
			Result = p->Init(p);
	}
	return Result;
}

int Format_Get(format_base* p,int No,void* Data,int Size)
{
	int Result = ERR_INVALID_PARAM;
	if (No >= FORMAT_COMMENT && No < FORMAT_COMMENT+p->StreamCount)  
	{
		GETVALUE(p->Streams[No-FORMAT_COMMENT]->Comment,pin)
	}
	else
	if (No >= FORMAT_STREAM && No < FORMAT_STREAM+p->StreamCount)
	{
		GETVALUE(p->Streams[No-FORMAT_STREAM]->Pin,pin)
	}
	else
	if (No >= (FORMAT_STREAM|PACKET_FORMAT) && No < (FORMAT_STREAM|PACKET_FORMAT)+p->StreamCount)
	{
		No -= FORMAT_STREAM|PACKET_FORMAT;
		if (Size == p->Streams[No]->FormatLength)
		{
			format_stream* s = p->Streams[No];
			memcpy(Data,s->Format,s->FormatLength);

			if (s->Pin.Node)
			{
				video Video;
				audio Audio;

				//try to fill unkown fields from codec
				switch (p->Streams[No]->FormatType)
				{
				case PACKET_VIDEO:
					if (s->Pin.Node->Get(s->Pin.Node,s->Pin.No|PACKET_FORMAT,&Video,sizeof(Video))==ERR_NONE)
						CombineVideo((video*)Data,&Video);
					break;
				case PACKET_AUDIO:
					if (s->Pin.Node->Get(s->Pin.Node,s->Pin.No|PACKET_FORMAT,&Audio,sizeof(Audio))==ERR_NONE)
						CombineAudio((audio*)Data,&Audio);
					break;
				}
			}
			Result = ERR_NONE;
		}
	}
	else
	switch (No)
	{
	case NODE_ID: GETVALUE(p->Id,int); break;
	case FORMAT_BUFFERSIZE: GETVALUE(p->BufferSize*CHUNKSIZE/1024,int); break;
	case FORMAT_INPUT: GETVALUE(p->Input,stream*); break;
	case FORMAT_EXTS: GETSTRING(p->Exts); break;
	case FORMAT_DURATION: GETVALUECOND(p->Duration,tick_t,p->Duration>=0); break;
	case FORMAT_FILEPOS: GETVALUECOND(p->FilePos,int,p->FilePos>=0); break;
	case FORMAT_UPDATESTREAMS: GETVALUE(p->UpdateStreams,pin); break;
	case FORMAT_FILESIZE:  GETVALUECOND(p->FileSize/1024,int,p->FileSize>=0); break;
	}
	return Result;
}

int Format_Set(format_base* p,int No,const void* Data,int Size)
{
	uint8_t* Ptr;
	int Result = ERR_INVALID_PARAM;

	if (No >= FORMAT_COMMENT && No < FORMAT_COMMENT+p->StreamCount)
	{
		No -= FORMAT_COMMENT;
		SETVALUE(p->Streams[No]->Comment,pin,ERR_NONE);
	}
	else
	if (No >= FORMAT_STREAM && No < FORMAT_STREAM+p->StreamCount)
	{
		No -= FORMAT_STREAM;
		if (Size == sizeof(pin))
		{
			if (p->Streams[No]->Pin.Node)
				p->ActiveStreams--;
			Result = NodeSetPin(&p->Streams[No]->Pin,(const pin*)Data,p->Streams[No]->Format,p->Streams[No]->FormatLength);
			if (!p->Streams[No]->Pin.Node)
				Format_ReleaseStream(p,No);
			else
				p->ActiveStreams++;
		}
	}
	else
	switch (No)
	{
	case FORMAT_INPUT: SETVALUE(p->Input,stream*,Format_Reset(p)); break;
	case FORMAT_UPDATESTREAMS: SETVALUE(p->UpdateStreams,pin,ERR_NONE); break;

	case NODE_REFRESH:
		p->DropTolerance = QueryAdvPlatform(ADVPLATFORM_DROPTOL);
		p->SkipTolerance = QueryAdvPlatform(ADVPLATFORM_SKIPTOL);
		p->AVOffset = QueryAdvPlatform(ADVPLATFORM_AVOFFSET);
		break;

	case FORMAT_DATAFEED:

		Ptr = (uint8_t*) Data;
		while (Size)
		{
			format_buffer* Buffer;

			Buffer = Format_BufferAlloc(p,2);
			if (Buffer)
			{
				Buffer->Length = Size;
				if (Buffer->Length > CHUNKSIZE)
					Buffer->Length = CHUNKSIZE;

				memcpy(Buffer->Buffer,Ptr,Buffer->Length);

				Ptr += Buffer->Length;
				Size -= Buffer->Length;

				Format_BufferInsert(p,Buffer);
			}
		}

		break;

	case FORMAT_BUFFERSIZE: 
		if (Size == sizeof(int))
		{
 			p->BufferSize = *(int*)Data * 1024;
			if (p->BufferSize < MINBUFFER)
				p->BufferSize = MINBUFFER;
			p->BufferSize /= CHUNKSIZE;
			Result = ERR_NONE;
			break;
		}
	}
	return Result;
}

format_packet* Format_PacketAlloc( format_base* p)
{
	format_packet* Packet = p->FreePackets;

	if (!Packet)
	{
		Packet = (format_packet*) Alloc(sizeof(format_packet));
		if (!Packet)
			return NULL;
	}
	else
		p->FreePackets = Packet->Next;

	Packet->Data = NULL;
	Packet->RefTime = -1;
	Packet->Stream = NULL;
	Packet->Next = NULL;
	Packet->Key = 1;

	return Packet;
}

int Format_Sync(format_base* p,tick_t Time,int FilePos)
{
	if (!p->Seek || !p->Input)
		return ERR_NOT_SUPPORTED;

	if (Time>=0 || FilePos>=0)
	{
		int Result = p->Seek(p,Time,FilePos);
		if (Result != ERR_NONE)
			return Result;
	}

	p->SyncMode = 1;
	p->SyncStream = NULL;
	return ERR_NONE;
}

int Format_Send(format_base* p,format_stream* Stream)
{
	int Result;
	tick_t CurrTime = p->ProcessTime;
	
	Stream->Packet.CurrTime = CurrTime;

	if (Stream->Packet.RefTime >= 0 && !Stream->DisableDrop)
	{
		if (CurrTime >= 0 && Stream->Packet.RefTime < CurrTime - p->DropTolerance)
		{
			if (Stream->Packet.RefTime < CurrTime - p->SkipTolerance - p->DropTolerance)
				Stream->Packet.DropLevel = 2;
			else
				Stream->Packet.DropLevel = 1;
		}
		else
			Stream->Packet.DropLevel = 0;
	}

	DEBUG_MSG4(DEBUG_FORMAT,T("Send stream:%d size:%d time:%d drop:%d"),Stream->No,Stream->Packet.Length,Stream->Packet.RefTime,Stream->Packet.DropLevel);

	// sending packet
	Result = Stream->Pin.Node->Set(Stream->Pin.Node,Stream->Pin.No,&Stream->Packet,sizeof(packet));

	// packet not processed?
	if (Result == ERR_BUFFER_FULL)
	{
		DEBUG_MSG1(DEBUG_FORMAT,T("Send stream:%d BufferFull"),Stream->No);
		return Result;
	}

	if (p->Sended)
		p->Sended(p,Stream);

	Stream->Pending = 0;

	// sync: found a valid frame?
	if (p->SyncMode && Result == ERR_NONE)
	{
		p->SyncMode = 0;
		p->SyncTime = Stream->Packet.RefTime;
		if (p->SyncTime < 0)
			p->SyncTime = Stream->LastTime;
		Result = ERR_SYNCED;
	}

	return Result;
}

int Format_FillQueue(format_base* p)
{
	int Result;
	format_packet* Packet;
	
	if (!p->BufferAvailable && !p->ReadBuffer)
		return ERR_NEED_MORE_DATA;

	Packet = Format_PacketAlloc(p);
	if (!Packet)
		return ERR_OUT_OF_MEMORY;

	Result = p->ReadPacket(p,Packet);

	if (Result == ERR_NONE && Packet->Stream && Packet->Stream->Pin.Node)
	{
		format_stream* s = Packet->Stream;

		if (Packet->RefTime>=0)
		{
			if (s->FormatType == PACKET_AUDIO)
			{
				Packet->RefTime += p->AVOffset;
				if (Packet->RefTime < 0)
					Packet->RefTime = 0;
			}

			if (s->LastTime<0)
				s->LastTime = Packet->RefTime; // found reftime
		}

		if (s->LastTime >= 0)
		{
			// add packet to stream
			if (!s->PacketFirst)
				s->PacketFirst = Packet;
			else
				s->PacketLast->Next = Packet;
			s->PacketLast = Packet;
		}
		else
			Format_PacketRelease(p,Packet);
	}
	else
		Format_PacketRelease(p,Packet);

	return Result;
}

int Format_ProcessStream(format_base* p,format_stream* Stream)
{
	format_packet* Packet;
	format_ref* Ref;
	int Result = ERR_NONE;
	int No;
	int Burst;

	if (Stream->Pending)
	{
		Result = Format_Send(p,Stream);

		if (Result == ERR_BUFFER_FULL || Result == ERR_SYNCED)
			return Result;
	}

	// process a limited number of packets at once (other streams need cpu time too)

	Burst = Stream->PacketBurst;

	for (No=0;No<Burst;++No)
	{
		while (!Stream->PacketFirst)
		{
			// prevent audio from going too much ahead
			if (p->BufferAvailable * CHUNKSIZE < (MINBUFFER/2) && !p->NoMoreInput)
				return ERR_NEED_MORE_DATA;

			Result = Format_FillQueue(p);

			if (Result != ERR_NONE || (p->Bench && (No || !Stream->PacketFirst)))
				return Result;
		}

		Packet = Stream->PacketFirst;

		// sync: all non SyncStream packets are stored (or dropped)
		if (p->SyncMode && Stream != p->SyncStream)
		{
			// catch up to syncstream (drop packets)
			while (Packet && Packet->RefTime < p->SyncStream->LastTime)
			{
				Stream->PacketFirst = Packet->Next;
				if (Stream->PacketLast == Packet)
					Stream->PacketLast = NULL;

				Format_PacketRelease(p,Packet);

				Packet = Stream->PacketFirst;
			}
			return ERR_NEED_MORE_DATA; // allow SyncStream to need more data
		}

		//DEBUG_MSG3(DEBUG_FORMAT,T("Packet stream:%d time:%d pos:%d"),Stream->No,Packet->RefTime,Packet->FilePos);

		// build output packet

		if (Packet->RefTime >= 0)
			Stream->LastTime = Packet->RefTime;

		Stream->Packet.RefTime = Packet->RefTime;
		Stream->Packet.Key = Packet->Key;
		Ref = Packet->Data;

		if (!Ref)
		{
			Stream->Packet.Data[0] = NULL;
			Stream->Packet.Length = 0;
			Stream->Pending = 1;
		}
		else
		if (!Stream->Fragmented)
		{
			if (Ref->Next || (Ref->Length + Ref->Begin + SAFETAIL > CHUNKSIZE) || Stream->ForceMerge)
			{
				// merge references
				MergeClear(&Stream->Merge);

				for (;Ref;Ref=Ref->Next)
					MergeWrite(&Stream->Merge,Ref->Buffer->Buffer + Ref->Begin, Ref->Length);

				Stream->Packet.Data[0] = Stream->Merge.Data;
				Stream->Packet.Length = Stream->Merge.WritePos;
			}
			else
			{
				// single reference
				Stream->Packet.Data[0] = Ref->Buffer->Buffer + Ref->Begin;
				Stream->Packet.Length = Ref->Length;
			}
			Stream->Pending = 1;
		}
		else
		{
			// send first reference (doesn't matter how long)
			Stream->Packet.Data[0] = Ref->Buffer->Buffer + Ref->Begin;
			Stream->Packet.Length = Ref->Length;
			Stream->Pending = 2; // indicate parital sending
		}

		Result = Format_Send(p,Stream);

		if (Result == ERR_BUFFER_FULL || Result == ERR_SYNCED)
			break;

		if (Stream->Packet.DropLevel==2)
			Burst = Stream->PacketBurst*2; // try to catch up
	}

	// buffer full: there was possibly some processing so don't allow sleeping
	// need data: burst count exceeded, but it doesn't mean there is no data left in buffers
	if (Result == ERR_BUFFER_FULL || Result == ERR_NEED_MORE_DATA)
		Result = ERR_NONE;

	return Result;
}

void Format_Sended(format_base* p, format_stream* Stream)
{
	format_packet* Packet = Stream->PacketFirst;

	if (Stream->Pending == 2 && Packet->Data)
	{
		// release sended references
		format_ref* Ref = Packet->Data;
		Packet->Data = Ref->Next;
		Format_SingleRefRelease(p,Ref);

		if (Packet->Data)
		{
			// packet still not finished
			Packet->RefTime = -1;
			return;
		}
	}

	// full packet is released
	Stream->PacketFirst = Packet->Next;
	if (Stream->PacketLast == Packet)
		Stream->PacketLast = NULL;

	Format_PacketRelease(p,Packet);
}

format_stream* Format_DefSyncStream( format_base* p )
{
	format_stream* Found = NULL;
	format_stream* Stream;
	int No;

	// prefer video streams

	for (No=0;No<p->StreamCount;++No)
	{
		Stream = p->Streams[No];
		if (Stream->Pin.Node)
		{
			if (Stream->FormatType == PACKET_VIDEO)
				return Stream;

			if (!Found)
				Found = Stream;
		}
	}

	return Found;
}

static INLINE bool_t Format_GetReadBuffer( format_base* p )
{
	// release old buffer
	if (p->ReadBuffer)
	{
		p->ReadPos -= p->ReadBuffer->Length;
		Format_BufferRelease(p,p->ReadBuffer);
	}

	// get new buffer from BufferFirst..BufferLast chain
	p->ReadBuffer = Format_BufferRemove(p);
	if (p->ReadBuffer)
		return 1;

	// load input now (can't wait for input thread)
	return Format_ReadBuffer(p,1);
}

int Format_Process( format_base* p,tick_t* Time,bool_t* WakeInput,bool_t Fill)
{
	int No;
	int Result;
	int AllNeedMoreData = 1;
	int AllEndOfFile = 1;
	int LastBufferUsed = p->BufferUsed;

	p->ProcessTime = *Time;
	p->Bench = p->ProcessTime < 0;

	if (!p->HeaderLoaded && p->FilePos <= HEADERLOAD)
	{
		// load format begining to find streams

		Result = p->FillQueue(p);

		if (p->FilePos > HEADERLOAD || (p->StreamCount && p->FilePos > HEADERLOAD2) || Format_Eof(p))
			p->HeaderLoaded = 1;

		*WakeInput = p->BufferUsed < LastBufferUsed;
		return Result;
	}

	if (p->SyncMode)
	{
		p->ProcessTime = -1;
		// we need to choose one stream (to find a non dropped frame)
		if (!p->SyncStream)
		{
			p->SyncStream = Format_DefSyncStream(p);
			if (!p->SyncStream)
			{
				// no sync needed
				p->SyncMode = 0;
				p->SyncTime = -1;

				// try to calc duration once (after first sync)
				if (p->NeedDuration)
				{
					p->NeedDuration = 0;
					CalcDuration(p);
				}

				// leave time as it is
				return ERR_SYNCED;
			}
		}
	}

	// if all streams are full: caller can go to sleep
	Result = ERR_BUFFER_FULL;

	for (No=0;No<p->StreamCount;++No)
		if (p->Streams[No]->Pin.Node)
		{
			int i = p->Process(p,p->Streams[No]);

			if (!p->SyncMode || p->SyncStream == p->Streams[No])
			{
				if (i == ERR_NEED_MORE_DATA)
				{
					if (!p->NoMoreInput)
					{
						// if less then 1/8 sec is left and stream needs data -> force need more data mode
						tick_t LastTime = p->Streams[No]->LastTime;
						if (Fill || (LastTime >= 0 && p->ProcessTime >= 0 && LastTime < p->ProcessTime + TICKSPERSEC/8))
							Result = ERR_NEED_MORE_DATA;
					}
				}
				else
					AllNeedMoreData = 0;

				if (i != ERR_END_OF_FILE)
					AllEndOfFile = 0;
			}

			if (i == ERR_SYNCED)
			{
				// try to calc duration once (after first sync)
				if (p->NeedDuration)
				{
					p->NeedDuration = 0;
					CalcDuration(p);
				}

				*Time = p->SyncTime;
				return i;
			}

			if (Result == ERR_BUFFER_FULL &&
				i != ERR_BUFFER_FULL && 
				i != ERR_END_OF_FILE && 
				i != ERR_NEED_MORE_DATA)
				Result = ERR_NONE; // no sleep
		}

	if (!p->ActiveStreams)
	{
		// release buffers
		format_buffer* Buffer;
		while ((Buffer = Format_BufferRemove(p))!=NULL)
		{
			p->FilePos += Buffer->Length;
			Format_BufferRelease(p,Buffer);
		}

		DEBUG_MSG1(DEBUG_FORMAT,T("NoStreams filepos:%d"),p->FilePos);

		*Time = -1; // player should use file position

		// return buffer full so fill mode doesn't go though file
		Result = p->NoMoreInput ? ERR_END_OF_FILE:ERR_BUFFER_FULL;
	}
	else
	{
		if (AllEndOfFile) // all end of file?
			Result = ERR_END_OF_FILE;
		else
		if (AllNeedMoreData) // all need data?
			Result = p->NoMoreInput ? ERR_END_OF_FILE:ERR_NEED_MORE_DATA;
	}

	if (p->SyncMode && Result == ERR_END_OF_FILE)
	{
		p->SyncMode = 0;
		if (p->Duration < 0)
			p->SyncTime = p->SyncStream->LastTime;
		else
			p->SyncTime = p->Duration;
		*Time = p->SyncTime;
		Result = ERR_SYNCED;
	}

	*WakeInput = p->BufferUsed < LastBufferUsed;
	return Result;
}

format_ref* Format_RefAlloc( format_base* p, format_buffer* To, int Begin, int Length )
{
	format_ref* Ref = p->FreeRefs;

	if (!Ref)
	{
		Ref = (format_ref*) Alloc(sizeof(format_ref));
		if (!Ref)
			return NULL;
	}
	else
		p->FreeRefs = Ref->Next;

	LockEnter(p->BufferLock);

	//DEBUG_MSG3(DEBUG_FORMAT,"Buffer:%d addref %d->%d",To->Id,To->RefCount,To->RefCount+1);
	++To->RefCount;

	LockLeave(p->BufferLock);

	Ref->Length = Length;
	Ref->Begin = Begin;
	Ref->Buffer = To;
	Ref->Next = NULL;
	return Ref;
}

bool_t Format_ReadBuffer( format_base* p, bool_t ToRead )
{
	bool_t Result;

	LockEnter(p->InputLock);
	LockEnter(p->BufferLock);

	if (!p->BufferFirst && !p->NoMoreInput)
	{
		format_buffer* Buffer = Format_BufferAlloc(p,1);

		if (Buffer)
		{
			Buffer->Length = p->Input->Read(p->Input,Buffer->Buffer,CHUNKSIZE);

			if (Buffer->Length < 0)
				Format_BufferRelease(p,Buffer);
			else
			{
				p->NoMoreInput = Buffer->Length == 0;
				if (p->NoMoreInput)
					Format_BufferRelease(p,Buffer);
				else
					Format_BufferInsert(p,Buffer);
			}
		}
	}

	Result = p->BufferFirst != NULL;

	if (Result && ToRead)
		p->ReadBuffer = Format_BufferRemove(p);

	LockLeave(p->BufferLock);
	LockLeave(p->InputLock);

	return Result;
}

format_ref* Format_DupRef(format_base* p, format_ref* Chain, int Offset, int Length )
{
	format_ref* Head = NULL;
	format_ref** Ptr = &Head;

	while (Chain && Offset >= Chain->Length)
	{
		Chain = Chain->Next;
		Offset -= Chain->Length;
	}

	while (Chain && Length > 0)
	{
		int Len = Chain->Length - Offset;
		if (Len > Length)
			Len = Length;

		*Ptr = Format_RefAlloc(p,Chain->Buffer,Chain->Begin+Offset,Len);
		if (!*Ptr)
			break;

		Chain = Chain->Next;
		Length -= Len;
		Offset = 0;

		Ptr = &(*Ptr)->Next;
	}

	return Head;
}

format_ref* Format_ReadAsRef( format_base* p, int Length )
{
	int n;
	format_ref* Ref;
	format_ref** RefPtr;

	Ref = NULL;
	RefPtr = &Ref;

	if (Length < 0) // read as much as possible from one buffer
	{
		while (!p->ReadBuffer || p->ReadPos >= p->ReadBuffer->Length)
			if (!Format_GetReadBuffer(p))
				return NULL;

		n = p->ReadBuffer->Length - p->ReadPos;
		if (n > -Length)
			n = -Length;
		Length = n;
	}

	while (Length > 0)
	{
		while (!p->ReadBuffer || p->ReadPos >= p->ReadBuffer->Length)
			if (!Format_GetReadBuffer(p))
				return Ref;

		n = p->ReadBuffer->Length - p->ReadPos;
		if (n > Length)
			n = Length;

		*RefPtr = Format_RefAlloc(p,p->ReadBuffer,p->ReadPos,n);
		if (!*RefPtr)
			break;

		RefPtr = &(*RefPtr)->Next;

		Length -= n;
		p->FilePos += n;
		p->ReadPos += n;
	}
	return Ref;
}

bool_t Format_Eof(format_base* p)
{
	return !p->BufferAvailable && !p->ReadBuffer && p->NoMoreInput;
}

int Format_Read( format_base* p, void* Data, int Length0 )
{
	uint8_t* Ptr = (uint8_t*)Data;
	int Length = Length0;
	int n;
	while (Length > 0)
	{
		while (!p->ReadBuffer || p->ReadPos >= p->ReadBuffer->Length)
			if (!Format_GetReadBuffer(p))
				return Length0 - Length;

		n = p->ReadBuffer->Length - p->ReadPos;
		if (n > Length)
			n = Length;

		memcpy(Ptr,p->ReadBuffer->Buffer+p->ReadPos,n);
		Ptr += n;
		Length -= n;
		p->FilePos += n;
		p->ReadPos += n;
	}

	return Length0;
}

void Format_Skip(format_base* p,int n)
{
	if (n > 0)
	{
		p->FilePos += n;
		p->ReadPos += n;
	}
}

int Format_Read8(format_base* p)
{
	while (!p->ReadBuffer || p->ReadPos >= p->ReadBuffer->Length)
		if (!Format_GetReadBuffer(p))
			return -1;

	p->FilePos++;
	return p->ReadBuffer->Buffer[p->ReadPos++];
}

int Format_ReadBE16(format_base* p)
{
	int v = Format_Read8(p) << 8;
	v |= Format_Read8(p);
	return v;
}

int Format_ReadLE16(format_base* p)
{
	int v = Format_Read8(p);
	v |= Format_Read8(p) << 8;
	return v;
}

int Format_ReadBE32(format_base* p)
{
	int v = Format_Read8(p) << 24;
	v |= Format_Read8(p) << 16;
	v |= Format_Read8(p) << 8;
	v |= Format_Read8(p);
	return v;
}

int Format_ReadLE32(format_base* p)
{
	int v = Format_Read8(p);
	v |= Format_Read8(p) << 8;
	v |= Format_Read8(p) << 16;
	v |= Format_Read8(p) << 24;
	return v;
}

int64_t Format_ReadBE64(format_base* p)
{
	int64_t v = (int64_t)Format_ReadBE32(p) << 32;
	v |= Format_ReadBE32(p);
	return v;
}

int64_t Format_ReadLE64(format_base* p)
{
	int64_t v = Format_ReadLE32(p);
	v |= (int64_t)Format_ReadLE32(p) << 32;
	return v;
}

int Format_ReadInput(format_base* p,int BufferPercent,int* BufferUsed)
{
	int Result = ERR_NONE;
	int MaxUsed;
	format_buffer* Buffer;

	LockEnter(p->BufferLock);

	BufferPercent *= p->BufferSize; // unit: CHUNKSIZE/PERCENT_ONE

	if (BufferPercent < (MINBUFFER/CHUNKSIZE)*PERCENT_ONE)
		BufferPercent = (MINBUFFER/CHUNKSIZE)*PERCENT_ONE;

	MaxUsed = BufferPercent/PERCENT_ONE;
	if (p->BufferUsed >= MaxUsed)
	{
		if (BufferUsed)
			*BufferUsed = p->BufferUsed*CHUNKSIZE/1024;
		LockLeave(p->BufferLock);
		return ERR_BUFFER_FULL;
	}

	// allocate buffer (optional)
	Buffer = Format_BufferAlloc(p,0);

	if (BufferUsed)
		*BufferUsed = p->BufferUsed*CHUNKSIZE/1024;

	//DEBUG_MSG2(-1,T("Used:%d Max:%d"),p->BufferUsed,MaxUsed);

	if (p->BufferUsed >= MaxUsed)
		Result = ERR_BUFFER_FULL;

	LockLeave(p->BufferLock);

	if (!Buffer)
		return ERR_BUFFER_FULL;

	LockEnter(p->InputLock);
	Buffer->Length = p->Input->Read(p->Input,Buffer->Buffer,CHUNKSIZE);

	if (Buffer->Length < 0)
	{
		Format_BufferRelease(p,Buffer);
		Result = ERR_DEVICE_ERROR;
	}
	else
	{
		p->NoMoreInput = Buffer->Length == 0;
		if (p->NoMoreInput)
		{
			Format_BufferRelease(p,Buffer);
			Result = ERR_END_OF_FILE;
		}
		else
		{
			//DEBUG_MSG2(DEBUG_FORMAT,"Buffer:%d input %d",Buffer->Id,Buffer->RefCount);
			LockEnter(p->BufferLock);
			Format_BufferInsert(p,Buffer);
			LockLeave(p->BufferLock);
		}
	}

	LockLeave(p->InputLock);
	return Result;
}

void Format_PrepairStream(format_base* p,format_stream* Stream)
{
	if (Stream->FormatType == PACKET_AUDIO && Stream->Format && 
		((audio*)Stream->Format)->Format == AUDIOFMT_PCM)
	{
		// there could be alignement problems (with memory and with audio.BlockSize)
		Stream->Fragmented = 0;
		Stream->ForceMerge = 1;
	}

	if (p->UpdateStreams.Node)
		p->UpdateStreams.Node->Set(p->UpdateStreams.Node,p->UpdateStreams.No,NULL,0);
}

void* Format_StreamFormat(format_stream* s, int Type, int Length )
{
	s->FormatLength = Length;
	s->FormatType = Type;
	s->Format = CAlloc(Length,1);
	s->PacketBurst = (Type == PACKET_VIDEO) ? VIDEO_BURST : DEFAULT_BURST;
	return s->Format;
}

void* Format_StreamExtra(format_stream* s, int Length )
{
	s->Extra = Length ? CAlloc(Length,1) : NULL;
	s->ExtraLength = Length;

	if (s->Format)
		switch (s->FormatType)
		{
		case PACKET_VIDEO:
			((video*)s->Format)->Extra = s->Extra;
			((video*)s->Format)->ExtraLength = s->ExtraLength;
			break;

		case PACKET_AUDIO:
			((audio*)s->Format)->Extra = s->Extra;
			((audio*)s->Format)->ExtraLength = s->ExtraLength;
			break;
		}

	return s->Extra;
}

int Format_BitmapInfoMem(format_stream* s,void* Data,int Length)
{
	char* p = (char*)Data;
	video* Video = (video*) Format_StreamFormat(s,PACKET_VIDEO,sizeof(video));
	if (Video && Length>=4)
	{
		int Size = INT32LE(*(int32_t*)p);
		if (Length >= Size)
		{
			Video->Width = INT32LE(*(int32_t*)(p+4));
			Video->Height = INT32LE(*(int32_t*)(p+8));
			Video->Pixel.BitCount = INT16LE(*(int16_t*)(p+14));
			Video->Pixel.FourCC = INT32LE(*(int32_t*)(p+16));
			Video->Aspect = ASPECT_ONE;

			if (Format_StreamExtra(s,Size-40))
				memcpy(s->Extra,p+40,s->ExtraLength);

			Format_DefaultVideo(Video);
		}
		else
			Size = 4;
		Length -= Size;
	}
	return Length;
}

int Format_BitmapInfo(format_base* p,format_stream* s,int Length)
{
	video* Video = (video*) Format_StreamFormat(s,PACKET_VIDEO,sizeof(video));
	if (Video && Length>=4)
	{
		int Size = Format_ReadLE32(p); //size
		if (Length >= Size)
		{
			Video->Width = Format_ReadLE32(p); 
			Video->Height = Format_ReadLE32(p); 
			Format_ReadLE16(p); //planes
			Video->Pixel.BitCount = Format_ReadLE16(p);
			Video->Pixel.FourCC = Format_ReadLE32(p);
			Video->Aspect = ASPECT_ONE;
			Format_Skip(p,20); //SizeImage,XPelsPerMeter,YPelsPerMeter,ClrUsed,ClrImportant

			if (Format_StreamExtra(s,Size-40))
				Format_Read(p,s->Extra,s->ExtraLength);

			Format_DefaultVideo(Video);
		}
		else
			Size = 4;

		Length -= Size;
	}
	return Length;
}

void Format_DefaultVideo( video* Video )
{
	if (!Video->Pixel.FourCC) // DIB?
	{
		switch (Video->Pixel.BitCount)
		{
		case 16:
			DefaultRGB(&Video->Pixel,Video->Pixel.BitCount,5,5,5,0,0,0);
			break;
		case 24:
		case 32:
			DefaultRGB(&Video->Pixel,Video->Pixel.BitCount,8,8,8,0,0,0);
			break;
		}
		Video->Direction = DIR_MIRRORUPDOWN;
	}
	else
	{
		Video->Pixel.Flags = PF_FOURCC;
		Video->Pixel.FourCC = UpperFourCC(Video->Pixel.FourCC);
		Video->Direction = 0;
	}
	DefaultPitch(Video);

	if (Video->Height<0)
	{
		Video->Height = -Video->Height;
		Video->Direction ^= DIR_MIRRORUPDOWN;
	}
}

void Format_DefaultAudio( audio* Audio )
{
	// detect fake PCM
	if (Audio->Format > 8192 && 
		Audio->BlockAlign > 0 &&
		Audio->ByteRate > 0 &&
		Audio->BlockAlign == ((Audio->Channels * Audio->Bits) >> 3) &&
		Audio->ByteRate == Audio->SampleRate * Audio->BlockAlign)
		Audio->Format = AUDIOFMT_PCM;

	if (Audio->Format == AUDIOFMT_PCM)
	{
		Audio->FracBits = Audio->Bits - 1;
		if (Audio->Bits <= 8)
			Audio->Flags = PCM_UNSIGNED;
		Audio->Bits = (Audio->Bits + 7) & ~7;
		Audio->BlockAlign = (Audio->Channels * Audio->Bits) >> 3;
	}
}

int Format_WaveFormatMem(format_stream* s,void* Data,int Length)
{
	char* p = (char*)Data;
	audio* Audio = (audio*) Format_StreamFormat(s,PACKET_AUDIO,sizeof(audio));
	if (Audio && Length >= 16)
	{
		Audio->Format = INT16LE(*(int16_t*)(p+0));
		Audio->Channels =INT16LE(*(int16_t*)(p+2));
		Audio->SampleRate = INT32LE(*(int32_t*)(p+4));
		Audio->ByteRate = INT32LE(*(int32_t*)(p+8));
		Audio->BlockAlign = INT16LE(*(int16_t*)(p+12));
		Audio->Bits = INT16LE(*(int16_t*)(p+14));
		Length -= 16;

		if (Length >= 2)
		{
			int Extra = INT16LE(*(int16_t*)(p+16));
			Length -= 2;
			if (Extra > 0 && Length >= Extra)
			{
				if (Format_StreamExtra(s,Extra))
					memcpy(s->Extra,p+18,s->ExtraLength);
				Length -= Extra;
			}
		}

		Format_DefaultAudio(Audio);
	}
	return Length;
}

int Format_WaveFormat(format_base* p,format_stream* s,int Length)
{
	audio* Audio = (audio*) Format_StreamFormat(s,PACKET_AUDIO,sizeof(audio));
	if (Audio && Length >= 16)
	{
		Audio->Format = Format_ReadLE16(p);
		Audio->Channels = Format_ReadLE16(p);
		Audio->SampleRate = Format_ReadLE32(p);
		Audio->ByteRate = Format_ReadLE32(p);
		Audio->BlockAlign = Format_ReadLE16(p);
		Audio->Bits = Format_ReadLE16(p);
		Length -= 16;

		if (Length >= 2)
		{
			int Extra = Format_ReadLE16(p);
			Length -= 2;
			if (Extra > 0 && Length >= Extra)
			{
				if (Format_StreamExtra(s,Extra))
					Format_Read(p,s->Extra,Extra);
				Length -= Extra;
			}
		}

		Format_DefaultAudio(Audio);
	}
	return Length;
}

void Format_RemoveStream(format_base* p)
{
	if (p->StreamCount)
	{
		format_stream* Stream = p->Streams[--p->StreamCount];
		if (p->FreeStream)
			p->FreeStream(p,Stream);
		Free(Stream->Format);
		if (Stream->ExtraLength > 0)
			Free(Stream->Extra);
		MergeDone(&Stream->Merge);
		Free(Stream);
	}
}

format_stream* Format_AddStream(format_base* p, int Length )
{
	format_stream* s;

	if (p->StreamCount == MAXSTREAM)
		return NULL;

	s = (format_stream*) CAlloc(Length,1);
	if (!s)
		return NULL;

	MergeInit(&s->Merge);
	s->No = p->StreamCount;
	s->LastTime = -1;
	p->Streams[p->StreamCount++] = s;
	return s;
}

format_base FormatBase =
{
	Format_Enum,
	Format_Get,
	Format_Set,

	NULL, // Probe
	Format_Sync,
	Format_ReadInput,
	Format_Process,
	NULL, //ReadList
	NULL, //WriteList

	NULL, //Init
	NULL, //Done
	NULL, //Seek
	Format_FillQueue, //FillQueue
	NULL, //ReadPacket
	NULL, //FreeStream
	NULL, //ReleaseStream
	NULL, //ReleaseBuffers

	Format_ProcessStream,
	Format_Sended,
	NULL, //BackupPacketState
};
