/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: idct.h 131 2004-12-04 20:36:04Z picard $
 *
 * BetaPlayer Core
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/
 
#ifndef __IDCT_H
#define __IDCT_H

//---------------------------------------------------------------
// motion vector macros
#define NOMV		0x80000000

#define MVX(a)		(((a)<<16)>>16)
#define MVY(a)		((a)>>16)
#define MAKEMV(x,y)	(((y)<<16)|((x)&0xFFFF))

#ifdef BIG_ENDIAN
#define MVXIDX	1
#define MVYIDX	0
#else
#define MVXIDX	0
#define MVYIDX	1
#endif

//---------------------------------------------------------------
// idct (abstract)

#define IDCT_CLASS			FOURCC('I','D','C','T')
// motion compesation rounding
#define IDCT_ROUNDING		0x50
// idct data format, see below (int)
#define IDCT_FORMAT			0x51
// fastest idct data format (int readonly)
#define IDCT_NATIVEFORMAT	0x52
// number of buffers (int)
#define IDCT_BUFFERCOUNT	0x53
// video output pin (TYPE_VIDEO)
#define IDCT_OUTPUT			0x54
// output direction, just a hint (int)
#define IDCT_DIRECTION		0x55
// showed buffer (int)
#define IDCT_SHOW			0x56
// frame no associated with a buffer (int) (array)
#define IDCT_FRAMENO		0x1000

extern DLL const datadef IDCTParams[];

//---------------------------------------------------------------
// input format for Data and MCompData functions
#define IDCTFORMAT_BLOCK	1
#define IDCTFORMAT_PACK		2

//---------------------------------------------------------------
// scan orders (not every data format requires)
#define IDCTSCAN_ZIGZAG		0
#define IDCTSCAN_ALT_HORI	1
#define IDCTSCAN_ALT_VERT	2

//---------------------------------------------------------------
// block format
//
// input is a idct_block_t[8][8] matrix
// which is directly passed to 8x8 idct
// attention: input matrix is overwritten(trashed) after the call
//
// length/scantype is hint how much data is in the matrix
// block[scan[length..63]] souhld be zero
// pass length=64 if you can't provide this information

typedef short int idct_block_t;

#define IDCT_BLOCK_DECLARE							\
	idct_block_t _block[2*64+16/sizeof(idct_block_t)];

#define IDCT_BLOCK_PREPARE(x,out)					\
	out = (idct_block_t*)ALIGN16(x->_block);		


//---------------------------------------------------------------
// idct virtual method table

// set all buffer's frameno to -1
typedef void (*idctdrop)(void* This);
// send packet to output (Packet->Data, Packet->LastData is filled by idct)
typedef	int (*idctsend)(void* This,packet* Packet,int* DroppedCounter);
// lock one of the buffers
typedef	int (*idctlock)(void* This,int No,planes Planes,int* Brightness,video* Format);
// unlock buffer
typedef	void (*idctunlock)(void* This,int No);
// start decoding a frame (result is the old frameno in dst buffer, -1 if content is totally lost )
typedef	int (*idctframestart)(void* This,int FrameNo,int Dst,int Back,int Fwd,int Show,bool_t Drop);
// end decoding a frame
typedef	void (*idctframeend)(void* This);

// start processing a macroblock
typedef	void (*idctprocess)(void* This,int x,int y);
// copy a macroblock from the backward or forward buffer
typedef void (*idctcopy)(void* This,int x,int y,int Forward);
// motion compensation data (MVBack and MVFwd has to be valid during Inter8x8 are called)
typedef void (*idctmcomp)(void* This,const int *MVBack,const int *MVFwd);
// passing idct data 
typedef void (*idctintra)(void* This,void* Data,int Length,int ScanType);
// passing idct data (always has to be called even if there is no data)
typedef void (*idctinter)(void* This,void* Data,int Length);

typedef struct idct
{
	nodeenum		Enum;
	nodeget			Get;
	nodeset			Set;

	idctsend		Send;
	idctdrop		Drop;
	idctlock		Lock;
	idctunlock		Unlock;
	idctframestart	FrameStart;
	idctframeend	FrameEnd;

	idctprocess		Process;
	idctcopy		Copy16x16;
	idctmcomp		MComp16x16;
	idctmcomp		MComp8x8;
	idctintra		Intra8x8;
	idctinter		Inter8x8; // data is optional (length=0)
} idct;

DLL void IDCTSetEmpty(idct*); // set empty decode functions

#endif
