/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: nullaudio.c 131 2004-12-04 20:36:04Z picard $
 *
 * BetaPlayer Core
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "stdafx.h"

typedef struct nullaudio
{
	node VMT;

	audio InputFormat;

} nullaudio;

static int Enum( nullaudio* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,FlowParams,OutputParams,AOutputParams,NULL);
}

static int Get( nullaudio* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(NULLAUDIO_ID,int); break;
	case AOUT_INPUT|PACKET_FORMAT: GETVALUE(p->InputFormat,audio); break;
	}
	return Result;
}

static int Set( nullaudio* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case AOUT_INPUT|PACKET_FORMAT:
		if (!Size || Size == sizeof(audio))
		{
			memset(&p->InputFormat,0,sizeof(audio));
			if (Size)
			{
				p->InputFormat = *(const audio*)Data;
				if (p->InputFormat.Format != AUDIOFMT_PCM) 
					return ERR_INVALID_DATA;
			}
			Result = ERR_NONE;
		}
		break;
	case AOUT_INPUT:
		if (Size == sizeof(packet))
		{
			const packet* Packet = (const packet*)Data;

			if (Packet->CurrTime != -1 && Packet->RefTime > Packet->CurrTime + SHOWAHEAD)
				Result = ERR_BUFFER_FULL;
			else
				Result = ERR_NONE;
		}
		break;
	}
	return Result;
}

static nullaudio NullAudio;

void NullAudio_Init()
{
	NodeFill(&NullAudio.VMT,sizeof(nullaudio),(nodeenum)Enum,(nodeget)Get,(nodeset)Set);
	NodeRegister(&NullAudio.VMT,PRI_MINIMUM);
}

void NullAudio_Done()
{
	NodeUnRegister(&NullAudio.VMT);
}
