/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: nullvideo.c 131 2004-12-04 20:36:04Z picard $
 *
 * BetaPlayer Core
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "stdafx.h"

typedef struct nullvideo
{
	node VMT;

	video InputFormat;

} nullvideo;

static int Enum( nullvideo* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,FlowParams,OutputParams,VOutputParams,NULL);
}

static int Get( nullvideo* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(NULLVIDEO_ID,int); break;
	case VOUT_PRIMARY: GETVALUE(0,bool_t); break;
	case VOUT_INPUT|PACKET_FORMAT: GETVALUE(p->InputFormat,video); break;
	}
	return Result;
}

static int Set( nullvideo* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case VOUT_INPUT|PACKET_FORMAT:
		if (!Size || Size == sizeof(video))
		{
			memset(&p->InputFormat,0,sizeof(video));
			if (Size)
			{
				p->InputFormat = *(const video*)Data;
				if (Compressed(&p->InputFormat.Pixel))
					return ERR_NOT_SUPPORTED;
			}
			Result = ERR_NONE;
		}
		break;
	case VOUT_INPUT:
		if (Size == sizeof(packet))
		{
			const packet* Packet = (const packet*)Data;

			if (Packet->CurrTime != -1 && Packet->RefTime > Packet->CurrTime + SHOWAHEAD)
				Result = ERR_BUFFER_FULL;
			else
				Result = ERR_NONE;
		}
		break;

	case FLOW_EMPTY:
		Result = ERR_NONE;
		break;
	}
	return Result;
}

static nullvideo NullVideo;

void NullVideo_Init()
{
	NodeFill(&NullVideo.VMT,sizeof(nullvideo),(nodeenum)Enum,(nodeget)Get,(nodeset)Set);
	NodeRegister(&NullVideo.VMT,PRI_MINIMUM);
}

void NullVideo_Done()
{
	NodeUnRegister(&NullVideo.VMT);
}
