/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: platform.h 142 2004-12-12 07:30:52Z picard $
 *
 * BetaPlayer Core
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#ifndef __PLATFORM_H
#define __PLATFORM_H

#define DEBUG_TEST			0x01
#define DEBUG_FORMAT		0x02
#define DEBUG_PLAYER		0x04
#define DEBUG_WIN			0x08
#define DEBUG_SYS			0x10
#define DEBUG_VIDEO			0x20
#define DEBUG_AUDIO			0x40
#define DEBUG_VCODEC		0x80
#define DEBUG_ACODEC		0x100
#define DEBUG_VCODEC2		0x200

#define PLATFORM_ID			FOURCC('P','L','A','T')

#define PLATFORM_LANG		0x10
#define PLATFORM_MODEL		0x12
#define PLATFORM_CAPS		0x13
#define PLATFORM_ICACHE		0x14
#define PLATFORM_DCACHE		0x15
#define PLATFORM_TYPE		0x17
#define PLATFORM_VER		0x19
#define PLATFORM_TYPENO		0x1A
#define PLATFORM_OEMINFO	0x18
#define PLATFORM_VERSION	0x1B
#define PLATFORM_CPU		0x11
#define PLATFORM_MHZ		0x22
#define PLATFORM_CPUID		0x24

#define PLATFORM_BATTERYTIMEOUT	0x2000
#define PLATFORM_ACTIMEOUT		0x2001
#define PLATFORM_SCREENSAVER	0x2002
#define PLATFORM_DISPPOWER		0x2003

#define ADVPLATFORM_ID		FOURCC('A','D','V','P')

#define ADVPLATFORM_OLDSHELL		0x1F
#define ADVPLATFORM_NOBACKLIGHT		0x20
#define ADVPLATFORM_NOWMMX			0x28
#define ADVPLATFORM_SLOW_VIDEO		0x25
#define ADVPLATFORM_IDCTSWAP		0x29
#define ADVPLATFORM_COLOR_LOOKUP	0x26
#define ADVPLATFORM_HOMESCREEN		0x2D
#define ADVPLATFORM_ALLKEYS			0x2E
#define ADVPLATFORM_BENCHFROMPOS	0x30
#define ADVPLATFORM_PIXELASPECT		0x31

#define ADVPLATFORM_DROPTOL			0x2A
#define ADVPLATFORM_SKIPTOL			0x2B
#define ADVPLATFORM_AVOFFSET		0x2C

// models
#define	MODEL_DEFAULT			0
#define	MODEL_IPAQ_H3600		1
#define	MODEL_IPAQ_H3700		2
#define	MODEL_IPAQ_H3800		3
#define	MODEL_IPAQ_H3900		4
#define	MODEL_O2_XDA			5
#define	MODEL_TOSHIBA_E740		6
#define	MODEL_TOSHIBA_E750		7
#define	MODEL_CASIO_E105		8
#define	MODEL_CASIO_E115		9
#define	MODEL_CASIO_E125		10
#define	MODEL_CASIO_EM500		11
#define	MODEL_CASIO_E200		12
#define MODEL_AMIGO_600C		13
#define MODEL_JVC_C33			14
#define MODEL_JORNADA_680		15
#define MODEL_JORNADA_690		16
#define MODEL_JORNADA_710		17
#define MODEL_JORNADA_720		18
#define MODEL_SIGMARION			19
#define MODEL_SIGMARION2		20
#define MODEL_SIGMARION3		21
#define MODEL_MOBILEGEAR2_450	22
#define MODEL_MOBILEGEAR2_430	23
#define MODEL_MOBILEGEAR2_520	24
#define MODEL_MOBILEGEAR2_530	25
#define MODEL_MOBILEGEAR2_550	26
#define MODEL_MOBILEGEAR2_700	27
#define MODEL_MOBILEGEAR2_730	28
#define MODEL_MOBILEPRO_770		29
#define MODEL_MOBILEPRO_780		30
#define MODEL_INTERMEC_6651		31
#define MODEL_JORNADA_540		32
#define	MODEL_TOSHIBA_E800		33
#define	MODEL_CASIO_BE300		34
#define MODEL_SPV_C500_ORIGROM	35
#define MODEL_BSQUARE_ALCHEMY	36
#define MODEL_IPAQ_HX4700		37
#define MODEL_AXIM_X5			38
#define MODEL_AXIM_X50			39

// special cpu and framebuffer features
#define CAPS_ARM_5E				0x0001
#define CAPS_ARM_XSCALE			0x1000
#define CAPS_ARM_WMMX			0x0002
#define CAPS_MIPS_VR41XX		0x0004
#define CAPS_X86_MMX			0x0008
#define CAPS_X86_MMX2			0x0100
#define CAPS_X86_SSE			0x0010
#define CAPS_X86_SSE2			0x0200
#define CAPS_X86_3DNOW			0x0400
#define CAPS_ONLY12BITRGB		0x0020
#define CAPS_LANDSCAPE			0x0040
#define CAPS_PORTRAIT			0x0080
#define CAPS_OLDSHELL			0x0800

// platform types
#define TYPE_UNKNOWN			0
#define TYPE_POCKETPC			1
#define TYPE_SMARTPHONE			2

DLL void Platform_Init();
DLL void Platform_Done();

DLL void Advanced_Init();
DLL void Advanced_Done();

DLL void CalcDropLevel( packet* );

DLL void Dump();

DLL int QueryPlatform( int Param );
DLL bool_t QueryAdvPlatform( int Param );
DLL void QueryDesktop( video* );
DLL int QueryOrientation();
DLL bool_t IsOrientationChanged();
DLL void GlobalInvalidate( const rect* );

DLL void* PhyMemBegin(uint32_t Addr,uint32_t Length);
DLL void PhyMemEnd(void*);

DLL void* LockFrameBuffer( video* );
DLL void UnlockFrameBuffer();

DLL const tchar_t* GetDocumentPath();
DLL const tchar_t* GetSystemPath();

DLL void SoftReset();
DLL bool_t KernelMode(bool_t);
DLL bool_t GetDisplayPower();
DLL void SetDisplayPower(bool_t State);
DLL void BacklightTimeout(bool_t Disable);
DLL void IdleTimerReset();

DLL void SetFileAssociation(const tchar_t* Ext,const tchar_t* Type,bool_t State);
DLL bool_t GetFileAssociation(const tchar_t* Ext,const tchar_t* Type);
DLL void SetFileType(const tchar_t* Type, const tchar_t* Open, const tchar_t* Icon );

DLL void* CodeAlloc(int Size);
DLL void CodeFree(void*,int Size);
DLL void CodeLock(void*,int Size);
DLL void CodeUnlock(void*,int Size);
DLL void CodeFindPages(void* Ptr,unsigned char** Min,unsigned char** Max,int* PageSize);

DLL	void* LockCreate();
DLL void LockDelete(void*);
DLL	void LockEnter(void*);
DLL void LockLeave(void*);

DLL void WaitBegin();
DLL void WaitEnd();

DLL void* ThreadCreate( int(*Start)(void*) ,void* Parameter,int Quantum);

DLL int AvailMemory();
DLL void* Alloc(int);
DLL void* CAlloc(int,int);
DLL void* ReAlloc(void*,int);
DLL void Free(void*);
DLL void* Alloc16(int); //aligned to 16 bytes
DLL void Free16(void*);
DLL void CheckHeap();
DLL extern bool_t ShowMemoryLow;

DLL int GetTime(); // in millisec

DLL void GetProc(void** Module,void* Ptr,const tchar_t* ProcName,int Optional);

DLL void ShowMessage(const tchar_t* Title,const tchar_t* Msg,...);
DLL void DebugMessage(const tchar_t*,...);
DLL void DebugTime(const tchar_t*);
DLL extern int DebugMask;

#ifndef NDEBUG
  #define DEBUG_MSG(m,x) if (DebugMask & (m)) DebugMessage(x)
  #define DEBUG_MSG1(m,x,a) if (DebugMask & (m)) DebugMessage(x,a)
  #define DEBUG_MSG2(m,x,a,b) if (DebugMask & (m)) DebugMessage(x,a,b)
  #define DEBUG_MSG3(m,x,a,b,c) if (DebugMask & (m)) DebugMessage(x,a,b,c)
  #define DEBUG_MSG4(m,x,a,b,c,d) if (DebugMask & (m)) DebugMessage(x,a,b,c,d)
  #define DEBUG_MSG5(m,x,a,b,c,d,e) if (DebugMask & (m)) DebugMessage(x,a,b,c,d,e)
  #define DEBUG_MSG6(m,x,a,b,c,d,e,f) if (DebugMask & (m)) DebugMessage(x,a,b,c,d,e,f)
  #define DEBUG_MSG7(m,x,a,b,c,d,e,f,g) if (DebugMask & (m)) DebugMessage(x,a,b,c,d,e,f,g)
#else
  #define DEBUG_MSG(m,x)
  #define DEBUG_MSG1(m,x,a)
  #define DEBUG_MSG2(m,x,a,b)
  #define DEBUG_MSG3(m,x,a,b,c)
  #define DEBUG_MSG4(m,x,a,b,c,d)
  #define DEBUG_MSG5(m,x,a,b,c,d,e)
  #define DEBUG_MSG6(m,x,a,b,c,d,e,f)
  #define DEBUG_MSG7(m,x,a,b,c,d,e,f,g)
#endif

DLL int SafeException(void*);

#if defined(_MSC_VER)
#define SAFE_BEGIN __try {
#define SAFE_END ;} __except (SafeException(GetExceptionInformation())) {}
#define TRY_BEGIN __try {
#define TRY_END ;} __except (1) {}
#else
#define SAFE_BEGIN !error!
#define SAFE_END !error!
#define TRY_BEGIN !error!
#define TRY_END !error!
#endif

#define PLATFORM_CRASH_MESSAGE		100
#define PLATFORM_CRASH_TITLE		101
#define PLATFORM_YES				102
#define PLATFORM_NO					103
#define PLATFORM_DUMP_MESSAGE		104
#define PLATFORM_DUMP_TITLE			105
#define PLATFORM_OK					106
#define PLATFORM_CANCEL				107
#define PLATFORM_ERROR				108
#define PLATFORM_MEMORY_LOW			109
#define PLATFORM_KEY_ENTER			110
#define PLATFORM_KEY_UP				111
#define PLATFORM_KEY_DOWN			112
#define PLATFORM_KEY_LEFT			113
#define PLATFORM_KEY_RIGHT			114
#define PLATFORM_KEY_SPACE			115
#define PLATFORM_KEY_APP			116
#define PLATFORM_KEY_CTRL			117
#define PLATFORM_KEY_SHIFT			118
#define PLATFORM_KEY_ALT			119
#define PLATFORM_KEY_WIN			120
#define PLATFORM_KEY_ACTION			123
#define PLATFORM_ASSIGN				121
#define PLATFORM_CLEAR				122
#define PLATFORM_RESET				130
#define PLATFORM_KEY_ESCAPE			131

#define HOTKEY_MASK					0x0FFFF
#define HOTKEY_SHIFT				0x10000
#define HOTKEY_CTRL					0x20000
#define	HOTKEY_ALT					0x40000
#define	HOTKEY_WIN					0x80000
#define	HOTKEY_KEEP					0x100000

#endif
