/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: softidct.c 140 2004-12-10 07:10:49Z picard $
 *
 * BetaPlayer SoftIDCT 
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/
 
#include "../stdafx.h"
#include "softidct.h"

#ifdef MIPS64

// 8 byte steps
static void edge_copy(uint8_t *p_border_top,uint8_t *p_border_top_ref,int width){
	__asm (	"srl	$8,$6,4;"
			"bucle: ldr	$10,0($5);"
			"ldr	$11,8($5);"
#ifdef MIPSVR41XX
			".set noreorder;"
			"cache	13,0($4);" //cache without loading
			".set reorder;"
#endif
			"addi	$5,$5,16;"
			"sdr	$10,0($4);"
			"sdr	$11,8($4);"
			"addi	$8,$8,-1;"
			"addi	$4,$4,16;"
			"bgtz	$8,bucle;"
			"andi	$6,$6,0x08;"
			"beq	$6,$0,final;"
			"ldr	$10,0($5);"
			"sdr	$10,0($4);"
			"final:"
			);
}

// 16 byte steps
static void edge_set(uint8_t *p_border_top,int p_border_top_ref,int width) {
	__asm (	"srl	$8,$6,4;"
			"dsll	$10,$5,56;"
			"dsrl	$11,$10,8;"
			"or		$10,$10,$11;"
			"dsrl	$11,$10,16;"
			"or		$10,$10,$11;"
			"dsrl	$11,$10,32;"
			"or		$10,$10,$11;"
			"bucle2:"
#ifdef MIPSVR41XX
			".set noreorder;"
			"cache	13,0($4);" //cache without loading
			".set reorder;"
#endif
			"sdr	$10,0($4);"
			"sdr	$10,8($4);"
			"addi	$8,$8,-1;"
			"addi	$4,$4,16;"
			"bgtz	$8,bucle2;"
			);
}
#else
#define edge_copy memcpy
#define edge_set memset
#endif

static void FillEdge(uint8_t *Ptr, int Width, int Height, int EdgeX, int EdgeY)
{
	int n;
	uint8_t *p;
	int InnerWidth = Width - EdgeX*2;
	int InnerHeight = Height - EdgeY*2;
		
	// left and right
	p = Ptr + EdgeX + EdgeY * Width;
	for (n=0;n<InnerHeight;++n)
	{
		edge_set(p-EdgeX, p[0], EdgeX);
		edge_set(p+InnerWidth,p[InnerWidth-1], EdgeX);
		p += Width;
	}

	// top
	p = Ptr;
	for (n=0;n<EdgeY;++n)
	{
		edge_copy(p, Ptr+EdgeY * Width, Width);
		p += Width;
	}

	// bottom
	p = Ptr + (Height-EdgeY)*Width;
	for (n=0;n<EdgeY;++n)
	{
		edge_copy(p, Ptr+(Height-EdgeY-1)*Width, Width);
		p += Width;
	}
}

static void FillEdgeYUV(softidct* p,uint8_t *Ptr)
{
	FillEdge(Ptr, p->BufWidth, p->BufHeight, EDGE, EDGE );
	Ptr += p->YToU;
	FillEdge(Ptr, p->BufWidth>>p->UVX2, p->BufHeight>>p->UVY2, EDGE>>p->UVX2, EDGE>>p->UVY2);
	Ptr += p->YToU >> (p->UVX2+p->UVY2);
	FillEdge(Ptr, p->BufWidth>>p->UVX2, p->BufHeight>>p->UVY2, EDGE>>p->UVX2, EDGE>>p->UVY2);
}

#if defined( MIPS64_WIN32 ) 
static int TouchPages(softidct* p)
{
	uint8_t* i;
	int Sum = 0;
	for (i=p->CodeBegin;i<p->CodeEnd;i+=p->CodePage)
		Sum += *i;
	return Sum;
}
#endif

static int Lock(softidct* p,int No,planes Planes,int* Brightness,video* Format)
{
	if (Brightness)
		*Brightness = 0;
	if (Format)
	{
		*Format = p->OutputFormat;
		Format->Pitch = p->BufWidth;
		Format->Width = p->BufWidth - 2*EDGE;
		Format->Height = p->BufHeight - 2*EDGE;
	}

	if (No<0 || No>=p->BufCount)
	{
		Planes[0] = NULL;
		Planes[1] = NULL;
		Planes[2] = NULL;
		return ERR_INVALID_PARAM;
	}

	Planes[0] = p->Buffer[No] + EDGE + EDGE*p->BufWidth;
	Planes[1] = p->Buffer[No] + p->YToU + (EDGE >> p->UVX2) + (EDGE >> p->UVY2)*(p->BufWidth >> p->UVX2);
	Planes[2] = (uint8_t*)Planes[1] + (p->YToU >> (p->UVX2+p->UVY2));
	return ERR_NONE;
}

static void Unlock(softidct* p,int No)
{
}

static int SetRounding(softidct* p,int Rounding)
{
	p->CopyBlock = p->AllCopyBlock[0][Rounding];
	p->CopyMBlock = p->AllCopyBlock[1][Rounding];
	p->Rounding = Rounding;
	return ERR_NONE;
}

static int SetBufferCount(softidct* p, int n, int Temp)
{
	int i;
	n += Temp;
	if (n>p->MaxCount) return ERR_NOT_SUPPORTED;

	for (i=n;i<p->BufCount;++i)
	{
		Free(p->_Buffer[i]);
		p->_Buffer[i] = NULL;
		p->Buffer[i] = NULL;
	}
	if (p->BufCount > n)
		p->BufCount = n;

	for (i=p->BufCount;i<n;++i)
	{
#ifndef MIPS
		int Align = 32;
		int Offset = 0;
#else
		int Align = 8192;
		int Offset = ((i>>1)+(i&1)*2)*2048;
#endif
		p->_Buffer[i] = (uint8_t*) Alloc(p->BufSize+Align+Offset);
		if (!p->_Buffer[i]) 
		{
			p->MaxCount = p->BufCount;
			break;
		}
		p->Buffer[i] = (uint8_t*)(((int)p->_Buffer[i] + Align-1) & ~(Align-1));
		p->Buffer[i] += Offset;
		p->BufBorder[i] = 0;
		p->BufFrameNo[i] = -1;
		p->BufCount = i+1;
	}

	if (p->ShowNext >= p->BufCount)
		p->ShowNext = -1;
	if (p->ShowCurr >= p->BufCount)
		p->ShowCurr = -1;
	p->LastTemp = Temp && p->BufCount == n;

	if (p->BufCount != n)
		return ERR_OUT_OF_MEMORY;

	return ERR_NONE;
}

static int Init(softidct* p)
{
	int AlignX,AlignY,Mode;

	PlanarYUV(&p->OutputFormat.Pixel,&p->UVX2,&p->UVY2,NULL);

	AlignX = (8 << p->UVX2) - 1;
	AlignY = (8 << p->UVY2) - 1;

	p->BufWidth = ((p->OutputFormat.Width+AlignX)&~AlignX) + 2*EDGE;
	p->BufHeight = ((p->OutputFormat.Height+AlignY)&~AlignY) + 2*EDGE;
	p->YToU = p->BufHeight * p->BufWidth;

	p->BufSize = p->YToU + 2*(p->YToU >> (p->UVX2+p->UVY2));
	p->BufWidthUV = p->BufWidth >> p->UVX2;

	p->OutputFormat.Pitch = p->BufWidth;

	if (p->OutputFormat.Pixel.Flags & PF_YUV420)
	{
		Mode = 0;
		p->Tab[0] = 2*(8);					//Y[0;0] -> Y[0;1]
		p->Tab[1] = 2*(8*p->BufWidth-8);	//Y[0;1] -> Y[1;0]
		p->Tab[2] = 2*(8);					//Y[1;0] -> Y[1;1]
		p->Tab[4] = 2*(p->YToU >> 2);		//U->V
		p->Tab[5] = 2*(0);	
	}
	else 
	if (p->OutputFormat.Pixel.Flags & PF_YUV422)
	{
		Mode = 1;
		p->Tab[0] = 2*(8);					//Y[0;0] -> Y[0;1]
		p->Tab[2] = 2*(p->YToU >> 1);		//U->V
		p->Tab[3] = 2*(0);	
	}
	else
	if (p->OutputFormat.Pixel.Flags & PF_YUV444)
	{
		Mode = 2;
		p->Tab[0] = 2*(p->YToU);			//Y->U
		p->Tab[1] = 2*(p->YToU);			//U->V
		p->Tab[2] = 2*(0);					
	}
	else
		return ERR_NOT_SUPPORTED;

#ifdef IDCTSWAP
	if (p->OutputFormat.Direction & DIR_SWAPXY)
	{
		p->VMT.Process = p->ProcessSwap[Mode];
		p->VMT.Copy16x16 = p->Copy16x16Swap[Mode];
		p->VMT.Intra8x8 = p->Intra8x8Swap;
		p->VMT.MComp8x8 = (idctmcomp) SoftMComp8x8Swap;
		p->VMT.MComp16x16 = (idctmcomp) SoftMComp16x16Swap;

		if (Mode == 0)
		{
			p->Tab[0] = 2*(8*p->BufWidth);		//Y[0;0] -> Y[1;0]
			p->Tab[1] = 2*(8-8*p->BufWidth);	//Y[1;0] -> Y[0;1]
			p->Tab[2] = 2*(8*p->BufWidth);		//Y[0;1] -> Y[1;1]
		}
	}
	else
#endif
	{
		p->VMT.Process = p->Process[Mode];
		p->VMT.Copy16x16 = p->Copy16x16[Mode];
		p->VMT.Intra8x8 = p->Intra8x8;
		p->VMT.MComp8x8 = (idctmcomp) SoftMComp8x8;
		p->VMT.MComp16x16 = (idctmcomp) SoftMComp16x16;
	}

#ifdef MIPS64
	p->VMT.Inter8x8 = Inter8x8Add;
#endif

	if (p->Output.Node)
	{
		int Result = p->Output.Node->Set(p->Output.Node,p->Output.No | PACKET_FORMAT,&p->OutputFormat,sizeof(video));
		if (Result != ERR_NONE)
			return Result;
	}

	return ERR_NONE;
}

#ifdef IDCTSWAP
static void ChangeSwap(softidct* p)
{
	//DEBUG_MSG(-1,T("IDCTSoft SWAP"));

	if (p->BufCount > 0)
	{
		blitfx FX;
		int No;
		planes TmpPlanes;
		video TmpFormat;
		int Brightness;
		video Format[2];
		planes Planes[2];

		memset(&FX,0,sizeof(FX));
		FX.ScaleX = SCALE_ONE;
		FX.ScaleY = SCALE_ONE;

		TmpFormat = p->OutputFormat;
		TmpFormat.Width = p->BufWidth - 2*EDGE;
		TmpFormat.Height = p->BufHeight - 2*EDGE;

		if (SurfaceAlloc(TmpPlanes,&TmpFormat) == ERR_NONE)
		{
			if (Lock(p,0,Planes[0],&Brightness,&Format[0])==ERR_NONE)
			{
				SurfaceCopy(&Format[0],&TmpFormat,Planes[0],TmpPlanes,&FX);
				Unlock(p,0);
			}

			FX.Direction = DIR_SWAPXY;

			for (No=1;No<p->BufCount;++No)
				if (Lock(p,No,Planes[0],&Brightness,&Format[0])==ERR_NONE)
				{
					SwapInt(&p->BufWidth,&p->BufHeight);
					if (Lock(p,0,Planes[1],&Brightness,&Format[1])==ERR_NONE)
					{
						SurfaceCopy(&Format[0],&Format[1],Planes[0],Planes[1],&FX);
						Unlock(p,0);

						SwapPByte(&p->Buffer[No],&p->Buffer[0]);
						SwapPByte(&p->_Buffer[No],&p->_Buffer[0]);
						p->BufBorder[No] = 0;
					}
					SwapInt(&p->BufWidth,&p->BufHeight);
					Unlock(p,No);
				}

			SwapInt(&p->BufWidth,&p->BufHeight);
			if (Lock(p,0,Planes[0],&Brightness,&Format[0])==ERR_NONE)
			{
				SurfaceCopy(&TmpFormat,&Format[0],TmpPlanes,Planes[0],&FX);
				Unlock(p,0);
				p->BufBorder[0] = 0;
			}
			SwapInt(&p->BufWidth,&p->BufHeight);

			SurfaceFree(TmpPlanes);
		}
	}

	p->OutputFormat.Direction ^= DIR_SWAPXY;
	SwapInt(&p->OutputFormat.Width,&p->OutputFormat.Height);
	Init(p);
}
#endif

static void CheckSwap(softidct* p)
{
#ifdef IDCTSWAP
	if (QueryAdvPlatform(ADVPLATFORM_IDCTSWAP) && !(p->OutputFormat.Pixel.Flags & PF_YUV422))
	{
		if (p->Output.Node)
		{
			video Prefer;
			if (p->Output.Node->Get(p->Output.Node,p->Output.No | PACKET_PREF_FORMAT,&Prefer,sizeof(video))==ERR_NONE)
			{
				if ((p->OutputFormat.Direction ^ Prefer.Direction) & DIR_SWAPXY)
					ChangeSwap(p);
			}
		}
	}
	else
	if (p->OutputFormat.Direction & DIR_SWAPXY)
		ChangeSwap(p);
#endif
}

static int SetFormat(softidct* p, const video* Format )
{
	SetBufferCount(p,0,0);
	memset(&p->OutputFormat,0,sizeof(p->OutputFormat));

	if (Format)
	{
		if (!(Format->Pixel.Flags & (PF_YUV420|PF_YUV422|PF_YUV444)))
			return ERR_NOT_SUPPORTED;

		p->OutputFormat = *Format;
		if (p->OutputFormat.Direction & DIR_SWAPXY)
			SwapInt(&p->OutputFormat.Width,&p->OutputFormat.Height);

		p->OutputFormat.Direction = 0;
		p->OutputFormat.Pixel.Flags |= PF_16ALIGNED | PF_SAFEBORDER;

		p->MaxCount = MAXBUF;
		return Init(p);
	}
	else
	{
		if (p->Output.Node)
			p->Output.Node->Set(p->Output.Node,p->Output.No | PACKET_FORMAT,NULL,0);

		return ERR_NONE;
	}
}

static int Send(softidct* p,packet* Packet,int* Dropped);

static int Set(softidct* p, int No, const void* Data, int Size )
{
	packet Packet;
	int Result = ERR_INVALID_PARAM;

	switch (No)
	{
	case IDCT_SHOW: 
		SETVALUE(p->ShowNext,int,ERR_NONE); 
		if (p->ShowNext >= p->BufCount) p->ShowNext = -1;
		break;

	case FLOW_UPDATE:
		CheckSwap(p);
		Result = ERR_NONE;
		break;

	case FLOW_EMPTY:
		p->ShowCurr = -1;
		p->ShowNext = -1;
		return ERR_NONE;

	case IDCT_FORMAT:
		//only block format is supported by softidct
		if (Size == sizeof(int))
			Result = (*(const int*)Data == IDCTFORMAT_BLOCK) ? ERR_NONE:ERR_NOT_SUPPORTED;
		break;

	case IDCT_OUTPUT|PACKET_FORMAT:
		if (!Size || Size == sizeof(video))
		{
			if (!Size) Data = NULL;
			Result = SetFormat(p,(const video*)Data);
		}
		break;

	case IDCT_OUTPUT:
		if (Size == sizeof(pin))
		{
			Result = NodeSetPin(&p->Output,(pin*)Data,&p->OutputFormat,sizeof(video));
			if (Result == ERR_NONE)
				CheckSwap(p);
		}
		break;

	case IDCT_ROUNDING:
		if (Size == sizeof(bool_t))
			Result = SetRounding(p,*(const bool_t*)Data);
		break;

	case IDCT_BUFFERCOUNT:
		if (Size == sizeof(int))
			Result = SetBufferCount(p,*(const int*)Data,0);
		break;

	case FLOWBUFFER_RESEND:

		Packet.CurrTime = -1;
		Packet.RefTime = -1;
		Packet.DropLevel = 0;
		Result = Send(p,&Packet,NULL);
		break;
	}
	return Result;
}

static int Get(softidct* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(p->Id,int); break;
	case IDCT_NATIVEFORMAT: GETVALUE(IDCTFORMAT_BLOCK,int); break;
	case IDCT_FORMAT: GETVALUE(IDCTFORMAT_BLOCK,int); break;
	case IDCT_OUTPUT|PACKET_FORMAT: GETVALUE(p->OutputFormat,video); break;
	case IDCT_OUTPUT: GETVALUE(p->Output,pin); break;
	case IDCT_ROUNDING: GETVALUE(p->Rounding,bool_t); break;
	case IDCT_BUFFERCOUNT: GETVALUE(p->BufCount - p->LastTemp,int); break;
	case IDCT_SHOW: GETVALUE(p->ShowNext,int); break;
	}
	return Result;
}

static void Drop(softidct* p)
{
	int i;
	for (i=0;i<p->BufCount;++i)
		p->BufFrameNo[i] = -1;
}

static int Send(softidct* p,packet* Packet,int* Dropped)
{
	int Result = ERR_INVALID_DATA;

	if (Packet->DropLevel)
	{
		if (Dropped)
			++(*Dropped);
		return ERR_NONE;
	}

	if (p->ShowNext < 0)
		return ERR_NEED_MORE_DATA;

	if (p->ShowCurr == p->ShowNext)
		p->ShowCurr = -1;

	if (Lock(p,p->ShowNext,Packet->Data,NULL,NULL) == ERR_NONE)
	{
		Lock(p,p->ShowCurr,Packet->LastData,NULL,NULL);

		//DEBUG_MSG2(DEBUG_TEST,T("IDCT Show Next:%d Curr:%d"),p->ShowNext,p->ShowCurr);

		Result = p->Output.Node->Set(p->Output.Node,p->Output.No,Packet,sizeof(packet));

		Unlock(p,p->ShowCurr);
		Unlock(p,p->ShowNext);
	}

	if (Result != ERR_BUFFER_FULL)
		p->ShowCurr = p->ShowNext;

	return Result;
}

static int FrameStart(softidct* p,int FrameNo,int DstNo,int BackNo,int FwdNo,int ShowNo,bool_t Drop)
{
	int OldFrameNo;
#if defined( MIPS64_WIN32 ) 
	// win32 doesn't support 64bit natively. thread and process
	// changes won't save upper part of 64bit registers.
	// we have to disable interrupt and make sure all 1KB
	// code pages are in memory so no page loading is called
	// during code execution 
	TouchPages(p);	
	DisableInterrupts();
	p->LastY = -1;
#endif

#ifdef ARM
	// ShowCurr will be the next "LastData" in output packet
	// we don't want to lose it's content
	if (p->ShowCurr == DstNo)
	{
		int SwapNo;
		for (SwapNo=0;SwapNo<p->BufCount;++SwapNo)
			if (SwapNo != DstNo && SwapNo != BackNo && SwapNo != FwdNo && SwapNo != ShowNo)
				break;

		if (SwapNo < p->BufCount || SetBufferCount(p,p->BufCount,1)==ERR_NONE)
		{
			//DEBUG_MSG2("IDCT Swap %d,%d",SwapNo,DstNo);

			SwapPByte(&p->Buffer[SwapNo],&p->Buffer[DstNo]);
			SwapPByte(&p->_Buffer[SwapNo],&p->_Buffer[DstNo]);
			SwapInt(&p->BufFrameNo[SwapNo],&p->BufFrameNo[DstNo]);
			SwapBool(&p->BufBorder[SwapNo],&p->BufBorder[DstNo]);

			p->ShowCurr = SwapNo;
		}
	}
#endif

	p->ShowNext = ShowNo;

	OldFrameNo = p->BufFrameNo[DstNo];
	p->BufFrameNo[DstNo] = FrameNo;

	p->Dst = p->Buffer[DstNo];
	p->Ref[0] = NULL;
	p->Ref[1] = NULL;
	p->RefMin[0] = NULL;
	p->RefMin[1] = NULL;
	p->RefMax[0] = NULL;
	p->RefMax[1] = NULL;

	if (BackNo>=0)
	{
		p->Ref[0] = p->Buffer[BackNo];
		if (!p->BufBorder[BackNo])
		{
			p->BufBorder[BackNo] = 1;
			FillEdgeYUV(p,p->Ref[0]);
		}
		p->RefMin[0] = p->Ref[0];
		p->RefMax[0] = p->Ref[0] + p->BufSize -8-8*p->BufWidthUV;
	}

	if (FwdNo>=0)
	{
		p->Ref[1] = p->Buffer[FwdNo];
		if (!p->BufBorder[FwdNo])
		{
			p->BufBorder[FwdNo] = 1;
			FillEdgeYUV(p,p->Ref[1]);
		}
		p->RefMin[1] = p->Ref[1];
		p->RefMax[1] = p->Ref[1] + p->BufSize -8-8*p->BufWidthUV;
	}

	p->BufBorder[DstNo] = 0; //invalidate border for dst

#ifndef MIPS64
#ifdef IDCTSWAP
	if (p->OutputFormat.Direction & DIR_SWAPXY)
		p->VMT.Inter8x8 = p->Inter8x8Swap[p->Ref[1]!=NULL];
	else
#endif
		p->VMT.Inter8x8 = p->Inter8x8[p->Ref[1]!=NULL];
#endif

	return OldFrameNo;
} 

static void FrameEnd(softidct* p)
{
#ifdef MIPS64_WIN32
	EnableInterrupts();
#endif
}

static int Enum( softidct* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,FlowParams,FlowBufferParams,IDCTParams,NULL);
}

static softidct SoftIDCT = 
{
	// VMT
	{
		(nodeenum)Enum,
		(nodeget)Get,
		(nodeset)Set,

		(idctsend)Send,
		(idctdrop)Drop,
		(idctlock)Lock,
		(idctunlock)Unlock,
		(idctframestart)FrameStart,
		(idctframeend)FrameEnd,
		NULL, //process
		NULL, //copy
		NULL, //mcomp16x16
		NULL, //mcomp8x8
		NULL, //intra8x8
		NULL, //inter8x8
	},

	//AllCopyBlock[2][2][4]
	{
	   {{ CopyBlock, CopyBlockHor, CopyBlockVer, CopyBlockHorVer },
		{ CopyBlock, CopyBlockHorRound, CopyBlockVerRound, CopyBlockHorVerRound }},
#ifdef MIPS64
	   {{ CopyMBlock, CopyMBlockHor, CopyMBlockVer, CopyMBlockHorVer },
		{ CopyMBlock, CopyMBlockHorRound, CopyMBlockVerRound, CopyMBlockHorVerRound }},
#endif
	},

#if !defined(MIPS64) && !defined(MIPS32)
	//AddBlock[4]
	{AddBlock, AddBlockHor, AddBlockVer, AddBlockHorVer },
#endif

	//Copy16x16[3]
	{(idctcopy)Copy420,NULL,NULL},
	//Process[3]
	{(idctprocess)Process420,(idctprocess)Process422,(idctprocess)Process444}, 
	//Intra8x8
	(idctintra)Intra8x8,
#ifndef MIPS64
	//Inter8x8[2]
	{(idctinter)Inter8x8Back,(idctinter)Inter8x8BackFwd},
#endif

#ifdef IDCTSWAP
	//Copy16x16Swap[3]
	{(idctcopy)CopySwap420,NULL,NULL},
	//ProcessSwap[3]
	{(idctprocess)ProcessSwap420,(idctprocess)ProcessSwap422,(idctprocess)ProcessSwap444},
	//Intra8x8Swap
	(idctintra)Intra8x8Swap,
#ifndef MIPS64
	//Inter8x8Swap[2]
	{(idctinter)Inter8x8BackSwap,(idctinter)Inter8x8BackFwdSwap},
#endif
#endif
};

void SoftIDCT_Done()
{
	NodeUnRegister((node*)&SoftIDCT);
}

idct* SoftIDCT_Init(int Id,int Priority)
{
	softidct* p = &SoftIDCT;

	// configure IDCT for current hardware configuration

#if defined(MMX)
	if (!(QueryPlatform(PLATFORM_CAPS) & CAPS_X86_MMX))
		return NULL;

#elif defined(ARM)

	if ((QueryPlatform(PLATFORM_CAPS) & CAPS_ARM_WMMX) && !QueryAdvPlatform(ADVPLATFORM_NOWMMX))
	{
		p->Intra8x8 = (idctintra)WMMXIntra8x8;
		p->Inter8x8[0] = (idctinter)WMMXInter8x8Back;
		p->Inter8x8[1] = (idctinter)WMMXInter8x8BackFwd;

#ifdef IDCTSWAP
		p->Intra8x8Swap = (idctintra)WMMXIntra8x8Swap;
		p->Inter8x8Swap[0] = (idctinter)WMMXInter8x8BackSwap;
		p->Inter8x8Swap[1] = (idctinter)WMMXInter8x8BackFwdSwap;
#endif

		p->AllCopyBlock[0][0][0] = WMMXCopyBlock;
		p->AllCopyBlock[0][0][1] = WMMXCopyBlockHor;
		p->AllCopyBlock[0][0][2] = WMMXCopyBlockVer;
		p->AllCopyBlock[0][0][3] = WMMXCopyBlockHorVer;
		p->AllCopyBlock[0][1][0] = WMMXCopyBlock;
		p->AllCopyBlock[0][1][1] = WMMXCopyBlockHorRound;
		p->AllCopyBlock[0][1][2] = WMMXCopyBlockVerRound;
		p->AllCopyBlock[0][1][3] = WMMXCopyBlockHorVerRound;
		p->AddBlock[0] = WMMXAddBlock;
		p->AddBlock[1] = WMMXAddBlockHor;
		p->AddBlock[2] = WMMXAddBlockVer;
		p->AddBlock[3] = WMMXAddBlockHorVer;
	}
	else
	if (QueryPlatform(PLATFORM_CAPS) & CAPS_ARM_5E)
	{
		if (QueryPlatform(PLATFORM_ICACHE) >= 32768)
		{
			p->AllCopyBlock[0][0][0] = FastCopyBlock;
			p->AllCopyBlock[0][0][1] = FastCopyBlockHor;
			p->AllCopyBlock[0][0][2] = FastCopyBlockVer;
			p->AllCopyBlock[0][0][3] = FastCopyBlockHorVer;
			p->AllCopyBlock[0][0][0] = FastCopyBlock;
			p->AllCopyBlock[0][1][1] = FastCopyBlockHorRound;
			p->AllCopyBlock[0][1][2] = FastCopyBlockVerRound;
			p->AllCopyBlock[0][1][3] = FastCopyBlockHorVerRound;
			p->AddBlock[0] = FastAddBlock;
			p->AddBlock[1] = FastAddBlockHor;
			p->AddBlock[2] = FastAddBlockVer;
			p->AddBlock[3] = FastAddBlockHorVer;
		}
		else
		{
			p->AllCopyBlock[0][0][0] = PreLoadCopyBlock;
			p->AllCopyBlock[0][0][1] = PreLoadCopyBlockHor;
			p->AllCopyBlock[0][0][2] = PreLoadCopyBlockVer;
			p->AllCopyBlock[0][0][3] = PreLoadCopyBlockHorVer;
			p->AllCopyBlock[0][0][0] = PreLoadCopyBlock;
			p->AllCopyBlock[0][1][1] = PreLoadCopyBlockHorRound;
			p->AllCopyBlock[0][1][2] = PreLoadCopyBlockVerRound;
			p->AllCopyBlock[0][1][3] = PreLoadCopyBlockHorVerRound;
			p->AddBlock[0] = PreLoadAddBlock;
			p->AddBlock[1] = PreLoadAddBlockHor;
			p->AddBlock[2] = PreLoadAddBlockVer;
			p->AddBlock[3] = PreLoadAddBlockHorVer;
		}
	}

#elif defined(MIPSVR41XX)
	if (!(QueryPlatform(PLATFORM_CAPS) & CAPS_MIPS_VR41XX))
		return NULL;
#endif

#if defined( MIPS64_WIN32 ) 
	CodeFindPages(*(void**)&p->VMT.Enum,&p->CodeBegin,&p->CodeEnd,&p->CodePage);
#endif

	p->Id = Id;
	p->Tmp = (uint8_t*)ALIGN16(p->_Tmp);
	p->MaxCount = MAXBUF;
	p->BufCount = 0;
	p->LastTemp = 0;
	p->ShowCurr = -1;
	p->ShowNext = -1;
	
	SetRounding(p,0);

	NodeRegister((node*)p,PRI_MAXIMUM);

	return &p->VMT;
}

