/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: stream.c 145 2004-12-12 18:18:28Z picard $
 *
 * BetaPlayer Core
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "stdafx.h"

const datadef StreamParams[] = 
{
	{ STREAM_CLASS,		TYPE_CLASS, },
	{ STREAM_MIME,		TYPE_STRING, DF_HIDDEN },
	{ STREAM_URL,		TYPE_STRING },
	{ STREAM_LENGTH,	TYPE_INT, DF_HIDDEN },
	{ STREAM_SILENT,	TYPE_BOOL, DF_HIDDEN },
	{ STREAM_CREATE,	TYPE_BOOL, DF_HIDDEN },
	{ STREAM_CONTENTTYPE, TYPE_STRING, DF_HIDDEN|DF_RDONLY },

	DATADEF_END
};

const tchar_t* GetMime( const tchar_t* URL, tchar_t* Mime )
{
	const tchar_t* s = tcschr(URL,':');
	if (s && s[1] == '/' && s[2] == '/')
	{
		if (Mime)
			TcsNCpy(Mime,URL,s-URL+1);
		s += 3;
	}
	else
	{
		if (Mime)
			tcscpy(Mime,T("file"));
		s = URL;
	}
	return s;
}

stream* GetStream( const tchar_t* URL, bool_t Silent )
{
	tchar_t Mime[MAXPATH];
	stream* Input;
	tchar_t m[MAXPATH];
	int No;

	GetMime(URL,Mime);

	for (No=0;(Input=(stream*)NodeEnum(STREAM_CLASS,No))!=NULL;++No)
		if (Input->Get(Input,STREAM_MIME,m,sizeof(m))==ERR_NONE && tcscmp(Mime,m)==0)
			break;

	if (!Input && !Silent)
	{
		tcsupr(Mime);
		NodeError(NULL,ERR_ID,ERR_MIME_NOT_FOUND,Mime);
	}

	if (Input)
		Input = Input->Create(Input);

	return Input;
}

