/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: stream.h 144 2004-12-12 15:28:39Z picard $
 *
 * BetaPlayer Core
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#ifndef __STREAM_H
#define __STREAM_H

#define STREAM_CLASS		FOURCC('S','T','R','M')

#define STREAM_MIME			0x90
#define STREAM_URL			0x91
#define STREAM_LENGTH		0x92
#define STREAM_SILENT		0x93
#define STREAM_CREATE		0x94
#define STREAM_CONTENTTYPE	0x95

extern DLL const datadef StreamParams[];

typedef struct streamdir
{
	tchar_t FileName[MAXPATH];
	int Type; // from Exts
	int Size; // -1 for directory
	int64_t Date;

	void* Private;

} streamdir;

// read from stream
typedef	int (*streamread)(void* This,void* Data,int Size);
// seek
typedef	int (*streamseek)(void* This,int Pos,int SeekMode);
// write to stream
typedef	int (*streamwrite)(void* This,void* Data,int Size);
// list directory 
typedef	int (*streamenumdir)(void* This,const tchar_t* URL,const tchar_t* Exts,bool_t ExtFilter,streamdir* Item);

typedef struct stream
{
	nodeenum		Enum;
	nodeget			Get;
	nodeset			Set;
	dynnodecreate	Create;
	dynnodedelete	Delete;

	streamread		Read;
	streamseek		Seek;
	streamenumdir	EnumDir;
	streamwrite		Write;

} stream;

DLL const tchar_t* GetMime( const tchar_t* URL, tchar_t* Mime );
DLL stream* GetStream( const tchar_t* URL, bool_t Silent );

DLL void StreamPrintf( stream* Stream, bool_t UTF8, const tchar_t* Msg,...);

#endif
