/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: tools.c 145 2004-12-12 18:18:28Z picard $
 *
 * BetaPlayer Core
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "stdafx.h"

void BoolToString( tchar_t* Out, bool_t Bool, int OutSize )
{
	OutSize /= sizeof(tchar_t);
	TcsNCpy(Out,StringDef(PLATFORM_ID,Bool ? PLATFORM_YES:PLATFORM_NO),OutSize);
};

int StringToFourCC( const tchar_t* In, bool_t Upper )
{
	char s[4+1];
	tchar_t Up[4+1];
	int i;

	if (Upper)
	{
		TcsNCpy(Up,In,4+1);
		tcsupr(Up);
		In = Up;
	}
	TcsToStr(s,In,5);

	for (i=0;i<4;++i)
		if (!s[i])
			for (;i<4;++i)
				s[i] = '_';

	return FOURCC(s[0],s[1],s[2],s[3]);
}

void FourCCToString( tchar_t* Out, int FourCC, int OutSize )
{
	char s[4+1];
	*(int*)s = FourCC;
	s[4] = 0;
	StrToTcs(Out,s,OutSize/sizeof(tchar_t));
}

void FractionToString( tchar_t* Out, const fraction* p, int OutSize, bool_t Percent, int Decimal )
{
	if (Percent)
		stprintf(Out,T("%0.*f%%"),Decimal,(100.0f*p->Num)/p->Den);
	else
		stprintf(Out,T("%0.*f"),Decimal,(float)p->Num/p->Den);
}

int StringToInt( const tchar_t* In, bool_t Hex )
{
	int v=0;
	stscanf(In,Hex ? T("%X"):T("%d"),&v);
	return v;
}

void IntToString( tchar_t* Out, int p, int OutSize, bool_t Hex )
{
	stprintf(Out,Hex ? T("0x%08X"):T("%d"),p);
}

void TickToString( tchar_t* Out, tick_t Tick, int OutSize, bool_t MS, bool_t Extended, bool_t Fix )
{
	if (!MS)
	{
		int Hour,Min,Sec;
		Hour = Tick / 3600 / TICKSPERSEC;
		Tick -= Hour * 3600 * TICKSPERSEC;
		Min = Tick / 60 / TICKSPERSEC;
		Tick -= Min * 60 * TICKSPERSEC;
		Sec = Tick / TICKSPERSEC;
		Tick -= Sec * TICKSPERSEC;
		if (Hour)
			stprintf(Out,T("%d:%02d"),Hour,Min);
		else
			stprintf(Out,Fix?T("%02d"):T("%d"),Min);
		stprintf(Out+tcslen(Out),T(":%02d"),Sec);
		if (Extended)
			stprintf(Out+tcslen(Out),T(".%03d"),(Tick*1000)/TICKSPERSEC);
	}
	else
		stprintf(Out,T("%0.2f ms"),(Tick*1000.f)/TICKSPERSEC);
}

int GCD(int a,int b)
{
	int c;
	while (b)
	{
		c = b;
		b = a % b;
		a = c;
	}
	return a;
}

void Simplify( fraction* f, int MaxNum, int MaxDen )
{
	int Den = abs(f->Den);
	int Num = abs(f->Num);

	if ((int64_t)Num*MaxDen < (int64_t)Den*MaxNum)
	{
		if (Den > MaxDen)
		{
			f->Num = Scale(f->Num,MaxDen,Den);
			f->Den = Scale(f->Den,MaxDen,Den);
		}
	}
	else
	{
		if (Num > MaxNum)
		{
			f->Num = Scale(f->Num,MaxNum,Num);
			f->Den = Scale(f->Den,MaxNum,Num);
		}
	}
}

bool_t EqAudio( const audio* a, const audio* b )
{
	return a->Bits == b->Bits &&
		   a->Channels == b->Channels &&
		   a->SampleRate == b->SampleRate &&
		   a->FracBits == b->FracBits &&
		   a->Flags == b->Flags &&
		   a->Format == b->Format;
}

int UpperFourCC(int FourCC)
{
	return (toupper((FourCC >> 0) & 255) << 0) |
		   (toupper((FourCC >> 8) & 255) << 8) |
		   (toupper((FourCC >> 16) & 255) << 16) |
		   (toupper((FourCC >> 24) & 255) << 24);
}

int IDCTCopy( idct* Src, idct* Dst, int MinCount )
{
	int Count;
	blitfx FX;
	video SrcFormat,DstFormat;
	planes SrcPlanes,DstPlanes;
	int SrcBrightness,DstBrightness;
	int No;
	int Result;

	if (!Dst)
	{
		Src->Set(Src,IDCT_OUTPUT|PACKET_FORMAT,NULL,0);
		return ERR_NONE;
	}

	Result = Src->Get(Src,IDCT_OUTPUT|PACKET_FORMAT,&SrcFormat,sizeof(SrcFormat));
	if (Result != ERR_NONE)	return Result;

	Result = Src->Get(Src,IDCT_BUFFERCOUNT,&Count,sizeof(Count));
	if (Result != ERR_NONE)	return Result;

	Result = Dst->Set(Dst,IDCT_OUTPUT|PACKET_FORMAT,&SrcFormat,sizeof(SrcFormat));
	if (Result != ERR_NONE)	return Result;

	Result = Dst->Set(Dst,IDCT_BUFFERCOUNT,&MinCount,sizeof(MinCount));
	if (Result != ERR_NONE)	return Result;

	if (Count > MinCount)
		Dst->Set(Dst,IDCT_BUFFERCOUNT,&Count,sizeof(Count));

	memset(&FX,0,sizeof(FX));
	FX.ScaleX = SCALE_ONE;
	FX.ScaleY = SCALE_ONE;

	for (No=0;No<Count;++No)
	{
		int FrameNo = -1;
		Src->Get(Src,IDCT_FRAMENO+No,&FrameNo,sizeof(FrameNo));
		Dst->Set(Dst,IDCT_FRAMENO+No,&FrameNo,sizeof(FrameNo));

		if (Src->Lock(Src,No,SrcPlanes,&SrcBrightness,&SrcFormat) == ERR_NONE)
		{
			if (Dst->Lock(Dst,No,DstPlanes,&DstBrightness,&DstFormat) == ERR_NONE)
			{
				FX.Direction = CombineDir( SrcFormat.Direction, 0, DstFormat.Direction );
				FX.Brightness = DstBrightness - SrcBrightness;
				
				SurfaceCopy(&SrcFormat,&DstFormat,SrcPlanes,DstPlanes,&FX);

				Dst->Unlock(Dst,No);
			}
			Src->Unlock(Src,No);
		}
	}

	if (Src->Get(Src,IDCT_SHOW,&No,sizeof(No))==ERR_NONE)
		Dst->Set(Dst,IDCT_SHOW,&No,sizeof(No));

	Src->Set(Src,IDCT_OUTPUT|PACKET_FORMAT,NULL,0);
	return ERR_NONE;
}

bool_t EqFrac(const fraction* a, const fraction* b )
{
	if (a->Den == b->Den && a->Num == b->Num) 
		return 1;
	return (int64_t)b->Den * a->Num == (int64_t)a->Den * b->Num;
}

bool_t EqBlitFX( const blitfx* a, const blitfx* b )
{
	return a->Flags == b->Flags &&
		   a->Contrast == b->Contrast &&
		   a->Saturation == b->Saturation &&
		   a->Brightness == b->Brightness &&
		   a->Direction == b->Direction &&
		   a->ScaleX == b->ScaleX &&
		   a->ScaleY == b->ScaleY;
}

void FillColor(uint8_t* Dst,int DstPitch,int x,int y,int Width,int Height,int BPP,int Value)
{
	if (Width>0 && Height>0)
	{
		uint16_t *p16,*p16e;
		uint32_t *p32,*p32e;
		uint8_t* End;

		Dst += y*DstPitch + (x*BPP)/8;
		End = Dst + Height * DstPitch;
		do
		{
			switch (BPP)
			{
			case 1:
				Value &= 1;
				memset(Dst,Value * 0xFF,Width >> 3);
				break;
			case 2:
				Value &= 3;
				memset(Dst,Value * 0x55,Width >> 2);
				break;
			case 4:
				Value &= 15;
				memset(Dst,Value * 0x11,Width >> 1);
				break;
			case 8:
				memset(Dst,Value,Width);
				break;
			case 16:
				p16 = (uint16_t*)Dst;
				p16e = p16+Width;
				for (;p16!=p16e;++p16)
					*p16 = (uint16_t)Value;
				break;
			case 32:
				p32 = (uint32_t*)Dst;
				p32e = p32+Width;
				for (;p32!=p32e;++p32)
					*p32 = Value;
				break;
			}
			Dst += DstPitch;
		}
		while (Dst != End);
	}
}

int SurfaceCopy(const video* SrcFormat, const video* DstFormat, 
				const planes Src, planes Dst, blitfx* FX )
{
	void* Blit;
	rect SrcRect;
	rect DstRect;

	VirtToPhy(NULL,&SrcRect,SrcFormat);
	VirtToPhy(NULL,&DstRect,DstFormat);

	Blit = BlitCreate(DstFormat,SrcFormat,FX,NULL);
	if (!Blit)
		return ERR_NOT_SUPPORTED;

	BlitAlign(Blit,&DstRect,&SrcRect);
	BlitImage(Blit,Dst,Src,NULL,-1,-1);
	BlitRelease(Blit);

	return ERR_NONE;
}

int SurfaceAlloc( planes Ptr, const video* p )
{
	int i;
	for (i=0;i<MAXPLANES;++i)
		Ptr[i] = NULL;

	if (p->Pixel.Flags & (PF_YUV420|PF_YUV422|PF_YUV444))
	{
		int x,y,s;
		PlanarYUV(&p->Pixel,&x,&y,&s);
		Ptr[0] = Alloc16(p->Height * p->Pitch);
		Ptr[1] = Alloc16((p->Height>>y) * (p->Pitch>>s));
		Ptr[2] = Alloc16((p->Height>>y) * (p->Pitch>>s));
		if (!Ptr[0] || !Ptr[1] || !Ptr[2])
		{
			SurfaceFree(Ptr);
			return ERR_OUT_OF_MEMORY;
		}
		return ERR_NONE;
	}
	Ptr[0] = Alloc16(GetImageSize(p));
	return Ptr[0] ? ERR_NONE : ERR_OUT_OF_MEMORY;
}

void SurfaceFree( planes p )
{
	int i;
	for (i=0;i<MAXPLANES;++i)
	{
		Free16(p[i]);
		p[i] = NULL;
	}
}

void CombineVideo(video* Acc, video* New )
{
	if (!Acc->Width && !Acc->Height)
	{
		Acc->Width = New->Width;
		Acc->Height = New->Height;
		Acc->Direction = New->Direction;
	}
	if (!Acc->Aspect)
		Acc->Aspect = New->Aspect;
	if (!Acc->ByteRate)
		Acc->ByteRate = New->ByteRate;
	if (!Acc->FrameRate.Num)
		Acc->FrameRate = New->FrameRate;
}

void CombineAudio(audio* Acc, audio* New )
{
	if (!Acc->Channels)
		Acc->Channels = New->Channels;
	if (!Acc->SampleRate)
		Acc->SampleRate = New->SampleRate;
	if (!Acc->ByteRate)
		Acc->ByteRate = New->ByteRate;
	if (!Acc->Bits)
	{
		Acc->Bits = New->Bits;
		Acc->FracBits = New->FracBits;
	}
}

int CombineDir( int Src, int Blit, int Dst )
{
	//order of transformations 
	//  SrcMirror
	//  SrcSwap
	//  BlitSwap
	//  BlitMirror
	//  DstSwap
	//  DstMirror

	//should be combined to a single
	//  Swap
	//  Mirror

	if (Dst & DIR_SWAPXY)
	{
		if (Blit & DIR_MIRRORLEFTRIGHT)
			Dst ^= DIR_MIRRORUPDOWN;
		if (Blit & DIR_MIRRORUPDOWN)
			Dst ^= DIR_MIRRORLEFTRIGHT;
	}
	else
		Dst ^= Blit & (DIR_MIRRORUPDOWN|DIR_MIRRORLEFTRIGHT);
	Dst ^= Blit & DIR_SWAPXY;
	Dst ^= Src & DIR_SWAPXY;
	if (Dst & DIR_SWAPXY)
	{
		if (Src & DIR_MIRRORLEFTRIGHT)
			Dst ^= DIR_MIRRORUPDOWN;
		if (Src & DIR_MIRRORUPDOWN)
			Dst ^= DIR_MIRRORLEFTRIGHT;
	}
	else
		Dst ^= Src & (DIR_MIRRORUPDOWN|DIR_MIRRORLEFTRIGHT);

	return Dst;
}

int BitMaskSize( uint32_t Mask )
{
	int i;
	for (i=0;Mask;++i)
		Mask &= Mask - 1;
	return i;
}

int BitMaskPos( uint32_t Mask )
{
	int i;
	for (i=0;Mask && !(Mask&1);++i) 
		Mask >>= 1;
	return i;
}

void DefaultPitch( video* p )
{
	p->Pitch = p->Width*GetBPP(&p->Pixel);
	if (p->Pixel.Flags & PF_RGB) 
		p->Pitch = ((p->Pitch+31)>>5)*4; // dword align
	else 
		p->Pitch = (p->Pitch+7)>>3; // byte align
}

void DefaultRGB( pixel* p, int BitCount, 
		         int RBits, int GBits, int BBits,
				 int RGaps, int GGaps, int BGaps )
{
	p->Flags = PF_RGB;
	p->BitCount = BitCount;
	p->RBitMask = ((1<<RBits)-1) << (RGaps+GBits+GGaps+BBits+BGaps);
	p->GBitMask = ((1<<GBits)-1) << (GGaps+BBits+BGaps);
	p->BBitMask = ((1<<BBits)-1) << (BGaps);
}

bool_t Compressed( const pixel* Fmt )
{
	return (Fmt->Flags & PF_FOURCC) && !AnyYUV(Fmt);
}

bool_t PlanarYUV( const pixel* Fmt, int* x, int* y,int *s)
{
	if (PlanarYUV420(Fmt))
	{
		if (x) *x = 1;
		if (y) *y = 1;
		if (s)
		{
			if (Fmt->Flags & PF_FOURCC &&
				((Fmt->FourCC == FOURCC_IMC2) || (Fmt->FourCC == FOURCC_IMC4)))
				*s = 0; // interleaved uv scanlines
			else
				*s = 1;
		}
		return 1;

	}
	if (PlanarYUV422(Fmt))
	{
		if (x) *x = 1;
		if (s) *s = 1;
		if (y) *y = 0;
		return 1;

	}
	if (PlanarYUV444(Fmt))
	{
		if (x) *x = 0;
		if (y) *y = 0;
		if (s) *s = 0;
		return 1;
	}

	if (x) *x = 0;
	if (y) *y = 0;
	if (s) *s = 0;
	return 0;
}

bool_t PlanarYUV420( const pixel* Fmt )
{
	if (Fmt->Flags & PF_YUV420)
		return 1;

	return (Fmt->Flags & PF_FOURCC) && 
	  	  ((Fmt->FourCC == FOURCC_YV12) ||
		   (Fmt->FourCC == FOURCC_IYUV) ||
		   (Fmt->FourCC == FOURCC_I420) ||
		   (Fmt->FourCC == FOURCC_IMC2) ||
		   (Fmt->FourCC == FOURCC_IMC4));
}

bool_t PlanarYUV422( const pixel* Fmt )
{
	if (Fmt->Flags & PF_YUV422)
		return 1;

	return (Fmt->Flags & PF_FOURCC) && (Fmt->FourCC == FOURCC_YV16);
}

bool_t PlanarYUV444( const pixel* Fmt )
{
	return (Fmt->Flags & PF_YUV444) != 0;
}

bool_t PackedYUV( const pixel* Fmt )
{
	return (Fmt->Flags & PF_FOURCC) && 
		  ((Fmt->FourCC == FOURCC_YUY2) ||
		   (Fmt->FourCC == FOURCC_YUNV) ||
		   (Fmt->FourCC == FOURCC_V422) ||
  		   (Fmt->FourCC == FOURCC_YUYV) ||
		   (Fmt->FourCC == FOURCC_VYUY) ||
		   (Fmt->FourCC == FOURCC_UYVY) ||
		   (Fmt->FourCC == FOURCC_Y422) ||
		   (Fmt->FourCC == FOURCC_UYNV));
}

bool_t AnyYUV( const pixel* Fmt )
{
	return PlanarYUV420(Fmt) || 
		PlanarYUV422(Fmt) || 
		PlanarYUV444(Fmt) || 
		PackedYUV(Fmt);
}

uint32_t RGBToFormat( uint32_t RGB, const pixel* Fmt )
{
	uint32_t v;
	int R,G,B;
	int Y,U,V;
	int Pos[3];

	R = (RGB >> 0) & 255;
	G = (RGB >> 8) & 255;
	B = (RGB >> 16) & 255;

	if (AnyYUV(Fmt))
	{
		Y = ( (2105 * R) + (4128 * G) + ( 802 * B))/0x2000 + 16;
		V = ( (3596 * R) - (3015 * G) - ( 582 * B))/0x2000 + 128;
		U = (-(1212 * R) - (2384 * G) + (3596 * B))/0x2000 + 128;

		if (Fmt->Flags & PF_INVERTED)
		{
			Y ^= 255;
			U ^= 255;
			V ^= 255;
		}
		v =  (Fmt->RBitMask / 255) * Y;
		v += (Fmt->GBitMask / 255) * U;
		v += (Fmt->BBitMask / 255) * V;
	}
	else
	{
		if (Fmt->Flags & PF_INVERTED)
		{
			R ^= 255;
			G ^= 255;
			B ^= 255;
		}

		Pos[0] = BitMaskPos(Fmt->RBitMask) + BitMaskSize(Fmt->RBitMask);
		Pos[1] = BitMaskPos(Fmt->GBitMask) + BitMaskSize(Fmt->GBitMask);
		Pos[2] = BitMaskPos(Fmt->BBitMask) + BitMaskSize(Fmt->BBitMask);

		v = ((R << Pos[0]) & (Fmt->RBitMask << 8)) |
			((G << Pos[1]) & (Fmt->GBitMask << 8)) |
			((B << Pos[2]) & (Fmt->BBitMask << 8));
		v >>= 8;
	}

	return v;
}

void FillInfo( pixel* Fmt )
{
	Fmt->BitCount = GetBPP(Fmt);
	if (PlanarYUV(Fmt,NULL,NULL,NULL))
	{
		switch (Fmt->FourCC)
		{
		case FOURCC_IMC4:
		case FOURCC_I420: 
		case FOURCC_IYUV: 
			Fmt->RBitMask = 0x000000FF;
			Fmt->GBitMask = 0x0000FF00;
			Fmt->BBitMask = 0x00FF0000;
			break;
		case FOURCC_IMC2:
		case FOURCC_YV16:
		case FOURCC_YV12: 
			Fmt->RBitMask = 0x000000FF;
			Fmt->GBitMask = 0x00FF0000;
			Fmt->BBitMask = 0x0000FF00;
			break;
		}
	}
	else
	if (PackedYUV(Fmt))
		switch (Fmt->FourCC)
		{
		case FOURCC_YUY2:
		case FOURCC_YUNV:
		case FOURCC_V422:
		case FOURCC_YUYV:
			Fmt->RBitMask = 0x00FF00FF;
			Fmt->GBitMask = 0x0000FF00;
			Fmt->BBitMask = 0xFF000000;
			break;
		case FOURCC_YVYU:
			Fmt->RBitMask = 0x00FF00FF;
			Fmt->GBitMask = 0xFF000000;
			Fmt->BBitMask = 0x0000FF00;
			break;
		case FOURCC_UYVY:
		case FOURCC_Y422:
		case FOURCC_UYNV:
			Fmt->RBitMask = 0xFF00FF00;
			Fmt->GBitMask = 0x000000FF;
			Fmt->BBitMask = 0x00FF0000;
			break;
		}	
}

int GetImageSize( const video* p )
{
	int BPP = p->Pixel.BitCount;
	if (PlanarYUV420(&p->Pixel))
		BPP = 12;
	else
	if (PackedYUV(&p->Pixel) || PlanarYUV422(&p->Pixel))
		BPP = 16;
	else
	if (PlanarYUV444(&p->Pixel))
		BPP = 24;
	return (BPP * p->Pitch * p->Height) >> 3;
}

int GetBPP( const pixel* Fmt)
{
	if (Fmt->Flags & (PF_RGB | PF_PALETTE)) 
		return Fmt->BitCount;
	if (PlanarYUV(Fmt,NULL,NULL,NULL))
		return 8;
	if (PackedYUV(Fmt))
		return 16;
	return 0;
}

bool_t EqPoint( const point* a, const point* b )
{
	return a->x==b->x && a->y==b->y;
}

bool_t EqRect( const rect* a, const rect* b )
{
	return a->x==b->x && a->y==b->y && a->Width==b->Width && a->Height==b->Height;
}

bool_t EqPixel( const pixel* a, const pixel* b )
{
	if (a->Flags != b->Flags) 
		return 0;

	if ((a->Flags & PF_PALETTE) &&
		a->BitCount != b->BitCount)
		return 0;

	if ((a->Flags & PF_RGB) &&
		(a->BitCount != b->BitCount ||
		 a->RBitMask != b->RBitMask ||
		 a->GBitMask != b->GBitMask ||
		 a->BBitMask != b->BBitMask))
		return 0;

	if ((a->Flags & PF_FOURCC) && a->FourCC != b->FourCC)
		return 0;

	return 1;
}

bool_t EqVideo( const video* a, const video* b )
{
	return a->Width == b->Width &&
		   a->Height == b->Height &&
		   a->Pitch == b->Pitch &&
		   EqPixel(&a->Pixel,&b->Pixel);
}

void VirtToPhy(const rect* Virtual, rect* Physical, const video* p )
{
	if (Virtual)
	{
		*Physical = *Virtual;

		if (p->Pixel.Flags & PF_PIXELDOUBLE)
		{
			Physical->x >>= 1;
			Physical->y >>= 1;
			Physical->Width >>= 1;
			Physical->Height >>= 1;
		}

		if (p->Direction & DIR_SWAPXY)
			SwapRect(Physical);

		if (p->Direction & DIR_MIRRORLEFTRIGHT)
			Physical->x = p->Width - Physical->x - Physical->Width;

		if (p->Direction & DIR_MIRRORUPDOWN)
			Physical->y = p->Height - Physical->y - Physical->Height;

		// clipping
		if (Physical->x < 0)
		{
			Physical->Width += Physical->x;
			Physical->x = 0;
		}
		if (Physical->y < 0)
		{
			Physical->Height += Physical->y;
			Physical->y = 0;
		}
		if (Physical->x + Physical->Width > p->Width)
		{
			Physical->Width = p->Width - Physical->x;
			if (Physical->Width < 0)
			{
				Physical->Width = 0;
				Physical->x = 0;
			}
		}
		if (Physical->y + Physical->Height > p->Height)
		{
			Physical->Height = p->Height - Physical->y;
			if (Physical->Height < 0)
			{
				Physical->Height = 0;
				Physical->y = 0;
			}
		}
	}
	else
	{
		Physical->x = 0;
		Physical->y = 0;
		Physical->Width = p->Width;
		Physical->Height = p->Height;
	}
}

void PhyToVirt(const rect* Physical, rect* Virtual, const video* p )
{
	if (Physical)
		*Virtual = *Physical;
	else
	{
		Virtual->x = 0;
		Virtual->y = 0;
		Virtual->Width = p->Width;
		Virtual->Height = p->Height;
	}

	if (p->Direction & DIR_MIRRORLEFTRIGHT)
		Virtual->x = p->Width - Virtual->x - Virtual->Width;

	if (p->Direction & DIR_MIRRORUPDOWN)
		Virtual->y = p->Height - Virtual->y - Virtual->Height;

	if (p->Direction & DIR_SWAPXY)
		SwapRect(Virtual);

	if (p->Pixel.Flags & PF_PIXELDOUBLE)
	{
		Virtual->x <<= 1;
		Virtual->y <<= 1;
		Virtual->Width <<= 1;
		Virtual->Height <<= 1;
	}
}

void SwapPByte( uint8_t** a, uint8_t** b)
{
	uint8_t* t = *a;
	*a = *b;
	*b = t;
}

void SwapPChar( tchar_t** a, tchar_t** b)
{
	tchar_t* t = *a;
	*a = *b;
	*b = t;
}

void SwapInt( int* a, int* b)
{
	int t = *a;
	*a = *b;
	*b = t;
}

void SwapLong( long* a, long* b)
{
	long t = *a;
	*a = *b;
	*b = t;
}

void SwapBool( bool_t* a, bool_t* b)
{
	bool_t t = *a;
	*a = *b;
	*b = t;
}

void SwapPoint(point* p)
{
	int t = p->x;
	p->x = p->y;
	p->y = t;
}

void SwapRect(rect* r)
{
	int t1 = r->x;
	int t2 = r->Width;
	r->x = r->y;
	r->Width = r->Height;
	r->y = t1;
	r->Height = t2;
}

void NodeFill(node* Node, int Size, nodeenum Enum, nodeget Get, nodeset Set)
{
	memset(Node,0,Size);
	Node->Enum = Enum;
	Node->Get = Get;
	Node->Set = Set;
}

int NodeEnumType( int* EnumNo, datadef* Out, ... )
{
	const datadef* p;
	const datadef** List = (const datadef**)&Out+1;
	int Class = 0;

	for (;*List;++List)
	{
		for (p=*List;p->No>0;++p)
		{
			if (p->Type == TYPE_CLASS)
				Class = p->No;

			if (--*EnumNo<0)
			{
				*Out = *p;
				switch (Out->Type)
				{
				case TYPE_BOOL: Out->Size = sizeof(bool_t); break;
				case TYPE_INT: Out->Size = sizeof(int); break;
				case TYPE_HOTKEY: Out->Size = sizeof(int); break;
				case TYPE_FRACTION: Out->Size = sizeof(fraction); break;
				case TYPE_STRING: Out->Size = MAXDATA; break;
				case TYPE_RECT: Out->Size = sizeof(rect); break;
				case TYPE_POINT: Out->Size = sizeof(point); break;
				case TYPE_RGB: Out->Size = sizeof(rgb); break;
				case TYPE_EQUALIZER: Out->Size = sizeof(equalizer); break;
				case TYPE_NODE: Out->Size = sizeof(node*); break;
				case TYPE_PIN: Out->Size = sizeof(pin); break;
				case TYPE_PACKET: Out->Size = sizeof(packet); break;
				case TYPE_TICK: Out->Size = sizeof(tick_t); break;
				case TYPE_BLITFX: Out->Size = sizeof(bool_t); break;
				default:
					Out->Size = 0;
					break;
				}
				Out->Class = Class;
				if (!NodeInsideIsClass())
					Out->Name = String(Class,Out->No);
				else
					Out->Name = T("");
				return ERR_NONE;
			}
		}
	}
	return ERR_INVALID_PARAM;
}

int NodeSetPin( pin* Dst, const pin* Src, void* Format, int FormatLength )
{
	if (Src->Node)
	{	
		int Result = Src->Node->Set(Src->Node,Src->No | PACKET_FORMAT,Format,FormatLength);
		if (Result != ERR_NONE)
			return Result;
	}
	*Dst = *Src;
	return ERR_NONE;
}

void MergeInit(merge* p)
{
	p->Data = NULL;
	p->Allocated = 0;
	p->ReadPos = 0;
	p->WritePos = 0;
}

void MergeDone(merge* p)
{
	Free(p->Data);
	p->Allocated = 0;
	p->ReadPos = 0;
	p->WritePos = 0;
	p->Data = NULL;
}

bool_t MergeAlloc( merge* p, int Size, int Align )
{
	uint8_t* Data;

	Align--;
	Size = (Size + SAFETAIL + Align) & ~Align;
	Data = ReAlloc(p->Data,Size);

	if (Data)
	{
		p->Data = Data;
		p->Allocated = Size - SAFETAIL;
		return 1;
	}

	p->WritePos = 0;
	p->ReadPos = 0;
	return 0;
}

void MergeClear(merge* p)
{
	//drop previous packets
	p->WritePos = 0;
	p->ReadPos = 0;
}

bool_t MergeStream(merge* p,stream* Stream)
{
	MergePack(p,0);

	if (p->Allocated > p->WritePos)
	{
		int n = Stream->Read(Stream,p->Data,p->Allocated - p->WritePos);
		if (n > 0)
			p->WritePos += n;
	}
	return p->WritePos > p->ReadPos;
}

void MergeWrite(merge* p, const uint8_t* Ptr, int Length)
{
	//append new data to buffer
	int WritePos = p->WritePos + Length;
	if (WritePos > p->Allocated && !MergeAlloc(p,WritePos,16384))
		return;
	memcpy(p->Data+p->WritePos,Ptr,Length);
	p->WritePos = WritePos;
}

bool_t MergeRead(merge* p, const uint8_t** Ptr, int Length )
{
	if (p->WritePos - p->ReadPos < Length)
		return 0;

	*Ptr = p->Data + p->ReadPos;
	p->ReadPos += Length;
	return 1;
}

void MergePack(merge* p, int Length )
{
	int ReadPos = p->ReadPos;
	ReadPos += Length;
	if (p->WritePos > ReadPos)
	{
		// move end part to the beginning
		p->WritePos = p->WritePos - ReadPos;
		if (ReadPos)
			memmove(p->Data,p->Data + ReadPos,p->WritePos);
	}
	else
		p->WritePos = 0;
	p->ReadPos = 0;
}

void SplitURL( const tchar_t* URL, tchar_t* Mime, tchar_t* Dir, tchar_t* Name, tchar_t* Ext )
{
	const tchar_t* p;

	// mime 

	if (!Dir || Mime)
		URL = GetMime(URL,Mime);

	// dir

	p = tcsrchr(URL,'\\');
	if (!p)
		p = tcsrchr(URL,'/');
	if (p)
	{
		if (Dir)
			TcsNCpy(Dir,URL,p-URL+1);
		URL = p+1;
	}
	else
	if (Dir)
		Dir[0] = 0;

	// name

	if (Name && Name == Ext)
		tcscpy(Name,URL);
	else
	{
		p = tcsrchr(URL,'.');
		if (p)
		{
			if (Name)
				TcsNCpy(Name,URL,p-URL+1);
			if (Ext)
				tcscpy(Ext,p+1);
		}
		else
		{
			if (Name)
				tcscpy(Name,URL);
			if (Ext)
				Ext[0] = 0;
		}
	}
}

void RelPath(tchar_t* Rel, const tchar_t* Any, const tchar_t* Base )
{
	unsigned int n = tcslen(Base);
	tcscpy(Rel,Any);
	if (n<tcslen(Rel) && (Rel[n]=='\\' || Rel[n]=='/'))
	{
		Rel[n] = 0;
		if (tcsicmp(Rel,Base)==0)
			Any += n+1;
		tcscpy(Rel,Any);
	}
}

void AbsPath(tchar_t* Abs, const tchar_t* Any, const tchar_t* Base )
{
	if (GetMime(Any,NULL)==Any && Any[0] != '/' && Any[0] != '\\' &&
		!(Any[0] && Any[1]==':' && Any[2]=='\\'))
	{	
		tcscpy(Abs,Base);
		tcscat(Abs,T("\\"));
		tcscat(Abs,Any);
	}
	else
		tcscpy(Abs,Any);
}

int CheckExts( const tchar_t* URL, const tchar_t* Exts )
{
	tchar_t Ext[MAXPATH];

	SplitURL(URL,NULL,NULL,NULL,Ext);

	while (Exts)
	{
		const tchar_t* p = tcschr(Exts,':');
		if (p)
		{
			tchar_t s[MAXPATH];
			TcsNCpy(s,Exts,p-Exts+1);
			if (tcsicmp(Ext,s)==0)
				return p[1]; // return type char
			p = tcschr(p,';');
			if (p) ++p;
		}
		Exts = p;
	}
	return 0;
}

void UpperUntil( tchar_t* s, tchar_t Delimiter )
{
	tchar_t Save = 0;
	tchar_t* p = tcschr(s,Delimiter);
	if (p)
	{
		Save = *p;
		*p = 0;
	}
	tcsupr(s);
	if (p)
		*p = Save;
}

bool_t ReadLine( merge* Buffer, stream* Stream, tchar_t* Out, int Len )
{
	bool_t Result = 0;
	char s[MAXLINE];
	char ch;
	int i=0;

	while (Buffer->WritePos>Buffer->ReadPos || MergeStream(Buffer,Stream))
	{
		ch = Buffer->Data[Buffer->ReadPos++];
		if (ch==10)
		{
			Result = 1;
			break;
		}
		else
		if (ch!=13 && i<sizeof(s)-1)
		{
			s[i++] = ch;
			Result = 1;
		}
	}

	s[i]=0;
	StrToTcs(Out,s,Len/sizeof(tchar_t));

	return Result;
}

void* Alloc16(int n)
{
	char* p = (char*) Alloc(n+sizeof(void*)+16);
	if (p)
	{
		int i;
		char* p0 = p;
		p += sizeof(void*);
		i = (int)p & 15;
		if (i) p += 16-i;
		((void**)p)[-1] = p0;
	}
	return p;
}

void Free16(void* p)
{
	if (p)
		Free(((void**)p)[-1]);
}
