/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: dmo.c 11 2004-07-15 13:46:20Z picard $
 *
 * BetaPlayer Core
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "../stdafx.h"

#define _WIN32_WINNT 0x0400
#define INITGUID
#include <objbase.h>

//#define static 

#if defined(_WIN32_WCE)
#define ACCESS_LOAD		0
#define ACCESS_SAVE		0 
#else
#define ACCESS_LOAD		KEY_READ
#define ACCESS_SAVE		KEY_READ | KEY_WRITE | KEY_SET_VALUE 
#endif

#if defined(_WIN32_WCE)
#define TWIN(a) L ## a
#else
#define TWIN(a) a
#endif

#define DMO_INPUT_STATUSF_ACCEPT_DATA				0x00000001

#define DMO_INPUT_DATA_BUFFERF_SYNCPOINT			0x00000001
#define DMO_INPUT_DATA_BUFFERF_TIME					0x00000002
#define DMO_INPUT_DATA_BUFFERF_TIMELENGTH			0x00000004

#define DMO_INPUT_STREAMF_WHOLE_SAMPLES				0x00000001
#define DMO_INPUT_STREAMF_SINGLE_SAMPLE_PER_BUFFER	0x00000002
#define DMO_INPUT_STREAMF_FIXED_SAMPLE_SIZE			0x00000004
#define DMO_INPUT_STREAMF_HOLDS_BUFFERS				0x00000008

#define DMO_SET_TYPEF_TEST_ONLY						0x00000001
#define DMO_SET_TYPEF_CLEAR							0x00000002

#define DMO_PROCESS_OUTPUT_DISCARD_WHEN_NO_BUFFER	0x00000001

#define	DMO_OUTPUT_DATA_BUFFERF_SYNCPOINT			0x00000001
#define	DMO_OUTPUT_DATA_BUFFERF_TIME				0x00000002
#define	DMO_OUTPUT_DATA_BUFFERF_TIMELENGTH			0x00000004
#define	DMO_OUTPUT_DATA_BUFFERF_INCOMPLETE			0x01000000

DEFINE_GUID(ID_IMEDIABUFFER,0x59eff8b9,0x938c,0x4a26,0x82,0xf2,0x95,0xcb,0x84,0xcd,0xc8,0x37);
DEFINE_GUID(ID_IUNKNOWN,0x00000000,0x0000,0x0000,0xC0,0x00,0x00,0x00,0x00,0x00,0x00,0x46);
DEFINE_GUID(ID_ICLASSFACTORY,00000001,0x0000,0x0000,0xC0,0x00,0x00,0x00,0x00,0x00,0x00,0x46);

DEFINE_GUID(ID_IDMOQUALITYCONTROL,0x65abea96,0xcf36,0x453f,0xaf,0x8a,0x70,0x5e,0x98,0xf1,0x62,0x60);

EXTERN_GUID(WMCMEDIATYPE_Video,0x73646976, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71);
EXTERN_GUID(WMCMEDIATYPE_Audio,0x73647561, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71);

EXTERN_GUID(WMCFORMAT_VideoInfo,0x05589f80, 0xc356, 0x11ce, 0xbf, 0x01, 0x00, 0xaa, 0x00, 0x55, 0x59, 0x5a); 
EXTERN_GUID(WMCFORMAT_WaveFormatEx,0x05589f81, 0xc356, 0x11ce, 0xbf, 0x01, 0x00, 0xaa, 0x00, 0x55, 0x59, 0x5a); 

EXTERN_GUID(WMCMEDIASUBTYPE, 0, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xAA, 0x00, 0x38, 0x9B, 0x71);

const IID IID_IMediaObject = { 0xd8ad0f58, 0x5494, 0x4102, 0x97, 0xc5, 0xec, 0x79, 0x8e, 0x59, 0xbc, 0xf4 };

typedef int64_t REFERENCE_TIME;

struct IMediaObject;
struct IMediaBuffer;
struct IDMOQualityControl;

typedef struct VIDEOINFOHEADER 
{
	RECT rcSource;
	RECT rcTarget;
	DWORD dwBitRate;
	DWORD dwBitErrorRate;
	REFERENCE_TIME AvgTimePerFrame;
	BITMAPINFOHEADER bmiHeader;

} VIDEOINFOHEADER;

typedef struct DMO_MEDIA_TYPE
{
    GUID majortype;
    GUID subtype;
    BOOL bFixedSizeSamples;
    BOOL bTemporalCompression;
    ULONG lSampleSize;
    GUID formattype;
    IUnknown* pUnk;
    ULONG cbFormat;
    BYTE* pbFormat;

} DMO_MEDIA_TYPE;

typedef struct DMO_OUTPUT_DATA_BUFFER
{
    struct IMediaBuffer* pBuffer;
    DWORD dwStatus;
    REFERENCE_TIME rtTimestamp;
    REFERENCE_TIME rtTimelength;

} DMO_OUTPUT_DATA_BUFFER;

typedef struct IMediaBufferVMT
{
    HRESULT ( STDCALL *QueryInterface )( void* This, REFIID riid, void **ppvObject);    
    ULONG ( STDCALL *AddRef )( void* This);     
    ULONG ( STDCALL *Release )( void* This);    
    HRESULT ( STDCALL *SetLength )( void* This, DWORD cbLength);    
    HRESULT ( STDCALL *GetMaxLength )( void* This, DWORD *pcbMaxLength);    
    HRESULT ( STDCALL *GetBufferAndLength )( void* This, BYTE **ppBuffer, DWORD *pcbLength);
    
} IMediaBufferVMT;

typedef struct IMediaBuffer
{
    struct IMediaBufferVMT *VMT;

} IMediaBuffer;

typedef struct IDMOQualityControlVMT
{
    HRESULT ( STDCALL *QueryInterface )( struct IDMOQualityControl* This, REFIID riid, void** ppvObject);
    ULONG ( STDCALL *AddRef )( struct IDMOQualityControl* This);
    ULONG ( STDCALL *Release )( struct IDMOQualityControl* This);
    HRESULT ( STDCALL *SetNow)( struct IDMOQualityControl* This, REFERENCE_TIME rtNow );
    HRESULT ( STDCALL *SetStatus)( struct IDMOQualityControl* This, DWORD dwFlags );
    HRESULT ( STDCALL *GetStatus)( struct IDMOQualityControl* This, DWORD *pdwFlags );

} IDMOQualityControlVMT;

typedef struct IDMOQualityControl
{
	IDMOQualityControlVMT *VMT;

} IDMOQualityControl;
 
typedef struct IMediaObjectVMT
{
    HRESULT ( STDCALL *QueryInterface )( struct IMediaObject * This, REFIID riid, void** ppvObject);
    ULONG ( STDCALL *AddRef )( struct IMediaObject * This);
    ULONG ( STDCALL *Release )( struct IMediaObject * This);
    HRESULT ( STDCALL *GetStreamCount )( struct IMediaObject * This, DWORD *pcInputStreams, DWORD *pcOutputStreams);
    HRESULT ( STDCALL *GetInputStreamInfo )( struct IMediaObject * This, DWORD dwInputStreamIndex, DWORD *pdwFlags);
    HRESULT ( STDCALL *GetOutputStreamInfo )( struct IMediaObject * This, DWORD dwOutputStreamIndex, DWORD *pdwFlags);    
	HRESULT ( STDCALL *GetInputType )( struct IMediaObject * This, DWORD dwInputStreamIndex, DWORD dwTypeIndex, DMO_MEDIA_TYPE *pmt);
    HRESULT ( STDCALL *GetOutputType )( struct IMediaObject * This, DWORD dwOutputStreamIndex, DWORD dwTypeIndex, DMO_MEDIA_TYPE *pmt);
    HRESULT ( STDCALL *SetInputType )( struct IMediaObject * This, DWORD dwInputStreamIndex, const DMO_MEDIA_TYPE *pmt, DWORD dwFlags);    
    HRESULT ( STDCALL *SetOutputType )( struct IMediaObject * This, DWORD dwOutputStreamIndex,const DMO_MEDIA_TYPE *pmt, DWORD dwFlags);    
    HRESULT ( STDCALL *GetInputCurrentType )( struct IMediaObject * This, DWORD dwInputStreamIndex, DMO_MEDIA_TYPE *pmt);    
    HRESULT ( STDCALL *GetOutputCurrentType )( struct IMediaObject * This, DWORD dwOutputStreamIndex, DMO_MEDIA_TYPE *pmt);    
    HRESULT ( STDCALL *GetInputSizeInfo )( struct IMediaObject * This, DWORD dwInputStreamIndex, DWORD *pcbSize, DWORD *pcbMaxLookahead, DWORD *pcbAlignment);
    HRESULT ( STDCALL *GetOutputSizeInfo )( struct IMediaObject * This, DWORD dwOutputStreamIndex, DWORD *pcbSize, DWORD *pcbAlignment);     
    HRESULT ( STDCALL *GetInputMaxLatency )( struct IMediaObject * This, DWORD dwInputStreamIndex, REFERENCE_TIME *prtMaxLatency);    
    HRESULT ( STDCALL *SetInputMaxLatency )( struct IMediaObject * This, DWORD dwInputStreamIndex, REFERENCE_TIME rtMaxLatency);    
    HRESULT ( STDCALL *Flush )( struct IMediaObject * This);
    HRESULT ( STDCALL *Discontinuity )( struct IMediaObject * This, DWORD dwInputStreamIndex);
    HRESULT ( STDCALL *AllocateStreamingResources )( struct IMediaObject * This);    
    HRESULT ( STDCALL *FreeStreamingResources )( struct IMediaObject * This);    
    HRESULT ( STDCALL *GetInputStatus )( struct IMediaObject * This,DWORD dwInputStreamIndex,DWORD *dwFlags);    
    HRESULT ( STDCALL *ProcessInput )( struct IMediaObject * This, DWORD dwInputStreamIndex, IMediaBuffer *pBuffer, DWORD dwFlags, REFERENCE_TIME rtTimestamp, REFERENCE_TIME rtTimelength);    
    HRESULT ( STDCALL *ProcessOutput )( struct IMediaObject * This, DWORD dwFlags, DWORD cOutputBufferCount, DMO_OUTPUT_DATA_BUFFER *pOutputBuffers, DWORD *pdwStatus);   
    HRESULT ( STDCALL *Lock )( struct IMediaObject * This, LONG bLock);
    
} IMediaObjectVMT;

typedef struct IMediaObject
{
    IMediaObjectVMT *VMT;

} IMediaObject;

static HRESULT (STDCALL* FuncMoInitMediaType)( DMO_MEDIA_TYPE*, DWORD Size );
static HRESULT (STDCALL* FuncMoFreeMediaType)( DMO_MEDIA_TYPE* );

#define DMO_INPUT		0x100
#define DMO_OUTPUT		0x101

typedef struct IOwnMediaBuffer
{
    struct IMediaBufferVMT *VMT;

	int RefCount;
	int Length;
	int Allocated; // -1 for external allocated buffer
	void* Data;

	struct IOwnMediaBuffer** Chain;
	struct IOwnMediaBuffer* Next;

} IOwnMediaBuffer;

typedef struct dmo
{
	node VMT;
	
	int Id;
	const CLSID *ClassId;
	HMODULE Module;
	int (STDCALL *CreateCodecDMO)(IMediaObject**);
	int (STDCALL *QueryUnloadCodecDMO)();
	int (STDCALL *DllGetClassObject)(REFCLSID,REFIID,IClassFactory**);
	int (STDCALL *DllCanUnloadNow)();

	int FourCCCount;
	int FourCC[32];

	IMediaObject* Media;
	
	IOwnMediaBuffer* InputBuffers;
	DMO_OUTPUT_DATA_BUFFER OutputBuffer[2];
	int OutputNoMask;
	int OutputNext;
	int OutputPrev;

	int WMPVersion;
	int FormatType;
	int FormatSize;
	void* InputFormat;
	void* OutputFormat;
	pin Output;

	DMO_MEDIA_TYPE InType;
	DMO_MEDIA_TYPE OutType;
	int ConstSize;
	BITMAPINFOHEADER* Const;

	int TotalBytes;
	int TotalPackets;
	int DroppedPackets;
	bool_t Pending;
	bool_t First;
	bool_t UseDropping;
	packet Packet;

	struct dmo* Chain;

} dmo;

//****************************************************

HRESULT STDCALL MBQueryInterface( IOwnMediaBuffer * p, REFIID riid, void **ppvObject)
{
	if (memcmp(&riid,&ID_IMEDIABUFFER,sizeof(GUID))==0 ||
		memcmp(&riid,&ID_IUNKNOWN,sizeof(GUID))==0)
	{
		if (ppvObject)
		{
			++p->RefCount;
			*ppvObject = p;
		}
		return 0;
	}
	return -1;
}

ULONG STDCALL MBAddRef( IOwnMediaBuffer * p)
{
	p->RefCount++;
	return 0;
}

HRESULT STDCALL MBSetMaxLength( IOwnMediaBuffer * p, DWORD cbLength)
{
	if (cbLength > 0)
	{
		if ((int)cbLength > p->Allocated)
		{
			p->Data = ReAlloc(p->Data,cbLength);
			p->Allocated = p->Data ? cbLength:0;
		}
	}
	else
	{
		Free(p->Data);
		p->Data = NULL;
		p->Allocated = 0;
	}
	if (p->Length > p->Allocated)
		p->Length = p->Allocated;
	return 0;
}

ULONG STDCALL MBRelease( IOwnMediaBuffer * p)
{
	if (--p->RefCount<=0)
	{
		p->Length = 0;
		if (p->Chain)
		{
			p->Next = *p->Chain;
			*p->Chain = p;
		}
		else
		{
			Free(p->Data);
			Free(p);
			return 0;
		}
	}
	return p->RefCount;
}

HRESULT STDCALL MBSetLength( IOwnMediaBuffer * p, DWORD cbLength)
{
	DEBUG_MSG2(DEBUG_VIDEO,T("MediaBuffer_SetLength %d (%d)"),cbLength,p->Allocated);
	if ((int)cbLength > p->Allocated)
		return -1;
	p->Length = cbLength;
	return 0;
}

HRESULT STDCALL MBGetMaxLength( IOwnMediaBuffer * p, DWORD *pcbMaxLength)
{
	DEBUG_MSG1(DEBUG_VIDEO,T("MediaBuffer_GetMaxLength %d"),p->Allocated);
	if (pcbMaxLength)
		*pcbMaxLength = p->Allocated;
	return 0;
}
HRESULT STDCALL MBGetBufferAndLength( IOwnMediaBuffer * p, BYTE **ppBuffer, DWORD *pcbLength)
{
	DEBUG_MSG1(DEBUG_VIDEO,T("MediaBuffer_GetBufferAndLength %d"),p->Length);
	if (ppBuffer)
		*ppBuffer = (BYTE*)p->Data;
	if (pcbLength)
		*pcbLength = p->Length;
	return 0;
}

static IMediaBufferVMT MediaBufferVMT =
{
    MBQueryInterface,
    MBAddRef,
    MBRelease,
    MBSetLength,
    MBGetMaxLength,
    MBGetBufferAndLength,
};

void FreeBuffers(IOwnMediaBuffer** Chain)
{
	while (*Chain)
	{
		IOwnMediaBuffer* p = (*Chain)->Next;
		MBSetMaxLength(*Chain,0);
		Free(*Chain);
		*Chain = p;
	}
}

static INLINE void AppendBuffer(IOwnMediaBuffer* p, int Length, void* Data )
{
	memcpy((char*)p->Data + p->Length, Data, Length );
	p->Length += Length;
}

IMediaBuffer* AllocBuffer(int MaxLength,IOwnMediaBuffer** Chain)
{
	IOwnMediaBuffer* p = Chain ? *Chain : NULL;

	if (MaxLength <= 0)
		return NULL;

	if (p)
		*Chain = p->Next;
	else
	{
		p = (IOwnMediaBuffer*) CAlloc(sizeof(IOwnMediaBuffer),1);
		if (!p)
			return NULL;
		p->VMT = &MediaBufferVMT;
	}

	p->Chain = Chain;
	p->Length = 0;

	MBSetMaxLength(p,MaxLength);

	p->VMT->AddRef(p);

	return (IMediaBuffer*)p;
}

//****************************************************

bool_t MatchOutVideo( dmo* p ) 
{
	int No;
    for (No=0;;++No)
	{
        if (p->Media->VMT->GetOutputType(p->Media, 0, No, &p->OutType) != S_OK)
			break;

        if (memcmp(&p->OutType.formattype,&WMCFORMAT_VideoInfo,sizeof(GUID))==0 &&
			p->OutType.cbFormat >= sizeof(VIDEOINFOHEADER) && 
			p->OutType.pbFormat)
		{
		    VIDEOINFOHEADER *Info = (VIDEOINFOHEADER*) p->OutType.pbFormat;
			video* Video = (video*) p->OutputFormat;

			Video->Pixel.FourCC = Info->bmiHeader.biCompression;
			Video->Pixel.BitCount = Info->bmiHeader.biBitCount;
			Video->Width = Info->bmiHeader.biWidth;
			Video->Height = Info->bmiHeader.biHeight;
			Video->Aspect = ASPECT_ONE;

			Format_DefaultVideo(Video);

			if (PlanarYUV420(&Video->Pixel))
				return 1;
		}
		
		if (p->OutType.pbFormat) 
			FuncMoFreeMediaType(&p->OutType);
		memset(&p->OutType,0,sizeof(p->OutType));
	}
	return 0;
}

bool_t MatchOutAudio( dmo* p ) 
{
	/* on WM2003 doesn't work this way...
	int No;
    for (No=0;;++No)
	{
        if (p->Media->VMT->GetOutputType(p->Media, 0, No, &p->OutType) != S_OK)
			break;

        if (memcmp(&p->OutType.formattype,&WMCFORMAT_WaveFormatEx,sizeof(GUID))==0 &&
			p->OutType.cbFormat >= sizeof(WAVEFORMATEX) && 
			p->OutType.pbFormat)
		{
		    WAVEFORMATEX *Info = (WAVEFORMATEX*) p->OutType.pbFormat;
			audio* Audio = (audio*) p->OutputFormat;

			Audio->Format = Info->wFormatTag;
			Audio->Channels = Info->nChannels;
			Audio->SampleRate = Info->nSamplesPerSec;
			Audio->ByteRate = Info->nAvgBytesPerSec;
			Audio->BlockAlign = Info->nBlockAlign;
			Audio->Bits = Info->wBitsPerSample;
			
			if (Audio->Format == AUDIOFMT_PCM)
			{
				Audio->FracBits = Audio->Bits - 1;
				if (Audio->Bits <= 8)
					Audio->Flags = PCM_UNSIGNED;
				Audio->Bits = (Audio->Bits + 7) & ~7;
				Audio->BlockAlign = (Audio->Channels * Audio->Bits) >> 3;
			}

			// accept PCM audio
			//if (Audio->Format == AUDIOFMT_PCM && Audio->Channels <= 2)
			//	return 1;
		}
		
		if (p->OutType.pbFormat) 
			FuncMoFreeMediaType(&p->OutType);
		memset(&p->OutType,0,sizeof(p->OutType));
	}
	return 0;
	*/

	// manual audio format guessing

	WAVEFORMATEX *WaveFormat;
	audio* Audio = (audio*) p->OutputFormat;
	memcpy(p->OutputFormat,p->InputFormat,sizeof(audio));

	if (FuncMoInitMediaType(&p->OutType,sizeof(WAVEFORMATEX)) != S_OK)
		return 0;

	WaveFormat = (WAVEFORMATEX*) p->OutType.pbFormat;

	Audio->Format = AUDIOFMT_PCM;
	if (Audio->Channels > 2)
		Audio->Channels = 2;
	if (Audio->Bits > 16)
		Audio->Bits = 16;
	if (Audio->Bits <= 8)
		Audio->Flags = PCM_UNSIGNED;
	Audio->Bits = (Audio->Bits + 7) & ~7;
	Audio->FracBits = Audio->Bits - 1;
	Audio->BlockAlign = (Audio->Channels * Audio->Bits) >> 3;
	Audio->ByteRate = Audio->BlockAlign * Audio->SampleRate;
	Audio->Extra = NULL;
	Audio->ExtraLength = 0;

	WaveFormat->wFormatTag = (WORD)Audio->Format;
	WaveFormat->nChannels = (WORD)Audio->Channels;
	WaveFormat->nSamplesPerSec = Audio->SampleRate;
	WaveFormat->nAvgBytesPerSec = Audio->ByteRate;
	WaveFormat->nBlockAlign = (WORD)Audio->BlockAlign;
	WaveFormat->wBitsPerSample = (WORD)Audio->Bits;
	WaveFormat->cbSize = 0;

	p->OutType.majortype = WMCMEDIATYPE_Audio;
	p->OutType.formattype = WMCFORMAT_WaveFormatEx;
	p->OutType.subtype = WMCMEDIASUBTYPE;
	p->OutType.subtype.Data1 = AUDIOFMT_PCM;
	p->OutType.bTemporalCompression = 0;
	p->OutType.bFixedSizeSamples = 1;
	p->OutType.lSampleSize = 0;

	return 1;
}

static int SetFormat( dmo* p, const void* Format, int Size )
{
	VIDEOINFOHEADER* VideoInfo;
	WAVEFORMATEX* WaveFormat;
	video* Video;
	audio* Audio;
	int Dummy;
	int BufferSize;
	int AutoBufferSize;
	int No,Extra;

	p->Pending = 0;
	p->TotalPackets = 0;
	p->TotalBytes = 0;
	p->DroppedPackets = 0;

	for (No=0;No<p->OutputNoMask+1;++No)
		if (p->OutputBuffer[No].pBuffer)
		{
			p->OutputBuffer[No].pBuffer->VMT->Release(p->OutputBuffer[No].pBuffer);
			p->OutputBuffer[No].pBuffer = NULL;
		}

	p->ConstSize = 0;
	p->Const = NULL;

	Free(p->InputFormat); 
	p->InputFormat = NULL;

	Free(p->OutputFormat); 
	p->OutputFormat = NULL;

 	if (p->InType.pbFormat) FuncMoFreeMediaType(&p->InType);
	memset(&p->InType,0,sizeof(p->InType));

	if (p->OutType.pbFormat) FuncMoFreeMediaType(&p->OutType);
	memset(&p->OutType,0,sizeof(p->OutType));

	if (p->Media)
	{
		p->Media->VMT->FreeStreamingResources(p->Media);
		p->Media->VMT->Release(p->Media);
		p->Media = NULL;
	}

	if (Format)
	{
		switch (p->FormatType)
		{
		case PACKET_VIDEO:

			if (Size != sizeof(video)) return ERR_INVALID_PARAM;

			Video = (video*)Format;
			if (!(Video->Pixel.Flags & PF_FOURCC))
				return ERR_INVALID_PARAM;
	
			for (No=0;No<p->FourCCCount;++No)
				if (p->FourCC[No] == Video->Pixel.FourCC)
					break;

			if (No==p->FourCCCount)
				return ERR_INVALID_PARAM;
			break;

		case PACKET_AUDIO:

			if (Size != sizeof(audio)) return ERR_INVALID_PARAM;
			Audio = (audio*)Format;

			for (No=0;No<p->FourCCCount;++No)
				if (p->FourCC[No] == Audio->Format)
					break;

			if (No==p->FourCCCount)
				return ERR_INVALID_PARAM;
			break;
		}

		if (p->CreateCodecDMO)
		{
			p->CreateCodecDMO(&p->Media);
		}
		else
		if (p->DllGetClassObject)
		{
			IClassFactory* Class = NULL;
			if (p->DllGetClassObject(p->ClassId,&ID_ICLASSFACTORY,&Class) == S_OK)
			{
				Class->lpVtbl->CreateInstance(Class,NULL,&IID_IMediaObject,&p->Media);
				Class->lpVtbl->Release(Class);
			}
		}
		if (!p->Media)
			return ERR_INVALID_PARAM;

		switch (p->FormatType)
		{
		case PACKET_VIDEO:

			Video = (video*)Format;
			p->FormatSize = sizeof(video);
			p->InputFormat = CAlloc(sizeof(video),1);
			p->OutputFormat = CAlloc(sizeof(video),1);

			// wmp10 wm video not supported yet
			if (p->WMPVersion > 9 && (
				Video->Pixel.FourCC == FOURCC('W','M','V','1') || 
				Video->Pixel.FourCC == FOURCC('W','M','V','2') || 
				Video->Pixel.FourCC == FOURCC('W','M','V','3')))
				return ERR_INVALID_PARAM;

			Extra = Video->ExtraLength;
			if ((Video->Pixel.FourCC == FOURCC('W','M','V','3') || 
				 Video->Pixel.FourCC == FOURCC('W','M','V','2')) && Extra>4)
				Extra = 4; //hope this is right...

			if (!p->InputFormat || !p->OutputFormat)
				return ERR_OUT_OF_MEMORY;

			if (FuncMoInitMediaType(&p->InType,sizeof(VIDEOINFOHEADER)+Extra) != S_OK)
				return ERR_OUT_OF_MEMORY;
			
			memcpy(p->InputFormat,Format,sizeof(video));

			VideoInfo = (VIDEOINFOHEADER*)p->InType.pbFormat;
			memset(VideoInfo,0,sizeof(VIDEOINFOHEADER));
			VideoInfo->bmiHeader.biSize = sizeof(VideoInfo->bmiHeader)+Extra;
			VideoInfo->bmiHeader.biWidth = Video->Width;
			VideoInfo->bmiHeader.biHeight = Video->Height;
			VideoInfo->bmiHeader.biCompression = Video->Pixel.FourCC;
			VideoInfo->bmiHeader.biBitCount = (WORD)Video->Pixel.BitCount;
			VideoInfo->bmiHeader.biPlanes = 1;
			VideoInfo->bmiHeader.biSizeImage = (Video->Width * Video->Height * Video->Pixel.BitCount)/8;
			if (p->WMPVersion > 9)
				VideoInfo->bmiHeader.biSizeImage = 0;
			if (Extra)
				memcpy(VideoInfo+1,Video->Extra,Extra);

			VideoInfo->rcSource.left   = 0;
			VideoInfo->rcSource.top    = 0;
			VideoInfo->rcSource.right  = Video->Width;
			VideoInfo->rcSource.bottom = Video->Height;
			VideoInfo->rcTarget        = VideoInfo->rcSource;

			if (Video->FrameRate.Num)
			{
				VideoInfo->AvgTimePerFrame = (REFERENCE_TIME)Video->FrameRate.Den * 10000000 / Video->FrameRate.Num;
				VideoInfo->dwBitRate = Scale(VideoInfo->bmiHeader.biSizeImage*8,Video->FrameRate.Num,Video->FrameRate.Den);
			}

			p->InType.majortype = WMCMEDIATYPE_Video;
			p->InType.formattype = WMCFORMAT_VideoInfo;
			p->InType.subtype = WMCMEDIASUBTYPE;
			p->InType.subtype.Data1 = Video->Pixel.FourCC;

			if (p->Media->VMT->SetInputType(p->Media, 0, &p->InType, 0) != S_OK)
				return ERR_INVALID_DATA;

			if (!MatchOutVideo(p))
			    return ERR_INVALID_DATA;
    
			AutoBufferSize = GetImageSize((video*)p->OutputFormat);

			if (p->WMPVersion)
			{
				AutoBufferSize += VideoInfo->bmiHeader.biSize;
				p->ConstSize = VideoInfo->bmiHeader.biSize;
				p->Const = &VideoInfo->bmiHeader;
			}

			p->OutputNoMask = 1;
			p->UseDropping = 1;
			break;

		case PACKET_AUDIO:

			Audio = (audio*)Format;
			p->FormatSize = sizeof(audio);
			p->InputFormat = CAlloc(sizeof(audio),1);
			p->OutputFormat = CAlloc(sizeof(audio),1);

			if (!p->InputFormat || !p->OutputFormat)
				return ERR_OUT_OF_MEMORY;

			if (FuncMoInitMediaType(&p->InType,sizeof(WAVEFORMATEX)+Audio->ExtraLength) != S_OK)
				return ERR_OUT_OF_MEMORY;

			memcpy(p->InputFormat,Format,sizeof(audio));

			WaveFormat = (WAVEFORMATEX*)p->InType.pbFormat;
			memset(WaveFormat,0,sizeof(WAVEFORMATEX));

			WaveFormat->wFormatTag = (WORD)Audio->Format;
			WaveFormat->nChannels = (WORD)Audio->Channels;
			WaveFormat->nSamplesPerSec = Audio->SampleRate;
			WaveFormat->nAvgBytesPerSec = Audio->ByteRate;
			WaveFormat->nBlockAlign = (WORD)Audio->BlockAlign;
			WaveFormat->wBitsPerSample = (WORD)Audio->Bits;
			WaveFormat->cbSize = (WORD)Audio->ExtraLength;
			if (Audio->ExtraLength)
			{
				if (p->WMPVersion == 9 && WaveFormat->wFormatTag == AUDIOFMT_WMA9 && Audio->ExtraLength==10)
				{
					char* p = (char*)(WaveFormat+1);
					p[0] = p[1] = 0;
					p[8] = p[9] = 0;
					memcpy(p+2,Audio->Extra,6);
				}
				else
					memcpy(WaveFormat+1,Audio->Extra,Audio->ExtraLength);
			}

			p->InType.majortype = WMCMEDIATYPE_Audio;
			p->InType.formattype = WMCFORMAT_WaveFormatEx;
			p->InType.subtype = WMCMEDIASUBTYPE;
			p->InType.subtype.Data1 = Audio->Format;
			p->InType.bTemporalCompression = 0;
			p->InType.bFixedSizeSamples = 1;
			p->InType.lSampleSize = 0;

			if (p->Media->VMT->SetInputType(p->Media, 0, &p->InType, 0) != S_OK)
				return ERR_INVALID_DATA;

			if (!MatchOutAudio(p))
			    return ERR_INVALID_DATA;
   
			// it seems pocketpc can't split output data
			if (WaveFormat->nBlockAlign && WaveFormat->nAvgBytesPerSec)
			{
				AutoBufferSize = Scale(((audio*)p->OutputFormat)->ByteRate,WaveFormat->nBlockAlign,WaveFormat->nAvgBytesPerSec);
				AutoBufferSize *= 4; // just to be safe
				if (AutoBufferSize < 128*1024)
					AutoBufferSize = 128*1024;
			}
			else
				AutoBufferSize = 128*1024;

			p->OutputNoMask = 0;
			p->UseDropping = 0;
			break;

		default:
			return ERR_INVALID_DATA;
		}

		if (p->Media->VMT->SetOutputType(p->Media,0, &p->OutType, 0) != S_OK)
			return ERR_INVALID_DATA;

		BufferSize = 0;
		if (p->Media->VMT->GetOutputSizeInfo(p->Media, 0, (DWORD*)&BufferSize, (DWORD*)&Dummy) != S_OK)
			return ERR_INVALID_DATA;

		if (BufferSize<=0)
			BufferSize = AutoBufferSize;
		else
		if (p->WMPVersion>9)
			BufferSize += 4096;

		for (No=0;No<p->OutputNoMask+1;++No)
		{
			p->OutputBuffer[No].pBuffer = AllocBuffer(BufferSize,NULL);
			if (!p->OutputBuffer[No].pBuffer)
				return ERR_OUT_OF_MEMORY;
		}

		p->First = p->UseDropping;
		p->OutputNext = 0;
		p->OutputPrev = -1;
		p->Media->VMT->AllocateStreamingResources(p->Media);
	}

	FreeBuffers(&p->InputBuffers);
	return ERR_NONE;
}

static int Resend(dmo* p)
{
	int Result = ERR_INVALID_DATA;
	if (p->OutputPrev >= 0 && p->OutputPrev != p->OutputNext)
	{
		IOwnMediaBuffer* Buffer = (IOwnMediaBuffer*) p->OutputBuffer[p->OutputPrev].pBuffer;
		packet Packet;

		memset(&Packet,0,sizeof(Packet));
		Packet.RefTime = -1;
		Packet.CurrTime = -1;

		if (p->ConstSize)
		{
			BITMAPINFOHEADER* Info = Buffer->Data;
			Packet.Length = Buffer->Length - Info->biSize;
			Packet.Data[0] = (char*)Buffer->Data + Info->biSize;
		}
		else
		{
			Packet.Length = Buffer->Length;
			Packet.Data[0] = Buffer->Data;
		}
		
		if (Packet.Length > 0)
			Result = p->Output.Node->Set(p->Output.Node,p->Output.No,&Packet,sizeof(packet));
	}
	return Result;
}

static bool_t DecodePacket( dmo* p )
{
	DWORD Status;
	IOwnMediaBuffer* Buffer;
	IOwnMediaBuffer* LastBuffer;

	LastBuffer = NULL;
	if (p->OutputPrev >= 0 && p->OutputPrev != p->OutputNext)
		LastBuffer = (IOwnMediaBuffer*) p->OutputBuffer[p->OutputPrev].pBuffer;

	if (p->Packet.DropLevel > 0)
	{
		DMO_OUTPUT_DATA_BUFFER Discard;
		Discard.dwStatus = 0;
		Discard.pBuffer = NULL;

		p->Media->VMT->ProcessOutput(p->Media, DMO_PROCESS_OUTPUT_DISCARD_WHEN_NO_BUFFER, 1, 
			&Discard, &Status);

		//DebugMessage(T("DMO%d CurrTime:%d Drop"),p->FormatType,p->Packet.CurrTime);
	}

	for (;;)
	{
		Buffer = (IOwnMediaBuffer*) p->OutputBuffer[p->OutputNext].pBuffer;
		Buffer->Length = 0;

		p->OutputBuffer[p->OutputNext].dwStatus = 0;
		if (p->Media->VMT->ProcessOutput(p->Media, 0, 1, &p->OutputBuffer[p->OutputNext], &Status)!=S_OK || Buffer->Length==0)
			break;

		if (p->Packet.DropLevel > 0)
		{
			continue;
		}

		if (p->OutputBuffer[p->OutputNext].dwStatus & DMO_OUTPUT_DATA_BUFFERF_TIME)
			p->Packet.RefTime = (tick_t)(p->OutputBuffer[p->OutputNext].rtTimestamp * TICKSPERSEC / 10000000);
		else
			p->Packet.RefTime = -1;

		if (p->ConstSize)
		{
			BITMAPINFOHEADER* Info = Buffer->Data;
			p->Packet.Length = Buffer->Length - Info->biSize;
			p->Packet.Data[0] = (char*)Buffer->Data + Info->biSize;
			p->Packet.LastData[0] = NULL;
			if (LastBuffer)
				p->Packet.LastData[0] = (char*)LastBuffer->Data + Info->biSize;
			if (p->Packet.Length <= 0)
				continue; // empty frame (i guess)
		}
		else
		{
			p->Packet.Length = Buffer->Length;
			p->Packet.Data[0] = Buffer->Data;
			p->Packet.LastData[0] = NULL;
			if (LastBuffer)
				p->Packet.LastData[0] = LastBuffer->Data;
		}

		//DebugMessage(T("DMO%d CurrTime:%d RefTime:%d Diff:%d"),p->FormatType,p->Packet.CurrTime,p->Packet.RefTime,p->Packet.CurrTime-p->Packet.RefTime); 

		DEBUG_MSG4(DEBUG_VIDEO,T("DMO%d Output Time:%d Length:%d LastData:%d"),
			p->FormatType,
			p->Packet.RefTime,p->Packet.Length,p->Packet.LastData[0] != NULL);

		p->First = 0;
		p->TotalBytes += p->Packet.Length;
		return 1;
	}
	return 0;
}

static int Decode( dmo* p, const packet* Packet )
{
	IMediaBuffer* Buffer;
	REFERENCE_TIME TimeStamp;
	DWORD Status;
	int Result = ERR_NONE;

	p->Packet.CurrTime = Packet->CurrTime;
	p->Packet.DropLevel = p->First ? 0:Packet->DropLevel;

	if (p->Pending)
	{
		do
		{
			if (p->Output.Node->Set(p->Output.Node,p->Output.No,
					&p->Packet,sizeof(packet)) == ERR_BUFFER_FULL)
				return ERR_BUFFER_FULL;

			p->OutputPrev = p->OutputNext;
			p->OutputNext = (p->OutputNext+1) & p->OutputNoMask;
		}
		while (DecodePacket(p));

		p->Pending = 0;
	}

	if (!Packet->Length)
		return ERR_NEED_MORE_DATA;

	p->TotalPackets++;

	if (Packet->DropLevel > 0)
	{
		++p->DroppedPackets;

		if (Packet->DropLevel > 1)
		{
			DEBUG_MSG3(DEBUG_VIDEO,T("DMO%d DropLevel2 Curr:%d Time:%d"),p->FormatType,Packet->CurrTime,Packet->RefTime);
			p->Media->VMT->Discontinuity(p->Media,0);
			p->First = p->UseDropping;
			return ERR_NONE;
		}
	}

    if (p->Media->VMT->GetInputStatus(p->Media, 0, &Status)==0 &&
		!(Status & DMO_INPUT_STATUSF_ACCEPT_DATA)) // this shouldn't happen
        p->Media->VMT->Discontinuity(p->Media,0);

	Status = 0;
	if (Packet->Key)
		Status |= DMO_INPUT_DATA_BUFFERF_SYNCPOINT;
	else
	if (p->First)
		return ERR_NEED_MORE_DATA;

	TimeStamp = 0;
	if (Packet->RefTime >= 0)
	{
		Status |= DMO_INPUT_DATA_BUFFERF_TIME;
		TimeStamp = ((REFERENCE_TIME)Packet->RefTime * 10000000) / TICKSPERSEC;
	}

	DEBUG_MSG3(DEBUG_VIDEO,T("DMO%d Input Time:%d Size:%d"),p->FormatType,Packet->RefTime,Packet->Length);

	Buffer = AllocBuffer(Packet->Length+p->ConstSize,&p->InputBuffers);
	if (Buffer)
	{
		if (p->Const)
		{
			if (p->WMPVersion>9) p->Const->biSizeImage = Packet->Length;
			AppendBuffer((IOwnMediaBuffer*)Buffer,p->ConstSize,p->Const);
		}

		if (Packet->Data[0])
			AppendBuffer((IOwnMediaBuffer*)Buffer,Packet->Length,Packet->Data[0]);

		p->Media->VMT->ProcessInput(p->Media,0,Buffer,Status,TimeStamp,0);
		Buffer->VMT->Release(Buffer);
	}

	if (DecodePacket(p))
	{
		Result = ERR_NONE;
		do
		{
			int i = p->Output.Node->Set(p->Output.Node,p->Output.No,
						 				 &p->Packet,sizeof(packet));
			if (i == ERR_BUFFER_FULL)
			{
				p->Pending = 1;
				//return last result
				break;
			}

			p->OutputPrev = p->OutputNext;
			p->OutputNext = (p->OutputNext+1) & p->OutputNoMask;
			Result = i;
		}
		while (DecodePacket(p));
	}
	else
		Result = ERR_NEED_MORE_DATA; // if no frame decoded: need more packet

	return Result;
}

static bool_t IsEmpty( dmo* p )
{
	packet Zero;

  	if (!p->InputFormat) 
		return 1;

    p->Media->VMT->Discontinuity(p->Media,0);

	Zero.CurrTime = -1;
	Zero.DropLevel = p->Packet.DropLevel;
	Zero.Length = 0;
	Zero.Data[0] = NULL;
	Zero.RefTime = -1;

	return Decode(p,&Zero) == ERR_NEED_MORE_DATA;
}

const datadef DMOParams[] = 
{
	{ DMO_CLASS,	TYPE_CLASS },
	{ DMO_INPUT,	TYPE_PACKET, DF_INPUT, -1 },
	{ DMO_OUTPUT,	TYPE_PIN, DF_OUTPUT, -1 },

	DATADEF_END
};

static int Enum( dmo* p, int No, datadef* Param )
{
	int Result = NodeEnumType(&No,Param,NodeParams,FlowParams,DMOParams,NULL);

	if (Result != ERR_NONE && p->FormatType == PACKET_VIDEO)
		Result = NodeEnumType(&No,Param,FlowBufferParams,NULL);

	if (Result == ERR_NONE && Param->Class == DMO_CLASS && 
		(Param->No == DMO_INPUT || Param->No == DMO_OUTPUT))
		Param->Format1 = p->FormatType;

	return Result;
}

static int GetTotal( dmo* p )
{
	if (p->FormatType == PACKET_VIDEO)
		return p->TotalPackets;

	if (p->FormatType == PACKET_AUDIO && p->OutputFormat)
	{
		int BlockAlign = ((audio*)p->OutputFormat)->BlockAlign;
		if (BlockAlign)
			return p->TotalBytes / BlockAlign;
	}
	return 0;
}

static int GetDropped( dmo* p )
{
	if (p->FormatType == PACKET_VIDEO)
		return p->DroppedPackets;
	return 0;
}

static int Get( dmo* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(p->Id,int); break;
	case DMO_INPUT | PACKET_FORMAT: 
		if (Size == p->FormatSize)
		{
			if (p->InputFormat)
				memcpy(Data,p->InputFormat,p->FormatSize);
			else
				memset(Data,0,p->FormatSize);
			Result = ERR_NONE;
		}
		break;
	case DMO_OUTPUT | PACKET_FORMAT:
		if (Size == p->FormatSize)
		{
			if (p->OutputFormat)
				memcpy(Data,p->OutputFormat,p->FormatSize);
			else
				memset(Data,0,p->FormatSize);
			Result = ERR_NONE;
		}
		break;
	case DMO_OUTPUT: GETVALUE(p->Output,pin); break;
	case FLOW_TOTAL:GETVALUE(GetTotal(p),int); break;
	case FLOW_DROPPED:GETVALUE(GetDropped(p),int); break;
	case FLOW_EMPTY:GETVALUE(IsEmpty(p),bool_t); break;
	}
	return Result;
}

static int Set( dmo* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case DMO_INPUT | PACKET_FORMAT:
		if (!Size) Data = NULL;
		Result = SetFormat(p,(const audio*)Data,Size);
		break;

	case DMO_INPUT:
		if (Size == sizeof(packet))
			Result = Decode(p,(const packet*)Data);
		break;

	case DMO_OUTPUT:
		if (Size == sizeof(pin))
			Result = NodeSetPin(&p->Output,(pin*)Data,p->OutputFormat,p->FormatSize);
		break;

	case FLOWBUFFER_RESEND:
		Result = Resend(p);
		break;

	case FLOW_TOTAL: 
		p->TotalPackets = 0;
		p->TotalBytes = 0;
		break;

	case FLOW_DROPPED: 
		p->DroppedPackets = 0;
		break;

	case FLOW_EMPTY:
		p->OutputPrev = -1;
		p->Media->VMT->Discontinuity(p->Media,0);
		p->Media->VMT->Flush(p->Media);
		p->Pending = 0;
		p->First = p->UseDropping;
		break;
	}
	return Result;
}

static dmo* Chain = NULL;

bool_t DMO_Load( const CLSID* ClassId,const tchar_t* FileNamePPC,const tchar_t* FileNameCOM, 
			     int Id,int FormatType, const tchar_t* Name, ... )
{
	dmo* p;
	const int* List = (const int*)&Name+1;

	if (!FuncMoInitMediaType || !FuncMoFreeMediaType)
		return 0;

	p = (dmo*) CAlloc(sizeof(dmo),1);
	p->Id = Id;
	p->ClassId = ClassId;
	p->FormatType = FormatType;
	for (;*List;++List)
		p->FourCC[p->FourCCCount++] = *List;

	if (FileNamePPC)
		p->Module = LoadLibrary(FileNamePPC);
	if (!p->Module && FileNameCOM)
		p->Module = LoadLibrary(FileNameCOM);

	if (p->Module)
	{
		*(FARPROC*)&p->CreateCodecDMO = GetProcAddress(p->Module,TWIN("CreateCodecDMO"));
		*(FARPROC*)&p->QueryUnloadCodecDMO = GetProcAddress(p->Module,TWIN("QueryUnloadCodecDMO"));
		*(FARPROC*)&p->DllGetClassObject = GetProcAddress(p->Module,TWIN("DllGetClassObject"));
		*(FARPROC*)&p->DllCanUnloadNow = GetProcAddress(p->Module,TWIN("DllCanUnloadNow"));
	}

	if (!p->CreateCodecDMO && !p->DllGetClassObject)
	{
		if (p->Module)
			FreeLibrary(p->Module);
		Free(p);
		return 0;
	}

#if defined(_WIN32_WCE)
	p->WMPVersion = 9;
	{
		HKEY Key;
		if (RegOpenKeyEx( HKEY_LOCAL_MACHINE, T("SOFTWARE\\Microsoft\\MediaPlayer\\ASFCodecs"), 0, ACCESS_LOAD, &Key ) == ERROR_SUCCESS)
		{
			p->WMPVersion = 10;
			RegCloseKey( Key );
		}
	}
#else
	p->WMPVersion = 0;
#endif


	p->VMT.Enum = Enum;
	p->VMT.Get = Get;
	p->VMT.Set = Set;
	p->Chain = Chain;
	Chain = p;

	StringAdd(LANG_DEFAULT,Id,0,Name);
	NodeRegister(&p->VMT,PRI_DEFAULT-100); // we prefer native codecs

	return 1;
}

static HMODULE MSDMODLL = NULL;

void DMO_Init()
{
	CoInitializeEx(NULL,COINIT_MULTITHREADED);

	MSDMODLL = LoadLibrary(T("msdmo.dll"));
	if (!MSDMODLL)
		return;

	*(FARPROC*)&FuncMoInitMediaType = GetProcAddress(MSDMODLL,TWIN("MoInitMediaType"));
	*(FARPROC*)&FuncMoFreeMediaType = GetProcAddress(MSDMODLL,TWIN("MoFreeMediaType"));
}

void DMO_Done()
{
	dmo* i = Chain;
	while (i)
	{
		dmo* j=i->Chain;
		NodeUnRegister(&i->VMT);
		SetFormat(i,NULL,0);

		if (i->Module && 
			(!i->QueryUnloadCodecDMO || i->QueryUnloadCodecDMO()) &&
			(!i->DllCanUnloadNow || i->DllCanUnloadNow()==S_OK))
			FreeLibrary(i->Module);

		Free(i);
		i=j;
	}
	Chain = NULL;

	if (MSDMODLL)
	{
		FreeLibrary(MSDMODLL);
		MSDMODLL = NULL;
	}
	CoUninitialize();
}

