/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: http_win32.c 145 2004-12-12 18:18:28Z picard $
 *
 * BetaPlayer Core
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "../stdafx.h"

#if defined(_WIN32)

#ifndef STRICT
#define STRICT
#endif
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

typedef HANDLE HINTERNET;
#define HTTP_QUERY_CONTENT_TYPE			1
#define HTTP_QUERY_CONTENT_LENGTH		5
#define HTTP_QUERY_STATUS_TEXT			20
#define HTTP_QUERY_RAW_HEADERS          21
#define HTTP_QUERY_RAW_HEADERS_CRLF     22
#define INTERNET_OPEN_TYPE_DIRECT		1
#define INTERNET_FLAG_NO_CACHE_WRITE    0x04000000

typedef struct http
{
	stream VMT;

	tchar_t URL[MAXPATH];
	HINTERNET Internet;
	HINTERNET Handle;
	int Length;
	int Pos;
	bool_t Silent;
	HMODULE DLL;
	tchar_t ContentType[MAXPATH];

	BOOL (WINAPI *InternetReadFile)(HINTERNET hFile,LPVOID lpBuffer,DWORD dwNumberOfBytesToRead,LPDWORD lpdwNumberOfBytesRead);
	HINTERNET (WINAPI *InternetOpen)(LPCTSTR lpszAgent, DWORD dwAccessType, LPCTSTR lpszProxy , LPCTSTR lpszProxyBypass , DWORD dwFlags);
	HINTERNET (WINAPI *InternetOpenUrl)( HINTERNET hInternetSession, LPCTSTR lpszUrl, LPCTSTR lpszHeaders , DWORD dwHeadersLength, DWORD dwFlags, DWORD dwContext);
	BOOL (WINAPI *InternetCloseHandle)(HINTERNET hInternet);
	BOOL (WINAPI *HttpQueryInfo)(HINTERNET hRequest,DWORD dwInfoLevel, LPVOID lpBuffer , LPDWORD lpdwBufferLength, LPDWORD lpdwIndex );

} http;

static int EnumType( http* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,StreamParams,NULL);
}

static int Get( http* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(HTTP_ID,int); break;
	case STREAM_MIME: GETSTRING(T("http")); break;
	case STREAM_URL: GETSTRING(p->URL); break;
	case STREAM_SILENT: GETVALUE(p->Silent,bool_t); break;
	case STREAM_LENGTH: GETVALUECOND(p->Length,int,p->Length>=0); break;
	case STREAM_CONTENTTYPE: GETSTRING(p->ContentType); break;
	}
	return Result;
}

static int Open( http* p, const tchar_t* URL, bool_t ReOpen )
{
	if (p->Handle)
	{
		p->InternetCloseHandle(p->Handle);
		p->Handle = 0;
	}
	if (p->Internet)
	{
		p->InternetCloseHandle(p->Internet);
		p->Internet = 0;
	}
	p->ContentType[0] = 0;
	p->Length = -1;
	if (!ReOpen)
		p->URL[0] = 0;
	else
		Sleep(200);

	if (URL && URL[0])
	{
		DWORD Length;
		DWORD n;

		p->Internet = p->InternetOpen(T("BetaPlayer"),INTERNET_OPEN_TYPE_DIRECT,NULL,NULL,0);

		p->Handle = p->InternetOpenUrl(p->Internet, URL, NULL, (DWORD)-1, INTERNET_FLAG_NO_CACHE_WRITE, 1);

		if (!p->Handle)
		{
			if (!ReOpen && !p->Silent)
				NodeError(NULL,ERR_ID,ERR_CONNECT_FAILED,URL);
			return ERR_FILE_NOT_FOUND;
		}

		TcsNCpy(p->URL,URL,MAXPATH);

		n = sizeof(p->ContentType);
		if (!p->HttpQueryInfo(p->Handle, HTTP_QUERY_CONTENT_TYPE , p->ContentType, &n, NULL))
			p->ContentType[0] = 0;

		n = sizeof(Length);
		if (p->HttpQueryInfo(p->Handle, HTTP_QUERY_CONTENT_LENGTH , &Length, &n, NULL))
			p->Length = Length;
		p->Pos = 0;
	}
	else
		p->Length = -1;
	return ERR_NONE;
}

static int Set( http* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case STREAM_SILENT: SETVALUE(p->Silent,bool_t,ERR_NONE); break;
	case STREAM_URL:
		Result = Open(p,(const tchar_t*)Data,0);
		break;
	}
	return Result;
}

static int Read(http* p,void* Data,int Size)
{
	DWORD Readed;
	DWORD Error;

	if (p->InternetReadFile(p->Handle,Data,Size,&Readed))
	{
		p->Pos += Readed;
		return Readed;
	}

	Error = GetLastError();

	if (Error == ERROR_NOT_CONNECTED || Error == ERROR_INVALID_HANDLE)
		Open(p,p->URL,1);

	return -1;
}

static int Seek(http* p,int Pos,int SeekMode)
{
	return ERR_NOT_SUPPORTED;
}

static int EnumDir(http* p,const tchar_t* URL,const tchar_t* Exts,bool_t ExtFilter,streamdir* Item)
{
	return ERR_NOT_DIRECTORY;
}

static void* Create( http* p )
{
	http* i = (http*) Alloc(sizeof(http));
	*i = *p;
	return i;
}

static void Delete( http* p )
{
	Open(p,NULL,0);
	Free(p);
}

static http HTTP = 
{
	EnumType,
	Get,
	Set,
	Create,
	Delete,

	Read,
	Seek,
	EnumDir,
	NULL,
};

void HTTP_Init()
{
	http* p = &HTTP;

	p->DLL = LoadLibrary(T("wininet.dll"));
	GetProc(&p->DLL,&p->InternetReadFile,T("InternetReadFile"),0);
	GetProc(&p->DLL,&p->InternetCloseHandle,T("InternetCloseHandle"),0);
#ifdef UNICODE
	GetProc(&p->DLL,&p->InternetOpen,T("InternetOpenW"),0);
	GetProc(&p->DLL,&p->InternetOpenUrl,T("InternetOpenUrlW"),0);
	GetProc(&p->DLL,&p->HttpQueryInfo,T("HttpQueryInfoW"),0);
#else
	GetProc(&p->DLL,&p->InternetOpen,T("InternetOpenA"),0);
	GetProc(&p->DLL,&p->InternetOpenUrl,T("InternetOpenUrlA"),0);
	GetProc(&p->DLL,&p->HttpQueryInfo,T("HttpQueryInfoA"),0);
#endif

	if (p->DLL)
		NodeRegister((node*)p,PRI_MINIMUM);
}

void HTTP_Done()
{
	http* p = &HTTP;
	if (p->DLL)
	{
		NodeUnRegister((node*)p);
		FreeLibrary(p->DLL);
	}
}

#endif
