/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: platform_win32.c 145 2004-12-12 18:18:28Z picard $
 *
 * BetaPlayer Core
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "../stdafx.h"

#if defined(_WIN32)

int DebugMask = DEBUG_SYS|DEBUG_FORMAT|DEBUG_PLAYER;//|DEBUG_VIDEO;//|DEBUG_AUDIO|DEBUG_VIDEO|DEBUG_PLAYER;//|DEBUG_VIDEO;//DEBUG_TEST;//|DEBUG_FORMAT|DEBUG_PLAYER;

// before including <windows.h> because it can override STDCALL
extern int STDCALL CheckARM5E();
extern int STDCALL CheckARMXScale();
extern void STDCALL GetCpuId(int,uint32_t*);
extern int64_t STDCALL CPUSpeedClk(int);

#if !defined(MIPS) && !defined(ARM) && !defined(_M_IX86)
void STDCALL GetCpuId(int Id,uint32_t* Val) {}
#if !defined(SH3)
int64_t STDCALL CPUSpeedClk(int v) { return 0; }
#endif
#endif

#ifndef STRICT
#define STRICT
#endif
#include <windows.h>

#if defined(_MSC_VER) && !defined(_WIN32_WCE)
#include <crtdbg.h>
#endif

bool_t ShowMemoryLow = 1;

static const datadef PlatformParams[] = 
{
	{ PLATFORM_ID,			TYPE_CLASS },
	{ PLATFORM_LANG,		TYPE_INT, DF_SETUP|DF_ENUMSTRING|DF_RESTART, LANG_ID },
	{ PLATFORM_TYPE,		TYPE_STRING, DF_SETUP|DF_RDONLY|DF_GAP },
	{ PLATFORM_VER,			TYPE_INT, DF_SETUP|DF_RDONLY|DF_HIDDEN },
	{ PLATFORM_VERSION,		TYPE_STRING, DF_SETUP|DF_RDONLY },
	{ PLATFORM_OEMINFO,		TYPE_STRING, DF_SETUP|DF_RDONLY },
	{ PLATFORM_TYPENO,		TYPE_INT, DF_SETUP|DF_RDONLY|DF_HIDDEN },
	{ PLATFORM_MODEL,		TYPE_INT, DF_SETUP|DF_RDONLY|DF_HIDDEN },
	{ PLATFORM_CAPS,		TYPE_INT, DF_SETUP|DF_RDONLY|DF_HIDDEN|DF_HEX },
	{ PLATFORM_CPU,			TYPE_STRING, DF_SETUP|DF_RDONLY },
	{ PLATFORM_MHZ,			TYPE_INT, DF_SETUP|DF_RDONLY|DF_MHZ },
	{ PLATFORM_ICACHE,		TYPE_INT, DF_SETUP|DF_RDONLY|DF_HIDDEN },
	{ PLATFORM_DCACHE,		TYPE_INT, DF_SETUP|DF_RDONLY|DF_HIDDEN },
	DATADEF_END
};

typedef struct platform
{
	node VMT;
	int Model;
	int Caps;
	int ICache;
	int DCache;
	int Type;
	int Ver;
	tchar_t OemInfo[256];
	tchar_t PlatformType[256];
	tchar_t CPU[64];
	int64_t Tick0;

} platform;

#if defined(_WIN32_WCE)

#ifndef DISP_CHANGE_SUCCESSFUL
#define DISP_CHANGE_SUCCESSFUL 0
#endif
#ifndef CDS_TEST
#define CDS_TEST            0x00000002
#endif
#ifndef DM_DISPLAYORIENTATION
#define DM_DISPLAYORIENTATION 0x00800000L
#endif
#ifndef DM_DISPLAYQUERYORIENTATION 
#define DM_DISPLAYQUERYORIENTATION 0x01000000L
#endif
#ifndef DMDO_0
#define DMDO_0      0
#endif
#ifndef DMDO_90
#define DMDO_90     1
#endif
#ifndef DMDO_180
#define DMDO_180    2
#endif
#ifndef DMDO_270
#define DMDO_270    4
#endif

#define SPI_GETOEMINFO				258
#define SPI_GETPLATFORMTYPE			257
#define CACHE_SYNC_INSTRUCTIONS		2

static HANDLE (WINAPI* FuncSetPowerRequirement)(PVOID,int,ULONG,PVOID,ULONG) = NULL;
static DWORD (WINAPI* FuncReleasePowerRequirement)(HANDLE) = NULL;
static BOOL (WINAPI* FuncSHGetDocumentsFolder)(LPCTSTR,LPTSTR) = NULL;
static BOOL (WINAPI* FuncSHGetSpecialFolderPath)(HWND,LPTSTR,int,BOOL) = NULL;
static LONG (WINAPI* FuncChangeDisplaySettingsEx)(LPCTSTR,LPDEVMODE,HWND,DWORD,LPVOID) = NULL;
static void (WINAPI* FuncCacheSync)(DWORD) = NULL;
static BOOL (WINAPI* FuncVirtualCopy)(LPVOID, LPVOID, DWORD, DWORD) = NULL;
static BOOL (WINAPI* FuncCeSetThreadQuantum)(HANDLE, DWORD ) = NULL;
static BOOL (WINAPI* FuncSetKMode)(BOOL) = NULL;
static BOOL (WINAPI* FuncRedrawWindow)(HWND,CONST RECT*,HRGN,UINT) = NULL;
static void (WINAPI* FuncSHIdleTimerReset)() = NULL;

static HMODULE CoreDLL = NULL;
static HMODULE CEShellDLL = NULL;

#define ACCESS_LOAD		0
#define ACCESS_SAVE		0 

#define MAXPHY		8

typedef struct phymem
{
	char* Virt;
	uint32_t Phy;
	uint32_t Length;
	uint32_t RefCount;

} phymem;

static phymem PhyMem[MAXPHY] = { 0 };

#else

#define ACCESS_LOAD		KEY_READ
#define ACCESS_SAVE		KEY_READ | KEY_WRITE | KEY_SET_VALUE 

#endif

static HMODULE AygShellDLL = NULL;
static tchar_t DocPath[MAXPATH] = T("");
static tchar_t SysPath[MAXPATH] = T("");

static const tchar_t RegBacklight[] = T("ControlPanel\\Backlight");
static const tchar_t RegBatteryTimeout[] = T("BatteryTimeout");
static const tchar_t RegACTimeout[] = T("ACTimeout");
static const tchar_t RegScreenSaver[] = T("ControlPanel\\ScreenSaver");
static const tchar_t RegMode[] = T("Mode");
static const tchar_t RegPower[] = T("ControlPanel\\Power");
static const tchar_t RegDisplay[] = T("Display");

static const tchar_t RegCASIOBacklight[] = T("Drivers\\CASIO\\BackLight");
static const tchar_t RegTimeoutBattery[] = T("TimeoutBattery");
static const tchar_t RegTimeoutExPower[] = T("TimeoutExPower");

#define GETVFRAMEPHYSICAL			6144
#define GETVFRAMELEN				6145
#define DBGDRIVERSTAT				6146
#define SETPOWERMANAGEMENT			6147
#define GETPOWERMANAGEMENT			6148

typedef struct VIDEO_POWER_MANAGEMENT {
    ULONG Length;
    ULONG DPMSVersion;
    ULONG PowerState;
} VIDEO_POWER_MANAGEMENT;

static rgb Palette[256];

static bool_t WaitCursor = 0;
static bool_t DisplayPower = 1;
static int Orientation = -1;
static void* PhyFrameBuffer = NULL;

#if defined( MIPS)
int GetTime()
{
	SYSTEMTIME t;
	GetSystemTime(&t);
	return (t.wDay*24+t.wHour)*60*60*1000+(t.wMinute*60+t.wSecond)*1000+t.wMilliseconds;
}
void GetTimeCycle(int* p)
{
	SYSTEMTIME i,j;
	int n=0;
	GetSystemTime(&i);
	do
	{
		++n;
		GetSystemTime(&j);
	}
	while (j.wMilliseconds==i.wMilliseconds);
	p[0] = (j.wDay*24+j.wHour)*60*60*1000+(j.wMinute*60+j.wSecond)*1000+j.wMilliseconds;
	p[1] = n;
}
#elif defined(_WIN32_WCE)
int GetTime()
{
	return GetTickCount();
}
void GetTimeCycle(int* p)
{
	int n=1;
	int j;
	int i = GetTickCount();
	while ((j = GetTickCount())==i)
		++n;
	p[0] = j;
	p[1] = n;
}
#else
int GetTime()
{
	return timeGetTime();
}
void GetTimeCycle(int* p)
{
	int n=1;
	int j;
	int i = timeGetTime();
	while ((j = timeGetTime())==i)
		++n;
	p[0] = j;
	p[1] = n;
}
#endif

static platform Platform;

static int Enum( platform* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,PlatformParams,NULL);
}

int CPUSpeed()
{
	//int64_t Freq;
	int64_t Clk,Best = 0;
	int BestCount = 0;
	int n,Count;
	int Speed = 0;
	HANDLE Thread = GetCurrentThread();
	int Priority = GetThreadPriority(Thread);
	SetThreadPriority(Thread,THREAD_PRIORITY_TIME_CRITICAL);
#if defined(_WIN32_WCE)
	if (FuncCeSetThreadQuantum)
		FuncCeSetThreadQuantum(Thread,250);
#endif

	Count = 800;
	{
		int Sub=1,SubTime=8;
		int c0[2];
		int c1[2];
		int t;

		for (n=0;n<16;++n)
		{
			GetTimeCycle(c0);
			GetTimeCycle(c1);
			if (c1[1]*SubTime > Sub*(c1[0]-c0[0]))
			{
				Sub = c1[1];
				SubTime = c1[0]-c0[0];
			}
			if (n>6 && SubTime >= 4)
				break;
		}

		if (SubTime < 4)
		{
			// Sub = approx number of possbile GetTime() calls per SubTime ms

			for (n=0;n<64;++n)
			{
				CPUSpeedClk(1); // load instruction cache
				GetTimeCycle(c0);
				Clk = CPUSpeedClk(Count);
				GetTimeCycle(c1);

				t = c1[0]-c0[0];
#if !defined(_WIN32_WCE)
				// we are assuming a 1ms timer interrupt with 
				// minimum 6500 cycles interrupt handler
				Clk += t*6500;
#else
				if (Platform.Type == TYPE_SMARTPHONE)
					Clk += t*2000;
#endif
				t *= 100000; 
				t -=(c1[1] * 100000 * SubTime)/Sub; // adjust with subtime

				if (t < 25000) 
					Count *= 2;
				else
				if (t < 50000) // at least 0.5ms
					Count += Count/8;
				else
				{
#ifdef ARM
					Clk = (Clk *  99600000) / t;
#else
					Clk = (Clk * 100000000) / t;
#endif
					if (Best < Clk)
					{
						Best = Clk;
						BestCount = 1;
					}
					else
					if ((Best * 127)/128 < Clk && ++BestCount>16)
						break;
				}
			}
		}
	}

	Speed = (int)((Best + 500000)/1000000);

	//rounding
	if ((Speed % 104)==1 || Speed==207) --Speed;
	else if ((Speed % 104)==103 || Speed==205) ++Speed;
	else if ((Speed % 10)==1) --Speed;
	else if ((Speed % 10)==9) ++Speed;
	else if (((Speed % 100)%33)==1) --Speed;
	else if (((Speed % 100)%33)==32) ++Speed;

#if defined(_WIN32_WCE)
	if (FuncCeSetThreadQuantum)
		FuncCeSetThreadQuantum(Thread,25);
#endif
	SetThreadPriority(Thread,Priority);
	return Speed;
}

static int Get( platform* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(PLATFORM_ID,int); break;
	case PLATFORM_LANG: GETVALUE(StringLang,int); break;
	case PLATFORM_TYPE:	GETSTRING(p->PlatformType[0] ? p->PlatformType : T("Windows")); break;
	case PLATFORM_CPU: GETSTRING(p->CPU); break;
	case PLATFORM_OEMINFO: if (p->OemInfo[0]) GETSTRING(p->OemInfo); break;
	case PLATFORM_VER: GETVALUE(p->Ver,int); break;
	case PLATFORM_TYPENO: GETVALUE(p->Type,int); break;
	case PLATFORM_MODEL: GETVALUE(p->Model,int); break;
	case PLATFORM_CAPS: GETVALUE(p->Caps,int); break;
	case PLATFORM_ICACHE: GETVALUECOND(p->ICache,int,p->ICache>0); break;
	case PLATFORM_DCACHE: GETVALUECOND(p->DCache,int,p->DCache>0); break;
	case PLATFORM_VERSION: 
		stprintf((tchar_t*)Data,T("%d.%02d"),p->Ver/100,p->Ver%100);
		Result = ERR_NONE;
		break;
	case PLATFORM_MHZ: 
		if (Size==sizeof(int))
		{
			int Speed = CPUSpeed();
			if (Speed)
			{
				*(int*)Data = Speed;
				Result = ERR_NONE;
			}
		}
		break;
	}
	return Result;
}

static int Set( platform* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case PLATFORM_LANG: SETVALUE(StringLang,int,ERR_NONE); break;
	}
	return ERR_INVALID_PARAM;
}

int QueryPlatform( int Param )
{
	int Value;
	Platform.VMT.Get(&Platform,Param,&Value,sizeof(Value));
	return Value;
}

void SafeGetCpuId( int Id, uint32_t* p )
{
	memset(p,0,4*sizeof(uint32_t));
	TRY_BEGIN
	{
		bool_t Mode = KernelMode(1);
		GetCpuId(Id,p); 
		KernelMode(Mode);
	}
	TRY_END
}

void GetProc(void** Module,void* Ptr,const tchar_t* ProcName,int Optional)
{
	FARPROC* Func = (FARPROC*)Ptr;
	if (*Module)
	{
#if !defined(_WIN32_WCE) && defined(UNICODE)
		char ProcName8[256];
		wcstombs(ProcName8,ProcName,sizeof(ProcName8));
		*Func = GetProcAddress(*Module,ProcName8);
#else
		*Func = GetProcAddress(*Module,ProcName);
#endif
		if (!*Func && !Optional)
		{
			FreeLibrary(*Module);
			*Module = NULL;
		}
	}
	else
		*Func = NULL;
}

void Platform_Init()
{
#if defined(_WIN32_WCE)
	tchar_t OemUpper[256];
	RECT WorkArea;
#endif

	char Lang8[16];
	tchar_t Lang[16];
	OSVERSIONINFO Ver;
	platform* p = &Platform;
#ifndef SH3
	uint32_t CpuId[4];
#endif

#if defined(_DEBUG) && defined(_MSC_VER) && !defined(_WIN32_WCE)
	int Flag = _CrtSetDbgFlag( _CRTDBG_REPORT_FLAG );
	Flag |= _CRTDBG_LEAK_CHECK_DF;
	//Flag |= _CRTDBG_CHECK_ALWAYS_DF;
	_CrtSetDbgFlag( Flag );
#endif

	NodeFill(&p->VMT,sizeof(platform),Enum,Get,Set);

	Ver.dwOSVersionInfoSize = sizeof(Ver);
	GetVersionEx(&Ver);

	p->Type = TYPE_UNKNOWN;
	p->Ver = Ver.dwMajorVersion*100 + Ver.dwMinorVersion;
	p->OemInfo[0] = 0;
	p->PlatformType[0] = 0;

#if defined(_WIN32_WCE)
	AygShellDLL = LoadLibrary(T("aygshell.dll"));
	if (AygShellDLL)
		*(FARPROC*)&FuncSHIdleTimerReset = GetProcAddress(AygShellDLL,MAKEINTRESOURCE(2006));

	CoreDLL = LoadLibrary(T("coredll.dll"));
	if (CoreDLL)
	{
		*(FARPROC*)&FuncChangeDisplaySettingsEx = GetProcAddress(CoreDLL,T("ChangeDisplaySettingsEx"));
		*(FARPROC*)&FuncCacheSync = GetProcAddress(CoreDLL,T("CacheSync"));
		*(FARPROC*)&FuncVirtualCopy = GetProcAddress(CoreDLL,T("VirtualCopy"));
		*(FARPROC*)&FuncCeSetThreadQuantum = GetProcAddress(CoreDLL,T("CeSetThreadQuantum"));
		*(FARPROC*)&FuncSetKMode = GetProcAddress(CoreDLL,T("SetKMode"));
		*(FARPROC*)&FuncRedrawWindow = GetProcAddress(CoreDLL,T("RedrawWindow"));
		*(FARPROC*)&FuncSHGetSpecialFolderPath = GetProcAddress(CoreDLL,T("SHGetSpecialFolderPath"));
		*(FARPROC*)&FuncSetPowerRequirement = GetProcAddress(CoreDLL,T("SetPowerRequirement"));
		*(FARPROC*)&FuncReleasePowerRequirement = GetProcAddress(CoreDLL,T("ReleasePowerRequirement"));
	}
	CEShellDLL = LoadLibrary(T("ceshell.dll"));
	if (CEShellDLL)
		*(FARPROC*)&FuncSHGetDocumentsFolder = GetProcAddress(CEShellDLL,MAKEINTRESOURCE(75));

	if (FuncSHGetSpecialFolderPath)
		FuncSHGetSpecialFolderPath(NULL,SysPath,0x24/*CSIDL_WINDOWS*/,FALSE);

	if (!FuncSHGetDocumentsFolder || !FuncSHGetDocumentsFolder(T("\\"),DocPath))
		DocPath[0] = 0;

	if (SystemParametersInfo(SPI_GETPLATFORMTYPE,sizeof(p->PlatformType),p->PlatformType,0)!=0)
	{
		if (tcsicmp(p->PlatformType,T("PocketPC"))==0)
			p->Type = TYPE_POCKETPC;
		else
		if (tcsicmp(p->PlatformType,T("Smartphone"))==0)
			p->Type = TYPE_SMARTPHONE;
	}
	else
	if (GetLastError()==ERROR_ACCESS_DENIED)
	{
		tcscpy(p->PlatformType,T("Smartphone"));
		p->Type = TYPE_SMARTPHONE;
	}
	else
		tcscpy(p->PlatformType,T("Windows CE"));

	SystemParametersInfo(SPI_GETOEMINFO,sizeof(p->OemInfo),p->OemInfo,0);

	if (!AygShellDLL && p->Ver < 300)
		p->Caps |= CAPS_OLDSHELL;

	// detect fake navigator bar and use old style menu
	if (p->Type != TYPE_SMARTPHONE && AygShellDLL && 
		SystemParametersInfo(SPI_GETWORKAREA,0,&WorkArea,0) && WorkArea.top==0)
		p->Caps |= CAPS_OLDSHELL;

	tcscpy(OemUpper,p->OemInfo);
	tcsupr(OemUpper);
#endif


	p->Model = MODEL_DEFAULT;

#if defined(ARM)
	tcscpy(p->CPU,T("ARM"));
	SafeGetCpuId(0,CpuId);

	if (CpuId[0])
	{
		p->ICache = 512 << ((CpuId[1] >> 6) & 7);
		p->DCache = 512 << ((CpuId[1] >> 18) & 7);
	}
	else
	{
		// when need to detect cpu features somehow
		TRY_BEGIN
		{	
			if (CheckARM5E())
			{
				int XScale;
				p->Caps |= CAPS_ARM_5E;

				XScale = CheckARMXScale();
				if (XScale)
				{
					p->ICache = p->DCache = 32768;
					p->Caps |= CAPS_ARM_XSCALE;
					if (XScale > 1)
						p->Caps |= CAPS_ARM_WMMX;
				}
			}
		}
		TRY_END
	}

	if ((CpuId[0] & 0xFF000000) == 0x54000000) //TI
	{
		tcscat(p->CPU,T(" TI"));
		switch ((CpuId[0] >> 4) & 0xFFF)
		{
		case 0x915: tcscat(p->CPU,T("915T")); break;
		case 0x925: tcscat(p->CPU,T("925T")); break;
		case 0x926: tcscat(p->CPU,T("926T")); break;
		}
	}
	else
	if ((CpuId[0] & 0xFF000000) == 0x41000000) //arm
	{
		switch ((CpuId[0] >> 4) & 0xFFF)
		{
		case 0x920: tcscat(p->CPU,T(" 920T")); break;
		case 0x922: tcscat(p->CPU,T(" 922T")); break;
		case 0x926: tcscat(p->CPU,T(" 926E")); break;
		case 0x940: tcscat(p->CPU,T(" 940T")); break;
		case 0x946: tcscat(p->CPU,T(" 946E")); break;
		case 0xA22: tcscat(p->CPU,T(" 1020E")); break;
		}
	}
	else
	if ((CpuId[0] & 0xFF000000) == 0x69000000) //intel
	{
		tcscat(p->CPU,T(" Intel"));

		if ((CpuId[0] & 0xFF0000) == 0x050000) //intel arm5e
			p->Caps |= CAPS_ARM_5E|CAPS_ARM_XSCALE;

		if (((CpuId[0] >> 4) & 0xFFF) == 0xB11)
		{
			tcscat(p->CPU,T(" SA1110"));
		}
		else
		{
			switch ((CpuId[0] >> 13) & 7)
			{
			case 0x2: p->Caps |= CAPS_ARM_WMMX; break;
			}

			switch ((CpuId[0] >> 4) & 31)
			{
			case 0x10: 
				tcscat(p->CPU,T(" PXA25x/26x")); break;
				break;
			case 0x11: 
				tcscat(p->CPU,T(" PXA27x")); break;
				break;
			case 0x12:
				tcscat(p->CPU,T(" PXA210"));
				break;
			}
		}
	}

#if defined(_WIN32_WCE)

	if (tcsstr(OemUpper,T("TYPHOON")))
		p->Model = MODEL_SPV_C500_ORIGROM;
	else
	if (tcsstr(OemUpper,T("PW10B")) ||
		tcsstr(OemUpper,T("PW10A")))
	{
		p->Model = MODEL_O2_XDA;
		p->Caps |= CAPS_ONLY12BITRGB;
	}
	else
	if (tcsstr(OemUpper,T("Compal P6C")))
		p->Model = MODEL_AMIGO_600C;
	else
	if (tcsstr(OemUpper,T("Pocket PC J710")))
		p->Model = MODEL_CASIO_E200;
	else
	if (tcsstr(OemUpper,T("JORNADA")))
	{
		p->Caps |= CAPS_LANDSCAPE;
		if (tcsstr(OemUpper,T("680")))
			p->Model = MODEL_JORNADA_680;
		else
		if (tcsstr(OemUpper,T("690")))
			p->Model = MODEL_JORNADA_690;
		else
		if (tcsstr(OemUpper,T("710")))
			p->Model = MODEL_JORNADA_710;
		else
		if (tcsstr(OemUpper,T("720")))
			p->Model = MODEL_JORNADA_720;
	}
	else
	if (tcsstr(OemUpper,T("IPAQ")))
	{
		p->Caps |= CAPS_PORTRAIT;
		if (tcsstr(OemUpper,T("H3600")))
		{
			p->Model = MODEL_IPAQ_H3600;
			p->Caps |= CAPS_ONLY12BITRGB;
		}
		else
		if (tcsstr(OemUpper,T("H3700")))
		{
			p->Model = MODEL_IPAQ_H3700;
			p->Caps |= CAPS_ONLY12BITRGB;
		}
		else
		if (tcsstr(OemUpper,T("H3800")))
			p->Model = MODEL_IPAQ_H3800;
		else
		if (tcsstr(OemUpper,T("H3900")))
			p->Model = MODEL_IPAQ_H3900;
		else
		if (tcsstr(OemUpper,T("HX4700")))
			p->Model = MODEL_IPAQ_HX4700;
	}
	else
	if (tcsstr(OemUpper,T("TOSHIBA")))
	{
		p->Caps |= CAPS_PORTRAIT;
		if (tcsstr(OemUpper,T("E740")))
			p->Model = MODEL_TOSHIBA_E740;
		else
		if (tcsstr(OemUpper,T("E750")))
			p->Model = MODEL_TOSHIBA_E750;
		else
		if (tcsstr(OemUpper,T("E80")))
			p->Model = MODEL_TOSHIBA_E800;
	}
	else
	if (tcsstr(OemUpper,T("AXIM")))
	{
		if (tcscmp(OemUpper+tcslen(OemUpper)-2,T("X5"))==0)
			p->Model = MODEL_AXIM_X5;

		if (tcscmp(OemUpper+tcslen(OemUpper)-3,T("X50"))==0)
			p->Model = MODEL_AXIM_X50;
	}

#endif

#elif defined(MIPS)
	SafeGetCpuId(0,CpuId);
	tcscpy(p->CPU,T("MIPS"));

	if (((CpuId[0] >> 8) & 255) == 0x0c)
	{
		p->Caps |= CAPS_MIPS_VR41XX;
		tcscat(p->CPU,T(" VR41XX"));
	}

#if defined(_WIN32_WCE)

	if (tcsstr(OemUpper,T("BSQUARE")))
	{
		tcscpy(OemUpper,p->PlatformType);
		tcsupr(OemUpper);

		if (tcsstr(OemUpper,T("ALCHEMY")))
			p->Model = MODEL_BSQUARE_ALCHEMY;

		tcscpy(OemUpper,p->OemInfo);
		tcsupr(OemUpper);
	}

	if (tcsstr(OemUpper,T("POCKET")))
	{
		if (tcsstr(OemUpper,T(" J74")))
		{
			p->Model = MODEL_CASIO_BE300;
			p->Caps |= CAPS_OLDSHELL;
		}
	}

	if (tcsstr(OemUpper,T("POCKET PC")))
	{
		p->Caps |= CAPS_PORTRAIT;
		if (tcsstr(OemUpper,T(" J67")))
			p->Model = MODEL_CASIO_E125;
		else
		if (tcsstr(OemUpper,T(" J76")))
			p->Model = MODEL_CASIO_EM500;
		else
		if (tcsstr(OemUpper,T(" J58")))
			p->Model = MODEL_CASIO_E115;
		else
		if (tcsstr(OemUpper,T(" J53")) || tcsstr(OemUpper,T(" JX53")))
			p->Model = MODEL_CASIO_E105;
	}

	if (tcsstr(OemUpper,T("JVC")))
	{
		if (tcsstr(OemUpper,T("C33")))
			p->Model = MODEL_JVC_C33;
	}
#endif

#elif defined(SH3)
	tcscpy(p->CPU,T("SH3"));

#if defined(_WIN32_WCE)
	if (tcsstr(OemUpper,T("JORNADA")))
	{
		if (tcsstr(OemUpper,T(" 43")))
		{
			p->Caps |= CAPS_ONLY12BITRGB;
		}
		if (tcsstr(OemUpper,T(" 54")))
		{
			p->Model = MODEL_JORNADA_540;
			p->Caps |= CAPS_ONLY12BITRGB | CAPS_PORTRAIT;
		}
	}
#endif

#elif defined(_M_IX86)
	tcscpy(p->CPU,T("x86"));
	SafeGetCpuId(0,CpuId);

    if (CpuId[1] == 0x756e6547 &&
        CpuId[3] == 0x49656e69 &&
        CpuId[2] == 0x6c65746e) 
	{
		tcscat(p->CPU,T(" Intel"));

Intel:
		SafeGetCpuId(1,CpuId);
        if (CpuId[3] & 0x00800000)
		{
			p->Caps |= CAPS_X86_MMX;
			if (CpuId[3] & 0x02000000) 
				p->Caps |= CAPS_X86_MMX2 | CAPS_X86_SSE;
			if (CpuId[3] & 0x04000000) 
				p->Caps |= CAPS_X86_SSE2;
		}
    } 
	else if (CpuId[1] == 0x68747541 &&
             CpuId[3] == 0x69746e65 &&
             CpuId[2] == 0x444d4163) 
	{
		tcscat(p->CPU,T(" AMD"));

		SafeGetCpuId(0x80000000,CpuId);
        if (CpuId[0] < 0x80000001)
            goto Intel;

		SafeGetCpuId(0x80000001,CpuId);
        if (CpuId[3] & 0x00800000)
		{
			p->Caps |= CAPS_X86_MMX;
			if (CpuId[3] & 0x80000000)
				p->Caps |= CAPS_X86_3DNOW;
			if (CpuId[3] & 0x00400000)
				p->Caps |= CAPS_X86_MMX2;
		}
    } 
	else if (CpuId[1] == 0x746e6543 &&
             CpuId[3] == 0x48727561 &&
             CpuId[2] == 0x736c7561) 
	{
		tcscat(p->CPU,T(" VIA C3"));

		SafeGetCpuId(0x80000000,CpuId);
        if (CpuId[0] < 0x80000001)
            goto Intel;

		SafeGetCpuId(0x80000001,CpuId);
		if (CpuId[3] & (1<<31))
			p->Caps |= CAPS_X86_3DNOW;
		if (CpuId[3] & (1<<23))
			p->Caps |= CAPS_X86_MMX;
		if (CpuId[3] & (1<<24))
			p->Caps |= CAPS_X86_MMX2;
	}
	else if (CpuId[1] == 0x69727943 &&
             CpuId[3] == 0x736e4978 &&
             CpuId[2] == 0x64616574) 
	{
		tcscat(p->CPU,T(" Cyrix"));

        if (CpuId[0] != 2) 
            goto Intel;

		SafeGetCpuId(0x80000001,CpuId);
        if (CpuId[3] & 0x00800000)
		{
			p->Caps |= CAPS_X86_MMX;
			if (CpuId[3] & 0x01000000)
				p->Caps |= CAPS_X86_MMX2;
		}
    }

#endif

	if (GetLocaleInfo(LOCALE_USER_DEFAULT,LOCALE_SABBREVLANGNAME,Lang,sizeof(Lang)/sizeof(tchar_t)) >= 2)
	{
		tcsupr(Lang);
		if (tcscmp(Lang,T("JPN"))==0) tcscpy(Lang,T("JA"));
		if (tcscmp(Lang,T("GER"))==0) tcscpy(Lang,T("DE"));
		if (tcscmp(Lang,T("SPA"))==0) tcscpy(Lang,T("ES"));

		// special three letter cases: Chinese traditional (CHT) and Chinese simplified (CHS)
		if (tcscmp(Lang,T("CHS"))==0 || tcscmp(Lang,T("ZHI"))==0 || tcscmp(Lang,T("ZHM"))==0)
			StringLang = FOURCC('C','H','S','_');
		else
		if (tcscmp(Lang,T("CHT"))==0 || tcscmp(Lang,T("ZHH"))==0)
			StringLang = FOURCC('C','H','T','_');
		else
		{
			// two letter version
			TcsToStr(Lang8,Lang,sizeof(Lang8));
			StringLang = FOURCC(Lang8[0],Lang8[1],'_','_');
		}
	}

	NodeRegister(&p->VMT,PRI_MAXIMUM+650);

	BacklightTimeout(0); // restore backlight settings (example if last time crashed)
}

void Platform_Done()
{
	NodeUnRegister(&Platform.VMT);

	UnlockFrameBuffer();
	SetDisplayPower(1);

#if defined(_WIN32_WCE)
	{
		phymem* p;
		for (p=PhyMem;p!=PhyMem+MAXPHY;++p)
			if (p->Length)
			{
				VirtualFree(p->Virt,0,MEM_RELEASE);
				p->Length = 0;
			}
	}

	if (AygShellDLL) FreeLibrary(AygShellDLL);
	if (CoreDLL) FreeLibrary(CoreDLL);
	if (CEShellDLL) FreeLibrary(CEShellDLL);
#endif
}

bool_t IsOrientationChanged()
{
	int Old = Orientation;
	Orientation = -1;
	return QueryOrientation() != Old;
}

int QueryOrientation()
{
#if defined(_WIN32_WCE)
	if (Orientation < 0)
	{
		HKEY Key;
		char Buffer[256];
		DEVMODE* Mode = (DEVMODE*)Buffer;
		int i,No;
		node* Node;

		Mode->dmSize = 192;
		Mode->dmFields = DM_DISPLAYQUERYORIENTATION;

		if (Platform.Ver >= 421 && // we don't trust this method on pre wm2003se systems
			FuncChangeDisplaySettingsEx &&
			FuncChangeDisplaySettingsEx(NULL, Mode, NULL, CDS_TEST, NULL) == DISP_CHANGE_SUCCESSFUL)
		{
			Mode->dmFields = DM_DISPLAYORIENTATION;
			FuncChangeDisplaySettingsEx(NULL, Mode, NULL, CDS_TEST, NULL);

			switch ((&Mode->dmDisplayFrequency)[1]) //(Mode->dmDisplayOrientation)
			{
			case DMDO_0: Orientation = 0; break;
			case DMDO_90: Orientation = DIR_SWAPXY | DIR_MIRRORUPDOWN; break;
			case DMDO_270: Orientation = DIR_SWAPXY | DIR_MIRRORLEFTRIGHT; break;
			case DMDO_180: Orientation = DIR_MIRRORUPDOWN | DIR_MIRRORLEFTRIGHT; break;
			}
		}

		if (Orientation < 0)
			for (No=0;(Node = NodeEnum(VOUT_CLASS,No))!=NULL;++No)
				if (Node->Get(Node,VOUT_ORIENTATION,&i,sizeof(int))==ERR_NONE)
				{
					Orientation = i;
					break;
				}

		if (Orientation < 0 && RegOpenKeyEx( HKEY_LOCAL_MACHINE, T("System\\GDI\\ROTATION"), 0, ACCESS_LOAD, &Key ) == ERROR_SUCCESS)
		{
			DWORD Value;
			DWORD RegSize = sizeof(Value);
			DWORD RegType;

			if (RegQueryValueEx(Key, T("Angle"), 0, &RegType, (LPBYTE) &Value, &RegSize) == ERROR_SUCCESS)
				switch (Value)
				{
				case 0: Orientation = 0; break;
				case 90: Orientation = DIR_SWAPXY | DIR_MIRRORUPDOWN; break;
				case 270: Orientation = DIR_SWAPXY | DIR_MIRRORLEFTRIGHT; break;
				case 180: Orientation = DIR_MIRRORUPDOWN | DIR_MIRRORLEFTRIGHT; break;
				}

			RegCloseKey( Key );
		}

		if (Orientation < 0)
			Orientation = 0;
	}
#else
	Orientation = 0;
#endif
	return Orientation;
}

void QueryDesktop( video* p )
{
	HDC DC = GetDC(0);

	memset(p,0,sizeof(video));

	p->Width = GetDeviceCaps(DC,HORZRES);
	p->Height = GetDeviceCaps(DC,VERTRES);
	p->Aspect = ASPECT_ONE;
	p->Direction = 0;

	if (GetDeviceCaps(DC,BITSPIXEL)*GetDeviceCaps(DC,PLANES) <= 8)
	{
		int n,i;

		p->Pixel.Flags = PF_PALETTE;
		p->Pixel.BitCount = 1;

		n=GetSystemPaletteEntries(DC,0,256,NULL);
		for (i=2;i<n;i=i*i)
			p->Pixel.BitCount <<= 1;

		if (n>16) //4,2,1 bits are grayscale (default palette required)
		{
			GetSystemPaletteEntries(DC,0,256,(PALETTEENTRY*)Palette);
			p->Pixel.Palette = Palette;
		}
	}
	else
	if (QueryPlatform(PLATFORM_CAPS) & CAPS_ONLY12BITRGB)
		DefaultRGB(&p->Pixel,GetDeviceCaps(DC,BITSPIXEL),4,4,4,1,2,1);
	else
	{
		int i;
		int RBits = 4;
		int GBits = 4;
		int BBits = 4;
		int BitCount = GetDeviceCaps(DC,BITSPIXEL);

		COLORREF Old = GetPixel(DC,0,0);
		for (i=3;i>=0;--i)
		{
			COLORREF c = SetPixel(DC,0,0,0x010101<<i);
			if (c == (COLORREF)-1)
			{
				if (BitCount > 16)
					RBits = GBits = BBits = 8;
				else
				{
					RBits = BBits = 5; GBits = 6;
				}
				break;
			}
			if (c & 0xFF)	  RBits++;
			if (c & 0xFF00)	  GBits++;
			if (c & 0xFF0000) BBits++;
		}
		SetPixel(DC,0,0,Old);

		DefaultRGB(&p->Pixel,BitCount,RBits,GBits,BBits,0,0,0);
	}

	ReleaseDC(0,DC);
}

void* PhyMemBegin(uint32_t Phy,uint32_t Length)
{
	if (!Length)
		return NULL;

#if defined(_WIN32_WCE)
	if (FuncVirtualCopy)
	{
		phymem* p;
		for (p=PhyMem;p!=PhyMem+MAXPHY;++p)
			if (p->Phy <= Phy && p->Phy+p->Length >= Phy+Length)
			{
				++p->RefCount;
				return p->Virt + (Phy - p->Phy);
			}

		for (p=PhyMem;p!=PhyMem+MAXPHY;++p)
		{
			if (!p->RefCount && p->Length)
			{
				VirtualFree(p->Virt,0,MEM_RELEASE);
				p->Length = 0;
			}

			if (!p->Length)
			{
				int Align = Phy & 65535;
				Phy -= Align;
				Length = (Length + Align + 65535) & ~65535;
				p->Virt = (char*) VirtualAlloc(0, Length, MEM_RESERVE, PAGE_NOACCESS);
				if (p->Virt)
				{
					if (FuncVirtualCopy(p->Virt,(LPVOID)(Phy >> 8), Length,
						PAGE_READWRITE | PAGE_NOCACHE | PAGE_PHYSICAL))
					{
						p->RefCount = 1;
						p->Phy = Phy;
						p->Length = Length;
						return p->Virt + Align;
					}
					VirtualFree(p->Virt,0,MEM_RELEASE);
				}
				break;
			}
		}
	}	
#endif
	return NULL;
}

void PhyMemEnd(void* Virt)
{
#if defined(_WIN32_WCE)
	if (Virt)
	{
		phymem* p;
		for (p=PhyMem;p!=PhyMem+MAXPHY;++p)
			if (p->RefCount && p->Virt <= (char*)Virt && p->Virt+p->Length > (char*)Virt)
			{
				--p->RefCount;
				break;
			}
	}
#endif
}

bool_t KernelMode(bool_t v)
{
#if defined(_WIN32_WCE)
	if (FuncSetKMode)
		return FuncSetKMode(v);
#endif
	return 0;
}

bool_t GetDisplayPower()
{
	return DisplayPower;
}

void SetDisplayPower(bool_t State)
{
	VIDEO_POWER_MANAGEMENT VPM;
	HDC DC;

	if (DisplayPower != State)
	{
		DisplayPower = State;

		DC = GetDC(NULL);
		VPM.Length = sizeof(VPM);
		VPM.DPMSVersion = 1;
		VPM.PowerState = State ? 1:4;

		ExtEscape(DC, SETPOWERMANAGEMENT, sizeof(VPM), (LPCSTR) &VPM, 0, NULL);
		ReleaseDC(NULL, DC);
	}
}

static void ChangeRegEntry( bool_t CurrentUser, const tchar_t* Reg, const tchar_t* Name, bool_t State, int NewValue, int BackupId ) 
{
	HKEY Key = 0;
	DWORD Size;
	DWORD Value;

	if (RegOpenKeyEx(CurrentUser ? HKEY_CURRENT_USER:HKEY_LOCAL_MACHINE,Reg, 0, ACCESS_SAVE, &Key ) == ERROR_SUCCESS) 
	{
		if (State)
		{
			// save old value 
			Size = sizeof(Value);
			if (RegQueryValueEx( Key, Name,NULL,NULL,(LPBYTE)&Value,&Size) == ERROR_SUCCESS)
				NodeSaveValue(BackupId,&Value,sizeof(Value),TYPE_INT);

			// set no screensaver
			Value = NewValue;
			Size = sizeof(Value);
			RegSetValueEx( Key, Name,0,REG_DWORD,(LPBYTE)&Value,Size);
		}
		else
		if (NodeLoadValue(BackupId,&Value,sizeof(Value),NULL))
		{
			// restore value (and delete old value)
			Size = sizeof(Value);
			RegSetValueEx( Key, Name,0,REG_DWORD,(LPBYTE)&Value,Size);

			NodeSaveValue(BackupId,NULL,0,TYPE_INT);
		}

		RegCloseKey( Key );
	}
}

static bool_t BacklightState = -1;

void IdleTimerReset()
{
#if defined(_WIN32_WCE)
	SystemIdleTimerReset();
	if (FuncSHIdleTimerReset)
		if (BacklightState || !QueryAdvPlatform(ADVPLATFORM_HOMESCREEN))
			FuncSHIdleTimerReset();
#endif
}

void BacklightTimeout(bool_t Disable)
{
	if (Disable && QueryAdvPlatform(ADVPLATFORM_NOBACKLIGHT))
		Disable = 0;

	if (Disable != BacklightState)
	{
		HANDLE Handle;
		int Value;

		BacklightState = Disable;

#ifdef MIPS
		if (Platform.Model == MODEL_CASIO_E125 ||
			Platform.Model == MODEL_CASIO_EM500 ||
			Platform.Model == MODEL_CASIO_E115 ||
			Platform.Model == MODEL_CASIO_BE300 ||
			Platform.Model == MODEL_CASIO_E105)
		{
			Value = 0;
			ChangeRegEntry(0,RegCASIOBacklight,RegTimeoutBattery,Disable,Value,PLATFORM_BATTERYTIMEOUT);
			ChangeRegEntry(0,RegCASIOBacklight,RegTimeoutExPower,Disable,Value,PLATFORM_ACTIMEOUT);
		}
		else
#endif
		{
			Value = (Platform.Type == TYPE_SMARTPHONE) ? 7199999 /*10*3600*1000*/ : 0x7FFFFFFF;
			if (Platform.Model == MODEL_AXIM_X5) 
				Value = 0;
			ChangeRegEntry(1,RegScreenSaver,RegMode,Disable,1,PLATFORM_SCREENSAVER);
			ChangeRegEntry(1,RegBacklight,RegBatteryTimeout,Disable,Value,PLATFORM_BATTERYTIMEOUT);
			ChangeRegEntry(1,RegBacklight,RegACTimeout,Disable,Value,PLATFORM_ACTIMEOUT);
			ChangeRegEntry(1,RegPower,RegDisplay,Disable,-1,PLATFORM_DISPPOWER);
		}
	
		Handle = CreateEvent( NULL, FALSE, FALSE, T("BackLightChangeEvent") );
		if (Handle) 
		{
			SetEvent(Handle);
			CloseHandle(Handle);
		}

#if defined(_WIN32_WCE)
		if (FuncSetPowerRequirement && FuncReleasePowerRequirement)
		{
			static HANDLE PowerHandle = NULL;
			if (Disable && !PowerHandle)
			{
				PowerHandle = FuncSetPowerRequirement(T("BKL1:"),0,1,NULL,0);
			}
			else if (!Disable && PowerHandle)
			{
				FuncReleasePowerRequirement(PowerHandle);
				PowerHandle = NULL;
			}
		}
#endif
	}
}

void* LockFrameBuffer( video* p )
{
	void* Ptr = NULL;

	memset(p,0,sizeof(video));

	//set default values (for most of the models)
	p->Width = 240;
	p->Height = 320;
	p->Pitch = 480;
	p->Aspect = ASPECT_ONE;
	p->Direction = 0;

	if (QueryPlatform(PLATFORM_CAPS) & CAPS_ONLY12BITRGB)
		DefaultRGB(&p->Pixel,16,4,4,4,1,2,1);
	else
		DefaultRGB(&p->Pixel,16,5,6,5,0,0,0);

	switch (QueryPlatform(PLATFORM_MODEL))
	{
	case MODEL_BSQUARE_ALCHEMY:
		Ptr = PhyFrameBuffer = PhyMemBegin(0x15000000,320*480);
		p->Width = 240;
		p->Height = 320;
		p->Pitch = 480;
		break;

	case MODEL_IPAQ_H3600:
	case MODEL_IPAQ_H3700:
		p->Direction = DIR_SWAPXY | DIR_MIRRORLEFTRIGHT;
		Ptr = (void*)0xAC050020;
		p->Width = 320;
		p->Height = 240;
		p->Pitch = 640;
		break;

	case MODEL_IPAQ_H3800:
		p->Direction = DIR_SWAPXY | DIR_MIRRORUPDOWN;
		Ptr = (void*)0xAC050020;
		p->Width = 320;
		p->Height = 240;
		p->Pitch = 640;
		break;

	case MODEL_CASIO_BE300:
	case MODEL_CASIO_E105:
	case MODEL_CASIO_E115:
		Ptr = (void*)0xAA200000;
		p->Pitch = 512;
		break;

	case MODEL_JORNADA_710:
	case MODEL_JORNADA_720:
		Ptr = PhyFrameBuffer = PhyMemBegin(0x48200000,240*1280);
		p->Width = 640;
		p->Height = 240;
		p->Pitch = 1280;
		break;

	case MODEL_JORNADA_680:
	case MODEL_JORNADA_690: 
		Ptr = (void*)0xB2000000; //only japan???
		p->Width = 640;
		p->Height = 240;
		p->Pitch = 1280;
		break;

	case MODEL_MOBILEGEAR2_550:
	case MODEL_MOBILEGEAR2_450:
	case MODEL_MOBILEGEAR2_530:
	case MODEL_MOBILEGEAR2_430:
	case MODEL_MOBILEPRO_780:
		Ptr = (void*)0xAA180100;
		p->Width = 640;
		p->Height = 240;
		p->Pitch = 1280;
		break;

	case MODEL_MOBILEPRO_770:
		Ptr = (void*)0xAA000000;
		p->Width = 640;
		p->Height = 240;
		p->Pitch = 1600;
		break;

	case MODEL_JVC_C33:
		Ptr = (void*)0x445C0000;
		p->Width = 1024;
		p->Height = 600;
		p->Pitch = 2048;
		break;

	case MODEL_SIGMARION:
		Ptr = (void*)0xAA000000;
		p->Width = 640;
		p->Height = 240;
		p->Pitch = 1280;
		break;

	case MODEL_SIGMARION2:
		Ptr = (void*)0xB0800000;
		p->Width = 640;
		p->Height = 240;
		p->Pitch = 1280;
		break;

	case MODEL_SIGMARION3:
		Ptr = PhyFrameBuffer = PhyMemBegin(0x14800000,480*1600);
		p->Width = 800;
		p->Height = 480;
		p->Pitch = 1600;
		break;

	case MODEL_INTERMEC_6651:
		Ptr = PhyFrameBuffer = PhyMemBegin(0x6C000000,480*1600);
		p->Width = 800;
		p->Height = 480;
		p->Pitch = 1600;
		break;
	}

	if (((int)Ptr & 15)==0 && (p->Pitch & 15)==0)
		p->Pixel.Flags |= PF_16ALIGNED;
	return Ptr;
}

void UnlockFrameBuffer()
{
	if (PhyFrameBuffer)
	{
		PhyMemEnd(PhyFrameBuffer);
		PhyFrameBuffer = NULL;
	}
}

void* CodeAlloc(int Size)
{
	return VirtualAlloc(NULL,Size,MEM_COMMIT,PAGE_EXECUTE_READ);
}

void CodeFree(void* Code,int Size)
{
	VirtualFree(Code,Size,MEM_DECOMMIT);
	VirtualFree(Code,0,MEM_RELEASE);
}

void CodeLock(void* Code,int Size)
{
	DWORD Protect;
	VirtualProtect(Code,Size,PAGE_READWRITE|PAGE_NOCACHE,&Protect);
}

void CodeUnlock(void* Code,int Size)
{
	DWORD Protect;
	VirtualProtect(Code,Size,PAGE_EXECUTE_READ,&Protect);
#if defined(_WIN32_WCE)
	if (FuncCacheSync)
		FuncCacheSync(CACHE_SYNC_INSTRUCTIONS);
#endif
}

void CodeFindPages(void* Ptr,uint8_t** Min,uint8_t** Max,int* PageSize)
{
#if defined(MIPS)
	*PageSize = 1024;
#else
	*PageSize = 4096;
#endif

	*Min = *Max = (uint8_t*) ((int)Ptr & ~(*PageSize-1));
	while (!IsBadCodePtr((FARPROC)(*Min-*PageSize)))
		*Min -= *PageSize;
	while (!IsBadCodePtr((FARPROC)*Max))
		*Max += *PageSize;
}

void Assert(const char* Exp,const char* File,int Line)
{
	TCHAR TExp[MAXPATH];
	TCHAR TFile[MAXPATH];
	StrToTcs(TExp,File,MAXPATH);
	stprintf(TFile,T("%d:%s"),Line,TExp);
	StrToTcs(TExp,Exp,MAXPATH);
	MessageBox(NULL,TExp,TFile,MB_OK);
}

void* LockCreate()
{
	void* p = Alloc(sizeof(CRITICAL_SECTION));
	InitializeCriticalSection((LPCRITICAL_SECTION)p);
	return p;
}

void LockDelete(void* p)
{
	if (p)
	{
		DeleteCriticalSection((LPCRITICAL_SECTION)p);
		Free(p);
	}
}

void LockEnter(void* p)
{
	EnterCriticalSection((LPCRITICAL_SECTION)p);
}

void LockLeave(void* p)
{
	LeaveCriticalSection((LPCRITICAL_SECTION)p);
}

void* ThreadCreate(int(*Start)(void*),void* Parameter,int Quantum)
{
	DWORD Id;
	HANDLE Handle = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)Start,Parameter,0,&Id);

#if defined(_WIN32_WCE)
	if (FuncCeSetThreadQuantum)
		FuncCeSetThreadQuantum(Handle,Quantum);
#endif
	return Handle;
}

void CheckHeap()
{
#if defined(_DEBUG) && defined(_MSC_VER)
	assert(_CrtCheckMemory());
#endif
}

//#define FIND 0x009F15A0

/*static INLINE void IsMemoryLow(void* p,int n)
{
	if (!p && n && ShowMemoryLow)
	{
		ShowMemoryLow = 0;
		NodeError((node*)&Platform,PLATFORM_ID,PLATFORM_MEMORY_LOW);
	}
}
*/

int AvailMemory()
{
	MEMORYSTATUS Status;
	Status.dwLength = sizeof(Status);

	GlobalMemoryStatus(&Status); 

	return Status.dwAvailPhys;
}

void* Alloc(int n)
{
	void* p = malloc(n);

	//DEBUG_MSG1(-1,T("Alloc %d"),n);
	//IsMemoryLow(p,n);

#ifdef FIND
	if ((int)p == FIND) DebugBreak();
#endif

	return p;
}

void* CAlloc(int n,int m)
{
	void* p;
	
	n *= m;
	p = malloc(n);
	//IsMemoryLow(p,n);
	if (p)
		memset(p,0,n);

#ifdef FIND
	if ((int)p == FIND) DebugBreak();
#endif

	return p;
}

void* ReAlloc(void* p,int n)
{
	p = realloc(p,n);
	//IsMemoryLow(p,n);

#ifdef FIND
	if ((int)p == FIND) DebugBreak();
#endif

	return p;
}

void Free(void* p)
{
	free(p);
}

void ShowMessage(const tchar_t* Title,const tchar_t* Msg,...)
{
	bool_t Wait = WaitCursor;
	tchar_t s[1024];
	vstprintf( s, Msg, (VA_LIST)(&Msg+1) );
	if (Wait) WaitEnd();
	MessageBox(NULL,s,Title,MB_OK|MB_SETFOREGROUND);
	if (Wait) WaitBegin();
}

void DebugTime(const tchar_t* Name)
{
	SYSTEMTIME t;
	GetSystemTime(&t);
	DebugMessage(T("%02d:%02d:%02d.%03d %s"),t.wHour,t.wMinute,t.wSecond,t.wMilliseconds,Name);
}

//#if !defined(_WIN32_WCE)
static file Debug = NULL;
//#endif

void DebugMessage(const tchar_t* Msg,...)
{
	tchar_t s[1024];

#ifndef NDEBUG
	if (!Debug)	
		Debug = FileOpen(T("\\debug.txt"),1);
#else
	if (!Debug)
	{
		stprintf(s,T("%s\\log.txt"),DocPath);
		Debug = FileOpen(s,1);
	}
#endif

	s[0]=0;
#ifndef NDEBUG
//	stprintf(s,T("%04x "),(int)GetCurrentThreadId());
#endif
	vstprintf( s+tcslen(s), Msg, (VA_LIST)(&Msg+1) );
	tcscat(s,T("\n"));

#ifndef NDEBUG
	OutputDebugString(s);
#endif
	FilePrintf(Debug,T("%s"),s);
}

typedef struct contextreg
{
	int Ofs;
	const tchar_t* Name;
} contextreg;

#if defined( _M_IX86 )
static contextreg Reg[] = 
{
	OFS(CONTEXT,Eax), T("eax"),
	OFS(CONTEXT,Ebx), T("ebx"),
	OFS(CONTEXT,Ecx), T("ecx"),
	OFS(CONTEXT,Edx), T("edx"),
	OFS(CONTEXT,Esi), T("esi"),
	OFS(CONTEXT,Edi), T("edi"),
	OFS(CONTEXT,Ebp), T("ebp"),
	OFS(CONTEXT,Esp), T("esp"),
	OFS(CONTEXT,Eip), T("eip"),
	OFS(CONTEXT,EFlags), T("flags"),

	OFS(CONTEXT,Esp), NULL, 
};
#elif defined( MIPS )
static contextreg Reg[] = 
{
	OFS(CONTEXT,IntZero), T("Zero"),
	OFS(CONTEXT,IntAt), T("At"),
	OFS(CONTEXT,IntV0), T("V0"),
	OFS(CONTEXT,IntV1), T("V1"),
	OFS(CONTEXT,IntA0), T("A0"),
	OFS(CONTEXT,IntA1), T("A1"),
	OFS(CONTEXT,IntA2), T("A2"),
	OFS(CONTEXT,IntA3), T("A3"),
	OFS(CONTEXT,IntT0), T("T0"),
	OFS(CONTEXT,IntT1), T("T1"),
	OFS(CONTEXT,IntT2), T("T2"),
	OFS(CONTEXT,IntT3), T("T3"),
	OFS(CONTEXT,IntT4), T("T4"),
	OFS(CONTEXT,IntT5), T("T5"),
	OFS(CONTEXT,IntT6), T("T6"),
	OFS(CONTEXT,IntT7), T("T7"),
	OFS(CONTEXT,IntS0), T("S0"),
	OFS(CONTEXT,IntS1), T("S1"),
	OFS(CONTEXT,IntS2), T("S2"),
	OFS(CONTEXT,IntS3), T("S3"),
	OFS(CONTEXT,IntS4), T("S4"),
	OFS(CONTEXT,IntS5), T("S5"),
	OFS(CONTEXT,IntS6), T("S6"),
	OFS(CONTEXT,IntS7), T("S7"),
	OFS(CONTEXT,IntT8), T("T8"),
	OFS(CONTEXT,IntT9), T("T9"),
	OFS(CONTEXT,IntK0), T("K0"),
	OFS(CONTEXT,IntK1), T("K1"),
	OFS(CONTEXT,IntGp), T("Gp"),
	OFS(CONTEXT,IntSp), T("Sp"),
	OFS(CONTEXT,IntS8), T("S8"),
	OFS(CONTEXT,IntRa), T("Ra"),
	OFS(CONTEXT,IntLo), T("Lo"),
	OFS(CONTEXT,IntHi), T("Hi"),
	OFS(CONTEXT,IntSp), NULL,
};
#elif defined( ARM )
static contextreg Reg[] = 
{
	OFS(CONTEXT,R0), T("R0"),
	OFS(CONTEXT,R1), T("R1"),
	OFS(CONTEXT,R2), T("R2"),
	OFS(CONTEXT,R3), T("R3"),
	OFS(CONTEXT,R4), T("R4"),
	OFS(CONTEXT,R5), T("R5"),
	OFS(CONTEXT,R6), T("R6"),
	OFS(CONTEXT,R7), T("R7"),
	OFS(CONTEXT,R8), T("R8"),
	OFS(CONTEXT,R9), T("R9"),
	OFS(CONTEXT,R10),T("R10"),
	OFS(CONTEXT,R11),T("R11"),
	OFS(CONTEXT,R12),T("R12"),
	OFS(CONTEXT,Sp), T("Sp"),
	OFS(CONTEXT,Lr), T("Lr"),
	OFS(CONTEXT,Pc), T("Pc"),
	OFS(CONTEXT,Psr),T("Psr"),
	OFS(CONTEXT,Sp), NULL,
};
#elif defined( SH3 )
static contextreg Reg[] = 
{
	OFS(CONTEXT,PR), T("PR"),
	OFS(CONTEXT,MACH), T("MACH"),
	OFS(CONTEXT,MACL), T("MACL"),
	OFS(CONTEXT,GBR), T("GBR"),
	OFS(CONTEXT,R0), T("R0"),
	OFS(CONTEXT,R1), T("R1"),
	OFS(CONTEXT,R2), T("R2"),
	OFS(CONTEXT,R3), T("R3"),
	OFS(CONTEXT,R4), T("R4"),
	OFS(CONTEXT,R5), T("R5"),
	OFS(CONTEXT,R6), T("R6"),
	OFS(CONTEXT,R7), T("R7"),
	OFS(CONTEXT,R8), T("R8"),
	OFS(CONTEXT,R9), T("R9"),
	OFS(CONTEXT,R10), T("R10"),
	OFS(CONTEXT,R11), T("R11"),
	OFS(CONTEXT,R12), T("R12"),
	OFS(CONTEXT,R13), T("R13"),
	OFS(CONTEXT,R14), T("R14"),
	OFS(CONTEXT,R15), T("R15"),
	OFS(CONTEXT,R15), NULL,
};
#else
static contextreg Reg[] = { -1, NULL };
#endif

file FileOpen( const tchar_t* Path, bool_t Write )
{
	return (file)CreateFile(Path,Write?GENERIC_WRITE:GENERIC_READ,FILE_SHARE_READ|FILE_SHARE_WRITE,NULL,Write?CREATE_ALWAYS:OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
}

int FileSeek( file File, int Ofs, int Mode)
{
	switch (Mode)
	{
	case SEEK_SET: Mode = FILE_BEGIN; break;
	case SEEK_CUR: Mode = FILE_CURRENT; break;
	case SEEK_END: Mode = FILE_END; break;
	default: return -1;
	}
	return SetFilePointer((HANDLE)File,Ofs,NULL,Mode);
}

int FileRead( file File, void* p, int n)
{
	DWORD m = (DWORD)-1;
	ReadFile((HANDLE)File,p,n,&m,NULL);
	return m;
}

int FileWrite( file File, const void* p, int n)
{
	DWORD m = (DWORD)-1;
	WriteFile((HANDLE)File,p,n,&m,NULL);
	return m;
}

void FileClose( file File )
{
	CloseHandle((HANDLE)File);
}

void FilePrintf( file File, const tchar_t* Msg,...)
{
	tchar_t* s0 = (tchar_t*)Alloc(2048*sizeof(tchar_t));
	tchar_t* s1 = s0;
	tchar_t* s2 = s0+1024;
	vstprintf( s1, Msg, (VA_LIST)(&Msg+1) );
	while (*s1)
	{
		if (*s1 == 10)
			*(s2++) = 13;
		*(s2++) = *(s1++);
	}
	*s2=0;
	TcsToStr((char*)s0,s0+1024,1024*sizeof(tchar_t));
	FileWrite(File,s0,strlen((char*)s0)*sizeof(char));
	Free(s0);
}

void StreamPrintf( stream* Stream, bool_t UTF8, const tchar_t* Msg,...)
{
	tchar_t* s0 = (tchar_t*)Alloc(2048*sizeof(tchar_t));
	tchar_t* s1 = s0;
	tchar_t* s2 = s0+1024;
	vstprintf( s1, Msg, (VA_LIST)(&Msg+1) );
	while (*s1)
	{
		if (*s1 == 10)
			*(s2++) = 13;
		*(s2++) = *(s1++);
	}
	*s2=0;

	if (UTF8)
		TcsToUTF8((char*)s0,s0+1024,1024*sizeof(tchar_t));
	else
		TcsToStr((char*)s0,s0+1024,1024*sizeof(tchar_t));

	Stream->Write(Stream,s0,strlen((char*)s0)*sizeof(char));
	Free(s0);
}

void Dump()
{
	file File;
	tchar_t Path[MAXPATH];
	stprintf(Path,T("%s\\dump.txt"),DocPath);

	File = FileOpen(Path,1);
	if (File)
	{
		NodeDump(File);
		FileClose(File);

		MessageBox(NULL,String(PLATFORM_ID,PLATFORM_DUMP_MESSAGE),
			String(PLATFORM_ID,PLATFORM_DUMP_TITLE),MB_OK|MB_SETFOREGROUND|MB_TOPMOST);
	}
}

int SafeException(void* p)
{
	EXCEPTION_POINTERS* Data = (EXCEPTION_POINTERS*)p;
	file* File;
	tchar_t Path[MAXPATH];
	stprintf(Path,T("%s\\crash.txt"),DocPath);

	File = FileOpen(Path,1);
	if (File)
	{
		void** Stack;
		contextreg* r;
		int No;
		const uint8_t* Context = (const uint8_t*) Data->ContextRecord;
		EXCEPTION_RECORD* Record = Data->ExceptionRecord;
		tchar_t* Name;
		int DllBase;
		tchar_t DllName[MAXPATH];

		FilePrintf(File,T("%s %s crash report\n----------------------------\n"),NodeProgramId(),NodeVersion());

		switch (Record->ExceptionCode)
		{
		case STATUS_ACCESS_VIOLATION:		Name = T("Access violation"); break;
		case STATUS_BREAKPOINT:				Name = T("Breakpoint"); break;
		case STATUS_DATATYPE_MISALIGNMENT:	Name = T("Datatype misalignment"); break;
		case STATUS_ILLEGAL_INSTRUCTION:	Name = T("Illegal instruction"); break;
		case STATUS_INTEGER_DIVIDE_BY_ZERO: Name = T("Int divide by zero"); break;
		case STATUS_INTEGER_OVERFLOW:		Name = T("Int overflow"); break;
		case STATUS_PRIVILEGED_INSTRUCTION: Name = T("Priv instruction"); break;
		case STATUS_STACK_OVERFLOW:			Name = T("Stack overflow"); break;
		default:							Name = T("Unknown"); break;
		}

		NodeLocate(Record->ExceptionAddress,DllName,&DllBase);
		FilePrintf(File,T("%s(%08x) at %08x (%s:%08x)"),Name,Record->ExceptionCode,Record->ExceptionAddress,DllName,DllBase);

		if (Record->ExceptionCode == STATUS_ACCESS_VIOLATION)
		{
			if (Record->ExceptionInformation[0])
				Name = T("Write to");
			else
				Name = T("Read from");
			FilePrintf(File,T("\n%s %08x"),Name,Record->ExceptionInformation[1]);
			if (NodeLocate((void*)Record->ExceptionInformation[1],DllName,&DllBase))
				FilePrintf(File,T(" (%s:%08x)"),DllName,DllBase);
		}
		
		// context

		FilePrintf(File,T("\n\ncpu dump:"));

		for (r=Reg;r->Name;++r)
		{
			void* Ptr = *(void**)(Context+r->Ofs);

			FilePrintf(File,T("\n%-5s = %08x"),r->Name,Ptr);

			if (NodeLocate(Ptr,DllName,&DllBase))
				FilePrintf(File,T(" (%s:%08x)"),DllName,DllBase);
		}

		if (r->Ofs >= 0)
		{
			FilePrintf(File,T("\n\nstack dump:"));

			Stack = *(void***)(Context+r->Ofs);
			for (No=0;No<256;++No,++Stack)
			{
				if (!IsBadReadPtr(Stack,sizeof(void*)))
				{
					void* Ptr = *Stack;

					FilePrintf(File,T("\n%08x %08x"),Stack,Ptr);

					if (NodeLocate(Ptr,DllName,&DllBase))
						FilePrintf(File,T(" (%s:%08x)"),DllName,DllBase);
				}
				else
					FilePrintf(File,T("\n%08x ????????"),Stack);
			}
		}

		FilePrintf(File,T("\n\nnode dump:\n"));

		NodeDump(File);

		FileClose(File);
	}

	MessageBox(NULL,String(PLATFORM_ID,PLATFORM_CRASH_MESSAGE),
		String(PLATFORM_ID,PLATFORM_CRASH_TITLE),MB_OK|MB_SETFOREGROUND|MB_TOPMOST|MB_ICONSTOP);

	TerminateProcess(GetCurrentProcess(),1);

	return 1;
}

void SetFileAssociation(const tchar_t* Ext,const tchar_t* Type,bool_t State)
{
	DWORD Disp=0;
	HKEY Key;
	DWORD RegSize;
	DWORD RegType;
	tchar_t Base[64];
	tchar_t Backup[256];

	stprintf(Base,T(".%s"),Ext);
	stprintf(Backup,T("%sBackup"),Type);

	if (RegOpenKeyEx( HKEY_CLASSES_ROOT, Base, 0, ACCESS_SAVE, &Key ) == ERROR_SUCCESS)
	{
		RegSize=sizeof(Base);
		if (RegQueryValueEx( Key, NULL, 0, &RegType, (LPBYTE)Base, &RegSize ) == ERROR_SUCCESS)
		{
			if (tcsicmp(Base,Type)!=0)
			{
				if (State)
				{
					RegSetValueEx( Key, NULL, 0, REG_SZ, (LPBYTE)Type, (tcslen(Type)+1)*sizeof(tchar_t) );
					RegSetValueEx( Key, Backup, 0, REG_SZ, (LPBYTE)Base, RegSize );
				}
			}
			else 
			if (!State)
			{
				RegSize = sizeof(Base);
				if (RegQueryValueEx( Key, Backup, 0, &RegType, (LPBYTE)Base, &RegSize ) != ERROR_SUCCESS)
				{
					Base[0] = 0;
					RegSize = sizeof(tchar_t);
				}

				RegSetValueEx( Key, NULL, 0, REG_SZ, (LPBYTE)Base, RegSize);
			}
		}
		RegCloseKey( Key );
	}
	else 
	if (State)
	{
		if (RegCreateKeyEx( HKEY_CLASSES_ROOT, Base, 0, NULL, 0, ACCESS_SAVE, NULL, &Key, &Disp ) == ERROR_SUCCESS)
		{
			RegSetValueEx( Key, NULL, 0, REG_SZ, (LPBYTE)Type, (tcslen(Type)+1)*sizeof(tchar_t) );
			RegCloseKey( Key );
		}
	}
}

bool_t GetFileAssociation(const tchar_t* Ext,const tchar_t* Type)
{
	bool_t Result = 0;
	HKEY Key;
	tchar_t Base[64];
	stprintf(Base,T(".%s"),Ext);

	if (RegOpenKeyEx( HKEY_CLASSES_ROOT, Base, 0, ACCESS_LOAD, &Key ) == ERROR_SUCCESS)
	{
		tchar_t s[64];
		DWORD RegSize = sizeof(s);
		DWORD RegType;

		if (RegQueryValueEx( Key, NULL, 0, &RegType, (LPBYTE)s, &RegSize ) == ERROR_SUCCESS && RegType == REG_SZ)
			Result = tcsicmp(s,Type)==0;

		RegCloseKey( Key );
	}

	return Result;
}

void SetFileType(const tchar_t* Type, const tchar_t* Open, const tchar_t* Icon )
{
	DWORD Disp=0;
	HKEY Key;
	tchar_t Base[256];

	if (RegCreateKeyEx( HKEY_CLASSES_ROOT, Type, 0, NULL, 0, ACCESS_SAVE, NULL, &Key, &Disp ) == ERROR_SUCCESS)
	{
		RegSetValueEx(Key, NULL, 0, REG_SZ, (LPBYTE)Type, (tcslen(Type)+1)*sizeof(tchar_t));
		RegCloseKey(Key);
	}

	stprintf(Base,T("%s\\DefaultIcon"),Type);
	if (RegCreateKeyEx( HKEY_CLASSES_ROOT, Base, 0, NULL, 0, ACCESS_SAVE, NULL, &Key, &Disp ) == ERROR_SUCCESS)
	{
		RegSetValueEx(Key, NULL, 0, REG_SZ, (LPBYTE)Icon, (tcslen(Icon)+1)*sizeof(tchar_t));
		RegCloseKey(Key);
	}

	stprintf(Base,T("%s\\Shell\\Open\\Command"),Type);
	if (RegCreateKeyEx( HKEY_CLASSES_ROOT, Base, 0, NULL, 0, ACCESS_SAVE, NULL, &Key, &Disp ) == ERROR_SUCCESS)
	{
		RegSetValueEx(Key, NULL, 0, REG_SZ, (LPBYTE)Open, (tcslen(Open)+1)*sizeof(tchar_t));
		RegCloseKey(Key);
	}
}

void TryInvalidate( HWND Wnd, RECT* r0 )
{
	RECT r;

	if (IsWindowVisible(Wnd))
	{
		GetWindowRect(Wnd,&r);
		if (r0->left < r.right && r0->right > r.left &&
			r0->top < r.bottom && r0->bottom > r.top)
		{
			// window is visible and intersects with r0

			HWND Child;
			POINT o;
			o.x = o.y = 0;
			ClientToScreen(Wnd,&o);

			r.left = r0->left - o.x;
			r.right = r0->right - o.x;
			r.top = r0->top - o.y;
			r.bottom = r0->bottom - o.y;
			InvalidateRect(Wnd,&r,0);

			Child = GetWindow(Wnd,GW_CHILD);
			while (Child && IsWindow(Child))
			{
				TryInvalidate(Child,r0);
				Child = GetWindow(Child,GW_HWNDNEXT);
			}
		}
	}
}

BOOL CALLBACK EnumInvalidate( HWND Wnd, LPARAM lParam )
{
	TryInvalidate(Wnd,(RECT*)lParam);
	return 1;
}

void GlobalInvalidate( const rect* Rect )
{
	RECT r;
	r.left = Rect->x;
	r.top = Rect->y;
	r.right = r.left + Rect->Width;
	r.bottom = r.top + Rect->Height;

#if defined(_WIN32_WCE)
	if (FuncRedrawWindow) // doesn't work for mio558???
		FuncRedrawWindow(NULL,&r,NULL,0x0001|0x0080); //RDW_INVALIDATE|RDW_ALLCHILDREN
	else
		EnumWindows(EnumInvalidate,(LPARAM)&r);
#else
	RedrawWindow(NULL,&r,NULL,RDW_INVALIDATE|RDW_ALLCHILDREN);
#endif
}

/*
#include <winioctl.h>

BOOL WINAPI KernelIoControl(
   DWORD dwIoControlCode,
   LPVOID lpInBuf,
   DWORD nInBufSize,
   LPVOID lpOutBuf,
   DWORD nOutBufSize,3
   LPDWORD lpBytesReturned);

#define IOCTL_HAL_REBOOT CTL_CODE(FILE_DEVICE_HAL, 15, METHOD_BUFFERED, FILE_ANY_ACCESS)

void SoftReset()
{
    KernelIoControl(IOCTL_HAL_REBOOT, NULL, 0, NULL, 0, NULL);
}
*/

void HotKeyToString( tchar_t* Out, int HotKey, int OutSize )
{
	int Id = 0;
	
	Out[0] = 0;
	if (HotKey)
	{
		bool_t Keep = (HotKey & HOTKEY_KEEP) != 0;
		if ((HotKey & HOTKEY_MASK) < 0xC1 || (HotKey & HOTKEY_MASK) > 0xCF)
		{
			if (HotKey & HOTKEY_WIN) { tcscat(Out,String(PLATFORM_ID,PLATFORM_KEY_WIN)); tcscat(Out,T("+")); }
			if (HotKey & HOTKEY_SHIFT) { tcscat(Out,String(PLATFORM_ID,PLATFORM_KEY_SHIFT)); tcscat(Out,T("+")); }
			if (HotKey & HOTKEY_CTRL) { tcscat(Out,String(PLATFORM_ID,PLATFORM_KEY_CTRL)); tcscat(Out,T("+")); }
			if (HotKey & HOTKEY_ALT) { tcscat(Out,String(PLATFORM_ID,PLATFORM_KEY_ALT)); tcscat(Out,T("+")); }
		}
		HotKey &= HOTKEY_MASK;

		if ((HotKey >= '0' && HotKey <= '9') ||
			(HotKey >= 'A' && HotKey <= 'Z'))
			stprintf(Out+tcslen(Out),T("%c"),HotKey);
		else
		if (HotKey >= 0x70 && HotKey < 0x80)
			stprintf(Out+tcslen(Out),T("F%d"),HotKey-0x70+1);
		else
		switch (HotKey)
		{
		case 0x86:
			Id = PLATFORM_KEY_ACTION;
			break;
		case VK_RETURN:
			Id = PLATFORM_KEY_ENTER;
			break;
		case VK_ESCAPE:
			Id = PLATFORM_KEY_ESCAPE;
			break;
		case VK_LEFT:
			Id = PLATFORM_KEY_LEFT;
			break;
		case VK_RIGHT:
			Id = PLATFORM_KEY_RIGHT;
			break;
		case VK_UP:
			Id = PLATFORM_KEY_UP;
			break;
		case VK_DOWN:
			Id = PLATFORM_KEY_DOWN;
			break;
		case VK_SPACE:
			Id = PLATFORM_KEY_SPACE;
			break;
		case 0xC1:
		case 0xC2:
		case 0xC3:
		case 0xC4:
		case 0xC5:
		case 0xC6:
		case 0xC7:
		case 0xC8:
		case 0xC9:
		case 0xCA:
		case 0xCB:
		case 0xCC:
		case 0xCD:
		case 0xCE:
		case 0xCF:
			stprintf(Out+tcslen(Out),String(PLATFORM_ID,PLATFORM_KEY_APP),HotKey-0xC0);
			break;
		default:
			stprintf(Out+tcslen(Out),T("#%02X"),HotKey);
			break;
		}
		if (Id)
			tcscat(Out,String(PLATFORM_ID,Id));
		if (Keep) { tcscat(Out,T("*")); }
	}
}

void WaitBegin()
{
	//DEBUG_MSG(DEBUG_SYS,T("WaitBegin"));
	SetCursor(LoadCursor (NULL, IDC_WAIT));
	WaitCursor = 1;
}

void WaitEnd()
{
	//DEBUG_MSG(DEBUG_SYS,T("WaitEnd"));
	SetCursor(LoadCursor (NULL, IDC_ARROW));
	WaitCursor = 0;
}

const tchar_t* GetDocumentPath() { return DocPath; }
const tchar_t* GetSystemPath() { return SysPath; }

#endif
