/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: mjpeg.c 11 2004-07-15 13:46:20Z picard $
 *
 * BetaPlayer MJPEG plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/
 
#include "stdafx.h"

static mjpeg_decode MJPEG;

const datadef MJPEGParams[] = 
{
	{ MJPEG_ID,		TYPE_CLASS },
	{ MJPEG_INPUT,	TYPE_PACKET, DF_INPUT, PACKET_VIDEO },
	{ MJPEG_IDCT,	TYPE_NODE, DF_OUTPUT, IDCT_CLASS },

	DATADEF_END
};

void MJPEG_SetEmpty( mjpeg_decode* p )
{
	p->Show = -1;
	if (p->IDCT)
		p->IDCT->Drop(p->IDCT);
}

static int MJPEG_Packet( mjpeg_decode* p, const packet* Packet )
{
	int Result;

	if (p->BufCount==0)
		return ERR_NOT_SUPPORTED;

	if (Packet->DropLevel > 1)
		MJPEG_SetEmpty(p);

	p->Packet.CurrTime = Packet->CurrTime; 

	if (p->Show>=0)
	{
		// pending frame
		
		Result = p->IDCT->Send(p->IDCT,&p->Packet,&p->Dropped);
		if (Result == ERR_BUFFER_FULL)
			return Result;

		p->Show = -1;
	}

	p->Packet.RefTime = Packet->RefTime;

	if (Packet->Length == 0)
		return ERR_INVALID_DATA;

	if (Packet->DropLevel)
	{
		p->Total++;
		p->Dropped++;
		return ERR_NONE;
	}

	p->Total++;

	Result = MJPEG_Decode(p,Packet->Data[0],Packet->Length);

	if (p->Show>=0)
	{
		if (p->IDCT->Send(p->IDCT,&p->Packet,&p->Dropped) != ERR_BUFFER_FULL)
			p->Show = -1;
	}

	return Result;
}

bool_t MJPEG_IDCTInit( mjpeg_decode* p, idct* IDCT )
{
	int Count;
	if (IDCT)
	{
		if (IDCT->Set(IDCT,IDCT_OUTPUT|PACKET_FORMAT,&p->Format,sizeof(p->Format)) != ERR_NONE)
			return 0;

		Count = 1;
		IDCT->Set(IDCT,IDCT_BUFFERCOUNT,&Count,sizeof(int));
		IDCT->Drop(IDCT);

		if (IDCT->Get(IDCT,IDCT_BUFFERCOUNT,&Count,sizeof(int)) != ERR_NONE || Count<1)
			return 0;
	}
	return 1;
}

int MJPEG_UpdateFormat( mjpeg_decode* p, int Width, int Height, int Flags)
{
	if (p->Format.Width != Width || p->Format.Height != Height || p->Format.Pixel.Flags != Flags)
	{
		p->Format.Width = Width;
		p->Format.Height = Height;
		p->Format.Pixel.Flags = Flags;

		if (p->IDCT)
		{
			if (!MJPEG_IDCTInit(p,p->IDCT))
			{
				p->BufCount = 0;
				return ERR_NOT_SUPPORTED;
			}
			p->BufCount = 1;
		}
	}
	return ERR_NONE;
}

static int MJPEG_SetIDCT( mjpeg_decode* p, idct* IDCT )
{
	int Format = IDCTFORMAT_BLOCK;
	if (IDCT && IDCT->Set(IDCT,IDCT_FORMAT,&Format,sizeof(Format)) != ERR_NONE)
		return ERR_NOT_SUPPORTED;

	if (p->IDCT)
	{
		if (IDCTCopy(p->IDCT,IDCT,1) != ERR_NONE)
			return ERR_NOT_SUPPORTED;
	}
	else
	{
		if (!MJPEG_IDCTInit(p,IDCT))
			return ERR_NOT_SUPPORTED;
	}

	p->BufCount = IDCT ? 1:0;
	p->IDCT = IDCT;
	return ERR_NONE;
}

bool_t MJPEG_IsEmpty( mjpeg_decode* p )
{
	if (!p->IDCT) 
		return 1;

	if (p->Show>=0)
	{
		packet Zero;

		Zero.CurrTime = -1;
		Zero.DropLevel = p->Packet.DropLevel;
		Zero.Length = 0;
		Zero.RefTime = -1;

		MJPEG_Packet(p,&Zero);
	}
	return p->Show<0;
}

int MJPEG_Get( mjpeg_decode* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(MJPEG_ID,int); break;
	case FLOW_TOTAL:GETVALUE(p->Total,int); break;
	case FLOW_DROPPED: GETVALUE(p->Dropped,int); break;
	case FLOW_EMPTY: GETVALUE(MJPEG_IsEmpty(p),bool_t); break;
	case MJPEG_INPUT|PACKET_FORMAT: GETVALUE(p->Format,video); break;
	case MJPEG_IDCT: GETVALUE(p->IDCT,idct*); break;
	}
	return Result;
}

int MJPEG_Set( mjpeg_decode* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case MJPEG_IDCT:
		if (Size == sizeof(idct*) && NodeIsClass(*(node**)Data,IDCT_CLASS))
			Result = MJPEG_SetIDCT(p,*(idct**)Data);
		break;

	case MJPEG_INPUT | PACKET_FORMAT:
		if (!Size || Size == sizeof(video))
		{
			if (!Size) Data = NULL;
			Result = MJPEG_SetFormat(p,(const video*)Data);
		}
		break;

	case MJPEG_INPUT:
		if (Size == sizeof(packet))
			Result = MJPEG_Packet(p,(const packet*)Data);
		break;

	case FLOW_TOTAL: SETVALUE(p->Total,int,ERR_NONE); break;
	case FLOW_DROPPED: SETVALUE(p->Dropped,int,ERR_NONE); break;

	case FLOW_EMPTY:
		MJPEG_SetEmpty(p);
		Result = ERR_NONE;
		break;
	}
	return Result;
}

int MJPEG_Enum( mjpeg_decode* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,FlowParams,MJPEGParams,NULL);
}

void MJPEG_Init()
{
	mjpeg_decode* p = &MJPEG;
	NodeFill(&p->VMT,sizeof(MJPEG),(nodeenum)MJPEG_Enum,(nodeget)MJPEG_Get,(nodeset)MJPEG_Set);
	NodeRegister(&p->VMT,PRI_DEFAULT);
}

void MJPEG_Done()
{
	mjpeg_decode* p = &MJPEG;
	
	MJPEG_SetIDCT(p,NULL);
	MJPEG_SetFormat(p,NULL);
	NodeUnRegister(&p->VMT);
}
