/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: mjpeg_decode.h 11 2004-07-15 13:46:20Z picard $
 *
 * BetaPlayer MJPEG plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 * defines from MJPEG encoder and decoder in libavcodec
 * Copyright (c) 2000, 2001 Fabrice Bellard. 
 *
 ****************************************************************************/
 
#ifndef __MJPEG_DECODE_H
#define __MJPEG_DECODE_H

#define SOF0   0xc0	// baseline
#define SOF1   0xc1	// extended sequential huffman
#define SOF2   0xc2	// progressive huffman
#define SOF3   0xc3	// lossless huffman

#define SOF5   0xc5	// differential sequential huffman
#define SOF6   0xc6	// differential progressive huffman
#define SOF7   0xc7	// differential lossless huffman
#define JPG    0xc8	// reserved for JPEG extension
#define SOF9   0xc9	// extended sequential arithmetic
#define SOF10  0xca	// progressive arithmetic
#define SOF11  0xcb	// lossless arithmetic

#define SOF13  0xcd	// differential sequential arithmetic
#define SOF14  0xce	// differential progressive arithmetic
#define SOF15  0xcf	// differential lossless arithmetic

#define DHT    0xc4	// define huffman tables

#define DAC    0xcc	// define arithmetic-coding conditioning

#define RST0   0xd0	// restart with modulo 8 count "m" 
#define RST1   0xd1
#define RST2   0xd2
#define RST3   0xd3
#define RST4   0xd4
#define RST5   0xd5
#define RST6   0xd6
#define RST7   0xd7

#define SOI    0xd8	// start of image
#define EOI    0xd9	// end of image
#define SOS    0xda	// start of scan
#define DQT    0xdb	// define quantization tables
#define DNL    0xdc	// define number of lines
#define DRI    0xdd	// define restart interval
#define DHP    0xde	// define hierarchical progression
#define EXP    0xdf	// expand reference components

#define APP0   0xe0
#define APP1   0xe1
#define APP2   0xe2
#define APP3   0xe3
#define APP4   0xe4
#define APP5   0xe5
#define APP6   0xe6
#define APP7   0xe7
#define APP8   0xe8
#define APP9   0xe9
#define APP10  0xea
#define APP11  0xeb
#define APP12  0xec
#define APP13  0xed
#define APP14  0xee
#define APP15  0xef

#define JPG0   0xf0
#define JPG1   0xf1
#define JPG2   0xf2
#define JPG3   0xf3
#define JPG4   0xf4
#define JPG5   0xf5
#define JPG6   0xf6
#define JPG7   0xf7
#define JPG8   0xf8
#define JPG9   0xf9
#define JPG10  0xfa
#define JPG11  0xfb
#define JPG12  0xfc
#define JPG13  0xfd

#define COM    0xfe	// comment

#define TEM    0x01	// temporary private use for arithmetic coding

typedef struct blockindex
{
	uint8_t quant;
	uint8_t comp;
	uint8_t ac;
	uint8_t dc;
} blockindex;

typedef struct mjpeg_decode 
{
	node VMT;
	packet Packet;
	int Total;
	int Dropped;
	idct *IDCT;
	video Format;
	bool_t ErrorShowed;
	int Show;
	int BufCount;

	merge buffer;
	int restart_interval;
	int restart_count;
	idct_block_t* blockptr;

	// bitstream
	bitstream s;
	vlc v[2][4];
    uint8_t quant[4][64];

	int hblock[4];
	int vblock[4];
	int compid[4];
	int quantidx[4];

	int blocks;
	int mb_width;
	int mb_height;
	int last_dc[4];
	int comp;

	blockindex *indexend;
	blockindex index[4*2*2];
	uint8_t zigzag[64];
	IDCT_BLOCK_DECLARE

} mjpeg_decode;

extern int MJPEG_Decode( mjpeg_decode*, const uint8_t* ptr, int len );
extern void MJPEG_VLC(vlc *v, const uint8_t *bits, const uint8_t *val);
extern int MJPEG_UpdateFormat( mjpeg_decode* p, int Width, int Height, int Flags);
extern int MJPEG_SetFormat( mjpeg_decode* p, const video* Format );

#endif
