/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: mp4_mblock_bvop.c 3 2004-07-13 11:26:13Z picard $
 *
 * BetaPlayer MPEG4 plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "stdafx.h"

#ifdef MPEG4

static void blockDirect( mp4_decode *dec, int pos, int dmv )
{
	int fmv[6];
	int bmv[6];

	int* mv = &dec->mv_buf[pos*4];
	
	int j;
	for (j=0;j<4;++j,++mv) {
		
		int64_t v;
		int dx,dy;

		v = dec->TRB * MVX(*mv);
		if ((int)(v>>32)<0) //best code for (v<0)
			dx = (int)((v-1) >> 32);
		else
			dx = (int)(v >> 32);
		if (dx<0) ++dx;

		v = dec->TRB * MVY(*mv);
		if ((int)(v>>32)<0) //best code for (v<0)
			dy = (int)((v-1) >> 32);
		else
			dy = (int)(v >> 32);
		if (dy<0) ++dy;

		dx += MVX(dmv);
		dy += MVY(dmv);

		fmv[j] = MAKEMV(dx,dy);
		
		if (dmv & 0xFFFF)
			dx -= MVX(*mv);
		else
		{
			v = dec->TRB_TRD * MVX(*mv);
			if ((int)(v>>32)<0) //best code for (v<0)
				dx = (int)((v-1) >> 32);
			else
				dx = (int)(v >> 32);
			if (dx<0) ++dx;
		}

		if (dmv & 0xFFFF0000)
			dy -= MVY(*mv);
		else
		{
			v = dec->TRB_TRD * MVY(*mv);
			if ((int)(v>>32)<0) //best code for (v<0)
				dy = (int)((v-1) >> 32);
			else
				dy = (int)(v >> 32);
			if (dy<0) ++dy;
		}

		bmv[j] = MAKEMV(dx,dy);
	}

	fmv[5]=fmv[4]=mvchroma4(fmv);
	bmv[5]=bmv[4]=mvchroma4(bmv);

	dec->idct->MComp8x8(dec->idct,bmv,fmv);

	blockInter(dec);
}

void BVOP_frame( mp4_decode *dec )
{
	int fmv[6];
	int bmv[6];
	int pos;

	dec->resync_marker_bits = 16 +
		(dec->fcode_back > dec->fcode_for ? dec->fcode_back : dec->fcode_for);

	for (pos=0;pos<dec->pos_end;pos+=MB_X-dec->mb_xsize) {

		int n;
		int fprev=0, bprev=0;

		for (;POSX(pos)<dec->mb_xsize;++pos) {

			inlineloadbits(dec);

			if (eofbits(dec))
				return;

			if (resync_marker(dec))
			{
				pos = resync(dec);
				if (pos>=dec->pos_end)
					return;
				fprev = bprev = 0;
			}

			n = dec->mapofs+(dec->framemap[pos] >> 1);

			// did the last refframe touch this block?
			if (dec->lastrefframe > n)
			{
				// no change: this block is the same in both backward and forward buffer

				// copy needed?
				if (dec->bufframe < n && dec->bufframe >= -dec->lastrefframe)
					dec->idct->Copy16x16(dec->idct,POSX(pos),POSY(pos),1);
			}
			else
			{
				//stream: 24bits available

			 	dec->idct->Process(dec->idct,POSX(pos),POSY(pos));

				if (!getbits1(dec))
				{
					int mb_type;
					n = getbits(dec,1);

					for (mb_type=0; mb_type<=3; ++mb_type)
						if (getbits1(dec))
							break;

					//stream: 19bits available

					if (!n)
						dec->cbp = getbits(dec,6);
					else
						dec->cbp = 0;

					//stream: 13bits available

					if (mb_type != DIRECT)
					{
						if (dec->cbp && getbits1(dec))
						{
							int q = dec->quantizer + (getbits1(dec) ? 2:-2);
							dec->quantizer = q<1 ? 1: (q>31 ? 31:q);
						}

						switch (mb_type) 
						{
						case BACKWARD:
							bmv[3]=bmv[2]=bmv[1]=bmv[0]=bprev = getMV( dec->fcode_back, bprev, dec );
							bmv[5]=bmv[4]=mvchroma(bprev);
							dec->idct->MComp16x16(dec->idct,bmv,NULL);
							blockInter(dec);
							break;

						case FORWARD:
							fmv[3]=fmv[2]=fmv[1]=fmv[0]=fprev = getMV( dec->fcode_for, fprev, dec );
							fmv[5]=fmv[4]=mvchroma(fprev);
							dec->idct->MComp16x16(dec->idct,NULL,fmv);
							blockInter(dec);
							break;

						default: //case INTERPOLATE:
							fmv[3]=fmv[2]=fmv[1]=fmv[0]=fprev = getMV( dec->fcode_for, fprev, dec );
							fmv[5]=fmv[4]=mvchroma(fprev);
							bmv[3]=bmv[2]=bmv[1]=bmv[0]=bprev = getMV( dec->fcode_back, bprev, dec );
							bmv[5]=bmv[4]=mvchroma(bprev);
							dec->idct->MComp16x16(dec->idct,bmv,fmv);
							blockInter(dec);
							break;
						}
					}
					else
					{
						blockDirect(dec,pos,getMV( 1, 0, dec ));
					}
				} 
				else {
					dec->cbp = 0;
					blockDirect(dec,pos,0);
				}
			}
		}
	}
}

#endif
