/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: avi.c 117 2004-11-28 02:45:11Z picard $
 *
 * BetaPlayer MPEG4AAC plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "../common/common.h"
#include "mpeg4aac.h"

typedef struct aac
{
	format_base p;
	int Head;
	int SeekPos;

} aac;

static int Init(aac* p)
{
	format_stream* s = Format_AddStream(&p->p,sizeof(format_stream));
	p->Head = 0;
	if (s)
	{
		audio* Audio = Format_StreamFormat(s,PACKET_AUDIO,sizeof(audio));
		if (Audio)
			Audio->Format = AUDIOFMT_AAC;

		s->Fragmented = 1;
		s->DisableDrop = 1;

		Format_PrepairStream(&p->p,s);

		if (s->Comment.Node)
		{
			// id3v1
			char Buffer[ID3TAG1SIZE];
			char Head[ID3TAGQUERYSIZE];
			int Save = p->p.Input->Seek(p->p.Input,0,SEEK_CUR);
			if (Save>=0 && p->p.Input->Seek(p->p.Input,-(int)sizeof(Buffer),SEEK_END)>=0)
			{
				if (p->p.Input->Read(p->p.Input,Buffer,sizeof(Buffer)) == sizeof(Buffer))
					Id3TagParse(Buffer,sizeof(Buffer),&s->Comment);

				p->p.Input->Seek(p->p.Input,Save,SEEK_SET);
			}

			// id3v2
			if (Format_Read(&p->p,Head,sizeof(Head))==sizeof(Head))
			{
				int TagSize = Id3TagQuery(Head,sizeof(Head));
				if (TagSize > 0)
				{
					void* Buffer = Alloc(TagSize);
					if (Buffer)
					{
						int Len = Format_Read(&p->p,Buffer,TagSize);
						Id3TagParse(Buffer,Len,&s->Comment);
						Free(Buffer);
					}
					p->Head = TagSize;
				}
			}

			Format_Seek(&p->p,p->Head,SEEK_SET);
		}
	}
	p->SeekPos = p->Head;
	return ERR_NONE;
}

static int Probe(aac* p, const void* Ptr, int Length, const tchar_t* URL, const tchar_t* ContentType )
{
	if (CheckExts(URL,p->p.Exts))
		return ERR_NONE;

	return ERR_INVALID_DATA;
}

static int ReadPacket(aac* p, format_packet* Packet)
{
	format_stream* Stream = p->p.Streams[0];
	audio* Format = (audio*)Stream->Format;

	if (p->p.FilePos<=p->Head)
		Packet->RefTime = 0;
	else
	if (Format->ByteRate)
		Packet->RefTime = Scale(p->p.FilePos-p->Head,TICKSPERSEC,Format->ByteRate);
	else
	if (p->p.FilePos<=p->SeekPos)
		Packet->RefTime = 1; // no timing possible, but don't want to return 0 either
	else
		Packet->RefTime = -1;

	Packet->Data = Format_ReadAsRef(&p->p,-CHUNKSIZE);
	Packet->Stream = Stream;
	return ERR_NONE;
}

static int Seek(aac* p, tick_t Time, int FilePos )
{
	format_stream* Stream = p->p.Streams[0];

	if (FilePos < 0)
	{
		if (Time > 0)
		{
			audio* Format = (audio*)Stream->Format;
			if (Format->ByteRate)
				FilePos = p->Head + Scale(Time,Format->ByteRate,TICKSPERSEC);
			else
				return ERR_NOT_SUPPORTED;
		}
		else
			FilePos = 0;
	}

	p->SeekPos = FilePos;
	return Format_Seek(&p->p,FilePos,SEEK_SET);
}

static aac AAC;

void AAC_Init()
{
	AAC.p = FormatBase;
	AAC.p.VMT.Probe = Probe;
	AAC.p.Init = Init;
	AAC.p.Seek = Seek;
	AAC.p.ReadPacket = ReadPacket;
	AAC.p.Id = AAC_ID;
	AAC.p.Exts = T("aac:A");

	NodeRegister((node*)&AAC.p.VMT,PRI_DEFAULT);
}

void AAC_Done()
{
	NodeUnRegister((node*)&AAC.p.VMT);
}

