/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: mp3.c 144 2004-12-12 15:28:39Z picard $
 *
 * BetaPlayer MPEGAudio plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "..\common\common.h"
#include "mpegaudio.h"

typedef struct mp3
{
	format_base p;

} mp3;

static int Init(mp3* p)
{
	format_stream* s = Format_AddStream(&p->p,sizeof(format_stream));
	if (s)
	{
		audio* Audio = Format_StreamFormat(s,PACKET_AUDIO,sizeof(audio));
		if (Audio)
			Audio->Format = AUDIOFMT_MP3;

		s->Fragmented = 1;
		s->DisableDrop = 1; // VBR MP3 could drop because timestamp and duration calculation is wrong

		Format_PrepairStream(&p->p,s);

		if (s->Comment.Node)
		{
			char Buffer[ID3TAG1SIZE];
			int Save = p->p.Input->Seek(p->p.Input,0,SEEK_CUR);
			if (Save>=0 && p->p.Input->Seek(p->p.Input,-(int)sizeof(Buffer),SEEK_END)>=0)
			{
				if (p->p.Input->Read(p->p.Input,Buffer,sizeof(Buffer)) == sizeof(Buffer))
					Id3TagParse(Buffer,sizeof(Buffer),&s->Comment);

				p->p.Input->Seek(p->p.Input,Save,SEEK_SET);
			}
		}
	}

	return ERR_NONE;
}

static int Probe(mp3* p, const void* Ptr, int Length, const tchar_t* URL, const tchar_t* ContentType )
{
	/* conflicts with AAC
	const uint8_t* Data = (const uint8_t*)Ptr;
	int Tag = Id3TagQuery(Ptr,Length);
	if (Tag<0) Tag=0;

	if (Tag+1<Length && Data[Tag] == 0xFF && (Data[Tag+1] & 0xE0) == 0xE0)
		return ERR_NONE;
	*/

	if (Length>7 && memcmp(Ptr,"ICY 200",7)==0) //!!! temporary solution
		return ERR_NONE;

	if (tcsicmp(ContentType,T("audio//mpeg"))==0)
		return ERR_NONE;

	if (CheckExts(URL,p->p.Exts))
		return ERR_NONE;

	return ERR_INVALID_DATA;
}

int Seek(mp3* p, tick_t Time, int FilePos )
{
	format_stream* Stream = p->p.Streams[0];

	if (FilePos < 0)
	{
		if (Time > 0)
		{
			audio* Format = (audio*)Stream->Format;
			if (Format->ByteRate)
				FilePos = Scale(Time,Format->ByteRate,TICKSPERSEC);
			else
				return ERR_NOT_SUPPORTED;
		}
		else
			FilePos = 0;
	}

	return Format_Seek(&p->p,FilePos,SEEK_SET);
}

int ReadPacket(mp3* p, format_packet* Packet)
{
	format_stream* Stream = p->p.Streams[0];
	audio* Format = (audio*)Stream->Format;

	if (p->p.FilePos==0)
		Packet->RefTime = 0;
	else
	if (Format->ByteRate)
		Packet->RefTime = Scale(p->p.FilePos,TICKSPERSEC,Format->ByteRate);
	else
		Packet->RefTime = -1;

	Packet->Data = Format_ReadAsRef(&p->p,-CHUNKSIZE);
	Packet->Stream = Stream;
	return ERR_NONE;
}

static mp3 MP3;

void MP3_Init()
{
	MP3.p = FormatBase;
	MP3.p.VMT.Probe = Probe;
	MP3.p.Init = Init;
	MP3.p.Seek = Seek;
	MP3.p.ReadPacket = ReadPacket;
	MP3.p.Id = MP3_ID;
	MP3.p.Exts = T("mp1:A;mp2:A;mp3:A;mpa:A");

	NodeRegister((node*)&MP3.p.VMT,PRI_DEFAULT);
}

void MP3_Done()
{
	NodeUnRegister((node*)&MP3.p.VMT);
}
