/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: mves.c 11 2004-07-15 13:46:20Z picard $
 *
 * BetaPlayer MPEGVideo plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "stdafx.h"

typedef struct mves
{
	format_base p;
	int LastSeekPos;
	tick_t LastSeekTime;
	bool_t MP4;

} mves;

static int Init(mves* p)
{
	format_stream* s = Format_AddStream(&p->p,sizeof(format_stream));
	if (s)
	{
		video* Video = (video*) Format_StreamFormat(s,PACKET_VIDEO,sizeof(video));
		Video->Pixel.Flags = PF_FOURCC | PF_FRAGMENTED;
		Video->Pixel.FourCC = p->MP4 ? FOURCC('M','P','4','V') : FOURCC('M','P','E','G');
		Video->FrameRate.Num = 25; //fixed frame rate (no codec timing support)
		Video->FrameRate.Den = 1;

		s->Fragmented = 1;
		Format_PrepairStream(&p->p,s);
	}
	p->LastSeekPos = 0;
	p->LastSeekTime = 0;

	return ERR_NONE;
}

static int Probe(mves* p, const void* Ptr, int Length, const tchar_t* URL, const tchar_t* ContentType )
{
	bool_t Ok = 0;
	const uint8_t* Data = (const uint8_t*)Ptr;
	int Code=-1;
    int i;

	for (i=0;i<20 && i<Length;i++)
	{
		Code = (Code << 8) | Data[i];
		if ((Code & ~0xFF) == 0x100) 
		{
			if (Code == PACK_START_CODE ||
				Code == SYSTEM_HEADER_START_CODE ||
				(Code >= 0x1E0 && Code <= 0x1EF) ||
				(Code >= 0x1C0 && Code <= 0x1DF) ||
				Code == PRIVATE_STREAM_2 ||
				Code == PROGRAM_STREAM_MAP ||
				Code == PRIVATE_STREAM_1 ||
				Code == PADDING_STREAM)
			{
				Ok = 0;
				break;
			}

			if (Code == SEQ_START_CODE)
				Ok = 1;

			break;
		}
	}

	if (Ok)
		return ERR_NONE;
	
	if (CheckExts(URL,p->p.Exts))
	{
		p->MP4 = CheckExts(URL,T("m4v:V")) != 0;
		return ERR_NONE;
	}

	return ERR_INVALID_DATA;
}

int Seek(mves* p, tick_t Time, int FilePos )
{
	video Format;
	format_stream* Stream = p->p.Streams[0];

	Format.ByteRate = 0;
	if (Stream->Pin.Node)
		Stream->Pin.Node->Get(Stream->Pin.Node,Stream->Pin.No|PACKET_FORMAT,&Format,sizeof(video));

	if (FilePos < 0)
	{
		if (Time > 0)
		{
			if (Format.ByteRate)
				FilePos = Scale(Time,Format.ByteRate,TICKSPERSEC);
			else
				return ERR_NOT_SUPPORTED;
		}
		else
			FilePos = 0;
	}
	else
	if (Format.ByteRate)
		Time = Scale(FilePos,TICKSPERSEC,Format.ByteRate);
	else
		Time = 1; //don't care (just not zero)

	p->LastSeekPos = FilePos;
	p->LastSeekTime = Time;

	return Format_Seek(&p->p,FilePos,SEEK_SET);
}

int ReadPacket(mves* p, format_packet* Packet)
{
	format_stream* Stream = p->p.Streams[0];

	if (p->p.FilePos==0)
		Packet->RefTime = 0;
	else
	if (p->p.FilePos==p->LastSeekPos)
		Packet->RefTime = p->LastSeekTime;
	else
		Packet->RefTime = -1;

	Packet->Data = Format_ReadAsRef(&p->p,-CHUNKSIZE);
	Packet->Stream = Stream;
	return ERR_NONE;
}

static mves MVES;

void MVES_Init()
{
	MVES.p = FormatBase;
	MVES.p.VMT.Probe = (fmtprobe)Probe;
	MVES.p.Init = (fmtfunc)Init;
	MVES.p.Seek = (fmtseek)Seek;
	MVES.p.ReadPacket = (fmtreadpacket)ReadPacket;
	MVES.p.Id = MVES_ID;
	MVES.p.Exts = T("m1v:V;m4v:V");

	NodeRegister((node*)&MVES.p.VMT,PRI_DEFAULT+1); // should come before mpeg
}

void MVES_Done()
{
	NodeUnRegister((node*)&MVES.p.VMT);
}
