/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: overlay_base.c 142 2004-12-12 07:30:52Z picard $
 *
 * BetaPlayer Overlay plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "stdafx.h"

void OvlInvalidate( overlay_base* p, const rect* Rect, bool_t Local )
{
	POINT o;
	RECT r;

	if (Rect->Height>0 && Rect->Width>0)
	{
		DEBUG_MSG5(DEBUG_VIDEO,T("Invalidate %d,%d,%d,%d (%d)"),Rect->x,Rect->y,Rect->Width,Rect->Height,Local);

		if (Local)
		{
			// only local invalidate
			o.x = 0;
			o.y = 0;
			ClientToScreen(p->Wnd,&o);

			DEBUG_MSG2(DEBUG_VIDEO,T("Invalidate local offset %d,%d"),o.x,o.y);

			r.left = Rect->x - o.x;
			r.top = Rect->y - o.y;
			r.right = r.left + Rect->Width;
			r.bottom = r.top + Rect->Height;
			InvalidateRect(p->Wnd,&r,0);
		}
		else
			GlobalInvalidate(Rect);
	}
}

void OvlValidate( overlay_base* p, const rect* Rect )
{
	POINT o;
	RECT r;

	if (Rect->Height>0 && Rect->Width>0)
	{
		DEBUG_MSG4(DEBUG_VIDEO,T("Validate %d,%d,%d,%d"),Rect->x,Rect->y,Rect->Width,Rect->Height);

		o.x = 0;
		o.y = 0;
		ClientToScreen(p->Wnd,&o);

		DEBUG_MSG2(DEBUG_VIDEO,T("Validate offset %d,%d"),o.x,o.y);

		r.left = Rect->x - o.x;
		r.top = Rect->y - o.y;
		r.right = r.left+Rect->Width;
		r.bottom = r.top+Rect->Height;

		ValidateRect(p->Wnd,&r);
	}
} 

void OvlInit(overlay_base* p,int Id,int UseDLL,int Priority)
{
	if (UseDLL && !p->DLL)
		return;

	p->Soft = NULL;
	p->Id = Id;
	p->Primary = 1;
	p->Overlay = 0;

	memset(&p->OrigFX,0,sizeof(blitfx));
	p->OrigFX.ScaleX = SCALE_ONE;
	p->OrigFX.ScaleY = SCALE_ONE;
	p->AutoPrerotate = 1;
	p->ColorKey = -1;
	p->Aspect.Num = p->Aspect.Den = 1;

	NodeRegister(&p->VMT,Priority);
}

void OvlDone(overlay_base* p)
{
	BlitRelease(p->Soft);
	NodeUnRegister(&p->VMT);

	if (p->DLL)
		FreeLibrary(p->DLL);
	p->DLL = NULL;
}

int OvlEnum(overlay_base* p,int No,datadef* DataType)
{
	return NodeEnumType(&No,DataType,NodeParams,FlowParams,OutputParams,VOutputParams,p->Params,NULL);
}

void OvlClearBorder( overlay_base* p )
{
	planes Planes;
	if (p->Lock(p,Planes,0) == ERR_NONE)
	{
		uint32_t c = RGBToFormat(0,&p->Output.Pixel);
		rect Viewport;
		VirtToPhy(&p->Viewport,&Viewport,&p->Output);

		FillColor(Planes[0],p->Output.Pitch,Viewport.x,Viewport.y,
			Viewport.Width,p->DstAlignedRect.y-Viewport.y,
			p->Output.Pixel.BitCount,c);

		FillColor(Planes[0],p->Output.Pitch,Viewport.x,p->DstAlignedRect.y,
			p->DstAlignedRect.x-Viewport.x,p->DstAlignedRect.Height,
			p->Output.Pixel.BitCount,c);

		FillColor(Planes[0],p->Output.Pitch,p->DstAlignedRect.x+p->DstAlignedRect.Width,p->DstAlignedRect.y,
			(Viewport.x+Viewport.Width)-(p->DstAlignedRect.x+p->DstAlignedRect.Width),p->DstAlignedRect.Height,
			p->Output.Pixel.BitCount,c);

		FillColor(Planes[0],p->Output.Pitch,Viewport.x,p->DstAlignedRect.y+p->DstAlignedRect.Height,
			Viewport.Width,(Viewport.y+Viewport.Height)-(p->DstAlignedRect.y+p->DstAlignedRect.Height),
			p->Output.Pixel.BitCount,c);

		p->Unlock(p);
	}
}

int OvlUpdateAlign( overlay_base* p )
{
	rect OldGUI = p->GUIAlignedRect;
	rect Old = p->DstAlignedRect;
	rect FullScreen;

	DEBUG_MSG4(DEBUG_TEST,T("BLIT Viewport:%d %d %d %d"),p->Viewport.x,p->Viewport.y,p->Viewport.Width,p->Viewport.Height);

	PhyToVirt(NULL,&FullScreen,&p->Output);
	p->FullScreenViewport = EqRect(&FullScreen,&p->Viewport);

	VirtToPhy(&p->Viewport,&p->DstAlignedRect,&p->Output);
	VirtToPhy(NULL,&p->SrcAlignedRect,&p->Input);

	BlitRelease(p->Soft);
	p->Soft = BlitCreate(&p->Output,&p->Input,&p->FX,&p->Caps);
	if (p->SetFX & BLITFX_AVOIDTEARING) // only suggest input dir avoid tearing set
		p->InputHintDir = p->Input.Direction ^ p->FX.Direction;

	BlitAlign(p->Soft,&p->DstAlignedRect, &p->SrcAlignedRect );
	PhyToVirt(&p->DstAlignedRect,&p->GUIAlignedRect,&p->Output);

	//DEBUG_MSG4(DEBUG_TEST,T("BLIT DstRect:%d %d %d %d"),p->DstAlignedRect.x,p->DstAlignedRect.y,p->DstAlignedRect.Width,p->DstAlignedRect.Height);
	//DEBUG_MSG4(DEBUG_TEST,T("BLIT SrcRect:%d %d %d %d"),p->SrcAlignedRect.x,p->SrcAlignedRect.y,p->SrcAlignedRect.Width,p->SrcAlignedRect.Height);

	if (!EqRect(&Old,&p->DstAlignedRect) && p->Show && p->Primary)
	{
		OvlInvalidate(p,&OldGUI,0);
		OvlInvalidate(p,&p->Viewport,1);
		OvlValidate(p,&p->GUIAlignedRect);
	}

	if (p->FullScreenViewport || !p->Primary)
		OvlClearBorder(p);

	return ERR_NONE;
}

int OvlGet(overlay_base* p,int No,void* Data,int Size)
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(p->Id,int); break;
	case VOUT_PRIMARY: GETVALUE(p->Primary,bool_t); break;
	case VOUT_OVERLAY: GETVALUE(p->Overlay,bool_t); break;
	case VOUT_INPUT|PACKET_FORMAT: GETVALUE(p->Input,video); break;
	case VOUT_INPUT|PACKET_PREF_FORMAT: GETVALUE(p->PrefInput,video); break;
	case VOUT_OUTPUT|PACKET_FORMAT: GETVALUE(p->Output,video); break;
	case VOUT_ACCEL_IDCT: GETVALUE(p->AccelIDCT,idct*); break;
	case VOUT_WND: GETVALUE(p->Wnd,void*); break;
	case VOUT_VISIBLE: GETVALUE(p->Visible,bool_t); break;
	case VOUT_CLIPPING: GETVALUE(p->Clipping,bool_t); break;
	case VOUT_FX: GETVALUE(p->OrigFX,blitfx); break;
	case VOUT_VIEWPORT: GETVALUE(p->Viewport,rect); break;
	case VOUT_OUTPUTRECT: GETVALUE(p->GUIAlignedRect,rect); break;
	case VOUT_AUTOPREROTATE: GETVALUE(p->AutoPrerotate,bool_t); break;
	case VOUT_UPDATING: GETVALUE(p->Updating,bool_t); break;
	case VOUT_PLAY: GETVALUE(p->Play,bool_t); break;
	case VOUT_DIRTY: GETVALUE(p->Dirty,bool_t); break;
	case VOUT_CAPS: GETVALUE(p->Caps,int); break;
	case VOUT_COLORKEY: GETVALUECOND(p->ColorKey,int32_t,!p->Disabled && p->ColorKey!=-1); break;
	case VOUT_ASPECT: GETVALUE(p->Aspect,fraction); break;
	case VOUT_INPUT_DIR: GETVALUE(p->InputHintDir,int); break;
	case VOUT_ORIENTATION:
		if (Size==sizeof(int) && p->QueryOrientation && !p->DisableOrientation)
		{
			*(int*)Data = p->QueryOrientation(p);
			Result = ERR_NONE;
		}
		break;
	}
	return Result;
}

int OvlUpdateFX(overlay_base* p, bool_t ForceUpdate)
{
	blitfx LastFX = p->FX;
	int Width,Height,ScaleX,ScaleY;

	if (!p->Inited) return ERR_INVALID_PARAM;
	if (p->Updating) return ERR_NONE;

	p->FX = p->OrigFX;
	p->FX.Flags |= p->SetFX;
	p->FX.Flags &= ~p->ClearFX;
	p->FX.Direction = CombineDir(p->InputDirection,p->OrigFX.Direction,p->Output.Direction);
	p->PrefInput.Direction = CombineDir(p->FX.Direction,0,p->Input.Direction);

	if (p->OrigFX.ScaleX<=0)
	{
		int v = p->OrigFX.ScaleX;

		Width = p->Viewport.Width;
		Height = p->Viewport.Height;

		if ((p->InputDirection ^ p->OrigFX.Direction) & DIR_SWAPXY)
			SwapInt(&Width,&Height);

		if (p->PreRotate)
		{
			ScaleX = Scale(Width,SCALE_ONE,p->Input.Width);
			ScaleY = Scale(Height,SCALE_ONE,Scale(p->Input.Height,p->Aspect.Num,p->Aspect.Den));
		}
		else
		{
			ScaleX = Scale(Width,SCALE_ONE,Scale(p->Input.Width,p->Aspect.Num,p->Aspect.Den));
			ScaleY = Scale(Height,SCALE_ONE,p->Input.Height);
		}

		if (p->PreRotate && (v == -1 || v == -2))
			v = -3-v;

		if ((v == -2) || (v != -1 && ScaleX>ScaleY))
			ScaleX = ScaleY;

		if (v<-2)
			ScaleX = Scale(ScaleX,-v,SCALE_ONE);

		p->FX.ScaleY = ScaleX;
		p->FX.ScaleX = Scale(ScaleX,p->Aspect.Num,p->Aspect.Den);

		if (p->PreRotate)
			SwapInt(&p->FX.ScaleX,&p->FX.ScaleY);
	}
	else
	{
		// prefer source horizontal scaling (because of smooth scale option)
		if (p->PreRotate)
			p->FX.ScaleY = Scale(p->FX.ScaleY,p->Aspect.Num,p->Aspect.Den);
		else
			p->FX.ScaleX = Scale(p->FX.ScaleX,p->Aspect.Num,p->Aspect.Den);
	}

	if (p->Output.Pixel.Flags & PF_PIXELDOUBLE)
	{
		p->FX.ScaleX >>= 1;
		p->FX.ScaleY >>= 1;
	}

	if ((p->FX.Flags & BLITFX_ONLYDIFF) && p->FX.ScaleX < (SCALE_ONE*2)/3)
		p->FX.Flags &= ~BLITFX_ONLYDIFF;

	if (ForceUpdate || !EqBlitFX(&p->FX,&LastFX))
	{
		p->Update(p);
		p->Dirty = 1;
		p->LastTime = -1;
	}

	return ERR_NONE;
}

static int UpdateInputDirection(overlay_base* p,bool_t Update)
{
	if (p->Inited && !p->Updating)
	{
		p->InputDirection = p->Input.Direction;
		p->PreRotate = 0;

		if (p->AutoPrerotate)
		{
			rect r;
			PhyToVirt(NULL,&r,&p->Input);

			// portrait?
			if (r.Width < r.Height)
			{
				p->PreRotate = 1;
				if (p->InputDirection & DIR_SWAPXY)
					p->InputDirection ^= DIR_SWAPXY | DIR_MIRRORLEFTRIGHT;
				else
					p->InputDirection ^= DIR_SWAPXY | DIR_MIRRORUPDOWN;
			}
		}

		if (Update)
			OvlUpdateFX(p,0);
	}
	return ERR_NONE;
}

static void Done(overlay_base* p)
{
	if (p->Inited)
	{
		if (p->Show)
		{
			bool_t Old = p->Visible;
			p->Visible = 0;
			p->Show = 0;
			if (p->UpdateShow)
				p->UpdateShow(p); // maybe calls OvlUpdateShow
			p->Visible = Old;
		}
		p->Done(p);
		BlitRelease(p->Soft);
		p->Soft = NULL;
		p->Inited = 0;
		memset(&p->GUIAlignedRect,0,sizeof(rect));

		if (p->Primary)
			OvlInvalidate(p,&p->Viewport,1);
	}
}

int OvlUpdateShow(overlay_base* p,bool_t Temp)
{
	bool_t Show = !p->Disabled && p->Visible && (p->ColorKey != -1 || !p->Clipping || !p->Primary);
	if (!p->Updating && p->Show != Show)
	{
		p->Show = Show;
		p->Dirty = 1;
		p->LastTime = -1;

		if (p->Inited && !Temp)
		{
			if (p->UpdateShow)
				p->UpdateShow(p);
			else
			if (p->Primary)
			{
				if (p->Show)
				{
					OvlInvalidate(p,&p->Viewport,1); // redraw border (zoom may have changed)
					OvlValidate(p,&p->GUIAlignedRect);
				}
				else
				{
					OvlInvalidate(p,&p->Viewport,0); // redraw other windows
					OvlValidate(p,&p->Viewport); // own window is fine
				}
			}
		}
	}
	return ERR_NONE;
}

static int Updating(overlay_base* p)
{
	if (!p->Updating)
	{
		UpdateInputDirection(p,0);
		OvlUpdateFX(p,p->ViewportChanged);
		OvlUpdateShow(p,0);
		p->ViewportChanged = 0;
	}
	return ERR_NONE;
}

static int Init(overlay_base* p)
{
	memset(&p->FX,0,sizeof(blitfx));
	memset(&p->Output,0,sizeof(video));
	memset(&p->DstAlignedRect,0,sizeof(rect));
	memset(&p->SrcAlignedRect,0,sizeof(rect));
	memset(&p->GUIAlignedRect,0,sizeof(rect));

	p->Disabled = 0;
	p->Updating = 0;
	p->Show = 0;
	p->Dirty = 1;
	p->LastTime = -1;
	p->ColorKey = -1;
	p->SetFX = 0;
	p->ClearFX = 0;
	p->InputHintDir = p->Input.Direction; // default: suggest no input direction change
	p->Caps = -1;

	if (p->Input.Pixel.Flags)
	{
		int Result;
		
		if (!p->Wnd)
			return ERR_INVALID_PARAM;

		if (Compressed(&p->Input.Pixel))
			return ERR_NOT_SUPPORTED;

		if (p->Input.Width<=0) 
		{
			p->Disabled = 1;
			p->Input.Width = 2;
		}

		if (p->Input.Height<=0) 
		{
			p->Disabled = 1;
			p->Input.Height = 2;
		}

		Result = p->Init(p);

		if (Result != ERR_NONE)
		{
			if (Result == ERR_DEVICE_ERROR)
				NodeError((node*)p,ERR_ID,ERR_DEVICE_ERROR);
			return Result;
		}

		p->Inited = 1;
			
		UpdateInputDirection(p,0);
		OvlUpdateFX(p,1);
		OvlUpdateShow(p,0);

		if (p->Primary)
			OvlInvalidate(p,&p->Viewport,1);
	}
	return ERR_NONE;
}

int OvlBlitImage(overlay_base* p, const planes Data, const planes DataLast )
{
	planes Planes;
	int Result = p->Lock(p,Planes,1);
	if (Result==ERR_NONE)
	{
		BlitImage(p->Soft,Planes,Data,DataLast,-1,-1);
		p->Unlock(p);
	}
	return Result;
}

int OvlBlit(overlay_base* p,const planes Data,const planes LastData)
{
	int Result = ERR_NONE;

	if (p->Soft && p->Inited && p->Show)
	{
		if (p->Dirty)
			LastData = NULL;

		Result = p->Blit(p,Data,LastData);
		
		if (Result == ERR_NONE)
			p->Dirty = 0;
	}

	return Result;
}

void OvlAdjustOrientation(overlay_base* p, bool_t Combine)
{
	int Width = GetSystemMetrics(SM_CXSCREEN);
	int Height = GetSystemMetrics(SM_CYSCREEN);
	int Orientation;
	rect Virt;

	p->DisableOrientation = 1;

	if ((Width == p->Output.Width*2 && Height == p->Output.Height*2) || 
		(Width == p->Output.Height*2 && Height == p->Output.Width*2))
		p->Output.Pixel.Flags |= PF_PIXELDOUBLE;

	Orientation = QueryOrientation();

	if (Combine)
		p->Output.Direction = CombineDir(0,p->Output.Direction,Orientation);

	// still rotated? 
	PhyToVirt(NULL,&Virt,&p->Output);
	if (Width != Height && Virt.Width == Height && Virt.Height == Width)
	{
		if (Orientation & DIR_MIRRORLEFTRIGHT)
			p->Output.Direction ^= DIR_SWAPXY | DIR_MIRRORLEFTRIGHT;
		else
			p->Output.Direction ^= DIR_SWAPXY | DIR_MIRRORUPDOWN;
	}

	p->DisableOrientation = 0;
}

bool_t OvlTiming( overlay_base* p, const packet* Packet )
{
	if (Packet->CurrTime >= 0)
	{
		if (!p->Play && Packet->CurrTime == p->LastTime)
			return 0;

		if (Packet->RefTime >= (Packet->CurrTime + SHOWAHEAD))
			return 0;

		p->LastTime = Packet->CurrTime;
	}
	else
		p->LastTime = Packet->RefTime;

	p->CurrTime = Packet->CurrTime;
	return 1;
}

int OvlSet(overlay_base* p,int No,const void* Data,int Size)
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case VOUT_WND:
		if (Size == sizeof(HWND))
		{
			p->Wnd = *(HWND*)Data;
			if (p->UpdateWnd)
				p->UpdateWnd(p);
			Result =ERR_NONE;
		}
		break;

	case VOUT_INPUT:
		if (Size == sizeof(packet))
		{
			const packet* Packet = (const packet*)Data;
			if (OvlTiming(p,Packet))
				Result = OvlBlit(p,Packet->Data,Packet->LastData);
			else
				Result = ERR_BUFFER_FULL;
		}
		break;

	case VOUT_INPUT | PACKET_FORMAT:
		if (!Size || Size == sizeof(video))
		{
			if (Size && p->Inited)
			{
				video Tmp = *(video*)Data;
				Tmp.Pitch = p->Input.Pitch;
				if ((p->Input.Direction ^ Tmp.Direction) & DIR_SWAPXY)
					SwapInt(&Tmp.Width,&Tmp.Height);
				Tmp.Direction = p->Input.Direction;
				if (EqVideo(&Tmp,&p->Input))
				{
					p->Input = *(video*)Data;
					UpdateInputDirection(p,0);
					OvlUpdateFX(p,1);
					Result = ERR_NONE;
					break;
				}
			}
			Done(p);
			if (Size)
				p->Input = *(video*)Data;
			else
				memset(&p->Input,0,sizeof(video));
			Result = Init(p);
		}
		break;

	case VOUT_PLAY: SETVALUE(p->Play,bool_t,ERR_NONE); break;
	case VOUT_DIRTY: SETVALUE(p->Dirty,bool_t,ERR_NONE); break;

	case VOUT_UPDATING: SETVALUECMP(p->Updating,bool_t,Updating(p),EqBool); break;
	case VOUT_CLIPPING: SETVALUE(p->Clipping,bool_t,OvlUpdateShow(p,0)); break;
	case VOUT_VISIBLE: SETVALUE(p->Visible,bool_t,OvlUpdateShow(p,0)); break;
	case VOUT_AUTOPREROTATE: SETVALUE(p->AutoPrerotate,bool_t,UpdateInputDirection(p,1)); break;
	case VOUT_ASPECT: SETVALUECMP(p->Aspect,fraction,OvlUpdateFX(p,0),EqFrac); break;
	case VOUT_FX: SETVALUE(p->OrigFX,blitfx,OvlUpdateFX(p,0)); break;
	case VOUT_VIEWPORT:
		if (Size == sizeof(rect))
		{
			Result = ERR_NONE;
			if (!EqRect(&p->Viewport,(const rect*)Data))
			{
				p->Viewport = *(const rect*)Data;
				if (p->Updating)
					p->ViewportChanged = 1;
				if (p->Inited)
					OvlUpdateFX(p,1);
			}
		}
		break;

	case VOUT_RESET:
		if (p->Inited && p->Reset)
		{
			p->Reset(p);
			p->Update(p);
			p->Dirty = 1;
			p->LastTime = -1;
		}
		Result = ERR_NONE;
		break;
	}
	return Result;
}

