/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: overlay_base.h 142 2004-12-12 07:30:52Z picard $
 *
 * BetaPlayer Overlay plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#ifndef __OVERLAY_BASE_H
#define __OVERLAY_BASE_H

// default colorkey to use
#define COLORKEY	0x200020

typedef int (*ovlfunc)(void* This);
typedef void (*ovldone)(void* This);
typedef int (*ovlblit)(void* This,const planes Data,const planes LastData);
typedef int (*ovllock)(void* This,planes Planes, bool_t OnlyAligned );

typedef struct overlay_base
{
	node VMT;

	ovlfunc Init;
	ovldone Done;
	ovlblit Blit;
	ovlfunc Update;
	ovlfunc UpdateShow;
	ovlfunc QueryOrientation;
	ovlfunc Reset;
	ovlfunc UpdateWnd;
	ovllock Lock;
	ovlfunc Unlock;

	idct* AccelIDCT;
	const datadef* Params;

	// these parameters should be filled during Init()
	int SetFX;	
	int ClearFX;
	int Caps;
	video Output;

	// these are handled by overlay_base
	int Id;
	bool_t Primary;
	bool_t Overlay;
	void* Wnd;
	bool_t Disabled; // created with width=0 or height=0
	bool_t Inited;
	bool_t Updating;
	bool_t ViewportChanged; //during Updating
	bool_t FullScreenViewport;
	bool_t Dirty;
	bool_t Visible;
	bool_t Clipping;
	bool_t AutoPrerotate;
	bool_t Show;
	bool_t InputHintDir;
	bool_t Play;
	tick_t LastTime;
	tick_t CurrTime;
	int32_t ColorKey;
	video Input;
	video PrefInput;
	void* Soft;
	rect Viewport;
	rect SrcAlignedRect;
	rect GUIAlignedRect;
	rect DstAlignedRect;
	blitfx FX;
	blitfx OrigFX;
	fraction Aspect;
	bool_t PreRotate;
	int InputDirection;
	bool_t DisableOrientation;
	void* DLL;

} overlay_base;

void OvlInit(overlay_base* p,int Id,int UseDLL,int Priority);
void OvlDone(overlay_base* p);
int OvlEnum(overlay_base* p,int EnumNo,datadef* DataType);
int OvlGet(overlay_base* p,int No,void* Data,int Size);
int OvlSet(overlay_base* p,int No,const void* Data,int Size);
int OvlUpdateAlign(overlay_base* p);
void OvlValidate( overlay_base* p, const rect* );
void OvlInvalidate( overlay_base* p, const rect*, bool_t Local );
void OvlAdjustOrientation( overlay_base* p, bool_t Combine );
int OvlUpdateShow( overlay_base* p, bool_t Temp );
int OvlBlitImage(overlay_base* p, const planes Data, const planes DataLast );
int OvlUpdateFX(overlay_base* p, bool_t ForceUpdate);

#endif
