/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: overlay_flytv.c 142 2004-12-12 07:30:52Z picard $
 *
 * BetaPlayer Overlay plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "stdafx.h"

#ifndef WIN32_PLATFORM_WFSP
// not for smartphones

#define OUTPUT_VGA		17
#define Format320X240	2

typedef struct flytv
{
	overlay_base p;
	planes Planes;
	bool_t Inited;
	bool_t ErrorInit;

	WCHAR* (*DriverVersion)();
	int (*Initial_FLY_TV)(int,int);
	void (*UnInitial_FLY_TV)();
	int (*Zoom)(int);
	int (*SendFrame)(const void*);
} flytv;

static int Init(flytv* p)
{
	p->p.Output.Width = 320;
	p->p.Output.Height = 240;
	p->p.Output.Direction = 0;
	p->p.Output.Aspect = ASPECT_ONE;
	p->p.Output.Pitch = p->p.Output.Width*2;
	DefaultRGB(&p->p.Output.Pixel,16,5,6,5,0,0,0);
	p->p.Output.Pixel.Flags |= PF_16ALIGNED;

	p->ErrorInit = 0;
	p->Inited = 0;
	p->Planes[0] = Alloc16(2 * p->p.Output.Height * p->p.Output.Width);
	if (!p->Planes[0])
		return ERR_OUT_OF_MEMORY;

	memset(p->Planes[0],0,p->p.Output.Width * p->p.Output.Height * 2);
	p->p.ClearFX = BLITFX_ONLYDIFF;
	return ERR_NONE;
}

static void Done(flytv* p)
{
	Free16(p->Planes[0]); 
	p->Planes[0] = NULL;
}

static int UpdateShow(flytv* p)
{
	if (p->p.Show)
	{
		if (!p->Initial_FLY_TV(OUTPUT_VGA,Format320X240))
		{
			if (!p->ErrorInit)
			{
				p->ErrorInit = 1;
				NodeError((node*)p,ERR_ID,ERR_DEVICE_ERROR);
			}
			return ERR_DEVICE_ERROR;
		}
		p->Inited = 1;
		p->Zoom(2);
	}
	else
	{
		p->UnInitial_FLY_TV();
		p->Inited = 0;
	}
	return ERR_NONE;
}

static int Lock(flytv* p, planes Planes, bool_t OnlyAligned )
{
	Planes[0] = p->Planes[0];
	return ERR_NONE;
}

static int Unlock(flytv* p )
{
	return ERR_NONE;
}

static int Blit(flytv* p, const planes Data, const planes DataLast )
{
	if (p->Inited)
	{
		BlitImage(p->p.Soft,p->Planes,Data,DataLast,-1,-1);
		p->SendFrame(p->Planes[0]);
	}
	return ERR_NONE;
}

static flytv FlyTV = 
{
	OvlEnum,
	OvlGet,
	OvlSet,
	Init,
	Done,
	Blit,
	OvlUpdateAlign,
	UpdateShow,
	NULL, //queryorientation
	NULL, //reset
	NULL, //updatewnd
	Lock,
	Unlock,
};

void FlyTV_Init() 
{ 
	flytv* p = &FlyTV;

	p->p.DLL = LoadLibrary(T("FlyPresenter_TV.DLL"));
	GetProc(&p->p.DLL,&p->DriverVersion,T("DriverVersion"),0);
	GetProc(&p->p.DLL,&p->Initial_FLY_TV,T("Initial_FLY_TV"),0);
	GetProc(&p->p.DLL,&p->UnInitial_FLY_TV,T("UnInitial_FLY_TV"),0);
	GetProc(&p->p.DLL,&p->Zoom,T("Zoom"),0);
	GetProc(&p->p.DLL,&p->SendFrame,T("SendFrame"),0);

	OvlInit(&p->p,FLYTV_ID,1,PRI_DEFAULT-10);
	p->p.Primary = 0;
}

void FlyTV_Done()
{
	OvlDone(&FlyTV.p);
}

#else
void FlyTV_Init() {}
void FlyTV_Done() {}
#endif