/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: overlay_raw.c 131 2004-12-04 20:36:04Z picard $
 *
 * BetaPlayer Overlay plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "stdafx.h"

#define ESC_QUERYESCSUPPORT 8
#define GETGXINFO			0x00020000
#define GETRAWFRAMEBUFFER   0x00020001

#define FORMAT_565 1
#define FORMAT_555 2
#define FORMAT_OTHER 3

typedef struct RawFrameBufferInfo
{
	WORD wFormat;
	WORD wBPP;
	char *pFramePointer;
	int  cxStride;
	int  cyStride;
	int  cxPixels;
	int  cyPixels;

} RawFrameBufferInfo;

#define kfDirect			0x20
#define kfDirect555			0x40
#define kfDirect565			0x80
#define kfLandscape			0x08

typedef struct GXDeviceInfo 
{
	long Version;           //00
	void * pvFrameBuffer;   //04
	unsigned long cbStride; //08
	unsigned long cxWidth;  //0c
	unsigned long cyHeight; //10
	unsigned long cBPP;     //14
	unsigned long ffFormat; //18
	char Unused[0x84-7*4];
} GXDeviceInfo;

typedef struct raw
{
	overlay_base p;
	int AdjustPtr;
	RawFrameBufferInfo Info;

} raw;

const datadef RAWParams[] = 
{
	{ RAW_ID,			TYPE_CLASS },
	{ RAW_WIDTH,		TYPE_INT, DF_SETUP | DF_RDONLY | DF_HIDDEN },
	{ RAW_HEIGHT,		TYPE_INT, DF_SETUP | DF_RDONLY | DF_HIDDEN },
	{ RAW_PITCHX,		TYPE_INT, DF_SETUP | DF_RDONLY | DF_HIDDEN },
	{ RAW_PITCHY,		TYPE_INT, DF_SETUP | DF_RDONLY | DF_HIDDEN },
	{ RAW_BPP,			TYPE_INT, DF_SETUP | DF_RDONLY | DF_HIDDEN },
	{ RAW_FORMAT,		TYPE_INT, DF_SETUP | DF_RDONLY | DF_HIDDEN },
	{ RAW_POINTER,		TYPE_INT, DF_SETUP | DF_RDONLY | DF_HEX | DF_HIDDEN },

	DATADEF_END
};

static void Query(raw* p,RawFrameBufferInfo* Info)
{
	HDC DC = GetDC(NULL);
	memset(Info,0,sizeof(RawFrameBufferInfo));

	ExtEscape(DC, GETRAWFRAMEBUFFER, 0, NULL, sizeof(RawFrameBufferInfo), (char*)Info);

	if (!Info->pFramePointer && 
		QueryPlatform(PLATFORM_VER) >= 421 && 
		QueryPlatform(PLATFORM_TYPENO) != TYPE_SMARTPHONE) // C500 should use GAPI
	{
		//try gxinfo
		DWORD Code = GETGXINFO;
		if (ExtEscape(DC, ESC_QUERYESCSUPPORT, sizeof(DWORD), (char*)&Code, 0, NULL) > 0)
		{
			DWORD DCWidth = GetDeviceCaps(DC,HORZRES);
			DWORD DCHeight = GetDeviceCaps(DC,VERTRES);
			GXDeviceInfo GXInfo;
			memset(&GXInfo,0,sizeof(GXInfo));
			GXInfo.Version = 100;
			ExtEscape(DC, GETGXINFO, 0, NULL, sizeof(GXInfo), (char*)&GXInfo);

			// avoid VGA devices
			if (GXInfo.cbStride>0 && !(GXInfo.ffFormat & kfLandscape) &&
				((DCWidth == GXInfo.cxWidth && DCHeight == GXInfo.cyHeight) ||
				 (DCWidth == GXInfo.cyHeight && DCHeight == GXInfo.cxWidth)))
			{
				bool_t Detect = 0;
				TRY_BEGIN
				{
					int* p = (int*)GXInfo.pvFrameBuffer;
					COLORREF Old = GetPixel(DC,0,0);
					*p ^= -1;
					Detect = GetPixel(DC,0,0) != Old;
					*p ^= -1;
				}
				TRY_END

				if (Detect)
				{
					Info->pFramePointer = GXInfo.pvFrameBuffer;
					Info->cxPixels = GXInfo.cxWidth;
					Info->cyPixels = GXInfo.cyHeight;
					Info->cxStride = GXInfo.cBPP/8;
					Info->cyStride = GXInfo.cbStride;
					Info->wBPP = (WORD)GXInfo.cBPP;
					Info->wFormat = (WORD)GXInfo.ffFormat;
				}
			}
		}
	}

	if (abs(Info->cyStride) < abs(Info->cxPixels * Info->cxStride))
		SwapInt(&Info->cxPixels,&Info->cyPixels); // correct possible orientation error

	ReleaseDC(NULL,DC);
}

static int Get(raw* p,int No,void* Data,int Size)
{
	RawFrameBufferInfo Info;
	int Result = OvlGet(&p->p,No,Data,Size);
	switch (No)
	{
	case RAW_WIDTH: Query(p,&Info); GETVALUE(Info.cxPixels,int); break;
	case RAW_HEIGHT: Query(p,&Info); GETVALUE(Info.cyPixels,int); break;
	case RAW_PITCHX: Query(p,&Info); GETVALUE(Info.cxStride,int); break;
	case RAW_PITCHY: Query(p,&Info); GETVALUE(Info.cyStride,int); break;
	case RAW_BPP: Query(p,&Info); GETVALUE(Info.wBPP,int); break;
	case RAW_FORMAT: Query(p,&Info); GETVALUE(Info.wFormat,int); break;
	case RAW_POINTER: Query(p,&Info); GETVALUE((int)Info.pFramePointer,int); break;
	}
	return Result;
}

static int Init(raw* p)
{
	Query(p,&p->Info);
	if (!p->Info.pFramePointer)
		return ERR_DEVICE_ERROR;

	memset(&p->p.Output,0,sizeof(p->p.Output));
	p->p.Output.Width = p->Info.cxPixels;
	p->p.Output.Height = p->Info.cyPixels;
	p->p.Output.Direction = 0;
	p->p.Output.Aspect = ASPECT_ONE;

	switch (p->Info.wFormat)
	{
	case kfDirect555:
	case kfDirect|kfDirect555:
	case kfDirect|kfDirect555|kfLandscape:
	case FORMAT_555:
		DefaultRGB(&p->p.Output.Pixel,p->Info.wBPP,5,5,5,0,0,0);
		break;
	case kfDirect565:
	case kfDirect|kfDirect565:
	case kfDirect|kfDirect565|kfLandscape:
	case FORMAT_565:
		DefaultRGB(&p->p.Output.Pixel,p->Info.wBPP,5,6,5,0,0,0);
		break;
	default:
		return ERR_DEVICE_ERROR;
	}

	// we need the physical start of the framebuffer
	p->AdjustPtr = 0;
	if (p->Info.cxStride<0) 
		p->AdjustPtr += (p->Info.cxStride * (p->Info.cxPixels-1));
	if (p->Info.cyStride<0) 
		p->AdjustPtr += (p->Info.cyStride * (p->Info.cyPixels-1));

	if (abs(p->Info.cyStride) < abs(p->Info.cxStride))
	{
		p->p.SetFX |= BLITFX_VMEMROTATED;
		p->p.Output.Direction |= DIR_SWAPXY;
		p->p.Output.Pitch = abs(p->Info.cxStride);
		SwapInt(&p->p.Output.Width,&p->p.Output.Height);

		if (p->Info.cxStride<0) 
			p->p.Output.Direction |= DIR_MIRRORUPDOWN;
		if (p->Info.cyStride<0) 
			p->p.Output.Direction |= DIR_MIRRORLEFTRIGHT;
	}
	else
	{
		p->p.Output.Pitch = abs(p->Info.cyStride);

		if (p->Info.cxStride<0) 
			p->p.Output.Direction |= DIR_MIRRORLEFTRIGHT;
		if (p->Info.cyStride<0) 
			p->p.Output.Direction |= DIR_MIRRORUPDOWN;
	}

	if (((int)(p->Info.pFramePointer + p->AdjustPtr) & 15)==0 && (p->p.Output.Pitch & 15)==0)
		p->p.Output.Pixel.Flags |= PF_16ALIGNED;

	p->p.SetFX = BLITFX_AVOIDTEARING;

	OvlAdjustOrientation(&p->p,1);

	return ERR_NONE;
}

static void Done(raw* p)
{
}

static int Reset(raw* p)
{
	Done(p);
	Init(p);
	return ERR_NONE;
}

static int Lock(raw* p, planes Planes, bool_t OnlyAligned )
{
	Planes[0] = p->Info.pFramePointer + p->AdjustPtr;
	return ERR_NONE;
}

static int Unlock(raw* p )
{
	return ERR_NONE;
}

static raw RAW = 
{
	OvlEnum,
	Get,
	OvlSet,
	Init,
	Done,
	OvlBlitImage,
	OvlUpdateAlign,
	NULL, //updateshow
	NULL, //queryorientation
	Reset,
	NULL, //updatewnd
	Lock,
	Unlock
};

void RAW_Init() 
{ 
	raw* p = &RAW;
	Query(p,&p->Info);
	if (p->Info.pFramePointer)
	{
		p->p.Params = RAWParams;
		OvlInit(&p->p,RAW_ID,0,PRI_DEFAULT+95);
	}
}

void RAW_Done()
{
	raw* p = &RAW;
	OvlDone(&p->p);
}
