/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: blit_soft.c 113 2004-11-25 06:44:11Z picard $
 *
 ****************************************************************************/

#include "stdafx.h"

#if defined(__ARM)

typedef struct xscaledesc
{
	uint32_t Next;
	uint32_t Base;
	uint32_t Id;
	uint32_t Cmd;

} xscaledesc;

#define LCCR0		0x000
#define LCCR2		0x008
#define FSADR0		0x204
#define FDADR0		0x200

static int XScaleVSync( overlay_base* p )
{
	int Pos = 0;
	uint32_t Addr = *(uint32_t*)(p->VSyncPtr + FSADR0);
	int No;
	for (No=0;No<VSYNCDATA;No+=2)
	{
		if (p->VSyncData[0] <= Addr && p->VSyncData[1] >= Addr)
			return Pos + (Addr - p->VSyncData[0]);
		Pos += p->VSyncData[1] - p->VSyncData[0];
	}

	// current address out of range. something went wrong
	return -1;
}

void OvlDetectVSync( overlay_base* p )
{
	p->VSync = NULL;

	if (QueryPlatform(PLATFORM_CAPS) & CAPS_ARM_XSCALE)
	{
		PhyMemEnd(p->VSyncPtr);
		p->VSyncPtr = PhyMemBegin(0x44000000,0x1000);
		if (p->VSyncPtr)
		{
			int Enabled = *(uint32_t*)(p->VSyncPtr + LCCR0) & 1;
			int Lines = *(uint32_t*)(p->VSyncPtr + LCCR2) & 1023;

			if (Enabled && Lines)
			{
				int No,NoEnd;
				int DescCount = 0;
				xscaledesc* Desc[VSYNCDATA/2];
				uint32_t Start = *(uint32_t*)(p->VSyncPtr + FDADR0);

				Desc[DescCount] = (xscaledesc*)PhyMemBegin(Start,16);
				++DescCount;

				while (Desc[DescCount-1])
				{
					if (Desc[DescCount-1]->Next == Start)
					{
						for (NoEnd=0;NoEnd<DescCount;++NoEnd)
							if (Desc[NoEnd]->Cmd & 0x200000)
							{
								int Total = 0;
								memset(p->VSyncData,0,sizeof(p->VSyncData));
								for (No=0;No<DescCount;++No)
								{
									xscaledesc* x = Desc[(NoEnd+No+1) % DescCount];
									p->VSyncData[No*2] = x->Base;
									p->VSyncData[No*2+1] = x->Base + (x->Cmd & 0x1FFFFF);
									Total += (x->Cmd & 0x1FFFFF);
								}
								p->VSync = XScaleVSync;
								p->VSyncTotal = Total + (Total*4)/Lines;
								p->VSyncBottom = Total - Total/60;
								p->VSyncRefresh = TICKSPERSEC / 60; // todo: find a way to get lcd refresh rate
								break;
							}
						break;
					}

					if (DescCount==VSYNCDATA/2) // too many descriptors
						break;

					Desc[DescCount] = (xscaledesc*)PhyMemBegin(Desc[DescCount-1]->Next,16);
					++DescCount;
				}

				for (No=0;No<DescCount;++No)
					PhyMemEnd(Desc[No]);
			}
		}
	}

	if (!p->VSync && p->VSyncPtr)
	{
		PhyMemEnd(p->VSyncPtr);
		p->VSyncPtr = NULL;
	}
}

	//!!!
	p->AdjustPtr = 0;
	p->Info.pFramePointer = PhyMemBegin(0xabf01000,640*480*2);


#endif
