/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: association.c 145 2004-12-12 18:18:28Z picard $
 *
 * BetaPlayer
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "stdafx.h"

static int Enum( node* p, int No, datadef* Param )
{
	int Id;
	node* Node;

	if (NodeEnumType(&No,Param,NodeParams,NULL)==ERR_NONE)
		return ERR_NONE;

	if (!NodeInsideIsClass())
	{
		Node = NodeEnum(FORMAT_CLASS,No);
		if (Node)
		{
			Node->Get(Node,NODE_ID,&Id,sizeof(Id));
			
			memset(Param,0,sizeof(datadef));
			Param->No = Id;
			Param->Type = TYPE_BOOL;
			Param->Size = sizeof(bool_t);
			Param->Flags = DF_SETUP|DF_OWNREG|DF_CHECKLIST;
			Param->Class = ASSOCIATION_ID;
			Param->Name = String(Id,0);

			return ERR_NONE;
		}
	}
	return ERR_INVALID_PARAM;
}

static int Get( node* p, int No, void* Data, int Size )
{
	node* Node;
	tchar_t s[16];
	tchar_t Exts[MAXPATH];
	tchar_t *r,*q;
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(ASSOCIATION_ID,int); break;
	default:
		Node = NodeFind(No);
		if (Node && NodeIsClass(Node,FORMAT_CLASS))
		{
			*(bool_t*)Data = 0;
			if (Node->Get(Node,FORMAT_EXTS,Exts,sizeof(Exts))==ERR_NONE)
			{
				*(bool_t*)Data = 1;
				for (r=Exts;r;r=q)
				{
					q = tcschr(r,':');
					if (q)
					{
						TcsNCpy(s,r,q-r+1);
						if (!GetFileAssociation(s,NodeProgramId()))
						{
							*(bool_t*)Data = 0;
							break;
						}
						q = tcschr(q,';');
						if (q) ++q;
					}
				}
			}
			return ERR_NONE;
		}
		break;
	}
	return Result;
}

static int Set( node* p, int No, const void* Data, int Size )
{
	node* Node;
	tchar_t Exts[MAXPATH];
	tchar_t s[16];
	tchar_t *r,*q;
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(ASSOCIATION_ID,int); break;
	default:
		Node = NodeFind(No);
		if (Node && NodeIsClass(Node,FORMAT_CLASS))
		{
			bool_t State = *(bool_t*)Data;
			if (Node->Get(Node,FORMAT_EXTS,Exts,sizeof(Exts))==ERR_NONE)
			{
				if (State)
				{
					tchar_t Path[MAXPATH];
					tchar_t Play[MAXPATH];
					tchar_t Icon[MAXPATH];

					GetModuleFileName(NULL,Path,MAXPATH);

					if (tcschr(Path,' '))
						stprintf(Play,T("\"%s\" %%1"),Path);
					else
						stprintf(Play,T("%s %%1"),Path);
					stprintf(Icon,T("%s, -%d"),Path,IDI_ICON);

					SetFileType(NodeProgramId(),Play,Icon);
				}

				for (r=Exts;r;r=q)
				{
					q = tcschr(r,':');
					if (q)
					{
						TcsNCpy(s,r,q-r+1);
						SetFileAssociation(s,NodeProgramId(),State);
						q = tcschr(q,';');
						if (q) ++q;
					}
				}
			}
			return ERR_NONE;
		}
		break;
	}
	return Result;
}

static node Association = 
{
	Enum,
	Get,
	Set,
};

void Association_Init()
{
	NodeRegister(&Association,PRI_MAXIMUM+500);
}

void Association_Done()
{
	NodeUnRegister(&Association);
}
