/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: main.c 144 2004-12-12 15:28:39Z picard $
 *
 * BetaPlayer
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "stdafx.h"

static tchar_t PROGRAM_TITLE[] = T("BetaPlayer");
static tchar_t PROGRAM_VERSION[] = 
#include "../version"
;

BOOL CALLBACK EnumWindowsProc(HWND Wnd,LPARAM Param)
{
	HWND* p = (HWND*)Param;
	if (GetWindow(Wnd,GW_OWNER) == *p)
	{
		*p = Wnd;
		return 0;
	}
	return 1;
}

bool_t FindRunning()
{
	HWND Wnd;
	Wnd = FindWindow(WinClassName(), NULL);
	if (Wnd)
	{
		if (NodeCmdLine()[0])
		{
			NodeSaveValue(IF_REG_OPEN,NodeCmdLine(),(tcslen(NodeCmdLine())+1)*sizeof(tchar_t),TYPE_STRING);
			SendMessage(Wnd,MSG_OPEN,0,0);
		}

		while (!IsWindowEnabled(Wnd))
		{
			HWND Last = Wnd;
			EnumWindows(EnumWindowsProc,(LPARAM)&Wnd);
			if (Wnd == Last)
				break;
		}
		SetForegroundWindow(Wnd);
		return 1;
	}
	return 0;
}

#if !defined(_WIN32_WCE) && defined(UNICODE)
int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE hParent,LPSTR CmdA,int CmdShow)
{
	WCHAR Cmd[MAXCMD];
	HANDLE Handle;
	bool_t Wait;
	mbstowcs(Cmd,CmdA,sizeof(Cmd)/sizeof(WCHAR));
#else
int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE hParent,TCHAR* Cmd,int CmdShow)
{
	HANDLE Handle;
	bool_t Wait;
#endif

#ifndef NDEBUG
	//DLLTest(); // just to help debugging plugins. comment out if not needed
#endif

	Node_Init(PROGRAM_TITLE,PROGRAM_VERSION,WinMain,Cmd);

	if (FindRunning())
		return 0;

	SAFE_BEGIN

	Handle = CreateMutex(NULL,FALSE,PROGRAM_TITLE);
	if (GetLastError() == ERROR_ALREADY_EXISTS)
		return 0;

	Wait = QueryPlatform(PLATFORM_TYPENO) == TYPE_SMARTPHONE;
	if (Wait) WaitBegin();

	NodePlugins();
	Player_Init();
	Win_Init(hInstance,IDI_ICON);
	MediaInfo_Init();
	BenchResult_Init();
	About_Init();
	QueryKey_Init();
	Settings_Init();
	OpenFile_Init();
	Association_Init();
	Interface_Init();
	PlayList_Init();
	PLS_Init();
	M3U_Init();
	ASX_Init();

	if (Wait) WaitEnd();

	// user input needs priority
	SetThreadPriority(GetCurrentThread(),THREAD_PRIORITY_ABOVE_NORMAL);

	WinOpen((win*)NodeFind(INTERFACE_ID),NULL,0);

	M3U_Done();
	PLS_Done();
	ASX_Done();
	PlayList_Done();
	Interface_Done();
	Association_Done();
	OpenFile_Done();
	Settings_Done();
	QueryKey_Done();
	About_Done();
	BenchResult_Done();
	MediaInfo_Done();
	Win_Done();
	Player_Done();
	Node_Done();

	CloseHandle(Handle);

	SAFE_END
	return 0;
}

