/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: mediainfo.c 141 2004-12-10 16:05:44Z picard $
 *
 * BetaPlayer
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "stdafx.h"

static int Comment(win* Win,player* Player,int y,int Stream)
{
	tchar_t s[256];
	int No[] = { COMMENT_TITLE, COMMENT_ARTIST, COMMENT_ALBUM, COMMENT_GENRE, 0, COMMENT_LANGUAGE, 0 };
	int i;

	for (i=0;No[i];++i)
		if (Player->CommentByName(Player,Stream,PlayerComment(No[i]),s,sizeof(s)))
		{
			if (Stream>=0 && Player->CommentByName(Player,-1,PlayerComment(No[i]),NULL,0))
				continue;

			y = WinPropLabel(Win,y,String(PLAYER_ID,No[i]),s,NULL);
		}

	if (Stream>=0)
		for (++i;No[i];++i)
			if (Player->CommentByName(Player,Stream,PlayerComment(No[i]),s,sizeof(s)))
				y = WinPropLabel(Win,y,String(PLAYER_ID,No[i]),s,NULL);

	return y;
}

static int Info(win* Win,node* Node,int y,int Name)
{
	int Id;
	int No;
	datadef DataDef;

	if (Name && Node->Get(Node,NODE_ID,&Id,sizeof(Id))==ERR_NONE)
		y = WinPropLabel(Win,y,String(MEDIAINFO_ID,Name),String(Id,0),NULL);

	for (No=0;Node->Enum(Node,No,&DataDef)==ERR_NONE;++No)
		if (!(DataDef.Flags & (DF_OUTPUT|DF_HIDDEN)))
		{
			tchar_t s[256];
			tick_t Tick;
			int i;
			bool_t Ok = 0;

			switch (DataDef.Type)
			{
			case TYPE_TICK:
				Ok = Node->Get(Node,DataDef.No,&Tick,sizeof(Tick))==ERR_NONE;
				TickToString(s,Tick,sizeof(s),0,1,0);
				break;

			case TYPE_INT:
				Ok = Node->Get(Node,DataDef.No,&i,sizeof(i))==ERR_NONE;
				IntToString(s,i,sizeof(s),(DataDef.Flags & DF_HEX)!=0);
				if (DataDef.Flags & DF_KBYTE)
					tcscat(s,T(" KB"));
				break;

			case TYPE_STRING:
				Ok = Node->Get(Node,DataDef.No,s,sizeof(s))==ERR_NONE;
				break;					
			}

			if (Ok)
				y = WinPropLabel(Win,y,DataDef.Name,s,NULL);
		}

	return y;
}

static bool_t Proc(win* Win,int Msg, WPARAM wParam, LPARAM lParam, int* Result)
{
	node* Player;
	node* Format;
	node* Input;
	int y;

	switch (Msg)
	{
	case WM_COMMAND:
		if (LOWORD(wParam) == PLATFORM_OK)
			PostMessage(Win->Wnd,WM_CLOSE,0,0);

		break;

	case WM_CREATE:
		WinCreateDialog(Win);

		Player = NodeFind(PLAYER_ID);
		Player->Get(Player,PLAYER_FORMAT,&Format,sizeof(Format));
		Player->Get(Player,PLAYER_INPUT,&Input,sizeof(Input));

		y = 4;

		if (Input)
			y = Info(Win,Input,y,0);

		if (Format)
		{
			int No;
			pin Pin;
			datadef DataDef;

			y = Info(Win,Format,y,MEDIAINFO_FORMAT);
			y = Comment(Win,(player*)Player,y,-1);

			for (No=0;Format->Get(Format,FORMAT_STREAM+No,&Pin,sizeof(Pin))==ERR_NONE;++No)
				if (NodeDataDef(Format,FORMAT_STREAM+No,&DataDef))
				{
					audio Audio;
					video Video;
					int Total;
					int Dropped;
					tchar_t s[256];

					y += 6;

					if (DataDef.Format1 == PACKET_VIDEO && Format->Get(Format,(FORMAT_STREAM+No)|PACKET_FORMAT,&Video,sizeof(Video))==ERR_NONE)
					{
						if (Video.Pixel.Flags & PF_FOURCC)
							FourCCToString(s,Video.Pixel.FourCC,sizeof(s));
						else
						if (AnyYUV(&Video.Pixel))
							tcscpy(s,T("YUV"));
						else
							stprintf(s,T("RGB %d bits"),Video.Pixel.BitCount);

						y = WinPropLabel(Win,y,String(MEDIAINFO_ID,MEDIAINFO_VIDEO),s,NULL);

						if (Pin.Node && Compressed(&Video.Pixel))
							y = Info(Win,Pin.Node,y,MEDIAINFO_CODEC);

						if (Video.Width && Video.Height)
						{
							stprintf(s,T("%d x %d"),Video.Width,Video.Height);
							y = WinPropLabel(Win,y,String(MEDIAINFO_ID,MEDIAINFO_SIZE),s,NULL);
						}
						if (Video.FrameRate.Num)
						{
							FractionToString(s,&Video.FrameRate,sizeof(s),0,3);
							y = WinPropLabel(Win,y,String(MEDIAINFO_ID,MEDIAINFO_FPS),s,NULL);
						}
						if (Video.ByteRate)
						{
							IntToString(s,(Video.ByteRate+62)/125,sizeof(s),0);
							tcscat(s,T(" kbit/s")); // 1000byte/sec
							y = WinPropLabel(Win,y,String(MEDIAINFO_ID,MEDIAINFO_BITRATE),s,NULL);
						}

						if (Pin.Node && Pin.Node->Get(Pin.Node,FLOW_TOTAL,&Total,sizeof(Total))==ERR_NONE)
						{
							IntToString(s,Total,sizeof(s),0);
							y = WinPropLabel(Win,y,String(MEDIAINFO_ID,MEDIAINFO_VIDEO_TOTAL),s,NULL);
						}
						else
							Total = -1;

						if (Pin.Node && Pin.Node->Get(Pin.Node,FLOW_DROPPED,&Dropped,sizeof(Dropped))==ERR_NONE)
						{
							IntToString(s,Dropped,sizeof(s),0);
							y = WinPropLabel(Win,y,String(MEDIAINFO_ID,MEDIAINFO_VIDEO_DROPPED),s,NULL);
						}
						else
							Dropped = -1;

						if (Total>0 && Dropped>=0 && Video.FrameRate.Num)
						{
							fraction Rate = Video.FrameRate;
							Simplify(&Rate,(1<<30)/Total,(1<<30)/Total);
							Rate.Num *= Total - Dropped;
							Rate.Den *= Total;
							FractionToString(s,&Rate,sizeof(s),0,3);
							y = WinPropLabel(Win,y,String(MEDIAINFO_ID,MEDIAINFO_AVG_FPS),s,NULL);
						}

					}
					else
					if (DataDef.Format1 == PACKET_AUDIO && Format->Get(Format,(FORMAT_STREAM+No)|PACKET_FORMAT,&Audio,sizeof(Audio))==ERR_NONE)
					{
						stprintf(s,T("%04X"),Audio.Format & 0xFFFF);
						y = WinPropLabel(Win,y,String(MEDIAINFO_ID,MEDIAINFO_AUDIO),s,NULL);

						if (Pin.Node && Audio.Format != 1)
							y = Info(Win,Pin.Node,y,MEDIAINFO_CODEC);

						if (Audio.SampleRate)
						{
							IntToString(s,Audio.SampleRate,sizeof(s),0);
							tcscat(s,T(" Hz"));
							y = WinPropLabel(Win,y,String(MEDIAINFO_ID,MEDIAINFO_SAMPLERATE),s,NULL);
						}
						if (Audio.Channels)
						{
							IntToString(s,Audio.Channels,sizeof(s),0);
							y = WinPropLabel(Win,y,String(MEDIAINFO_ID,MEDIAINFO_CHANNELS),s,NULL);
						}
						if (Audio.ByteRate)
						{
							IntToString(s,(Audio.ByteRate+62)/125,sizeof(s),0);
							tcscat(s,T(" kbit/s")); // 1000bit/sec
							y = WinPropLabel(Win,y,String(MEDIAINFO_ID,MEDIAINFO_BITRATE),s,NULL);
						}
					}
					else
					if (Pin.Node)
						y = Info(Win,Pin.Node,y,MEDIAINFO_CODEC);

					y = Comment(Win,(player*)Player,y,No);
				}
		}

		break;
	}
	return 0;
}

static menudef MenuDef[] =
{
	{ 0,PLATFORM_ID, PLATFORM_OK },

	MENUDEF_END
};

static win MediaInfo;

void MediaInfo_Init()
{
	memset(&MediaInfo,0,sizeof(MediaInfo));
	MediaInfo.Id = MEDIAINFO_ID;
	MediaInfo.Width = 240;
	MediaInfo.Height = 320;
	MediaInfo.Dialog = 1;
	MediaInfo.Title = String(MEDIAINFO_ID,0);
	MediaInfo.MenuDef = MenuDef;
	MediaInfo.Proc = Proc;
	MediaInfo.LabelWidth = 64;

	WinRegister(&MediaInfo,PRI_DEFAULT);
}

void MediaInfo_Done()
{
	WinUnRegister(&MediaInfo);
}