/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: player.c 145 2004-12-12 18:18:28Z picard $
 *
 * BetaPlayer
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "stdafx.h"

#define AUDIO_BUFFER_WITH_VIDEO		4*TICKSPERSEC
#define AUDIO_BUFFER_ALONE			1*TICKSPERSEC

#define MAXCOMMENT	256

#define PROBE_SIZE	2048

#define DEFAULT_TIMER_UNIT 32

typedef struct comment
{
	int Stream;
	tchar_t Name[64];
	tchar_t Value[120];
} comment;

typedef struct item
{
	tchar_t URL[MAXPATH];
	tchar_t Title[256];
	tick_t Length;
	bool_t Changed;
} item;

typedef struct ref
{
	node* Node;
	bool_t Flow;
	bool_t Used;
} ref;

typedef struct player_win32
{
	player VMT;
	node EqVMT; 
	node ColVMT;

	// setup

	bool_t BufferWarning;
//	bool_t BufferUnderrun;
	int BufferSize;
	int MDBufferSize;
	int CurrBufferSize;
	bool_t Background;
	bool_t MicroDrive;
	bool_t Repeat;
	bool_t Shuffle;
	bool_t PlayAtOpen;
	bool_t PlayAtOpenFull;
	bool_t KeepPlayVideo;
	bool_t KeepPlayAudio;
	bool_t KeepList;
	tick_t MoveBack;
	tick_t MoveFFwd;
	int BurstStart;
	int UnderRun;
	equalizer Eq;
	fraction Zoom;
	fraction Aspect;
	int FullDir;
	bool_t SmoothZoom50;
	bool_t SmoothZoomAlways;
	bool_t VideoAccel;
	bool_t AutoPreRotate;
	bool_t AudioSwap;
	int AudioQuality;
	int Brightness;
	int Contrast;
	int Saturation;
	fraction SeekAfterSync;
	fraction PlaySpeed;
	fraction FFwdSpeed;
//	DWORD ThreadId;

	node* AOutput;
	node* VOutput;
	node* VIDCT;
	node* FlowBuffer;
	idct* IDCT;

	// states

	pin ListChanged;
	rect SkinViewport;
	int SkinDir;
	void* Wnd;
	bool_t FullScreen;
	bool_t Clipping;
	bool_t Rotating;
	bool_t Dither;
	rect Viewport;
	bool_t Primary;
	bool_t Overlay;
	bool_t VideoUsed;
	bool_t AudioUsed;
	int Stereo;
	int StreamNo;

	// current media

	bool_t PreRotate;
	int BufferPercent;
	stream* Input;
	node* Timer;
	format* Format;
	ref* RefEnd;
	int AudioStream; 
	int VideoStream; 

	// threads and timings

	bool_t NoMoreInput;
	bool_t Play;
	bool_t FFwd;
	bool_t Sync;
	bool_t Fill;
	bool_t LoadMode;
	bool_t Bench;
	bool_t SaveMicroDrive;
	bool_t Exit;
	bool_t WaitForProcess;
	bool_t TimerPlay;
	tick_t Position;
	tick_t Time;
	fraction Speed;
	fraction Unit;
	tick_t BenchTime;
	int BenchStart;
	int IOError;
	int Used;
	int UsedEnough;

	int LockCount;
	void* Lock;			 // parameter read/write
	void* LockInput;
	void* LockProcess;

	HANDLE EventProcess; // ProcessThread has processed some packets (InputThread can continue)

	bool_t RunInput;
	bool_t RunProcess;
	HANDLE EventRunInput;	
	HANDLE EventRunProcess; 
	
	HANDLE ProcessThread;
	HANDLE InputThread;

	// playlist

	int Current;
	bool_t CurrentChanged;
	int PlayListCount;
	item* PlayList;
	int* PlayIndex;

	// comments
	int CommentCount;

	// misc

	int32_t ColorKey; // colorkeybursh color
	HBRUSH ColorKeyBrush;
	HBRUSH BlackBrush;

	// arrays

	ref Ref[MAXNODE];
	comment* Comment[MAXCOMMENT];
	tchar_t Title[256];
	bool_t StreamSkip[MAXSTREAM];

} player_win32;

const enumdef EnumQuality[] = 
{
	PLAYER_ID,PLAYER_LOW,
	PLAYER_ID,PLAYER_MEDIUM,
	PLAYER_ID,PLAYER_HIGH,

	ENUMDEF_END
};

static void EqReset( win* p, backup* Chain )
{
	for (;Chain;Chain=Chain->Chain)
		if (Chain->DataDef.Class == EQUALIZER_ID && 
			Chain->DataDef.Type == TYPE_INT)
			*(int*)Chain->Data = 0;
}

static void ColReset( win* p, backup* Chain )
{
	for (;Chain;Chain=Chain->Chain)
		if (Chain->DataDef.Class == COLOR_ID && 
			Chain->DataDef.Type == TYPE_INT)
			*(int*)Chain->Data = 0;
}

const datadef ColorParams[] = 
{
	{ COLOR_ID,				TYPE_CLASS },
	{ COLOR_BRIGHTNESS,		TYPE_INT, DF_SETUP|DF_MINMAX, -128,127 },
	{ COLOR_CONTRAST,		TYPE_INT, DF_SETUP|DF_MINMAX, -128,127 },
	{ COLOR_SATURATION,		TYPE_INT, DF_SETUP|DF_MINMAX, -128,127 },

	{ COLOR_RESET,			TYPE_FUNC,DF_SETUP|DF_OWNREG|DF_GAP,(int)ColReset },

	DATADEF_END
};

const datadef EqParams[] = 
{
	{ EQUALIZER_ID,			TYPE_CLASS },
	{ EQUALIZER_ENABLED,	TYPE_BOOL,DF_SETUP|DF_OWNREG|DF_CHECKLIST|DF_RESYNC },
	{ EQUALIZER_PREAMP,		TYPE_INT, DF_SETUP|DF_OWNREG|DF_MINMAX|DF_SHORTLABEL|DF_RESYNC, -20, 20 },

	{ EQUALIZER_1,			TYPE_INT, DF_SETUP|DF_OWNREG|DF_MINMAX|DF_NOWRAP|DF_SHORTLABEL|DF_RESYNC|DF_GAP, -20, 20 },
	{ EQUALIZER_2,			TYPE_INT, DF_SETUP|DF_OWNREG|DF_MINMAX|DF_NOWRAP|DF_SHORTLABEL|DF_RESYNC, -20, 20 },
	{ EQUALIZER_3,			TYPE_INT, DF_SETUP|DF_OWNREG|DF_MINMAX|DF_NOWRAP|DF_SHORTLABEL|DF_RESYNC, -20, 20 },
	{ EQUALIZER_4,			TYPE_INT, DF_SETUP|DF_OWNREG|DF_MINMAX|DF_NOWRAP|DF_SHORTLABEL|DF_RESYNC, -20, 20 },
	{ EQUALIZER_5,			TYPE_INT, DF_SETUP|DF_OWNREG|DF_MINMAX|DF_NOWRAP|DF_SHORTLABEL|DF_RESYNC, -20, 20 },
	{ EQUALIZER_6,			TYPE_INT, DF_SETUP|DF_OWNREG|DF_MINMAX|DF_NOWRAP|DF_SHORTLABEL|DF_RESYNC, -20, 20 },
	{ EQUALIZER_7,			TYPE_INT, DF_SETUP|DF_OWNREG|DF_MINMAX|DF_NOWRAP|DF_SHORTLABEL|DF_RESYNC, -20, 20 },
	{ EQUALIZER_8,			TYPE_INT, DF_SETUP|DF_OWNREG|DF_MINMAX|DF_NOWRAP|DF_SHORTLABEL|DF_RESYNC, -20, 20 },
	{ EQUALIZER_9,			TYPE_INT, DF_SETUP|DF_OWNREG|DF_MINMAX|DF_NOWRAP|DF_SHORTLABEL|DF_RESYNC, -20, 20 },
	{ EQUALIZER_10,			TYPE_INT, DF_SETUP|DF_OWNREG|DF_MINMAX|DF_NOWRAP|DF_SHORTLABEL|DF_RESYNC, -20, 20 },

	{ EQUALIZER_RESET,		TYPE_FUNC,DF_SETUP|DF_OWNREG|DF_GAP,(int)EqReset },

	DATADEF_END
};

const datadef PlayerParams[] = 
{
	{ PLAYER_ID,			TYPE_CLASS },
	{ PLAYER_AUTOPREROTATE,	TYPE_BOOL, DF_SETUP },
	{ PLAYER_REPEAT,		TYPE_BOOL, DF_SETUP|DF_HIDDEN },
	{ PLAYER_SHUFFLE,		TYPE_BOOL, DF_SETUP|DF_HIDDEN },
	{ PLAYER_KEEPPLAY_AUDIO,TYPE_BOOL, DF_SETUP },
	{ PLAYER_KEEPPLAY_VIDEO,TYPE_BOOL, DF_SETUP },
	{ PLAYER_KEEPLIST,		TYPE_BOOL, DF_SETUP },
	{ PLAYER_PLAYATOPEN,	TYPE_BOOL, DF_SETUP },
	{ PLAYER_PLAYATOPEN_FULL,TYPE_BOOL, DF_SETUP },
	{ PLAYER_ZOOM,			TYPE_FRACTION, DF_SETUP|DF_PERCENT },
	{ PLAYER_PLAY_SPEED,	TYPE_FRACTION, DF_SETUP|DF_PERCENT|DF_HIDDEN },
	{ PLAYER_FFWD_SPEED,	TYPE_FRACTION, DF_SETUP|DF_PERCENT },
	{ PLAYER_MOVEFFWD,		TYPE_TICK, DF_SETUP|DF_MINMAX, 0, 600*TICKSPERSEC },
	{ PLAYER_MOVEBACK,		TYPE_TICK, DF_SETUP|DF_MINMAX, 0, 600*TICKSPERSEC },

	{ PLAYER_BUFFER_SIZE,	TYPE_INT, DF_SETUP|DF_KBYTE|DF_GAP, 512, 128*1024 },
	{ PLAYER_UNDERRUN,		TYPE_INT, DF_SETUP|DF_PERCENT,0,PERCENT_ONE },

	{ PLAYER_MICRODRIVE,	TYPE_BOOL, DF_SETUP|DF_GAP },
	{ PLAYER_MD_BUFFER_SIZE,TYPE_INT, DF_SETUP|DF_KBYTE, 512, 128*1024 },
	{ PLAYER_BURSTSTART,	TYPE_INT, DF_SETUP|DF_PERCENT,0,PERCENT_ONE },

	{ PLAYER_AOUTPUT,		TYPE_NODE, DF_SETUP|DF_HIDDEN, AOUT_CLASS },
	{ PLAYER_VOUTPUT,		TYPE_NODE, DF_SETUP|DF_HIDDEN, VOUT_CLASS },
	{ PLAYER_VIDEO_ACCEL,	TYPE_BOOL, DF_SETUP|DF_HIDDEN },
	{ PLAYER_LIST_COUNT,	TYPE_INT, DF_SETUP|DF_HIDDEN },
	{ PLAYER_LIST_CURRENT,	TYPE_INT, DF_SETUP|DF_HIDDEN },
	{ PLAYER_EQUALIZER,		TYPE_EQUALIZER, DF_SETUP|DF_HIDDEN },
	{ PLAYER_EQENABLE,		TYPE_BOOL, DF_SETUP|DF_HIDDEN|DF_OWNREG },
	{ PLAYER_SMOOTH50,		TYPE_BOOL, DF_SETUP|DF_HIDDEN },
	{ PLAYER_SMOOTHALWAYS,	TYPE_BOOL, DF_SETUP|DF_HIDDEN },
	{ PLAYER_DITHER,		TYPE_BOOL, DF_SETUP|DF_HIDDEN },
	{ PLAYER_STEREO,		TYPE_INT,  DF_SETUP|DF_HIDDEN },
	{ PLAYER_ASPECT,		TYPE_FRACTION, DF_SETUP|DF_HIDDEN },
	{ PLAYER_FULL_DIR,		TYPE_INT, DF_SETUP|DF_HIDDEN },
	{ PLAYER_SKIN_DIR,		TYPE_INT, DF_HIDDEN },
	{ PLAYER_AUDIO_QUALITY,	TYPE_INT, DF_SETUP|DF_ENUM|DF_HIDDEN, (int)&EnumQuality },
	{ PLAYER_AUDIO_SWAP,	TYPE_BOOL, DF_SETUP|DF_HIDDEN },
	{ PLAYER_ASTREAM,		TYPE_INT, DF_HIDDEN|DF_SETUP },
	{ PLAYER_VSTREAM,		TYPE_INT, DF_HIDDEN|DF_SETUP },
	{ PLAYER_PERCENT,		TYPE_FRACTION, DF_HIDDEN|DF_SETUP },

	{ PLAYER_PLAY,			TYPE_BOOL, DF_HIDDEN },
	{ PLAYER_FFWD,			TYPE_BOOL, DF_HIDDEN },
	{ PLAYER_POSITION,		TYPE_TICK, DF_HIDDEN },
	{ PLAYER_DURATION,		TYPE_TICK, DF_HIDDEN },
	{ PLAYER_TITLE,			TYPE_STRING },
	{ PLAYER_INPUT,			TYPE_NODE, DF_HIDDEN, STREAM_CLASS },
	{ PLAYER_FORMAT,		TYPE_NODE, DF_HIDDEN, FORMAT_MEDIA_CLASS },
	{ PLAYER_WND,			TYPE_INT, DF_HIDDEN|DF_HEX },
	{ PLAYER_SKIN_VIEWPORT,	TYPE_RECT, DF_HIDDEN },
	{ PLAYER_CLIPPING,		TYPE_BOOL, DF_HIDDEN },
	{ PLAYER_FULLSCREEN,	TYPE_BOOL, DF_HIDDEN },
	{ PLAYER_VOLUME,		TYPE_INT, DF_MINMAX|DF_HIDDEN, 0, 100 },
	{ PLAYER_MUTE,			TYPE_BOOL, DF_HIDDEN },
	{ PLAYER_BACKGROUND,	TYPE_BOOL, DF_HIDDEN },
	{ PLAYER_AUDIOUSED,		TYPE_BOOL, DF_HIDDEN | DF_RDONLY },
	{ PLAYER_VIDEOUSED,		TYPE_BOOL, DF_HIDDEN | DF_RDONLY },

	DATADEF_END
};

#define PLAYEREQ(p) ((player_win32*)((char*)(p)-(int)&(((player_win32*)0)->EqVMT)))
#define PLAYERCOL(p) ((player_win32*)((char*)(p)-(int)&(((player_win32*)0)->ColVMT)))

static int EqEnum( void* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,EqParams,NULL);
}

static int ColEnum( void* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,ColorParams,NULL);
}

static int Enum( player_win32* p, int No, datadef* Param )
{
	if (NodeEnumType(&No,Param,NodeParams,OutputParams,PlayerParams,NULL) == ERR_NONE)
		return ERR_NONE;

	if (No>=0 && No<p->PlayListCount*3)
	{
		memset(Param,0,sizeof(datadef));
		if (No >= 2*p->PlayListCount)
		{
			Param->No = PLAYER_LIST_LENGTH + (No - 2*p->PlayListCount);
			Param->Type = TYPE_TICK;
			Param->Size = sizeof(tick_t);
		}
		else
		{
			if (No >= p->PlayListCount)
				Param->No = PLAYER_LIST_TITLE + (No - p->PlayListCount);
			else
				Param->No = PLAYER_LIST_URL + No;
			Param->Type = TYPE_STRING;
			Param->Size = MAXDATA;
		}
		Param->Flags = DF_HIDDEN | DF_SETUP;
		Param->Class = PLAYER_ID;
		return ERR_NONE;
	}

	return ERR_INVALID_PARAM;
}

static int UpdateSpeed(player_win32* p)
{
	if (p->Bench)
	{
		p->Speed.Num = 0;
		p->Speed.Den = 1;
	}
	else
	if (p->FFwd)
		p->Speed = p->FFwdSpeed;
	else
		p->Speed = p->PlaySpeed;

	if (p->Timer)
		p->Timer->Set(p->Timer,TIMER_SPEED,&p->Speed,sizeof(p->Speed));

	return ERR_NONE;
}

static void UpdateList(player_win32* p)
{
	if (p->ListChanged.Node)
		p->ListChanged.Node->Set(p->ListChanged.Node,p->ListChanged.No,NULL,0);
}

static void UpdateListLength(player_win32* p,tick_t Length)
{
	if (p->Current < p->PlayListCount && !p->PlayList[p->Current].Changed && 
		p->PlayList[p->Current].Length != Length)
	{
		p->PlayList[p->Current].Length = Length;
		UpdateList(p);
	}
}

static const tchar_t* GetFirstComment( player_win32* p, int Id )
{
	const tchar_t* Name = PlayerComment(Id);
	int i;
	for (i=0;i<p->CommentCount;++i)
		if (tcsicmp(p->Comment[i]->Name,Name)==0)
			return p->Comment[i]->Value;
	return T("");
}

static void UpdateListTitle(player_win32* p)
{
	const tchar_t* Title = GetFirstComment(p,COMMENT_TITLE);
	const tchar_t* Artist = GetFirstComment(p,COMMENT_ARTIST);

	if (Artist[0] && Title[0])
		stprintf(p->Title,T("%s - %s"),Artist,Title);
	else
	{
		if (!Title[0])
			Title = Artist;
		tcscpy(p->Title,Title);
	}
	PostMessage(p->Wnd,MSG_UPDATE,0,0);

	if (p->Current < p->PlayListCount && !p->PlayList[p->Current].Changed)
	{
		tcscpy(p->PlayList[p->Current].Title,p->Title);
		UpdateList(p);
	}
}

static void UpdateBacklight(player_win32* p)
{
	BacklightTimeout((p->Play || p->FFwd) && p->Primary && p->VideoUsed && p->Format);
}

static void UpdateRunInput(player_win32* p)
{
	// we don't want deadlocks so LoadMode will restart InputThread even with NoMoreInput
	p->RunInput = p->Format && p->Input && (!p->NoMoreInput || p->LoadMode);
	if (p->RunInput)
		SetEvent(p->EventRunInput);
	else
		ResetEvent(p->EventRunInput);
}

static void UpdateRunProcess(player_win32* p)
{
	p->RunProcess = p->Format && p->Input && p->Timer && (((p->FFwd || p->Play) && !p->LoadMode) || p->Sync || p->Fill);
	if (p->RunProcess)
		SetEvent(p->EventRunProcess);
	else
		ResetEvent(p->EventRunProcess);
}

static void UpdatePlay(player_win32* p)
{
	p->TimerPlay = (p->Play || p->FFwd) && !p->Sync && !p->Fill && !p->LoadMode;
	p->Timer->Set(p->Timer,TIMER_PLAY,&p->TimerPlay,sizeof(bool_t));
	if (p->VOutput)
		p->VOutput->Set(p->VOutput,VOUT_PLAY,&p->TimerPlay,sizeof(bool_t));
}

static void ReSync(player_win32* p);
static int UpdateFormat(player_win32* p,bool_t BufferChanged);
static int SetPlay(player_win32* p,bool_t Stop);
static int SetPlayListCount( player_win32* p, int Count );

static void UpdateTimerUnit( player_win32* p)
{
	if (p->Format && p->Timer)
	{
		pin Pin;
		video Video;
		datadef DataDef;
		int No = FORMAT_STREAM + p->VideoStream;

		if (p->Format->Get(p->Format,No,&Pin,sizeof(Pin))==ERR_NONE &&
			Pin.Node && NodeDataDef((node*)p->Format,No,&DataDef) && 
			DataDef.Format1 == PACKET_VIDEO && 
			p->Format->Get(p->Format,No|PACKET_FORMAT,&Video,sizeof(Video))==ERR_NONE &&
			Video.FrameRate.Num>0)

			p->Unit = Video.FrameRate;
		else
		{
			p->Unit.Num = DEFAULT_TIMER_UNIT;
			p->Unit.Den = 1;
		}

		p->Timer->Set(p->Timer,TIMER_UNIT,&p->Unit,sizeof(p->Unit));
	}
}

static void UpdateFlow(player_win32* p)
{
	ref* Ref;
	for (Ref=p->Ref;Ref!=p->RefEnd;++Ref)
		if (Ref->Used && Ref->Flow)
			Ref->Node->Set(Ref->Node,FLOW_UPDATE,NULL,0);
}

static int UpdateVideo( player_win32* p )
{
	node* Node = p->VOutput;
	bool_t FullScreen;
	bool_t Visible;
	bool_t Refresh;
	bool_t Updating;
	blitfx FX;
	
	if (Node)
	{
		p->Overlay = 0;
		Node->Get(Node,VOUT_PRIMARY,&p->Primary,sizeof(bool_t));
		Node->Get(Node,VOUT_OVERLAY,&p->Overlay,sizeof(bool_t));

		FullScreen = p->FullScreen || !p->Primary;
			
		if (FullScreen)
		{
			video Mode;
			Mode.Width = 0;
			Mode.Height = 0;
			Node->Get(Node,VOUT_OUTPUT|PACKET_FORMAT,&Mode,sizeof(video));

			PhyToVirt(NULL,&p->Viewport,&Mode);

			DEBUG_MSG4(DEBUG_TEST,T("UpdateVideo Viewport:%d %d %d %d"),p->Viewport.x,p->Viewport.y,p->Viewport.Width,p->Viewport.Height);
		}
		else
			p->Viewport = p->SkinViewport;


		FX.ScaleY = FX.ScaleX = Scale(SCALE_ONE,p->Zoom.Num,p->Zoom.Den);

		FX.Flags = 0;
		if (p->Dither)
			FX.Flags |= BLITFX_DITHER;
		if (p->SmoothZoom50)
			FX.Flags |= BLITFX_ARITHSTRETCH50;
		if (p->SmoothZoomAlways)
			FX.Flags |= BLITFX_ARITHSTRETCHALWAYS;
		if (QueryAdvPlatform(ADVPLATFORM_SLOW_VIDEO))
			FX.Flags |= BLITFX_ONLYDIFF;
		if (QueryAdvPlatform(ADVPLATFORM_COLOR_LOOKUP))
			FX.Flags |= BLITFX_COLOR_LOOKUP;

		FX.Brightness = p->Brightness;
		FX.Contrast = p->Contrast;
		FX.Saturation = p->Saturation;

		if (FullScreen)
			if (p->Primary)
				FX.Direction = CombineDir(QueryOrientation(),0,p->FullDir);
			else
				FX.Direction = p->FullDir;
		else
			FX.Direction = p->SkinDir;

		Visible = !p->Background && !p->Rotating;

		Updating = 1;
		Node->Set(Node,VOUT_UPDATING,&Updating,sizeof(bool_t));

		Node->Set(Node,VOUT_VIEWPORT,&p->Viewport,sizeof(rect));
		Node->Set(Node,VOUT_FX,&FX,sizeof(blitfx));
		Node->Set(Node,VOUT_ASPECT,&p->Aspect,sizeof(fraction));
		Node->Set(Node,VOUT_CLIPPING,&p->Clipping,sizeof(p->Clipping));
		Node->Set(Node,VOUT_VISIBLE,&Visible,sizeof(Visible));
		Node->Set(Node,VOUT_AUTOPREROTATE,&p->AutoPreRotate,sizeof(bool_t));

		Updating = 0;
		Node->Set(Node,VOUT_UPDATING,&Updating,sizeof(bool_t));

		UpdateFlow(p);

		// try to redraw overlay

		Refresh = p->VideoUsed && !p->Sync;

		if (Refresh && p->FlowBuffer)
			Refresh = p->FlowBuffer->Set(p->FlowBuffer,FLOWBUFFER_RESEND,NULL,0) != ERR_NONE;

		if (Refresh && p->Format)
		{
			DEBUG_MSG(DEBUG_TEST,T("ReSync"));
			ReSync(p); // redraw the hard way: resync
		}
	}
	else
	{
		p->Primary = 1;
		p->Overlay = 0;
		p->Viewport = p->SkinViewport;
	}

	return ERR_NONE;
}

static int UpdateAudioBuffer(player_win32* p)
{
	node* Node = p->AOutput;
	if (Node)
	{
		tick_t t = p->VideoUsed ? AUDIO_BUFFER_WITH_VIDEO : AUDIO_BUFFER_ALONE;
		Node->Set(Node,AOUT_BUFFER,&t,sizeof(t));
	}
	return ERR_NONE;
}

static int UpdateAudio( player_win32* p )
{
	node* Node = p->AOutput;

	if (Node)
	{
		Node->Set(Node,AOUT_STEREO,&p->Stereo,sizeof(int));
		Node->Set(Node,AOUT_SWAP,&p->AudioSwap,sizeof(bool_t));
		Node->Set(Node,AOUT_QUALITY,&p->AudioQuality,sizeof(int));
		Node->Set(Node,AOUT_BACKGROUND,&p->Background,sizeof(bool_t));

		UpdateFlow(p);
	}
	return ERR_NONE;
}

static int UpdateBackground( player_win32* p )
{
	if (p->Background)
		if ((p->VideoUsed && !p->KeepPlayVideo) ||
			(!p->VideoUsed && !p->KeepPlayAudio))
			SetPlay(p,1);

	UpdateVideo(p);
	UpdateAudio(p);

	return ERR_NONE;
}

static int BuildPlayIndex( player_win32* p )
{
	int No,No2;
	int* Index = (int*) ReAlloc(p->PlayIndex,sizeof(int)*p->PlayListCount);

	if (!Index && p->PlayListCount)
	{
		SetPlayListCount(p,0);
		return ERR_OUT_OF_MEMORY;
	}

	p->PlayIndex = Index;

	for (No=0;No<p->PlayListCount;++No)
		Index[No] = No;

	if (p->Shuffle)
	{
		// randomize order

		for (No=0;No<p->PlayListCount;++No)
		{ 
			No2 = rand() % p->PlayListCount;
			SwapInt(&Index[No],&Index[No2]);
		}
	}

	return ERR_NONE;
}

static int SetPlayListCount( player_win32* p, int Count )
{
	item* List = (item*) ReAlloc(p->PlayList,sizeof(item)*Count);

	if (!List && Count)
		return ERR_OUT_OF_MEMORY;

	while (p->PlayListCount < Count)
	{
		memset(List+p->PlayListCount,0,sizeof(item));
		List[p->PlayListCount].Length = -1;
		List[p->PlayListCount].Changed = 1;
		++p->PlayListCount;
	}

	p->PlayList = List;
	p->PlayListCount = Count;

	if (p->Current >= Count)
	{
		p->Current = 0;
		p->CurrentChanged = 1;
	}

	UpdateList(p);

	return BuildPlayIndex(p);
}

static void StopLoadMode(player_win32* p)
{
	if (p->LoadMode)
	{
		DEBUG_MSG(DEBUG_TEST,T("LoadMode Leaving"));

		p->LoadMode = 0;
		PostMessage(p->Wnd,MSG_WAIT,0,0);
		//PostThreadMessage(p->ThreadId,MSG_WAIT,0,0);
		UpdatePlay(p);
		UpdateRunProcess(p);
		UpdateRunInput(p);
	}
}

static void StopBench(player_win32* p)
{
	p->BenchTime = Scale(GetTime() - p->BenchStart,TICKSPERSEC,1000);
	p->Bench = 0;
	p->MicroDrive = p->SaveMicroDrive;

	UpdateFormat(p,0);
	UpdatePlay(p);
	UpdateSpeed(p);
	ReSync(p);
}

static void ReSync(player_win32* p)
{
	DEBUG_MSG(DEBUG_PLAYER,T("Sync"));
	if (p->Format->Sync(p->Format,-1,-1) == ERR_NONE)
	{
		p->Sync = 1;
		StopLoadMode(p);
		UpdatePlay(p);
		UpdateRunProcess(p);

		// wakeup input threads, just in case
		SetEvent(p->EventProcess); 
	}
}

static int SetPlay(player_win32* p,bool_t Stop)
{
	if (Stop)
	{
		p->Play = 0;
		p->FFwd = 0;
		p->Fill = 0;
		p->Sync = 0;
	}

	DEBUG_MSG2(DEBUG_PLAYER,T("Play %d,%d"),p->Play,p->FFwd);

	StopLoadMode(p);

	if (p->Bench && !p->Play)
		StopBench(p);

	if (p->Format && p->Input && p->Timer)
	{
		if (p->TimerPlay)
		{
			UpdatePlay(p);
			UpdateSpeed(p);
		}
		else
		{
			UpdateSpeed(p);
			UpdatePlay(p);
		}
		UpdateRunProcess(p);
	}

	UpdateBacklight(p);

	return ERR_NONE;
}

static int Seek(player_win32* p,tick_t Time,const fraction* Percent)
{
	int Result = ERR_INVALID_DATA;

	if (p->Format && p->Input && p->Timer)
	{
		tick_t Duration;
		int FileSize;
		int FilePos = -1;

		if (Percent)
		{
			if (p->Format->Get(p->Format,FORMAT_DURATION,&Duration,sizeof(Duration))==ERR_NONE)
			{
				Time = Scale(Duration,Percent->Num,Percent->Den);
				if (Time < 0)
					Time = 0;
			}
			else
			if (p->Input->Get(p->Input,STREAM_LENGTH,&FileSize,sizeof(FileSize))==ERR_NONE)
			{
				Time = -1;
				FilePos = Scale(FileSize,Percent->Num,Percent->Den);
				if (FilePos < 0)
					FilePos = 0;
			}
			else
				return ERR_NOT_SUPPORTED;
		}

		DEBUG_MSG2(DEBUG_PLAYER,T("Sync time:%d filepos:%d"),Time,FilePos);

		Result = p->Format->Sync(p->Format,Time,FilePos);
		if (Result == ERR_NONE)
		{
			ref* Ref;
			bool_t b = 1;

			// clear all buffers
			for (Ref=p->Ref;Ref!=p->RefEnd;++Ref)
				if (Ref->Used && Ref->Flow)
					Ref->Node->Set(Ref->Node,FLOW_EMPTY,&b,sizeof(b));

			p->NoMoreInput = 0;
			p->BufferPercent = PERCENT_ONE;

			p->Position = Time;
			p->Sync = 1;

			StopLoadMode(p);
			UpdatePlay(p);
			UpdateRunProcess(p);
			UpdateRunInput(p);

			// wakeup input threads (if needed)
			SetEvent(p->EventProcess); 
		}
	}
	else
	if (!p->Wnd)
		p->SeekAfterSync = *Percent;

	return Result;
}

static void StartBench(player_win32* p)
{
	ref* Ref;
	int Zero=0;

	p->SaveMicroDrive = p->MicroDrive;
	p->MicroDrive = 0;
	UpdateFormat(p,0);

	if (!QueryAdvPlatform(ADVPLATFORM_BENCHFROMPOS))
		Seek(p,0,NULL);

	p->BufferPercent = PERCENT_ONE;
	while (p->Format->Read(p->Format,p->BufferPercent,&p->Used) == ERR_NONE &&
		p->Used < p->UsedEnough);

	for (Ref=p->Ref;Ref!=p->RefEnd;++Ref)
		if (Ref->Used && Ref->Flow)
		{
			Ref->Node->Set(Ref->Node,FLOW_TOTAL,&Zero,sizeof(Zero));
			Ref->Node->Set(Ref->Node,FLOW_DROPPED,&Zero,sizeof(Zero));
		}
				
	p->Bench = 1;
	p->BenchStart = GetTime();
	p->Play = 1;
	p->FFwd = 0;
	p->Fill = 0;
	SetPlay(p,0);
}

static void SetTimerNode(player_win32* p,node* Timer)
{
	tick_t Time = 0;

	if (p->Timer)
		p->Timer->Get(p->Timer,TIMER_TIME,&Time,sizeof(Time));

	p->Timer = Timer;

	if (p->Timer)
	{
		p->Timer->Set(p->Timer,TIMER_SPEED,&p->Speed,sizeof(p->Speed));
		p->Timer->Set(p->Timer,TIMER_UNIT,&p->Unit,sizeof(p->Unit));
		p->Timer->Set(p->Timer,TIMER_TIME,&Time,sizeof(Time));
		UpdatePlay(p);
	}

	UpdateRunProcess(p);
}

static void ReleaseNode(player_win32* p,node* Node);
static bool_t UpdateNode(player_win32* p,node* Node);

static void ReleaseComment(player_win32* p,node* From,const datadef* FromDef)
{
	pin Pin;
	Pin.Node = NULL;
	Pin.No = 0;

	From->Set(From,FromDef->No,&Pin,sizeof(Pin));
}

static void TryComment(player_win32* p,node* From,const datadef* FromDef)
{
	pin Pin;
	Pin.Node = (node*)p;
	Pin.No = PLAYER_COMMENT+p->StreamNo;

	From->Set(From,FromDef->No,&Pin,sizeof(Pin));
}

static void ReleasePin(player_win32* p,node* Node,datadef* DataDef)
{
	pin Pin;
	node* Ptr;

	if (DataDef->Type == TYPE_NODE && 
		Node->Get(Node,DataDef->No,&Ptr,sizeof(Ptr))==ERR_NONE && Ptr)
	{
		ReleaseNode(p,Ptr);

		if (!(DataDef->Flags & DF_RDONLY))
		{
			Ptr = NULL;
			Node->Set(Node,DataDef->No,&Ptr,sizeof(Ptr));
		}
	}

	if (DataDef->Type == TYPE_PIN && DataDef->Format1 == PACKET_COMMENT)
		ReleaseComment(p,Node,DataDef);
	else
	if (DataDef->Type == TYPE_PIN &&
		Node->Get(Node,DataDef->No,&Pin,sizeof(Pin))==ERR_NONE && Pin.Node)
	{
		ReleaseNode(p,Pin.Node);

		if (!(DataDef->Flags & DF_RDONLY))
		{
			if (Pin.Node != (node*)&p->VMT)
				Pin.Node->Set(Pin.Node,Pin.No|PACKET_FORMAT,NULL,0);

			Pin.Node = NULL;
			Pin.No = 0;
			Node->Set(Node,DataDef->No,&Pin,sizeof(Pin));
		}
	}
}

static void ReleaseNode(player_win32* p,node* Node)
{
	datadef DataDef;
	int No;
	ref* Ref;

	assert(Node != (node*)p);

	for (Ref=p->Ref;Ref!=p->RefEnd;++Ref)
		if (Ref->Node == Node)
		{
			if (!Ref->Used)
				return;

			Ref->Used = 0;
			break;
		}

	if (Ref==p->RefEnd)
		return;

	if (NodeIsClass(Node,VOUT_CLASS))
		p->VideoUsed = 0;

	if (NodeIsClass(Node,AOUT_CLASS))
		p->AudioUsed = 0;

	if (p->FlowBuffer == Node)
		p->FlowBuffer = NULL;

	if (NodeIsClass(Node,IDCT_CLASS))
		p->IDCT = NULL;

	if (Node == p->Timer)
		SetTimerNode(p,NodeEnum(TIMER_CLASS,0)); // restore global timer

	for (No=0;Node->Enum(Node,No,&DataDef)==ERR_NONE;++No)
		if (DataDef.Flags & DF_OUTPUT)
			ReleasePin(p,Node,&DataDef);
}

static bool_t Used(player_win32* p,node* Node)
{
	ref* Ref;
	for (Ref=p->Ref;Ref!=p->RefEnd;++Ref)
		if (Ref->Node == Node)
			return Ref->Used;
	return 0;
}

static bool_t TryPin(player_win32* p,node* From,const datadef* FromDef,node* To,bool_t* FoundFirst,pin* Failed)
{
	pin Pin;
	int DataNo;
	datadef ToDef;

	if (To && !Used(p,To))
		for (DataNo=0;To->Enum(To,DataNo,&ToDef)==ERR_NONE;++DataNo)
			if (ToDef.Type == TYPE_PACKET &&
				ToDef.Format1 == FromDef->Format1 && 
				(Failed->Node != To || Failed->No != ToDef.No) &&
				(ToDef.Flags & DF_INPUT) && !(ToDef.Flags & DF_RDONLY))
			{
				Pin.Node = To;
				Pin.No = ToDef.No;

				// try to connect
				if (From->Set(From,FromDef->No,&Pin,sizeof(Pin))==ERR_NONE)
				{
					if (FoundFirst)
						*FoundFirst = 1;

					if (UpdateNode(p,To))
						return 1;

					if (Pin.Node != (node*)&p->VMT)
						Pin.Node->Set(Pin.Node,Pin.No|PACKET_FORMAT,NULL,0);
				}
			}

	return 0;
}

static bool_t UpdatePin(player_win32* p,node* From,const datadef* FromDef,bool_t* FoundFirst)
{
	pin Pin;
	node* To;
	int NodeNo;

	if (FoundFirst)
		*FoundFirst = 0;

	// node pointer or pin?

	if (FromDef->Type == TYPE_NODE)
	{
		node* Force = NULL;
		node* Skip = NULL;
		node* Failed = NULL;
		node* Old = NULL;

		if (FromDef->Format1 == IDCT_CLASS && !(FromDef->Flags & DF_RDONLY))
			if (p->VideoAccel)
				Force = p->VIDCT;
			else
				Skip = p->VIDCT;
			
		// already connected pointer?

		if (From->Get(From,FromDef->No,&To,sizeof(To))==ERR_NONE && To)
		{
			if (FoundFirst)
				*FoundFirst = 1;

			if ((!Force || To == Force) && (!Skip || To != Skip))
			{
				if (UpdateNode(p,To))
					return 1;

				Failed = To;
			}
			else
			{
				Old = To;
				ReleaseNode(p,To);
			}
		}

		if (FromDef->Flags & DF_RDONLY)
			return 0;

retry:	for (NodeNo=0;(To=NodeEnum(FromDef->Format1,NodeNo))!=NULL;++NodeNo)
			if (!Used(p,To) && (!Force || To == Force) && (!Skip || To != Skip) && Failed != To)
			{
				// try to connect
				if (From->Set(From,FromDef->No,&To,sizeof(To))==ERR_NONE)
				{
					if (FoundFirst)
						*FoundFirst = 1;

					if (UpdateNode(p,To))
						return 1;
				}
			}

		if (Old && UpdateNode(p,Old))
			return 1;
		Old = NULL;

		if (Force)
		{
			Force = NULL;
			goto retry;
		}

		// disconnect
		To = NULL;
		From->Set(From,FromDef->No,&To,sizeof(To));
	}

	if (FromDef->Type == TYPE_PIN)
	{
		if (FromDef->Format1 == PACKET_COMMENT)
		{
			TryComment(p,From,FromDef);

			if (FoundFirst)
				*FoundFirst = 1;

			return 1;
		}
		else
		{
			pin Failed;
			Failed.Node = NULL;
			Failed.No = 0;

			// already connected pin?

			if (From->Get(From,FromDef->No,&Pin,sizeof(Pin))==ERR_NONE && Pin.Node)
			{
				if (!NodeIsClass(Pin.Node,OUTPUT_CLASS) || (Pin.Node == p->VOutput || Pin.Node == p->AOutput))
				{
					if (FoundFirst)
						*FoundFirst = 1;

					if (UpdateNode(p,Pin.Node))
						return 1;

					Failed = Pin;
				}
				else
					ReleaseNode(p,Pin.Node);

				// disconnect

				if (FromDef->Flags & DF_RDONLY)
					return 0;

				if (Pin.Node != (node*)&p->VMT)
					Pin.Node->Set(Pin.Node,Pin.No|PACKET_FORMAT,NULL,0);
			}

			if (FromDef->Flags & DF_RDONLY)
				return 0;

			for (NodeNo=0;(To=NodeEnum(NODE_CLASS,NodeNo))!=NULL;++NodeNo)
				if (!Used(p,To) && !NodeIsClass(To,OUTPUT_CLASS))
					if (TryPin(p,From,FromDef,To,FoundFirst,&Failed))
						return 1;

			// try outputs
			
			if (TryPin(p,From,FromDef,p->VOutput,FoundFirst,&Failed))
				return 1;

			if (TryPin(p,From,FromDef,p->AOutput,FoundFirst,&Failed))
				return 1;

			// disconnect
			Pin.Node = NULL;
			Pin.No = 0;
			From->Set(From,FromDef->No,&Pin,sizeof(Pin));
		}
	}
	return 0;
}

static void UpdateEqualizer(player_win32* p,node* Node)
{
	int No;
	datadef DataDef;

	for (No=0;Node->Enum(Node,No,&DataDef)==ERR_NONE;++No)
		if (DataDef.Type == TYPE_EQUALIZER)
			Node->Set(Node,DataDef.No,&p->Eq,sizeof(p->Eq));
}

static int UpdateAllEqualizer(player_win32* p)
{
	ref* Ref;

	for (Ref=p->Ref;Ref!=p->RefEnd;++Ref)
		if (Ref->Used)
			UpdateEqualizer(p,Ref->Node);

	return ERR_NONE;
}

static bool_t UpdateNode(player_win32* p,node* Node)
{
	datadef DataDef;
	int No;
	bool_t Success = 1;
	ref* Ref;
	bool_t Used;

	assert(Node != (node*)p);

	for (Ref=p->Ref;Ref!=p->RefEnd;++Ref)
		if (Ref->Node == Node)
			break;

	if (Ref==p->RefEnd)
	{
		++p->RefEnd;
		Ref->Node = Node;
		Ref->Flow = NodeIsClass(Node,FLOW_CLASS);
		Ref->Used = 0;
	}

	Used = Ref->Used;
	Ref->Used = 1;

	if (!NodeIsClass(Node,OUTPUT_CLASS))
	{
		// update output pins
		for (No=0;Node->Enum(Node,No,&DataDef)==ERR_NONE;++No)
			if (DataDef.Flags & DF_OUTPUT)
				if (!UpdatePin(p,Node,&DataDef,NULL))
					Success = 0;
	}

	if (!Success)
	{
		if (!Used)
			Ref->Used = 0;
		else
			ReleaseNode(p,Node);
	}
	else
	if (!Used)
	{
		if (NodeIsClass(Node,VOUT_CLASS))
			p->VideoUsed = 1;

		if (NodeIsClass(Node,AOUT_CLASS))
			p->AudioUsed = 1;

		if (NodeIsClass(Node,TIMER_CLASS))
			SetTimerNode(p,Node);

		if (NodeIsClass(Node,FLOWBUFFER_CLASS))
			p->FlowBuffer = Node;

		if (NodeIsClass(Node,IDCT_CLASS))
			p->IDCT = (idct*)Node;

		UpdateEqualizer(p,Node);
	}

	return Success;
}

static void ReleaseStreams(player_win32* p)
{
	datadef DataDef;
	int No;

	for (No=0;p->Format->Enum(p->Format,No,&DataDef)==ERR_NONE;++No)
	{
		if (DataDef.No >= FORMAT_COMMENT && DataDef.No < FORMAT_COMMENT+MAXSTREAM)
			ReleaseComment(p,(node*)p->Format,&DataDef);

		if (DataDef.No >= FORMAT_STREAM && DataDef.No < FORMAT_STREAM+MAXSTREAM)
		{
			p->StreamNo = DataDef.No - FORMAT_STREAM;
			if (p->StreamNo != p->AudioStream && p->StreamNo != p->VideoStream)
				ReleasePin(p,(node*)p->Format,&DataDef);
		}
	}

	UpdateBacklight(p);
	UpdateAudioBuffer(p);
}

static int UpdateStreams(player_win32* p)
{
	datadef DataDef;
	int No;
	bool_t FoundCodec;

	if (!p->Format)
		return ERR_INVALID_PARAM;

	ReleaseStreams(p);

	for (No=0;p->Format->Enum(p->Format,No,&DataDef)==ERR_NONE;++No)
	{
		if (DataDef.No >= FORMAT_COMMENT && DataDef.No < FORMAT_COMMENT+MAXSTREAM)
		{
			p->StreamNo = DataDef.No - FORMAT_COMMENT;
			TryComment(p,(node*)p->Format,&DataDef);
		}

		if (DataDef.No >= FORMAT_STREAM && DataDef.No < FORMAT_STREAM+MAXSTREAM)
		{
			p->StreamNo = DataDef.No - FORMAT_STREAM;
			if (!p->StreamSkip[p->StreamNo] && DataDef.Type==TYPE_PIN)
				switch (DataDef.Format1)
				{
				case PACKET_AUDIO:
					if (p->AudioStream<0)
						p->AudioStream = p->StreamNo;

					if (p->AudioStream == p->StreamNo && 
						!UpdatePin(p,(node*)p->Format,&DataDef,&FoundCodec) && !FoundCodec)
					{
						audio Format;
						p->Format->Get(p->Format,DataDef.No|PACKET_FORMAT,&Format,sizeof(Format));

						if (Format.Format != AUDIOFMT_PCM)
						{
							NodeError((node*)p,PLAYER_ID,PLAYER_AUDIO_NOT_FOUND,Format.Format);
							p->AudioStream = -1;
							p->StreamSkip[p->StreamNo] = 1;
						}
					}
					break;

				case PACKET_VIDEO:
					if (p->VideoStream<0)
						p->VideoStream = p->StreamNo;
						
					if (p->VideoStream == p->StreamNo && 
						!UpdatePin(p,(node*)p->Format,&DataDef,&FoundCodec) && !FoundCodec)
					{
						video Format;
 						p->Format->Get(p->Format,DataDef.No|PACKET_FORMAT,&Format,sizeof(Format));

						if (Compressed(&Format.Pixel))
						{
							tchar_t s[8];
							FourCCToString(s,Format.Pixel.FourCC,sizeof(s));

							NodeError((node*)p,PLAYER_ID,PLAYER_VIDEO_NOT_FOUND,s);
							p->VideoStream = -1;
							p->StreamSkip[p->StreamNo] = 1;
						}

					}
					break;
				}
		}
	}

	UpdateVideo(p);
	UpdateTimerUnit(p);
	UpdateBacklight(p);
	UpdateAudioBuffer(p);

	return ERR_NONE;
}

static int UpdateFormat(player_win32* p,bool_t BufferChanged)
{
	if (BufferChanged)
		p->BufferWarning = 0;
	p->CurrBufferSize = p->MicroDrive ? p->MDBufferSize : p->BufferSize;
	if (p->Format)
	{
		p->Format->Set(p->Format,FORMAT_BUFFERSIZE,&p->CurrBufferSize,sizeof(int));
	}
	return ERR_NONE;
}

static void FreeComments( player_win32* p )
{
	int n;
	for (n=0;n<p->CommentCount;++n)
		Free(p->Comment[n]);
	p->CommentCount = 0;
}

static void Unload(player_win32* p,bool_t KeepPlay)
{
	node* NullPtr = NULL;

	p->Sync = 0;
	p->Fill = 0;

	if (!KeepPlay)
		SetPlay(p,1);

	p->AudioStream = -1;
	p->VideoStream = -1;

	if (p->Format)
	{
		ReleaseStreams(p);
		p->Format->Set(p->Format,FORMAT_INPUT,&NullPtr,sizeof(NullPtr));
	}

	if (p->Input)
		p->Input->Delete(p->Input);

	FreeComments(p);

	p->Title[0] = 0;
	PostMessage(p->Wnd,MSG_UPDATE,1,0);

	p->Input = NULL;
	p->Format = NULL;
	p->Timer = NULL;
	p->RefEnd = p->Ref;
	p->Position = -1;
	p->NoMoreInput = 0;
	p->CurrentChanged = 1;
	p->BufferWarning = 0;

	UpdateBacklight(p);
	UpdateRunInput(p);
	UpdateRunProcess(p);
}

static void URLToTitle( tchar_t* Title, const tchar_t* URL )
{
	tchar_t Ext[MAXPATH];
	tchar_t *i,*j;
	SplitURL(URL,NULL,NULL,Title,Ext);
	
	// replace %20 and '_' with space
	for (j=i=Title;*i;++i)
	{
		if (*i=='_') 
			*j++ = ' ';
		else
		if (i[0]=='%' && i[1]=='2' && i[2]=='0')
		{
			i += 2;
			*j++ = ' ';
		}
		else
			*j++ = *i;
	}
	*j=0;
}

static int Load(player_win32* p,bool_t Silent,bool_t OnlyLocal)
{
	tchar_t ContentType[MAXPATH];
	char* Probe;
	tchar_t* URL;
	stream* Input;
	int ProbeSize;
	int Result;
	int No;
	pin Pin;
	ref* Ref;
	bool_t b;
	fraction Pos = p->SeekAfterSync;
	p->SeekAfterSync.Num = -1;

	if (p->CurrentChanged || p->PlayList[p->Current].Changed || p->IOError)
	{
		Unload(p,1);

		// find input

		URL = p->PlayList[p->Current].URL;
		if (!URL[0])
			return ERR_NONE;

		if (OnlyLocal)
		{
			tchar_t Mime[MAXPATH];
			GetMime(URL,Mime);

			if (tcsicmp(Mime,T("file"))!=0)
				return ERR_FILE_NOT_FOUND;

			if (URL[0]=='\\' && URL[1]=='\\')
				return ERR_FILE_NOT_FOUND;
		}

		Input = GetStream(URL,Silent);
		if (!Input)
			return ERR_NOT_SUPPORTED;

		URLToTitle(p->Title,URL);

		memset(p->StreamSkip,0,sizeof(p->StreamSkip));
		p->Unit.Den = 1;
		p->Unit.Num = DEFAULT_TIMER_UNIT;

		SetTimerNode(p,NodeEnum(TIMER_CLASS,0));
		p->Input = Input;

		p->Input->Set(p->Input,STREAM_SILENT,&Silent,sizeof(Silent));

		Result = p->Input->Set(p->Input,STREAM_URL,URL,sizeof(tchar_t)*MAXPATH);

		b = 0;
		p->Input->Set(p->Input,STREAM_SILENT,&b,sizeof(b));

		if (Result != ERR_NONE)
		{
			Unload(p,Silent);
			return Result;
		}
		
		// find format
		Probe = Alloc(PROBE_SIZE);
		if (!Probe)
		{
			Unload(p,Silent);
			return ERR_OUT_OF_MEMORY;
		}

		if (p->Input->Get(p->Input,STREAM_CONTENTTYPE,ContentType,sizeof(ContentType))!=ERR_NONE)
			ContentType[0] = 0;

		ProbeSize = p->Input->Read(p->Input,Probe,PROBE_SIZE);
		if (ProbeSize > 0)
			for (No=0;(p->Format=(format*)NodeEnum(FORMAT_MEDIA_CLASS,No))!=NULL;++No)
			{
				if (p->Format->Probe(p->Format,Probe,ProbeSize,URL,ContentType) == ERR_NONE)
				{
					UpdateFormat(p,0); // set buffersize as soon as possible
					p->Format->Set(p->Format,FORMAT_DATAFEED,Probe,ProbeSize);

					Pin.Node = (node*)p;
					Pin.No = PLAYER_UPDATESTREAMS;
					p->Format->Set(p->Format,FORMAT_UPDATESTREAMS,&Pin,sizeof(Pin));

					if (p->Format->Set(p->Format,FORMAT_INPUT,&p->Input,sizeof(p->Input)) == ERR_NONE)
						break;
				}
			}

		if (!p->Format)
		{
			Free(Probe);
			Unload(p,0);
			if (!CheckExts(URL,T("exe:P")))
				NodeError((node*)p,PLAYER_ID,PLAYER_FORMAT_NOT_FOUND,URL);
			return ERR_NOT_SUPPORTED;
		}

		Free(Probe);

		p->PlayList[p->Current].Changed = 0;
		p->CurrentChanged = 0;
		p->NoMoreInput = 0;
		p->Position = 0;
		p->BufferPercent = PERCENT_ONE;
		p->Sync = 1; // start with syncing
		PostMessage(p->Wnd,MSG_UPDATE,1,0);

		// clear all buffers
		b = 1;
		for (Ref=p->Ref;Ref!=p->RefEnd;++Ref)
			if (Ref->Used && Ref->Flow)
				Ref->Node->Set(Ref->Node,FLOW_EMPTY,&b,sizeof(b));

		if (Pos.Num >= 0)
			p->SeekAfterSync = Pos; // restore syncaftersync (now we are sure there will a synced event)

		StopLoadMode(p);
		UpdatePlay(p);
	}
	else
	{
		UpdateFormat(p,0);
		UpdateStreams(p);

		if (Pos.Num >= 0)
			Seek(p,-1,&Pos);
		else
			Seek(p,0,NULL);
	}

	UpdateRunInput(p);
	UpdateRunProcess(p);
	UpdateBacklight(p);

	SetEvent(p->EventProcess); // wakeup input threads (if needed)
	return ERR_NONE;
}

static int UpdateWnd( player_win32* p )
{
	// set wnd for all vout devices
	int No;
	node* VOut;
	for (No=0;(VOut=NodeEnum(VOUT_CLASS,No))!=NULL;++No)
		VOut->Set(VOut,VOUT_WND,&p->Wnd,sizeof(void*));

	if (p->Wnd)
	{
		if (p->PlayListCount)
			Load(p,1,1);
		else
			p->SeekAfterSync.Num = -1;
	}
	else
	{
		if (!p->KeepList)
			SetPlayListCount(p,0);
		Unload(p,0);
	}

	return ERR_NONE;
}

static void Unlock( player_win32* p )
{
	//DEBUG_MSG2(DEBUG_PLAYER,T("Unlocked %d %08x"),p->LockCount,GetCurrentThreadId());
	LockLeave(p->LockInput);
	LockLeave(p->LockProcess);
	--p->LockCount;
	LockLeave(p->Lock);
}

static void Lock( player_win32* p )
{
	//DEBUG_MSG2(DEBUG_PLAYER,T("Locked Start %d %08x"),p->LockCount,GetCurrentThreadId());
	LockEnter(p->Lock);
	p->LockCount++;
	LockEnter(p->LockInput);
	LockEnter(p->LockProcess);
	//DEBUG_MSG(DEBUG_PLAYER,T("Locked End"));
}

static int NextPrev(player_win32* p,int Dir,bool_t StopIfEnd,bool_t NotTheSame)
{
	if (p->PlayListCount > 0)
	{
		int Index;
		int Tries;
		
		for (Index=0;Index<p->PlayListCount;++Index)
			if (p->PlayIndex[Index] == p->Current)
				break;

		if (Index == p->PlayListCount)
			Index = 0;

		for (Tries=NotTheSame?1:0;Tries<p->PlayListCount;++Tries)
		{
			Index += Dir;
			if (Index >= p->PlayListCount || Index < 0)
			{
				if (!p->Repeat)
				{
					if (StopIfEnd)
						SetPlay(p,1);
					return ERR_NONE;
				}
				Index -= Dir*p->PlayListCount;
			}
			if (p->Current != p->PlayIndex[Index])
			{
				p->Current = p->PlayIndex[Index];
				p->CurrentChanged = 1;
			}
			if (Load(p,1,0) == ERR_NONE)
				return ERR_NONE;
			if (!Dir)
				break;
		}

		SetPlay(p,1);
	}
	else
		Unload(p,0);

	return ERR_NEED_MORE_DATA;
}

static bool_t AllEmpty(player_win32* p)
{
	ref* Ref;
	bool_t b;

	if (p->Bench) // bench mode: no buffering
		return 1;

	for (Ref=p->Ref;Ref!=p->RefEnd;++Ref)
		if (Ref->Used && Ref->Flow &&
			Ref->Node->Get(Ref->Node,FLOW_EMPTY,&b,sizeof(b))==ERR_NONE && !b)
			return 0;

	return 1;
}

static void AddComment( player_win32* p,int Stream,const tchar_t* Comment)
{
	const tchar_t* s = tcschr(Comment,'=');
	if (s)
	{
		int n;
		comment c;

		c.Stream = Stream;

		n = s-Comment+1;
		if (n>sizeof(c.Name)/sizeof(tchar_t))
			n=sizeof(c.Name)/sizeof(tchar_t);
		
		TcsNCpy(c.Name,Comment,n);
		TcsNCpy(c.Value,s+1,sizeof(c.Value)/sizeof(tchar_t));

		for (n=0;n<p->CommentCount;++n)
			if (p->Comment[n]->Stream == c.Stream &&
				tcsicmp(p->Comment[n]->Name,c.Name)==0)
			{
				tcscpy(p->Comment[n]->Value,c.Value);
				break;
			}

		if (n==p->CommentCount && p->CommentCount < MAXCOMMENT)
		{
			p->Comment[p->CommentCount] = (comment*) Alloc(sizeof(comment));
			*p->Comment[p->CommentCount] = c;
			p->CommentCount++;
		}

		if (tcsicmp(c.Name,PlayerComment(COMMENT_TITLE))==0 || tcsicmp(c.Name,PlayerComment(COMMENT_ARTIST))==0)
			UpdateListTitle(p);
	}
}

int ProcessThread(player_win32* p)
{
	SAFE_BEGIN

	while (!p->Exit)
	{
		LockEnter(p->LockProcess);

		if (p->RunProcess)
		{
			bool_t WakeInput;
			int Result;

			p->Timer->Get(p->Timer,TIMER_TIME,&p->Time,sizeof(tick_t));

			DEBUG_MSG1(DEBUG_PLAYER,T("Process Time:%d"),p->Time);

			Result = p->Format->Process(p->Format,&p->Time,&WakeInput,p->Fill);

			if (Result == ERR_SYNCED)
			{
				DEBUG_MSG1(DEBUG_PLAYER,T("Synced: %d"),p->Time);

				p->Sync = 0;
				p->Position = p->Time;
				p->Timer->Set(p->Timer,TIMER_TIME,&p->Time,sizeof(tick_t));

				if (!p->Bench && p->SeekAfterSync.Num<0)
				{
					// enter fill mode (don't stop processing, even if playing if off)
					p->Fill = 1;
				}

				UpdateTimerUnit(p);
				UpdatePlay(p);
				UpdateRunProcess(p);
			}
			else
			if (p->Fill && (Result == ERR_END_OF_FILE || Result == ERR_BUFFER_FULL
				|| (Result == ERR_NEED_MORE_DATA && p->Used >= p->CurrBufferSize-32)))
			{
				DEBUG_MSG1(DEBUG_PLAYER,T("Filled: %d"),p->Time);
				
				if (Result == ERR_END_OF_FILE || Result == ERR_NEED_MORE_DATA)
					Result = ERR_BUFFER_FULL;

				p->Fill = 0;
				UpdatePlay(p);
				UpdateRunProcess(p);
			}

			if (Result == ERR_NEED_MORE_DATA)
			{
				p->BufferPercent = PERCENT_ONE;

				if (!p->LoadMode && (p->Play || p->FFwd) && !p->Bench && !p->Sync && !p->Fill)
				{
					if (p->Used >= (p->CurrBufferSize>>1))
					{
						if (!p->BufferWarning && p->Used >= p->CurrBufferSize-32)
						{
							p->BufferWarning = 1;
							NodeError((node*)p,PLAYER_ID,PLAYER_BUFFER_WARNING);
						}
					}
					else
					{
/*						if (!p->BufferUnderrun)
						{
							p->BufferUnderrun = 1;
							NodeError((node*)p,PLAYER_ID,PLAYER_BUFFER_UNDERRUN);
						}
						else
*/
						{
							DEBUG_MSG(DEBUG_TEST,T("LoadMode Enter"));

							p->LoadMode = 1;
							p->Fill = 1;
							PostMessage(p->Wnd,MSG_WAIT,1,0);
							//PostThreadMessage(p->ThreadId,MSG_WAIT,1,0);
							UpdatePlay(p);
							UpdateRunProcess(p);
							UpdateRunInput(p);
						}
					}
				}
			}

			if (!p->Sync)
				p->Position = p->Time;

			if (Result == ERR_END_OF_FILE && (p->Play || p->FFwd) && !AllEmpty(p))
				Result = ERR_BUFFER_FULL; // wait for all buffers to clear

			LockLeave(p->LockProcess);

			switch (Result)
			{
			case ERR_SYNCED:

				PostMessage(p->Wnd,MSG_UPDATE,0,0);
				if (p->Current < p->PlayListCount && !p->PlayList[p->Current].Changed && 
					p->PlayList[p->Current].Length<0)
				{
					tick_t Duration;
					if (p->Format->Get(p->Format,FORMAT_DURATION,&Duration,sizeof(Duration))==ERR_NONE)
						UpdateListLength(p,Duration);
				}

				if (p->FullScreen && !p->VideoUsed)
					PostMessage(p->Wnd,MSG_FULLSCREEN,0,0);

				if (p->SeekAfterSync.Num>=0)
				{
					fraction Pos;
					Lock(p);
					Pos = p->SeekAfterSync;
					p->SeekAfterSync.Num = -1;
					Seek(p,-1,&Pos);
					Unlock(p);
				}
				break;

			case ERR_END_OF_FILE:
				if (p->Play || p->FFwd || p->Sync)
				{
					Lock(p);
					if (p->Bench)
						SetPlay(p,1); // stop benchmarking and show results
					else
						NextPrev(p,1,1,p->Sync || p->Time < TICKSPERSEC);
					Unlock(p);
				}
				break;

			case ERR_NEED_MORE_DATA:
				DEBUG_MSG(DEBUG_PLAYER,T("Process Needs More Data Sleep"));

				SetEvent(p->EventProcess); // signal InputThread to coninue
				Sleep(25);
				break;

			case ERR_BUFFER_FULL: // no packets processed, we should wait
				DEBUG_MSG(DEBUG_PLAYER,T("Process Buffer Full Sleep"));

				Sleep(4); // this should be more intelligent...
				break;

			default:
				if (p->WaitForProcess && WakeInput)
				{
					SetEvent(p->EventProcess); // signal InputThread to coninue
					if (p->Sync)
						Sleep(5); // invalid media could block everything (going through the file)
				}

				if (p->LockCount > 0 || p->Used < p->UsedEnough)
				{
					DEBUG_MSG(DEBUG_PLAYER,T("Process Sleep"));
					Sleep(0);
				}
				break;
			}
		}
		else
		{
			LockLeave(p->LockProcess);
			WaitForSingleObject(p->EventRunProcess,INFINITE);
		}
	}

	SAFE_END

	return 0;
}

int InputThread(player_win32* p)
{
	SAFE_BEGIN

	p->Used = p->UsedEnough;
	p->IOError = 0;
	while (!p->Exit)
	{
		LockEnter(p->LockInput);

		if (p->RunInput)
		{
			int Result;

			Result = p->Format->Read(p->Format,p->BufferPercent,&p->Used);

			if (Result == ERR_DEVICE_ERROR)
			{
				if (++p->IOError >= 20)
					Result = ERR_END_OF_FILE;
			}
			else
				p->IOError = 0;

			if (Result == ERR_END_OF_FILE)
			{
				p->NoMoreInput = 1;
				UpdateRunInput(p);
			}

			DEBUG_MSG3(DEBUG_PLAYER,T("Input read %dKB (%d) percent:%d"),p->Used,Result,p->BufferPercent);

			LockLeave(p->LockInput);

			switch (Result)
			{
			case ERR_DEVICE_ERROR:
				Sleep(100);
				break;

			case ERR_BUFFER_FULL:

				if (p->LoadMode)
				{
					Lock(p);
					StopLoadMode(p);
					Unlock(p);
				}

				if (!p->BufferPercent)
					p->BufferPercent = PERCENT_ONE;

				if (p->MicroDrive && p->BufferPercent==PERCENT_ONE)
				{
					// buffer fully loaded, wait until it's almost empty
					p->BufferPercent = p->BurstStart;

					DEBUG_MSG1(DEBUG_TEST,T("MicroDrive Full: new percent %d"),p->BufferPercent);
				}

				DEBUG_MSG(DEBUG_PLAYER,T("Input waiting for process"));

				p->WaitForProcess = 1;
				WaitForSingleObject(p->EventProcess,2000);
				p->WaitForProcess = 0;

				DEBUG_MSG(DEBUG_PLAYER,T("Input waiting over"));
				break;

			case ERR_END_OF_FILE:
				if (p->LoadMode)
				{
					Lock(p);
					StopLoadMode(p);
					Unlock(p);
				}
				
				//no break;
			default:
				if (p->BufferPercent != PERCENT_ONE)
				{
					// microdrive started loading, change back to full buffer
					p->BufferPercent = PERCENT_ONE;

					DEBUG_MSG(DEBUG_TEST,T("MicroDrive needs data"));
				}

				if (p->LoadMode && !p->Fill && p->Used >= (p->CurrBufferSize * p->UnderRun)/PERCENT_ONE)
				{
					Lock(p);
					StopLoadMode(p);
					Unlock(p);
				}

				if (p->MicroDrive && p->Used >= (MINBUFFER/1024))
					Sleep(2); // try to give some time to ProcessThread during read bursts
				else
				if (p->LockCount > 0 || p->Used >= p->UsedEnough)
					Sleep(0);
				break;
			}
		}
		else
		{
			LockLeave(p->LockInput);
			WaitForSingleObject(p->EventRunInput,INFINITE);
		}
	}

	SAFE_END

	return 0;
}

static bool_t VOutputCaps( player_win32* p, int Caps )
{
	int i;
	return !p->VOutput || p->VOutput->Get(p->VOutput,VOUT_CAPS,&i,sizeof(i))!=ERR_NONE || (i & Caps);
}

static int Get( player_win32* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;

	if (No == NODE_ID) 
	{ 
		GETVALUE(PLAYER_ID,int); 
		return Result; 
	}

	LockEnter(p->Lock);

	if (No >= PLAYER_ARRAY)
	{
		if (No >= PLAYER_LIST_URL && No < PLAYER_LIST_URL+p->PlayListCount)
		{
			No -= PLAYER_LIST_URL;
			TcsNCpy((tchar_t*)Data,p->PlayList[No].URL,Size/sizeof(tchar_t));
			Result = ERR_NONE;
		}
		else
		if (No >= PLAYER_LIST_TITLE && No < PLAYER_LIST_TITLE+p->PlayListCount)
		{
			No -= PLAYER_LIST_TITLE;
			TcsNCpy((tchar_t*)Data,p->PlayList[No].Title,Size/sizeof(tchar_t));
			Result = ERR_NONE;
		}
		else
		if (No >= PLAYER_LIST_AUTOTITLE && No < PLAYER_LIST_AUTOTITLE+p->PlayListCount)
		{
			No -= PLAYER_LIST_AUTOTITLE;
			if (p->PlayList[No].Title[0])
				TcsNCpy((tchar_t*)Data,p->PlayList[No].Title,Size/sizeof(tchar_t));
			else
				URLToTitle((tchar_t*)Data,p->PlayList[No].URL);
			Result = ERR_NONE;
		}
		else
		if (No >= PLAYER_LIST_LENGTH && No < PLAYER_LIST_LENGTH+p->PlayListCount)
		{
			No -= PLAYER_LIST_LENGTH;
			*(tick_t*)Data = p->PlayList[No].Length;
			Result = ERR_NONE;
		}
	}
	else
	switch (No)
	{
	case PLAYER_BUFFER_SIZE: GETVALUE(p->BufferSize,int); break;
	case PLAYER_MD_BUFFER_SIZE: GETVALUE(p->MDBufferSize,int); break;
	case PLAYER_MICRODRIVE: GETVALUE(p->MicroDrive,bool_t); break;
	case PLAYER_BURSTSTART: GETVALUE(p->BurstStart,int); break;
	case PLAYER_UNDERRUN: GETVALUE(p->UnderRun,int); break;
	case PLAYER_REPEAT: GETVALUE(p->Repeat,bool_t); break;
	case PLAYER_SHUFFLE: GETVALUE(p->Shuffle,bool_t); break;
	case PLAYER_PLAYATOPEN: GETVALUE(p->PlayAtOpen,bool_t); break;
	case PLAYER_PLAYATOPEN_FULL: GETVALUE(p->PlayAtOpenFull,bool_t); break;
	case PLAYER_KEEPPLAY_VIDEO: GETVALUE(p->KeepPlayVideo,bool_t); break;
	case PLAYER_KEEPPLAY_AUDIO: GETVALUE(p->KeepPlayAudio,bool_t); break;
	case PLAYER_KEEPLIST: GETVALUE(p->KeepList,bool_t); break;
	case PLAYER_MOVEBACK: GETVALUE(p->MoveBack,tick_t); break;
	case PLAYER_MOVEFFWD: GETVALUE(p->MoveFFwd,tick_t); break;
	case PLAYER_PLAY: GETVALUE(p->Play,bool_t); break;
	case PLAYER_FFWD: GETVALUE(p->FFwd,bool_t); break;
	case PLAYER_PLAY_SPEED: GETVALUE(p->PlaySpeed,fraction); break;
	case PLAYER_FFWD_SPEED: GETVALUE(p->FFwdSpeed,fraction); break;
	case PLAYER_TITLE: GETSTRING(p->Title); break;
	case PLAYER_INPUT: GETVALUE(p->Input,stream*); break;
	case PLAYER_FORMAT: GETVALUE(p->Format,format*); break;
	case PLAYER_AOUTPUT: GETVALUE(p->AOutput,node*); break;
	case PLAYER_VOUTPUT: GETVALUE(p->VOutput,node*); break;
	case PLAYER_LIST_COUNT: GETVALUE(p->PlayListCount,int); break;
	case PLAYER_LIST_CURRENT: GETVALUE(p->Current,int); break;
	case PLAYER_EQUALIZER: GETVALUE(p->Eq,equalizer); break;
	case PLAYER_EQENABLE: GETVALUE(p->Eq.Enabled,bool_t); break;
	case PLAYER_VIDEO_ACCEL: GETVALUE(p->VideoAccel,bool_t); break;
	case PLAYER_AUDIO_SWAP: GETVALUE(p->AudioSwap,bool_t); break;
	case PLAYER_AUDIO_QUALITY: GETVALUE(p->AudioQuality,int); break;
	case PLAYER_FULLSCREEN: GETVALUE(p->FullScreen,bool_t); break;
	case PLAYER_AUTOPREROTATE: GETVALUE(p->AutoPreRotate,bool_t); break;
	case PLAYER_SKIN_VIEWPORT: GETVALUE(p->SkinViewport,rect); break;
	case PLAYER_CLIPPING: GETVALUE(p->Clipping,bool_t); break;
	case PLAYER_WND: GETVALUE(p->Wnd,void*); break;
	case PLAYER_ASPECT: GETVALUE(p->Aspect,fraction); break;
	case PLAYER_STEREO: GETVALUE(p->Stereo,int); break;
	case PLAYER_DITHER: GETVALUE(p->Dither,bool_t); if (!VOutputCaps(p,VC_DITHER)) Result = ERR_NOT_SUPPORTED; break;
	case PLAYER_ZOOM: GETVALUE(p->Zoom,fraction); break;
	case PLAYER_SMOOTH50: GETVALUE(p->SmoothZoom50,bool_t); break;
	case PLAYER_SMOOTHALWAYS: GETVALUE(p->SmoothZoomAlways,bool_t); break;
	case PLAYER_FULL_DIR: GETVALUE(p->FullDir,bool_t); break;
	case PLAYER_SKIN_DIR: GETVALUE(p->SkinDir,bool_t); break;
	case PLAYER_BENCHMARK: GETVALUECOND(p->BenchTime,tick_t,!p->Bench); break;
	case PLAYER_POSITION: GETVALUECOND(p->Position,tick_t,p->Position>=0); break;
	case PLAYER_BACKGROUND: GETVALUE(p->Background,bool_t); break;
	case PLAYER_AUDIOUSED: GETVALUE(p->AudioUsed,bool_t); break;
	case PLAYER_VIDEOUSED: GETVALUE(p->VideoUsed,bool_t); break;
	case PLAYER_VSTREAM: GETVALUE(p->VideoStream,int); break;
	case PLAYER_ASTREAM: GETVALUE(p->AudioStream,int); break;
	case PLAYER_LIST_CHANGED: GETVALUE(p->ListChanged,pin); break;
	case PLAYER_SYNCING: GETVALUE(p->Sync || !p->Format,bool_t); break;

	case PLAYER_PERCENT:
		if (p->Format && p->Input)
		{
			fraction* Percent = (fraction*)Data;
			tick_t Duration;
			int FilePos;
			int FileSize;

			if (p->Position >= 0 && p->Format->Get(p->Format,FORMAT_DURATION,&Duration,sizeof(Duration))==ERR_NONE)
			{
				UpdateListLength(p,Duration);
				Percent->Num = p->Position;
				Percent->Den = Duration;
				Result = ERR_NONE;
			}
			else
			if (p->Position==0)
			{
				Percent->Num = 0;
				Percent->Den = 0;
				Result = ERR_NONE;
			}
			else
			if (p->Input->Get(p->Input,STREAM_LENGTH,&FileSize,sizeof(FileSize))==ERR_NONE &&
				p->Format->Get(p->Format,FORMAT_FILEPOS,&FilePos,sizeof(FilePos))==ERR_NONE)
			{
				Percent->Num = FilePos;
				Percent->Den = FileSize;
				Result = ERR_NONE;
			}
		}
		break;
	case PLAYER_MUTE: 
		if (p->AOutput)
			Result = p->AOutput->Get(p->AOutput,AOUT_MUTE,Data,Size);
		break;

	case PLAYER_VOLUME:
		if (p->AOutput)
			Result = p->AOutput->Get(p->AOutput,AOUT_VOLUME,Data,Size);
		break;

	case PLAYER_DURATION:
		if (p->Format)
		{
			Result = p->Format->Get(p->Format,FORMAT_DURATION,Data,Size);
			if (Result == ERR_NONE) 
				UpdateListLength(p,*(tick_t*)Data);
		}
		break;
	}

	LockLeave(p->Lock);
	return Result;
}

static int ColGet( void* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	LockEnter(PLAYERCOL(p)->Lock);

	switch (No)
	{
	case NODE_ID: GETVALUE(COLOR_ID,int); break;
	case COLOR_RESET: Result = ERR_NONE; break;
	case COLOR_BRIGHTNESS: GETVALUE(PLAYERCOL(p)->Brightness,int); if (!VOutputCaps(PLAYERCOL(p),VC_BRIGHTNESS)) Result = ERR_NOT_SUPPORTED; break;
	case COLOR_CONTRAST: GETVALUE(PLAYERCOL(p)->Contrast,int); if (!VOutputCaps(PLAYERCOL(p),VC_CONTRAST)) Result = ERR_NOT_SUPPORTED; break;
	case COLOR_SATURATION: GETVALUE(PLAYERCOL(p)->Saturation,int); if (!VOutputCaps(PLAYERCOL(p),VC_SATURATION)) Result = ERR_NOT_SUPPORTED; break;
	}
	LockLeave(PLAYERCOL(p)->Lock);
	return Result;
}

static int EqGet( void* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	LockEnter(PLAYEREQ(p)->Lock);

	switch (No)
	{
	case NODE_ID: GETVALUE(EQUALIZER_ID,int); break;
	case EQUALIZER_RESET: Result = ERR_NONE; break;
	case EQUALIZER_ENABLED: GETVALUE(PLAYEREQ(p)->Eq.Enabled,bool_t); break;
	case EQUALIZER_PREAMP: GETVALUE(PLAYEREQ(p)->Eq.Amplify,int); break;
	case EQUALIZER_1: GETVALUE(PLAYEREQ(p)->Eq.Eq[0],int); break;
	case EQUALIZER_2: GETVALUE(PLAYEREQ(p)->Eq.Eq[1],int); break;
	case EQUALIZER_3: GETVALUE(PLAYEREQ(p)->Eq.Eq[2],int); break;
	case EQUALIZER_4: GETVALUE(PLAYEREQ(p)->Eq.Eq[3],int); break;
	case EQUALIZER_5: GETVALUE(PLAYEREQ(p)->Eq.Eq[4],int); break;
	case EQUALIZER_6: GETVALUE(PLAYEREQ(p)->Eq.Eq[5],int); break;
	case EQUALIZER_7: GETVALUE(PLAYEREQ(p)->Eq.Eq[6],int); break;
	case EQUALIZER_8: GETVALUE(PLAYEREQ(p)->Eq.Eq[7],int); break;
	case EQUALIZER_9: GETVALUE(PLAYEREQ(p)->Eq.Eq[8],int); break;
	case EQUALIZER_10:GETVALUE(PLAYEREQ(p)->Eq.Eq[9],int); break;
	}
	LockLeave(PLAYEREQ(p)->Lock);
	return Result;
}

static void SetAOutput( player_win32* p, node* AOutput )
{
	p->AOutput = AOutput;

	UpdateAudio(p);
	UpdateStreams(p);
}

static void SetVOutput( player_win32* p, node* VOutput )
{
	bool_t Save;

	p->VOutput = VOutput;
	p->VIDCT = NULL;

	if (VOutput)
	{
		VOutput->Get(VOutput,VOUT_ACCEL_IDCT,&p->VIDCT,sizeof(p->VIDCT));
		VOutput->Set(VOutput,VOUT_PLAY,&p->TimerPlay,sizeof(bool_t));
	}

	if (!p->VIDCT)
		p->VideoAccel = 0;

	Save = p->VideoUsed;
	p->VideoUsed = 0; // disable refresh
	UpdateVideo(p); // first set FX,etc...
	p->VideoUsed = Save;

	UpdateStreams(p); // init overlay (it will call UpdateVideo and UpdateBacklight)
}

static int Set( player_win32* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	
	if (No == PLAYER_UPDATESTREAMS)
	{
		// called by process thread
		// we are inside the LockProcess so Lock/Unlock not needed
		// inputprocess doesn't bother us here
		UpdateStreams(p);
		return ERR_NONE;
	}

	if (No >= PLAYER_COMMENT && No < PLAYER_COMMENT+MAXSTREAM)
	{
		AddComment(p,No-PLAYER_COMMENT,(const tchar_t*)Data);
		return ERR_NONE;
	}

	Lock(p);
	if (No >= PLAYER_LIST_URL && No <= PLAYER_LIST_URL+p->PlayListCount)
	{
		No -= PLAYER_LIST_URL;
		if (No == p->PlayListCount)
			SetPlayListCount(p,No+1);
		if (No < p->PlayListCount)
		{
			if (!Size) Data = T("");
			p->PlayList[No].Changed = 1;
			TcsNCpy(p->PlayList[No].URL,(tchar_t*)Data,sizeof(p->PlayList[No].URL)/sizeof(tchar_t));
			UpdateList(p);
			Result = ERR_NONE;
		}
	}
	else
	if (No >= PLAYER_LIST_TITLE && No <= PLAYER_LIST_TITLE+p->PlayListCount)
	{
		No -= PLAYER_LIST_TITLE;
		if (No == p->PlayListCount)
			SetPlayListCount(p,No+1);
		if (No < p->PlayListCount)
		{
			if (!Size) Data = T("");
			TcsNCpy(p->PlayList[No].Title,(tchar_t*)Data,sizeof(p->PlayList[No].Title)/sizeof(tchar_t));
			UpdateList(p);
			Result = ERR_NONE;
		}
	}
	else
	if (No >= PLAYER_LIST_LENGTH && No <= PLAYER_LIST_LENGTH+p->PlayListCount)
	{
		No -= PLAYER_LIST_LENGTH;
		if (No == p->PlayListCount)
			SetPlayListCount(p,No+1);
		if (No < p->PlayListCount)
		{
			p->PlayList[No].Length = *(tick_t*)Data;
			UpdateList(p);
			Result = ERR_NONE;
		}
	}
	else
	switch (No)
	{
	case PLAYER_BUFFER_SIZE: SETVALUE(p->BufferSize,int,UpdateFormat(p,1)); break;
	case PLAYER_MD_BUFFER_SIZE: SETVALUE(p->MDBufferSize,int,UpdateFormat(p,1)); break;
	case PLAYER_PLAY_SPEED: SETVALUE(p->PlaySpeed,fraction,UpdateSpeed(p)); break;
	case PLAYER_FFWD_SPEED: SETVALUE(p->FFwdSpeed,fraction,UpdateSpeed(p)); break;
	case PLAYER_EQENABLE: SETVALUECMP(p->Eq.Enabled,bool_t,UpdateAllEqualizer(p),EqBool); break;
	case PLAYER_EQUALIZER: SETVALUE(p->Eq,equalizer,UpdateAllEqualizer(p)); break;

	case PLAYER_BENCHMARK:
		if (p->Format && p->Input)
			StartBench(p);
		break;

	case PLAYER_PLAY:
		if (Size == sizeof(bool_t))
		{
			if (*(bool_t*)Data)
			{
				if (!p->Format && p->Wnd && p->PlayListCount)
					Load(p,0,0);

				if (!p->Format)
				{
					Result = ERR_INVALID_DATA;
					break;
				}

				p->Play = 1;
				p->FFwd = 0;
			}
			else
			{
				p->Play = 0;
				p->FFwd = 0;
				p->Fill = 1;
			}
			Result = SetPlay(p,0);
		}
		break;

	case PLAYER_FFWD:
		if (Size == sizeof(bool_t))
		{
			if (*(bool_t*)Data)
			{
				if (!p->Format && p->Wnd && p->PlayListCount)
					Load(p,0,0);

				if (!p->Format)
				{
					Result = ERR_INVALID_DATA;
					break;
				}

				p->FFwd = 1;
			}
			else
			{
				p->FFwd = 0;
				if (!p->Play && !p->Bench) 
					p->Fill = 1;
			}
			Result = SetPlay(p,0);
		}
		break;

	case PLAYER_LIST_COUNT:
		if (Size == sizeof(int))
			Result = SetPlayListCount(p,*(int*)Data);
		break;

	case PLAYER_LIST_CURRENT:
		if (Size == sizeof(int) && *(int*)Data<p->PlayListCount)
		{
			if (*(int*)Data<0)
				Unload(p,0);
			else
			{
				if (p->Current != *(int*)Data)
				{
					p->Current = *(int*)Data;
					p->CurrentChanged = 1;
				}
				if (p->Wnd && Load(p,0,0) == ERR_NONE && (p->PlayAtOpen || p->PlayAtOpenFull))
				{
					if (p->PlayAtOpenFull)
						PostMessage(p->Wnd,MSG_FULLSCREEN,1,0);

					p->Play = 1;
					p->FFwd = 0;
					SetPlay(p,0);
				}
			}
			Result = ERR_NONE;
		}
		break;

	case PLAYER_PERCENT:
		if (Size == sizeof(fraction))
			Result = Seek(p,-1,(const fraction*)Data);									
		break;

	case PLAYER_POSITION:
		if (Size == sizeof(tick_t))
		{
			tick_t t = *(const tick_t*)Data;
			if (t<0) t=0;
			Result = Seek(p,t,NULL);
		}
		break;

	case PLAYER_MICRODRIVE: SETVALUE(p->MicroDrive,bool_t,UpdateFormat(p,1)); break;
	case PLAYER_BURSTSTART: SETVALUE(p->BurstStart,int,ERR_NONE); break;
	case PLAYER_UNDERRUN: SETVALUE(p->UnderRun,int,ERR_NONE); break;
	case PLAYER_REPEAT: SETVALUE(p->Repeat,bool_t,ERR_NONE); break;
	case PLAYER_PLAYATOPEN: SETVALUE(p->PlayAtOpen,bool_t,ERR_NONE); break;
	case PLAYER_PLAYATOPEN_FULL: SETVALUE(p->PlayAtOpenFull,bool_t,ERR_NONE); break;
	case PLAYER_KEEPPLAY_AUDIO: SETVALUE(p->KeepPlayAudio,bool_t,ERR_NONE); break;
	case PLAYER_KEEPPLAY_VIDEO: SETVALUE(p->KeepPlayVideo,bool_t,ERR_NONE); break;
	case PLAYER_KEEPLIST: SETVALUE(p->KeepList,bool_t,ERR_NONE); break;
	case PLAYER_MOVEBACK: SETVALUE(p->MoveBack,tick_t,ERR_NONE); break;
	case PLAYER_MOVEFFWD: SETVALUE(p->MoveFFwd,tick_t,ERR_NONE); break;
	case PLAYER_SHUFFLE: SETVALUE(p->Shuffle,bool_t,BuildPlayIndex(p)); break;
	case PLAYER_AUDIO_SWAP: SETVALUECMP(p->AudioSwap,bool_t,UpdateAudio(p),EqBool); break;
	case PLAYER_AUDIO_QUALITY: SETVALUECMP(p->AudioQuality,int,UpdateAudio(p),EqInt); break;
	case PLAYER_FULLSCREEN: SETVALUECMP(p->FullScreen,bool_t,UpdateVideo(p),EqBool); break;
	case PLAYER_AUTOPREROTATE: SETVALUECMP(p->AutoPreRotate,bool_t,UpdateVideo(p),EqBool); break;
	case PLAYER_SKIN_VIEWPORT: SETVALUECMP(p->SkinViewport,rect,UpdateVideo(p),EqRect); break;
	case PLAYER_CLIPPING: SETVALUECMP(p->Clipping,bool_t,UpdateVideo(p),EqBool); break;
	case PLAYER_WND: SETVALUECMP(p->Wnd,void*,UpdateWnd(p),EqPtr); break;
	case PLAYER_ASPECT: SETVALUECMP(p->Aspect,fraction,UpdateVideo(p),EqFrac); break;
	case PLAYER_STEREO: SETVALUECMP(p->Stereo,int,UpdateAudio(p),EqInt); break;
	case PLAYER_DITHER: SETVALUECMP(p->Dither,bool_t,UpdateVideo(p),EqBool); break;
	case PLAYER_ZOOM: SETVALUECMP(p->Zoom,fraction,UpdateVideo(p),EqFrac); break;
	case PLAYER_SMOOTH50: SETVALUECMP(p->SmoothZoom50,bool_t,UpdateVideo(p),EqBool); break;
	case PLAYER_SMOOTHALWAYS: SETVALUECMP(p->SmoothZoomAlways,bool_t,UpdateVideo(p),EqBool); break;
	case PLAYER_FULL_DIR: SETVALUECMP(p->FullDir,int,UpdateVideo(p),EqInt); break;
	case PLAYER_SKIN_DIR: SETVALUECMP(p->SkinDir,int,UpdateVideo(p),EqInt); break;
	case PLAYER_VIDEO_ACCEL: SETVALUE(p->VideoAccel,bool_t,ERR_NONE); break;
	case PLAYER_BACKGROUND: SETVALUECMP(p->Background,bool_t,UpdateBackground(p),EqBool); break;
	case PLAYER_VSTREAM: SETVALUECMP(p->VideoStream,int,UpdateStreams(p),EqInt); break;
	case PLAYER_ASTREAM: SETVALUECMP(p->AudioStream,int,UpdateStreams(p),EqInt); break;
	case PLAYER_LIST_CHANGED: SETVALUE(p->ListChanged,pin,ERR_NONE); break;

	case PLAYER_MUTE:
		if (p->AOutput)
			Result = p->AOutput->Set(p->AOutput,AOUT_MUTE,Data,Size);
		break;

	case PLAYER_VOLUME:
		if (p->AOutput)
			Result = p->AOutput->Set(p->AOutput,AOUT_VOLUME,Data,Size);
		break;

	case PLAYER_UPDATEVIDEO: 
		Result = UpdateVideo(p); 
		break;

	case PLAYER_ROTATEEND: 
		p->Rotating = 0;
		Result = UpdateVideo(p); 
		break;

	case PLAYER_RESETVIDEO:
		if (p->VOutput)
			p->VOutput->Set(p->VOutput,VOUT_RESET,NULL,0);
		break;

	case PLAYER_ROTATEBEGIN:
		p->Rotating = 1;
		if (p->VOutput)
		{
			bool_t Visible = 0;
			p->VOutput->Set(p->VOutput,VOUT_VISIBLE,&Visible,sizeof(Visible));
			p->VOutput->Set(p->VOutput,VOUT_RESET,NULL,0);
		}
		break;

	case PLAYER_VOUTPUT:
		if (Size == sizeof(node*) && NodeIsClass(*(node**)Data,VOUT_CLASS))
		{
			SetVOutput(p,*(node**)Data);
			Result = ERR_NONE;
		}
		break;

	case PLAYER_AOUTPUT:
		if (Size == sizeof(node*) && NodeIsClass(*(node**)Data,AOUT_CLASS))
		{
			SetAOutput(p,*(node**)Data);
			Result = ERR_NONE;									
		}
		break;
	}
	Unlock(p);
	return Result;
}


static int ColSet( void* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;

	Lock(PLAYERCOL(p));

	switch (No)
	{
	case COLOR_BRIGHTNESS: SETVALUECMP(PLAYERCOL(p)->Brightness,int,UpdateVideo(PLAYERCOL(p)),EqInt); break;
	case COLOR_CONTRAST: SETVALUECMP(PLAYERCOL(p)->Contrast,int,UpdateVideo(PLAYERCOL(p)),EqInt); break;
	case COLOR_SATURATION: SETVALUECMP(PLAYERCOL(p)->Saturation,int,UpdateVideo(PLAYERCOL(p)),EqInt); break;
	}
	Unlock(PLAYERCOL(p));
	return Result;
}

static int EqSet( void* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;

	Lock(PLAYEREQ(p));

	switch (No)
	{
	case EQUALIZER_ENABLED: SETVALUE(PLAYEREQ(p)->Eq.Enabled,bool_t,UpdateAllEqualizer(PLAYEREQ(p))); break;
	case EQUALIZER_PREAMP: SETVALUE(PLAYEREQ(p)->Eq.Amplify,int,UpdateAllEqualizer(PLAYEREQ(p))); break;
	case EQUALIZER_1: SETVALUE(PLAYEREQ(p)->Eq.Eq[0],int,UpdateAllEqualizer(PLAYEREQ(p))); break;
	case EQUALIZER_2: SETVALUE(PLAYEREQ(p)->Eq.Eq[1],int,UpdateAllEqualizer(PLAYEREQ(p))); break;
	case EQUALIZER_3: SETVALUE(PLAYEREQ(p)->Eq.Eq[2],int,UpdateAllEqualizer(PLAYEREQ(p))); break;
	case EQUALIZER_4: SETVALUE(PLAYEREQ(p)->Eq.Eq[3],int,UpdateAllEqualizer(PLAYEREQ(p))); break;
	case EQUALIZER_5: SETVALUE(PLAYEREQ(p)->Eq.Eq[4],int,UpdateAllEqualizer(PLAYEREQ(p))); break;
	case EQUALIZER_6: SETVALUE(PLAYEREQ(p)->Eq.Eq[5],int,UpdateAllEqualizer(PLAYEREQ(p))); break;
	case EQUALIZER_7: SETVALUE(PLAYEREQ(p)->Eq.Eq[6],int,UpdateAllEqualizer(PLAYEREQ(p))); break;
	case EQUALIZER_8: SETVALUE(PLAYEREQ(p)->Eq.Eq[7],int,UpdateAllEqualizer(PLAYEREQ(p))); break;
	case EQUALIZER_9: SETVALUE(PLAYEREQ(p)->Eq.Eq[8],int,UpdateAllEqualizer(PLAYEREQ(p))); break;
	case EQUALIZER_10:SETVALUE(PLAYEREQ(p)->Eq.Eq[9],int,UpdateAllEqualizer(PLAYEREQ(p))); break;
	}
	Unlock(PLAYEREQ(p));
	return Result;
}

static void CmdReSync(player_win32* p)
{
	Lock(p);
	if (p->Format && p->Input && p->Timer)
	{
		if (p->Position>=0)
			Seek(p,p->Position,NULL);
		else
			ReSync(p);
	}
	Unlock(p);
}

static void CmdUnload(player_win32* p)
{
	Lock(p);
	Unload(p,0);
	Unlock(p);
}

static void CmdStop(player_win32* p)
{
	Lock(p);
	SetPlay(p,1);
	Unlock(p);
}

static void CmdSwap(player_win32* p,int a,int b)
{
	Lock(p);
	if (a < p->PlayListCount && b < p->PlayListCount && a!=b)
	{
		item Tmp;

		Tmp = p->PlayList[a];
		p->PlayList[a] = p->PlayList[b];
		p->PlayList[b] = Tmp;

		if (p->Current == a)
			p->Current = b;
		else
		if (p->Current == b)
			p->Current = a;

		p->PlayList[a].Changed = 1;
		p->PlayList[b].Changed = 1;

		UpdateList(p);
	}
	Unlock(p);
}

static void CmdGo(player_win32* p,int Delta)
{
	Lock(p);
	NextPrev(p,Delta,0,0);
	Unlock(p);
}

static bool_t CmdComment(player_win32* p,int Stream,const tchar_t* Name,tchar_t* Value,int Size)
{
	int i;
	bool_t Result = 0;

	Lock(p);

	for (i=0;i<p->CommentCount;++i)
		if ((Stream<0 || p->Comment[i]->Stream == Stream) &&
			tcsicmp(p->Comment[i]->Name,Name)==0)
		{
			if (Result)
			{
				Result = 0;
				break;
			}
			if (Size)
				TcsNCpy(Value,p->Comment[i]->Value,Size/sizeof(tchar_t));
			Result = 1;
			if (Stream>=0)
				break;
		}

	Unlock(p);
	return Result;
}

static void Paint(player_win32* p,void* DC,int x0,int y0)
{
	RECT R;
	rect r;
	int32_t ColorKey;

	if (p->Primary)
	{
		if (p->VOutput && p->VOutput->Get(p->VOutput,VOUT_OUTPUTRECT,&r,sizeof(r))==ERR_NONE && r.Width>0 && r.Height>0)
		{
			R.left = r.x - x0;
			R.right = R.left + r.Width;
			R.top = r.y - y0;
			R.bottom = R.top + r.Height;

			if (p->VOutput->Get(p->VOutput,VOUT_COLORKEY,&ColorKey,sizeof(ColorKey)) == ERR_NONE)
			{
				if (ColorKey != p->ColorKey)
				{
					p->ColorKey = ColorKey;
					p->ColorKeyBrush = CreateSolidBrush(ColorKey);			
				}
				FillRect(DC,&R,p->ColorKeyBrush);
				ExcludeClipRect(DC,R.left,R.top,R.right,R.bottom);
			}
			else
			if (!p->Clipping && !p->Overlay)
				ExcludeClipRect(DC,R.left,R.top,R.right,R.bottom);
		}

		r = p->Viewport;
	}
	else
		r = p->SkinViewport;

	R.left = r.x - x0;
	R.right = R.left + r.Width;
	R.top = r.y - y0;
	R.bottom = R.top + r.Height;

	DEBUG_MSG4(DEBUG_VIDEO,T("Player Paint %d,%d,%d,%d"),R.left,R.top,R.right,R.bottom);

	FillRect(DC,&R,p->BlackBrush);
}

static player_win32 Player_Win32 = 
{
	{
		Enum,
		Get,
		Set,
		Paint,
		CmdStop,
		CmdUnload,
		CmdReSync,
		CmdComment,
		CmdGo,
		CmdSwap,
	},
	{
		EqEnum,
		EqGet,
		EqSet,
	},
	{
		ColEnum,
		ColGet,
		ColSet,
	}
};

void Player_Init()
{
	video Desktop;
	bool_t SwapXY;
	player_win32* p = &Player_Win32;
	QueryDesktop(&Desktop);

	// default values
	if (QueryPlatform(PLATFORM_TYPENO) == TYPE_SMARTPHONE)
	{
		p->UsedEnough = 256;
		p->BufferSize = 600;
	}
	else
	{
		p->UsedEnough = 512;
		p->BufferSize = QueryPlatform(PLATFORM_VER) >= 421 ? 2400 : 1200;
	}
	p->MDBufferSize = 16000;
	p->BufferPercent = 0;
	p->BurstStart = PERCENT_ONE/10;
	p->UnderRun = PERCENT_ONE*7/10;
	p->MoveFFwd = p->MoveBack = TICKSPERSEC*10;
	p->PlayAtOpen = 0;
	p->PlayAtOpenFull = 0;
	p->KeepList = 1;
	p->KeepPlayAudio = 1;
	p->KeepPlayVideo = 0;
	p->AutoPreRotate = 1;
	p->AudioSwap = 0;
	p->Zoom.Den = 1;
	p->Zoom.Num = 0;
#if !defined(SH3) && !defined(MIPS)
	p->Dither = Desktop.Width*Desktop.Height < 480*320;
	p->AudioQuality = 2;
	p->SmoothZoom50 = 1;
	p->SmoothZoomAlways = 0;
#else
	p->Dither = 0;
	p->AudioQuality = 1;
	p->SmoothZoom50 = 0;
	p->SmoothZoomAlways = 0;
#if !defined(MIPS)
	p->Zoom.Num = 1;
#endif
#endif
	p->Repeat = 1;
	p->Aspect.Den = 1;
	p->Aspect.Num = 1;
	p->Contrast = 0;
	p->Brightness = 14;
	p->Saturation = 14;
	p->PlaySpeed.Num = 1;
	p->PlaySpeed.Den = 1;
	p->FFwdSpeed.Num = 2;
	p->FFwdSpeed.Den = 1;
	p->SeekAfterSync.Num = -1;
	p->VideoAccel = 1;
	p->Stereo = 0;

	p->Rotating = 0;
	p->Clipping = 0;
	p->Wnd = 0;
	p->FullScreen = 0;
	p->Primary = 1;

	SwapXY = (QueryOrientation() & DIR_SWAPXY)!=0;

	if (SwapXY != (Desktop.Width >= Desktop.Height))
		p->FullDir = 0;
	else
		p->FullDir = DIR_SWAPXY | DIR_MIRRORUPDOWN;

	p->SkinDir = 0;

	p->PlayIndex = NULL;
	p->PlayList = NULL;
	p->PlayListCount = 0;
	p->Current = 0;
	p->CurrentChanged = 1;

	p->Format = NULL;
	p->Input = NULL;
	p->Timer = NULL;
	p->RefEnd = p->Ref;
	p->AudioStream = -1;
	p->VideoStream = -1;
	p->VideoUsed = 0;
	p->AudioUsed = 0;
	p->Position = -1;
	
	SetAOutput(p,NodeEnum(AOUT_CLASS,0));
	SetVOutput(p,NodeEnum(VOUT_CLASS,0));

	p->ColorKey = -1;
	p->ColorKeyBrush = NULL;
	p->BlackBrush = (HBRUSH)GetStockObject(BLACK_BRUSH);

	//p->ThreadId = GetCurrentThreadId();
	p->BenchTime = 0;
	p->Bench = 0;
	p->Play = 0;
	p->FFwd = 0;
	p->Sync = 0;
	p->Fill = 0;
	p->Background = 0;
	p->NoMoreInput = 0;
	p->LockCount = 0;
	p->Lock = LockCreate();
	p->LockInput = LockCreate();
	p->LockProcess = LockCreate();
	p->EventProcess = CreateEvent(NULL,FALSE,FALSE,NULL);
	p->EventRunProcess = CreateEvent(NULL,TRUE,FALSE,NULL);
	p->EventRunInput = CreateEvent(NULL,TRUE,FALSE,NULL);
	p->InputThread = ThreadCreate(InputThread,p,25);
	p->ProcessThread = ThreadCreate(ProcessThread,p,25);

	UpdateFormat(p,0);
	UpdateRunProcess(p);
	UpdateRunInput(p);

	NodeRegister((node*)&p->VMT,PRI_MAXIMUM+600);
	NodeRegister((node*)&p->EqVMT,PRI_MAXIMUM+580);
	NodeRegister((node*)&p->ColVMT,PRI_MAXIMUM+575);
}

void Player_Done()
{
	player_win32* p = &Player_Win32;
	HANDLE EventProcess = p->EventProcess;
	HANDLE EventRunInput = p->EventRunInput;
	HANDLE EventRunProcess = p->EventRunProcess;

	Lock(p); // lock before saving setup

	NodeUnRegister((node*)&p->VMT);
	NodeUnRegister((node*)&p->EqVMT);
	NodeUnRegister((node*)&p->ColVMT);

	p->Exit = 1;

	// don't allow threads to go into waiting state again
	p->EventProcess = NULL;
	p->EventRunInput = NULL;
	p->EventRunProcess = NULL;

	Unlock(p);

	// wake them up
	SetEvent(EventProcess);
	SetEvent(EventRunProcess);
	SetEvent(EventRunInput);

	if (p->InputThread && WaitForSingleObject(p->InputThread,5000) == WAIT_TIMEOUT)
		TerminateThread(p->InputThread,0);

	if (p->ProcessThread && WaitForSingleObject(p->ProcessThread,5000) == WAIT_TIMEOUT)
		TerminateThread(p->ProcessThread,0);

	Unload(p,0);

	CloseHandle(EventRunInput);
	CloseHandle(EventRunProcess);
	CloseHandle(EventProcess);

	LockDelete(p->Lock);
	LockDelete(p->LockInput);
	LockDelete(p->LockProcess);

	p->InputThread = NULL;
	p->ProcessThread = NULL;
	p->Lock = NULL;
	p->LockInput = NULL;
	p->LockProcess = NULL;


	Free(p->PlayList);
	Free(p->PlayIndex);

	if (p->ColorKeyBrush)
		DeleteObject(p->ColorKeyBrush);
}

void PlayerSaveList(player* Player,const tchar_t* Path,format* Format)
{
	int Count;
	if (Format && Format->WriteList && Player->Get(Player,PLAYER_LIST_COUNT,&Count,sizeof(Count))==ERR_NONE)
	{
		stream* Input = GetStream(Path,0);

		if (Input)
		{
			bool_t b = 1;

			if (Input->Set(Input,STREAM_CREATE,&b,sizeof(b)) == ERR_NONE &&
				Input->Set(Input,STREAM_URL,Path,sizeof(tchar_t)*(tcslen(Path)+1)) == ERR_NONE)
			{
				if (Format->Set(Format,FORMAT_INPUT,&Input,sizeof(Input)) == ERR_NONE)
				{
					tchar_t Base[MAXPATH];
					tchar_t Rel[MAXPATH];
					tchar_t URL[MAXPATH];
					tchar_t Title[256];
					tick_t Length;
					int No;
				
					SplitURL(Path,NULL,Base,NULL,NULL);
					
					for (No=0;No<Count;++No)
					{
						if (Player->Get(Player,PLAYER_LIST_URL+No,URL,sizeof(URL))==ERR_NONE)
						{
							RelPath(Rel,URL,Base);
							Title[0] = 0;
							Length = -1;

							Player->Get(Player,PLAYER_LIST_TITLE+No,Title,sizeof(Title));
							Player->Get(Player,PLAYER_LIST_LENGTH+No,&Length,sizeof(Length));

							Format->WriteList(Format,Rel,Title,Length);
						}
					}

					Format->WriteList(Format,NULL,NULL,0);
				}
			}

			Input->Delete(Input);
		}
	}
}

int PlayerAdd(player* Player,int Index, const tchar_t* Path)
{
	node* NullPtr = NULL;
	tchar_t Exts[MAXPATH];
	tchar_t Base[MAXPATH];
	tchar_t Abs[MAXPATH];
	tchar_t ItemPath[MAXPATH];
	tchar_t ItemTitle[256];
	tick_t ItemLength;
	format* Format;
	stream* Input;
	int No;
	bool_t PlayList = 0;

	// process playlist files
	for (No=0;(Format=(format*)NodeEnum(FORMAT_PLAYLIST_CLASS,No))!=NULL;++No)
		if (Format->Get(Format,FORMAT_EXTS,Exts,sizeof(Exts))==ERR_NONE && CheckExts(Path,Exts))
		{
			// try to load playlist
			Input = GetStream(Path,0);

			if (Input)
			{
				if (Input->Set(Input,STREAM_URL,Path,sizeof(tchar_t)*(tcslen(Path)+1)) == ERR_NONE)
				{
					if (Format->Set(Format,FORMAT_INPUT,&Input,sizeof(Input)) == ERR_NONE)
					{
						SplitURL(Path,NULL,Base,NULL,NULL);

						PlayList = 1;

						while (Format->ReadList(Format,ItemPath,sizeof(ItemPath)/sizeof(tchar_t),ItemTitle,sizeof(ItemTitle)/sizeof(tchar_t),&ItemLength)==ERR_NONE)
						{
							AbsPath(Abs,ItemPath,Base);
							Player->Set(Player,PLAYER_LIST_URL+Index,Abs,sizeof(Abs));
							Player->Set(Player,PLAYER_LIST_TITLE+Index,ItemTitle,sizeof(ItemTitle));
							Player->Set(Player,PLAYER_LIST_LENGTH+Index,&ItemLength,sizeof(ItemLength));
							++Index;
						}

						Format->Set(Format,FORMAT_INPUT,&NullPtr,sizeof(NullPtr));
					}
				}
				Input->Delete(Input);
			}
		}

	if (!PlayList)
	{
		Player->Set(Player,PLAYER_LIST_URL+Index,Path,sizeof(tchar_t)*(tcslen(Path)+1));
		Player->Set(Player,PLAYER_LIST_TITLE+Index,NULL,0);
		++Index;
	}

	return Index;
}

const tchar_t* PlayerComment(int Code)
{
	switch (Code)
	{
	case COMMENT_TITLE: return T("TITLE");
	case COMMENT_ARTIST: return T("ARTIST");
	case COMMENT_ALBUM: return T("ALBUM");
	case COMMENT_GENRE: return T("GENRE");
	case COMMENT_LANGUAGE: return T("LANGUAGE");
	}
	return NULL;
}
