/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: querykey.c 131 2004-12-04 20:36:04Z picard $
 *
 * BetaPlayer
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "stdafx.h"

typedef struct querykey
{
	win Win;
	int Key;
	int Key2;
	bool_t Keep;

} querykey;

static bool_t DialogProc(querykey* p,int Msg, WPARAM wParam, LPARAM lParam, int* Result)
{
	switch (Msg)
	{
	case WM_SYSKEYDOWN:
	case WM_KEYDOWN:

		if (wParam != VK_CONTROL && wParam != VK_MENU && 
			wParam != VK_LWIN && wParam != VK_RWIN && wParam != VK_SHIFT)
		{
			int Key;
			if ((wParam == VK_F1 || wParam == VK_F2) &&
				QueryPlatform(PLATFORM_TYPENO)==TYPE_SMARTPHONE)
			{
				// can't loose menu
				WinAllKeys(0);
				keybd_event((BYTE)wParam,1,0,0);
				break; 
			}

			Key = WinKeyState(wParam);
			p->Key2 = Key;
			if (!p->Key) 
				p->Key = Key;
		}
		break;

	case WM_SYSKEYUP:
	case WM_KEYUP:
/*		if ((wParam == VK_F1 || wParam == VK_F2) &&
			QueryPlatform(PLATFORM_TYPENO)==TYPE_SMARTPHONE)
		{
			// can't loose menu
			WinAllKeys(0);
			keybd_event((BYTE)wParam,1,KEYEVENTF_KEYUP,0);
			WinAllKeys(1);
			break; 
		}
*/
		if (p->Key)
			PostMessage(p->Win.Wnd,WM_CLOSE,0,0);
		break;

	case WM_KILLFOCUS:
		WinAllKeys(0);
		break;

	case WM_SETFOCUS:
		WinAllKeys(1);
		break;
	}
	return 0;
}

static bool_t Proc(querykey* p,int Msg, WPARAM wParam, LPARAM lParam, int* Result)
{
	switch (Msg)
	{
	case WM_ACTIVATE:
		if (LOWORD(wParam)==WA_INACTIVE) 
			PostMessage(p->Win.Wnd,WM_CLOSE,0,0);
		break;

	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case PLATFORM_OK:
			PostMessage(p->Win.Wnd,WM_CLOSE,0,0);
			break;
		case PLATFORM_CANCEL:
			p->Key = p->Key2 = 0;
			PostMessage(p->Win.Wnd,WM_CLOSE,0,0);
			break;
		case QUERYKEY_KEEP:
			p->Keep = !p->Keep;
			CheckMenuItem(p->Win.Menu,QUERYKEY_KEEP,p->Keep ? MF_CHECKED:MF_UNCHECKED);
			break;
		}
		break;

	case WM_CLOSE:
		if (p->Keep)
		{
			if (p->Key) p->Key |= HOTKEY_KEEP;
			if (p->Key2) p->Key2 |= HOTKEY_KEEP;
		}
		break;

	case WM_CREATE:
		p->Keep = 0;
		p->Key = 0;
		p->Key2 = 0;
		WinCreateDialog(&p->Win);
		WinLabel(&p->Win,(p->Win.DlgHeight-12) >> 1,0,p->Win.DlgWidth,String(QUERYKEY_ID,QUERYKEY_MSG),
			12,LABEL_CENTER,NULL);
		break;
	}
	return 0;
}

static int Get(querykey* p,int No,void* Data,int Size)
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case QUERYKEY_KEY: GETVALUE(p->Key,int); break;
	case QUERYKEY_KEY2: GETVALUE(p->Key2,int); break;
	}
	return Result;
}

static menudef MenuDef[] =
{
	{ 0, PLATFORM_ID, PLATFORM_CANCEL },
	{ 0, QUERYKEY_ID, QUERYKEY_OPTIONS },
	{ 1, QUERYKEY_ID, QUERYKEY_KEEP },

	MENUDEF_END
};

static querykey QueryKey;

void QueryKey_Init()
{
	memset(&QueryKey,0,sizeof(QueryKey));
	QueryKey.Win.Id = QUERYKEY_ID;
	QueryKey.Win.Width = 200;
	QueryKey.Win.Height = 200;
	QueryKey.Win.MenuDef = MenuDef;
	QueryKey.Win.Dialog = 1;
	QueryKey.Win.Title = String(QUERYKEY_ID,QUERYKEY_TITLE);
	QueryKey.Win.Proc = Proc;
	QueryKey.Win.DialogProc = DialogProc;
	QueryKey.Win.Get = Get;
	WinRegister(&QueryKey.Win,PRI_DEFAULT);
}

void QueryKey_Done()
{
	WinUnRegister(&QueryKey.Win);
}