/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: settings.c 144 2004-12-12 15:28:39Z picard $
 *
 * BetaPlayer
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "stdafx.h"

#define SETTINGS_PAGES		50000

typedef struct settings 
{
	win Win;
	node* Page;

	int Count;
	node* Node[MAXNODE];
	backup* Backup[MAXNODE];

} settings;

static void Clear(settings* p)
{
	backup *i,*j;
	int No;

	for (No=0;No<p->Count;++No)
	{
		for (i=p->Backup[No];i;i=j)
		{
			j=i->Next;
			Free(i);
		}

		p->Node[No] = NULL;
		p->Backup[No] = NULL;
	}

	p->Count = 0;
}

static void Load(settings* p)
{
	int No,PNo;
	node* Node;
	backup* Backup;
	backup** Ptr;
	backup New;

	Clear(p);

	for (No=0;(Node = NodeEnum(NODE_CLASS,No))!=NULL;++No)
	{
		Backup = NULL;
		Ptr = &Backup;
		memset(&New,0,sizeof(backup));

		for (PNo=0;Node->Enum(Node,PNo,&New.DataDef)==ERR_NONE;++PNo)
			if ((New.DataDef.Flags & DF_SETUP) && !(New.DataDef.Flags & DF_HIDDEN))
			{
				int Result = Node->Get(Node,New.DataDef.No,New.Data,New.DataDef.Size);
				if (Result == ERR_NONE || Result == ERR_NOT_SUPPORTED)
				{
					*Ptr = (backup*) Alloc(sizeof(backup));
					if (!*Ptr)
					{
						player* Player = (player*) NodeFind(PLAYER_ID);
						if (Player)
						{
							Player->Unload(Player);
							*Ptr = (backup*) Alloc(sizeof(backup));
						}
						if (!*Ptr)
							break; // out of memory
					}

					memcpy(New.Old,New.Data,New.DataDef.Size);
					New.Next = NULL;
					New.Disabled = Result == ERR_NOT_SUPPORTED;
					**Ptr = New;
					Ptr = &((*Ptr)->Next);
				}
			}

		if (Backup)
		{
			p->Node[p->Count] = Node;
			p->Backup[p->Count] = Backup;
			++p->Count;
		}
	}
}

static void Save(settings* p)
{
	bool_t SoftReset = 0;
	bool_t Restart = 0;
	bool_t ReSync = 0;
	bool_t Save;
	backup *i;
	int No;

	for (No=0;No<p->Count;++No)
	{
		node* Node = p->Node[No];
		Save = 0;

		for (i=p->Backup[No];i;i=i->Next)
			if (memcmp(i->Old,i->Data,i->DataDef.Size)!=0)
			{
				Node->Set(Node,i->DataDef.No,i->Data,i->DataDef.Size);

				if (i->DataDef.Flags & DF_SOFTRESET)
					SoftReset = 1;
				if (i->DataDef.Flags & DF_RESTART)
					Restart = 1;
				if (i->DataDef.Flags & DF_RESYNC)
					ReSync = 1;
				Save = 1;
			}

		if (Save)
			NodeSaveSetup(Node);
	}

	if (SoftReset)
		ShowMessage(String(SETTINGS_ID,0),String(SETTINGS_ID,SETTINGS_SOFTRESET));
	else
	if (Restart)
		ShowMessage(String(SETTINGS_ID,0),String(SETTINGS_ID,SETTINGS_RESTART));

	if (ReSync)
	{
		player* Player = (player*) NodeFind(PLAYER_ID);
		if (Player)
			Player->ReSync(Player);
	}

	ShowMemoryLow = 1;
}

static void UpdatePage(settings* p)
{
	int Id = 0;
	int No;
	backup* i;
	int y;
	bool_t CheckList;

	ShowWindow(p->Win.WndDialog,SW_HIDE);
	WinPropClear(&p->Win);
	
	if (!p->Page && p->Count>0)
		p->Page = p->Node[0];

	// find page
	for (No=0;No<p->Count;++No)
	{
		CheckMenuItem(p->Win.Menu,SETTINGS_PAGES+No,p->Node[No]==p->Page ? MF_CHECKED:MF_UNCHECKED);
		if (p->Node[No]==p->Page)
		{
			p->Page->Get(p->Page,NODE_ID,&Id,sizeof(Id));
			i = p->Backup[No];
		}
	}

	SetWindowText(p->Win.Wnd,String(Id,0));

	if (Id)
	{
		p->Win.LabelWidth = (Id == PLATFORM_ID) ? 60:
			(QueryPlatform(PLATFORM_TYPENO) == TYPE_SMARTPHONE)?80:
			(Id == PLAYER_ID || Id == ADVPLATFORM_ID)?120:90;

		CheckList = Id == ASSOCIATION_ID;
		y = 2;

		// if the menu is in bottom of screen. print a hint before platform settings (first page shown)
		if (Id == PLATFORM_ID && p->Win.AygShellTB && QueryPlatform(PLATFORM_TYPENO) != TYPE_SMARTPHONE)
		{
			y = WinLabel(&p->Win,y,2,p->Win.DlgWidth-4,String(SETTINGS_ID,SETTINGS_HINT),PROPSIZE,0,NULL);
			y += 10;
		}

		for (;i;i=i->Next)
		{
			if (i->DataDef.Flags & DF_GAP)
				y += 7;

			if (!(i->DataDef.Flags & DF_RDONLY))
				y = WinPropControl(&p->Win,y,i);
			else
			{
				int No;
				tchar_t s[256];

				switch (i->DataDef.Type)
				{
				case TYPE_LABEL:
					y = WinLabel(&p->Win,y,2,p->Win.DlgWidth-4,i->DataDef.Name,PROPSIZE,0,NULL);
					break;

				case TYPE_TICK:
					TickToString(s,*(tick_t*)i->Data,sizeof(s),0,1,0);
					y = WinPropLabel(&p->Win,y,i->DataDef.Name,s,i);
					break;

				case TYPE_INT:
					No = *(int*)i->Data;
					s[0] = 0;

					if ((i->DataDef.Flags & DF_ENUM) && i->DataDef.Format1)
					{
						enumdef* Enum = (enumdef*)i->DataDef.Format1;
						while (No>0 && Enum->Class!=-1)
						{
							--No;
							++Enum;
						}
						if (No==0 && Enum->Class!=-1)
							tcscpy(s,String(Enum->Class,Enum->No));
					}
					else
					if (i->DataDef.Flags & DF_ENUMSTRING)
						tcscpy(s,String(i->DataDef.Format1,No));

					if (!s[0])
						IntToString(s,No,sizeof(s),(i->DataDef.Flags & DF_HEX)!=0);
					if (i->DataDef.Flags & DF_KBYTE)
						tcscat(s,T(" KB"));
					if (i->DataDef.Flags & DF_MHZ)
						tcscat(s,T(" Mhz"));
					y = WinPropLabel(&p->Win,y,i->DataDef.Name,s,i);
					break;

				case TYPE_STRING:
					y = WinPropLabel(&p->Win,y,i->DataDef.Name,(tchar_t*)i->Data,i);
					break;
					
				case TYPE_HOTKEY:
					HotKeyToString(s,*(int*)i->Data,sizeof(s));
					y = WinPropLabel(&p->Win,y,i->DataDef.Name,s,i);
					break;
					
				case TYPE_BOOL:
					y = WinPropLabel(&p->Win,y,i->DataDef.Name,
						String(PLATFORM_ID,*(bool_t*)i->Data ? PLATFORM_YES:PLATFORM_NO),i);
					break;

				default:
					y = WinPropLabel(&p->Win,y,i->DataDef.Name,NULL,i);
					break;
				}
			}
		}
	}

	ShowWindow(p->Win.WndDialog,SW_SHOWNA);
	WinUpdateScroll(&p->Win);
}

static void UpdateMenu(settings* p)
{
	int No;
	int Id;

	for (No=0;No<p->Count;++No)
		if (p->Node[No]->Get(p->Node[No],NODE_ID,&Id,sizeof(Id))==ERR_NONE)
			InsertMenu(p->Win.Menu,SETTINGS_PAGES-1,MF_BYCOMMAND|MF_STRING,
				SETTINGS_PAGES+No,String(Id,0));

	DeleteMenu(p->Win.Menu,SETTINGS_PAGES-1,MF_BYCOMMAND);
}

static bool_t Proc(settings* p,int Msg, WPARAM wParam, LPARAM lParam, int* Result)
{
	int Id;
	switch (Msg)
	{
		case WM_CREATE:
			WinCreateDialog(&p->Win);
			Load(p);
			UpdateMenu(p);
			UpdatePage(p);
			break;

		case WM_DESTROY:
			Clear(p);
			break;

		case WM_COMMAND:
			Id = LOWORD(wParam);
			if (Id==PLATFORM_OK)
			{
				Save(p);
				PostMessage(p->Win.Wnd,WM_CLOSE,0,0);
			}
			if (Id==PLATFORM_CANCEL)
				PostMessage(p->Win.Wnd,WM_CLOSE,0,0);

			if (Id >= SETTINGS_PAGES && Id < SETTINGS_PAGES+p->Count)
			{
				p->Page = p->Node[Id-SETTINGS_PAGES];
				UpdatePage(p);
			}
			break;
	}
	return 0;
}

static menudef MenuDef[] =
{
	{ 0, PLATFORM_ID, PLATFORM_OK },
	{ 0, SETTINGS_ID, SETTINGS_SELECTPAGE },
	{ 1, SETTINGS_ID, SETTINGS_PAGES-1 }, // just place holder (will be deleted)
	{ 0, PLATFORM_ID, PLATFORM_CANCEL },

	MENUDEF_END
};

static settings Settings;

void Settings_Init()
{
	memset(&Settings,0,sizeof(Settings));
	Settings.Win.Id = SETTINGS_ID;
	Settings.Win.Width = 240;
	Settings.Win.Height = 320;
	Settings.Win.Dialog = 1;
	Settings.Win.MenuDef = MenuDef;
	Settings.Win.Proc = Proc;

	WinRegister(&Settings.Win,PRI_DEFAULT);
}

void Settings_Done()
{
	WinUnRegister(&Settings.Win);
}