/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: win32.c 145 2004-12-12 18:18:28Z picard $
 *
 * BetaPlayer
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "stdafx.h"

#define MAXFONTSIZE	32

#define TRACKMAX		30000
#define TRACKTHUMB		9

#ifndef UDS_EXPANDABLE
#define UDS_EXPANDABLE  0x0200
#endif

#if defined(_WIN32_WCE)
#define DIALOG_COLOR COLOR_STATIC
#else
#define DIALOG_COLOR COLOR_BTNFACE
#define MOD_KEYUP	 0
#endif

static WNDCLASS WinClass;
static WNDCLASS DialogClass;
static tchar_t ClassName[256];
static HFONT FontCache[MAXFONTSIZE][2];
static bool_t MenuSmall = 0;
static int LogPixelSX = -1;
static int LogPixelSY = -1;

static void* (WINAPI* FuncDeferWindowPos)( void* Handle, HWND, HWND, int x, int y, int cx, int cy, UINT Flags ) = NULL;
static void* (WINAPI* FuncBeginDeferWindowPos)( int nNumWindows ) = NULL;
static BOOL (WINAPI* FuncEndDeferWindowPos)( void* Handle ) = NULL;
static BOOL (WINAPI* FuncUnregisterFunc1)( int Flags,int HoyKey ) = NULL;
static BOOL (WINAPI* FuncAllKeys)( int State ) = NULL;

#define TOOLBAR_ID			1000

static menudef* MenuDef(win* Win,menudef* p,bool_t Popup,HMENU* Menu)
{
	int Level = p->Level;
	if (!*Menu)
		*Menu = Popup ? CreatePopupMenu() : CreateMenu();

	while (p->Level==Level && Level>=0)
	{
		if (p->Id==-1)
		{
			if (!MenuSmall)
				AppendMenu(*Menu,MF_SEPARATOR,0,NULL);
			++p;
		}
		else
		{
			tchar_t s[256];
			bool_t Skip = 0;
			int Flags = MF_STRING;
			int Id = p->Id;

			tcscpy(s,String(p->Class,p->Id));

			if (p->Flags & MENU_ADV)
				Skip = QueryAdvPlatform(p->Flags >> 16) == 0;

			if (p->Class == PLATFORM_ID)
			{
				if (p->Level==0 && p->Id == PLATFORM_CANCEL && Win->TitleClose)
				{
					++p;
					continue;
				}
				if (p->Id == PLATFORM_OK)
				{
					Win->NeedOK = 1;
					if (Win->TitleOK)
					{
						++p;
						continue;
					}
				}
			}

			++p;
			if (p->Level > Level)
			{
				if (((p[-1].Flags & MENU_SMALL) && !MenuSmall) ||
					((p[-1].Flags & MENU_NOTSMALL) && MenuSmall))
				{
					if (p[-1].Flags & MENU_GRAYED)
					{
						AppendMenu(*Menu,Flags,Id,s);
						EnableMenuItem(*Menu,Id,MF_GRAYED|MF_BYCOMMAND);
					}
			
					// skip
					p = MenuDef(Win,p,1,Menu);
					continue;
				}
				else
				if (p[-1].Flags & MENU_GRAYED)
				{ 
					int i = tcslen(s);
					if (i && s[i-1]==':')
						s[i-1] = 0;
				}

				Id = 0;
				p = MenuDef(Win,p,1,(HMENU*)&Id);
				Flags |= MF_POPUP;
			}

			if (!Skip)
			{
				AppendMenu(*Menu,Flags,Id,s);

				if (!(Flags & MF_POPUP) && (p[-1].Flags & MENU_GRAYED))
					EnableMenuItem(*Menu,Id,MF_GRAYED|MF_BYCOMMAND);
			}
			else
			{
				if (Flags & MF_POPUP)
					DestroyMenu((HMENU)Id);
			}
		}
	}
	return p;
}

static void ReleaseFonts()
{
	int No;
	for (No=0;No<MAXFONTSIZE;++No)
	{
		if (FontCache[No][0])
			DeleteObject(FontCache[No][0]);
		if (FontCache[No][1])
			DeleteObject(FontCache[No][1]);
	}
	memset(&FontCache,0,sizeof(FontCache));
}

static void UpdateDPI()
{
	int DlgHeight;
	int y,x;
	HDC DC = GetDC(NULL);
	x = GetDeviceCaps(DC, LOGPIXELSY);
	y = GetDeviceCaps(DC, LOGPIXELSX);
	if (LogPixelSY != y || LogPixelSX != x)
	{
		ReleaseFonts();
		LogPixelSY = y;
		LogPixelSX = x;
	}
	DlgHeight = Scale(GetDeviceCaps(DC,VERTRES),72,LogPixelSY);
	MenuSmall = DlgHeight < 210;
	ReleaseDC(NULL,DC);
}

int WinPixelSX() { return LogPixelSX; }
int WinPixelSY() { return LogPixelSY; }


static void VScroll(win* p,int Cmd,int Delta);
static void HScroll(win* p,int Cmd,int Delta);

static void UpdateDialogSize(win* p)
{
	RECT r;
	GetClientRect(p->WndDialog,&r);

	if (!p->ScrollV)
		r.right -= GetSystemMetrics(SM_CXVSCROLL);

	if (!p->ScrollH)
		r.bottom -= GetSystemMetrics(SM_CYHSCROLL);

	p->DlgWidth = Scale(r.right,72,LogPixelSX);
	p->DlgHeight = Scale(r.bottom,72,LogPixelSY);
}

static void WinScroll(win* p,int dx,int dy)
{
	void* Handle;
	HWND Child;
	POINT Offset;

	Offset.x = Offset.y = 0;
	ClientToScreen(p->WndDialog,&Offset);

	ScrollWindowEx(p->WndDialog, dx, dy, NULL, NULL, NULL, NULL, 0 );

	if (FuncBeginDeferWindowPos)
		Handle = FuncBeginDeferWindowPos(16);

	Child = GetWindow(p->WndDialog,GW_CHILD);
	while (Child && IsWindow(Child))
	{
		RECT r;
		GetWindowRect(Child,&r);

		r.left -= Offset.x - dx; 
		r.top -= Offset.y - dy;

		if (FuncDeferWindowPos)
			Handle = FuncDeferWindowPos(Handle,Child,NULL,r.left,r.top,0,0,SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE);
		else
			SetWindowPos(Child,NULL,r.left,r.top,0,0,SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE);

		Child = GetWindow(Child,GW_HWNDNEXT);
	}

	if (Handle && FuncEndDeferWindowPos)
		FuncEndDeferWindowPos(Handle);

	UpdateWindow(p->WndDialog);
}

void WinUpdateScroll(win* p)
{
	SCROLLINFO Scroll;

	POINT Offset;
	RECT Page;
	int MaxWidth,MinWidth;
	int MaxHeight,MinHeight;
	HWND Child;
	int XVScroll = GetSystemMetrics(SM_CXVSCROLL);
	int YHScroll = GetSystemMetrics(SM_CYHSCROLL);

	if (p->InUpdateScroll || p->OwnDialogSize)
		return;

	p->InUpdateScroll = 1;
	
	Offset.x = Offset.y = 0;
	ClientToScreen(p->WndDialog,&Offset);
	GetClientRect(p->WndDialog,&Page);

	if (p->ScrollV)
		Page.right += XVScroll;

	if (p->ScrollH)
		Page.bottom += YHScroll;

	MinWidth = 0;
	MinHeight = 0;
	MaxWidth = 0;
	MaxHeight = 0;

	Child = GetWindow(p->WndDialog,GW_CHILD);
	while (Child && IsWindow(Child))
	{
		RECT r;
		GetWindowRect(Child,&r);

		r.left -= Offset.x;
		r.right -= Offset.x; 
		r.top -= Offset.y;
		r.bottom -= Offset.y; 
		r.bottom += YHScroll;

		if (r.left < MinWidth) MinWidth = r.left;
		if (r.right > MaxWidth) MaxWidth = r.right;
		if (r.top < MinHeight) MinHeight = r.top;
		if (r.bottom > MaxHeight) MaxHeight = r.bottom;

		Child = GetWindow(Child,GW_HWNDNEXT);
	}

	Scroll.cbSize = sizeof(Scroll);
	Scroll.fMask =  SIF_PAGE | SIF_RANGE | SIF_POS;

	if (MaxHeight < Page.bottom)
		MaxHeight = Page.bottom-1;

	p->ScrollV = MaxHeight-MinHeight > Page.bottom;
	Scroll.nMin = MinHeight;
	Scroll.nMax = MaxHeight;
	Scroll.nPos = 0;
	Scroll.nPage = Page.bottom;
	SetScrollInfo(p->WndDialog,SB_VERT,&Scroll,TRUE);

	if (p->ScrollV)
		Page.right -= XVScroll;

	if (MaxWidth < Page.right)
		MaxWidth = Page.right-1;

	p->ScrollH = MaxWidth-MinWidth > Page.right;
	Scroll.nMin = MinWidth;
	Scroll.nMax = MaxWidth;
	Scroll.nPos = 0;
	Scroll.nPage = Page.right;
	SetScrollInfo(p->WndDialog,SB_HORZ,&Scroll,TRUE);

	p->InUpdateScroll = 0;
}

#if !defined(_WIN32_WCE)
#define TBSTYLE_EX_MIXEDBUTTONS         0x00000008
#define BTNS_SHOWTEXT					0x40
#else
#define BTNS_SHOWTEXT					0
#endif

static void AddButton(win* p,int Pos,int Id,int Bitmap,tchar_t* Name,bool_t Check)
{
	TBBUTTON Button;
	memset(&Button,0,sizeof(Button));
	Button.iBitmap = p->BitmapNo + Bitmap;
	Button.idCommand = Id;
	Button.fsState = TBSTATE_ENABLED;
	Button.fsStyle = (BYTE)((Check ? TBSTYLE_CHECK : TBSTYLE_BUTTON) | TBSTYLE_AUTOSIZE | BTNS_SHOWTEXT);
	Button.dwData = 0;
	Button.iString = 0;

	if (Id<0)
	{
		Button.idCommand = 0;
		Button.fsStyle = TBSTYLE_SEP;

		if (!p->AygShellTB && !p->Dialog && GetSystemMetrics(SM_CXSCREEN)<=240)
			return;
	}

	SendMessage(p->WndTB,TB_INSERTBUTTON,Pos,(LPARAM)&Button);

	if (Name && Id>0)
	{
		TBBUTTONINFO Button;
		Button.cbSize = sizeof(Button);
		Button.dwMask = TBIF_TEXT;
		Button.pszText = Name;
		SendMessage(p->WndTB,TB_SETBUTTONINFO,Id,(LPARAM)&Button);
	}
}

#if defined(_WIN32_WCE)

#define SIPF_OFF	0x00000000
#define SIPF_ON 	0x00000001

typedef struct SHACTIVATEINFO
{
	DWORD cbSize;
	HWND hwndLastFocus;
	UINT fSipUp:1;
	UINT fSipOnDeactivation:1;
	UINT fActive:1;
	UINT fReserved:29;
} SHACTIVATEINFO;

#define SHMBOF_NODEFAULT    0x00000001
#define SHMBOF_NOTIFY       0x00000002

#define SHCMBM_GETMENU      (WM_USER + 402)
#define SHCMBM_OVERRIDEKEY  (WM_USER + 403)

#define SHCMBF_EMPTYBAR      0x0001
#define SHCMBF_HIDDEN        0x0002
#define SHCMBF_HIDESIPBUTTON 0x0004
#define SHCMBF_COLORBK       0x0008
#define SHCMBF_HMENU         0x0010

typedef struct SHMENUBARINFO
{
    DWORD cbSize;
	HWND hwndParent;
    DWORD dwFlags;
    UINT nToolBarId;
    HINSTANCE hInstRes;
    int nBmpId;
    int cBmpImages;
	HWND hwndMB;
    // COLORREF clrBk; //be as compatible as possible
} SHMENUBARINFO;

#define SHIDIM_FLAGS 0x0001 

#define SHIDIF_DONEBUTTON           0x0001
#define SHIDIF_SIZEDLG              0x0002
#define SHIDIF_SIZEDLGFULLSCREEN    0x0004
#define SHIDIF_SIPDOWN              0x0008
#define SHIDIF_FULLSCREENNOMENUBAR  0x0010
#define SHIDIF_EMPTYMENU            0x0020

typedef struct SHINITDLGINFO
{
    DWORD dwMask;
    HWND  hDlg;
    DWORD dwFlags;
} SHINITDLGINFO;

#define SPI_GETSIPINFO          225
#define SPI_APPBUTTONCHANGE     228

typedef struct SIPINFO
{
    DWORD   cbSize;
    DWORD   fdwFlags;
    RECT    rcVisibleDesktop;
    RECT    rcSipRect;
    DWORD   dwImDataSize;
    void   *pvImData;
} SIPINFO;

#define SHFS_SHOWTASKBAR            0x0001
#define SHFS_HIDETASKBAR            0x0002
#define SHFS_SHOWSIPBUTTON          0x0004
#define SHFS_HIDESIPBUTTON          0x0008
#define SHFS_SHOWSTARTICON          0x0010
#define SHFS_HIDESTARTICON          0x0020

extern int LinkVer;
static HMODULE AygShell = NULL;
static HMODULE CoreDLL = NULL;
static BOOL (WINAPI* FuncSipShowIM)(DWORD) = NULL;
static BOOL (WINAPI* FuncSHCreateMenuBar)( SHMENUBARINFO *pmb ) = NULL;
static BOOL (WINAPI* FuncSHInitDialog)( SHINITDLGINFO *shdi ) = NULL;
static BOOL (WINAPI* FuncSHFullScreen)( HWND, DWORD ) = NULL;
static BOOL (WINAPI* FuncSHHandleWMActivate)( HWND, WPARAM, LPARAM, void*, DWORD ) = NULL;
static BOOL (WINAPI* FuncSHHandleWMSettingChange)(HWND, WPARAM, LPARAM, void*) = NULL;
static BOOL (WINAPI* FuncSHSendBackToFocusWindow)(UINT, WPARAM, LPARAM ) = NULL;

void WinSetFullScreen(win* p,bool_t State)
{
	if (State != p->FullScreen)
	{
		static bool_t HHTaskbarTop = 0;
		// no need to hide HHTaskbar because we use TOPMOST and with aygshell we use SHFS_HIDETASKBAR
		HWND WndTaskbar = FindWindow(T("HHTaskbar"),NULL);
		HWND WndTray = FindWindow(T("Tray"),NULL); //for smartphone

		DEBUG_MSG1(DEBUG_WIN,T("FULLSCREEN %d"),State);

		p->FullScreen = State; // set flag before SetWindowPos so WM_SIZE,WM_MOVE has valid info

		if (State)
		{
			HWND Mode;
			if (FuncSHFullScreen)
			{
				Mode = HWND_TOP; //incoming calls on smartphones doesn't like topmost
				FuncSHFullScreen(p->Wnd,SHFS_HIDETASKBAR);
				FuncSHFullScreen(p->Wnd,SHFS_HIDESIPBUTTON);
			}
			else
			{
				Mode = HWND_TOPMOST;
				if (WndTaskbar)
				{
					RECT r;
					GetWindowRect(WndTaskbar,&r);
					HHTaskbarTop = r.top == 0 && r.left == 0;
					if (HHTaskbarTop)
						ShowWindow(WndTaskbar,SW_HIDE); // old-style with pocketpc
				}		
			}

			if (p->WndTB)
				ShowWindow(p->WndTB,SW_HIDE);
			if (WndTray)
				ShowWindow(WndTray,SW_HIDE);

			p->SaveScreen.x = GetSystemMetrics(SM_CXSCREEN);
			p->SaveScreen.y = GetSystemMetrics(SM_CYSCREEN);
			GetWindowRect(p->Wnd,&p->SaveRect);
			SetWindowPos(p->Wnd,Mode, 0, 0, p->SaveScreen.x,p->SaveScreen.y,0);
			
			ShowCursor(FALSE);
		}
		else
		{
			ShowCursor(TRUE);
			
			p->SaveRect.right += GetSystemMetrics(SM_CXSCREEN) - p->SaveScreen.x;
			p->SaveRect.bottom += GetSystemMetrics(SM_CYSCREEN) - p->SaveScreen.y;
			SetWindowPos(p->Wnd,HWND_NOTOPMOST,p->SaveRect.left,p->SaveRect.top,
				p->SaveRect.right-p->SaveRect.left,p->SaveRect.bottom-p->SaveRect.top,0);

			if (FuncSHFullScreen)
				FuncSHFullScreen(p->Wnd,SHFS_SHOWTASKBAR);
			else
			if (WndTaskbar && HHTaskbarTop)
				ShowWindow(WndTaskbar,SW_SHOWNA); // old-style with pocketpc

			if (p->WndTB)
				ShowWindow(p->WndTB,SW_SHOWNA);
			if (WndTray)
				ShowWindow(WndTray,SW_SHOWNA);
		}
	}
}

static void DestroyToolBar(win* p)
{
	if (p->WndTB)
		CommandBar_Destroy(p->WndTB);
}

void WinBitmap(win* p,int BitmapId16, int BitmapId32, int BitmapNum)
{
	if (p->WndTB)
	{
		if (LogPixelSX >= 96*2)
			p->BitmapNo = CommandBar_AddBitmap(p->WndTB,p->Module,BitmapId32,BitmapNum,32,32);
		else
			p->BitmapNo = CommandBar_AddBitmap(p->WndTB,p->Module,BitmapId16,BitmapNum,16,16);
	}
}

static menudef* SkipButtons(win* p,menudef* i)
{
	while (p->TitleOK && i->Level==0 && i->Class == PLATFORM_ID)
	{
		if (i->Id == PLATFORM_OK)
		{
			p->NeedOK = 1;
			++i;
		}
		else
			break;
	}
	return i;
}

static void CreateToolBar(win* p)
{
	RECT Rect,RectTB;
	bool_t Smart = QueryPlatform(PLATFORM_TYPENO) == TYPE_SMARTPHONE;

	p->WndTB = NULL;
	p->Menu = NULL;
	p->AygShellTB = 0;
	p->TitleClose = Smart && p->Dialog;
	p->TitleOK = !Smart;
	p->NeedOK = 0;

	if (FuncSHCreateMenuBar)
	{
		SHMENUBARINFO Info;
		memset(&Info,0,sizeof(Info));
		Info.cbSize = sizeof(Info);
		if (Smart)
			Info.cbSize += sizeof(COLORREF); //clrBk
		Info.hwndParent = p->Wnd;
		Info.nToolBarId = 0;
		Info.hInstRes = p->Module;
		Info.dwFlags = 0;
		if (!p->MenuDef)
			Info.dwFlags |= SHCMBF_EMPTYBAR;
		else
		{
			menudef* i = p->MenuDef;
			int Mode = 0;

			i = SkipButtons(p,i);
			if (i->Level==0)
			{
				++i;
				for (;i->Level>0;++i)
					Mode |= 1;  // submenu under first menu
			}
			i = SkipButtons(p,i);
			if (i->Level==0)
			{
				++i;
				for (;i->Level>0;++i)
					Mode |= 2;  // submenu under second menu
			}

			switch (Mode)
			{
			case 0: Info.nToolBarId = IDR_MENU00; break;
			case 1: Info.nToolBarId = IDR_MENU10; break;
			case 2: Info.nToolBarId = IDR_MENU01; break;
			case 3: Info.nToolBarId = IDR_MENU11; break;
			}
		}
					
		if (!p->Dialog)
			Info.dwFlags |= SHCMBF_HIDESIPBUTTON;

		FuncSHCreateMenuBar(&Info);

		p->WndTB = Info.hwndMB;
		p->ToolBarHeight = 0;

		GetWindowRect(p->Wnd, &Rect);
		if (p->WndTB)
		{
			p->Menu = (HMENU)SendMessage(p->WndTB,SHCMBM_GETMENU,0,0);

			if (p->MenuDef && !p->Menu)
			{
				CommandBar_Destroy(p->WndTB);
				p->WndTB = NULL;
			}
			else
			{
				if (p->MenuDef)
				{
					TBBUTTONINFO Button;
					menudef* i = p->MenuDef;
					int No = 0;

					while (i->Level>=0)
					{
						i = SkipButtons(p,i);
						if (i->Level<0)
							break;
						if (No<2)
						{
							int Id = No ? WIN_B : WIN_A;
							p->WinCmd[No] = i->Id;
							Button.cbSize = sizeof(Button);
							Button.dwMask = TBIF_TEXT;
							Button.pszText = (tchar_t*)String(i->Class,i->Id);
							SendMessage(p->WndTB,TB_SETBUTTONINFO,Id,(LPARAM)&Button);
							++i;

							if (i->Level>0)
							{
								HMENU SubMenu = GetSubMenu(p->Menu,No);
								DeleteMenu(SubMenu,No ? WIN_B_B:WIN_A_A,MF_BYCOMMAND);
								i = MenuDef(p,i,0,&SubMenu);
							}

							if (No==1 && !Smart)
							{
								TBBUTTON Sep;
								memset(&Sep,0,sizeof(Sep));
								Sep.fsStyle = TBSTYLE_SEP;
								SendMessage(p->WndTB,TB_INSERTBUTTON,No,(LPARAM)&Sep);
								++No;
							}
						}
						else
						{
							if (Smart) 
								break;
							AddButton(p,No,i->Id,-2,(tchar_t*)String(i->Class,i->Id),0);
							++i;
							assert(i->Level<=0); // no submenu allowed (only for WIN_A,WIN_B)
						}
						++No;
					}

					while (!Smart && SendMessage(p->WndTB,TB_BUTTONCOUNT,0,0)>No)
						SendMessage(p->WndTB,TB_DELETEBUTTON,No,0);
				}

				GetWindowRect(p->WndTB, &RectTB);
				if (Rect.bottom > RectTB.top)
					Rect.bottom = RectTB.top;
				//Rect.bottom -= (RectTB.bottom - RectTB.top);
				MoveWindow(p->Wnd,Rect.left,Rect.top,Rect.right - Rect.left,Rect.bottom - Rect.top,0);

				p->AygShellTB = 1;

				if (p->Dialog)
					SendMessage(p->WndTB,SHCMBM_OVERRIDEKEY,VK_ESCAPE,
						MAKELONG(SHMBOF_NODEFAULT|SHMBOF_NOTIFY,SHMBOF_NODEFAULT|SHMBOF_NOTIFY));
			}
		}
	}

	if (!p->WndTB)
	{
		p->TitleClose = 1;
		p->WndTB = CommandBar_Create(p->Module,p->Wnd,TOOLBAR_ID);
		p->ToolBarHeight = CommandBar_Height(p->WndTB);
		if (p->MenuDef)
		{
			MenuDef(p,p->MenuDef,0,&p->Menu);
			CommandBar_InsertMenubarEx(p->WndTB,NULL,(tchar_t*)p->Menu,0);
		}
	}

	if (p->WndTB)
		CommandBar_Show(p->WndTB, TRUE);
}

#else

WNDPROC OldToolBarProc = NULL;

LRESULT CALLBACK ToolBarProc( HWND Wnd, UINT Msg, WPARAM wParam, LPARAM lParam )
{
	if (Msg == WM_MOUSEACTIVATE)
		return MA_NOACTIVATE;	

	if (Msg == WM_LBUTTONUP)
	{
		int i;
		int n=SendMessage(Wnd,TB_BUTTONCOUNT,0,0);
		for (i=0;i<n;++i)
		{
			TBBUTTON Button;
			RECT r;
			POINT p;

			SendMessage(Wnd,TB_GETBUTTON,i,(LPARAM)&Button);
			if ((Button.fsState & TBSTATE_PRESSED))
			{
				if (IsMenu((HMENU)Button.dwData))
				{
					SendMessage(Wnd,TB_GETRECT,Button.idCommand,(LPARAM)&r);
					p.x = LOWORD(lParam);
					p.y = HIWORD(lParam);
					if (PtInRect(&r,p))
					{
						p.x = r.left;
						p.y = r.bottom;
						ClientToScreen(Wnd,&p);
						TrackPopupMenu((HMENU)Button.dwData,TPM_LEFTALIGN|TPM_TOPALIGN,
							p.x,p.y,0,GetParent(Wnd),NULL);
					}
				}
			}
		}
	}
	return CallWindowProc(OldToolBarProc,Wnd,Msg,wParam,lParam);
}

static void DestroyToolBar(win* p)
{
	DestroyWindow(p->WndTB);
}

static void AddMenu(win* p)
{
	TBBUTTONINFO Info;
	TBBUTTON Button;
	tchar_t Name[256];
	int Pos;

	p->TitleClose = 1;
	p->TitleOK = 0;
	p->NeedOK = 0;

	if (p->MenuDef)
	{
		MenuDef(p,p->MenuDef,0,&p->Menu);

		for (Pos=0;Pos<GetMenuItemCount(p->Menu);++Pos)
		{
			GetMenuString(p->Menu,Pos,Name,255,MF_BYPOSITION);

			memset(&Button,0,sizeof(Button));
			Button.iBitmap = -2;
			Button.idCommand = GetMenuItemID(p->Menu,Pos);
			if (Button.idCommand < 0)
				Button.idCommand = TOOLBAR_ID+Pos+1;
			Button.fsState = TBSTATE_ENABLED;
			Button.fsStyle = (BYTE)(TBSTYLE_BUTTON | TBSTYLE_AUTOSIZE | BTNS_SHOWTEXT);
			if (!Button.idCommand)
				Button.fsStyle = TBSTYLE_SEP;

			Button.dwData = (DWORD) GetSubMenu(p->Menu,Pos);
			Button.iString = 0;

			SendMessage(p->WndTB,TB_INSERTBUTTON,Pos,(LPARAM)&Button);

			Info.cbSize = sizeof(TBBUTTONINFO);
			Info.dwMask = TBIF_TEXT;
			Info.pszText = Name;

			SendMessage(p->WndTB,TB_SETBUTTONINFO,Button.idCommand,(LPARAM)&Info);
		}
	}

	SendMessage(p->WndTB,TB_AUTOSIZE,0,0);
}

void WinBitmap(win* p,int BitmapId16, int BitmapId32, int BitmapNum)
{
	TBADDBITMAP Add;
	Add.hInst = p->Module;
	Add.nID = BitmapId16;
	p->BitmapNo = SendMessage(p->WndTB,TB_ADDBITMAP,BitmapNum,(LPARAM)&Add);
}

static void CreateToolBar(win* p)
{
	RECT r;

    p->WndTB = CreateWindowEx(0, TOOLBARCLASSNAME, (LPSTR) NULL, 
       WS_CHILD|TBSTYLE_FLAT|TBSTYLE_LIST|WS_VISIBLE, 
       0, 0, 0, 0, p->Wnd, (HMENU)TOOLBAR_ID, p->Module, NULL); 

    SendMessage(p->WndTB, TB_BUTTONSTRUCTSIZE, (WPARAM) sizeof(TBBUTTON), 0); 

	SendMessage(p->WndTB,TB_SETEXTENDEDSTYLE,0,TBSTYLE_EX_MIXEDBUTTONS);

	OldToolBarProc = (WNDPROC) GetWindowLong(p->WndTB,GWL_WNDPROC);
	SetWindowLong(p->WndTB,GWL_WNDPROC,(LONG)ToolBarProc);

	GetWindowRect(p->WndTB,&r);
	p->ToolBarHeight = r.bottom - r.top;

	AddMenu(p);
}

void WinSetFullScreen(win* p,bool_t State)
{
	if (State != p->FullScreen)
	{
		p->FullScreen = State;
		if (State)
		{
			p->Style = GetWindowLong(p->Wnd,GWL_STYLE);
			GetWindowRect(p->Wnd,&p->SaveRect);

			SetWindowLong(p->Wnd,GWL_STYLE,WS_TABSTOP|WS_VISIBLE|WS_POPUP|WS_GROUP);
#ifdef NDEBUG
			SetWindowPos(p->Wnd,HWND_TOPMOST,0,0,
#else
			SetWindowPos(p->Wnd,HWND_NOTOPMOST,0,0,
#endif
				GetSystemMetrics(SM_CXSCREEN),GetSystemMetrics(SM_CYSCREEN),SWP_FRAMECHANGED);
		}
		else
		{
			SetWindowLong(p->Wnd,GWL_STYLE,p->Style);
			SetWindowPos(p->Wnd,HWND_NOTOPMOST,p->SaveRect.left,p->SaveRect.top,
				p->SaveRect.right-p->SaveRect.left,p->SaveRect.bottom-p->SaveRect.top,SWP_FRAMECHANGED);
		}

		ShowWindow(p->WndTB,State ? SW_HIDE:SW_SHOWNA);
		InvalidateRect(p->Wnd,NULL,0);
	}
}

#endif

const datadef WinParams[] = 
{
	{ WIN_CLASS,	TYPE_CLASS },

	DATADEF_END
};

static int WinEnum( win* p, int No, datadef* Param )
{
	if (NodeEnumType(&No,Param,NodeParams,WinParams,NULL)==ERR_NONE)
		return ERR_NONE;

	if (p->Enum)
		return p->Enum(&p->VMT,No,Param);

	return ERR_INVALID_PARAM;
}

static int WinGet( win* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(p->Id,int); break;
	default:
		if (p->Get)
			Result = p->Get(p,No,Data,Size);	
		break;
	}
	return Result;
}

static int WinSet( win* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	default:
		if (p->Set)
			Result = p->Set(p,No,Data,Size);	
		break;
	}
	return Result;
}

void WinButtonEnable(win* p,int Id,bool_t State)
{
	SendMessage(p->WndTB,TB_ENABLEBUTTON,Id,MAKELONG(State,0));
}

void WinButtonBitmap(win* p,int Id,int Bitmap)
{
	TBBUTTONINFO Info;
	Info.cbSize = sizeof(TBBUTTONINFO);
	Info.dwMask = TBIF_IMAGE;
	Info.iImage = p->BitmapNo + Bitmap;

	SendMessage(p->WndTB,TB_SETBUTTONINFO,Id,(LPARAM)&Info);
}

int FindButton(win* p,int Id)
{
	int i;
	int n=SendMessage(p->WndTB,TB_BUTTONCOUNT,0,0);
	for (i=0;i<n;++i)
	{
		TBBUTTON Button;
		SendMessage(p->WndTB,TB_GETBUTTON,i,(LPARAM)&Button);
		if (Button.idCommand == Id)
			return i;
	}
	return -1;
}

void WinDeleteButton(win* p,int Id)
{
	int No = FindButton(p,Id);
	if (No>=0)
		SendMessage(p->WndTB,TB_DELETEBUTTON,No,0);
}

void WinAddButton(win* p,int Id,int Bitmap,tchar_t* Name,bool_t Check)
{
	AddButton(p,SendMessage(p->WndTB,TB_BUTTONCOUNT,0,0),Id,Bitmap,Name,Check);
}

void WinCreateDialog(win* p)
{
	RECT r;
	GetClientRect(p->Wnd,&r);
	CreateWindow(DialogClass.lpszClassName,NULL,WS_CLIPCHILDREN|WS_CHILD|WS_VISIBLE,
		0,p->ToolBarHeight,r.right,r.bottom-p->ToolBarHeight,
		p->Wnd,NULL,DialogClass.hInstance,p);
}

void WinPropClear( win* p)
{
	// remove old controls
	HWND Wnd = GetWindow(p->WndDialog,GW_CHILD);
	while (Wnd)
	{
		DestroyWindow(Wnd);
		Wnd = GetWindow(p->WndDialog,GW_CHILD);
	}

	p->LabelCount = 0;
	p->Chain = NULL;
	p->Focus = NULL;
}

void WinPropFocus(win* p,backup* New,bool_t Set)
{
	int i;

	if (p->Focus == New)
		return;

	if (New && New->Disabled)
		return;

	if (p->Focus)
		for (i=0;i<p->LabelCount;++i)
			if (p->Focus == (backup*)GetWindowLong(p->Label[i],GWL_USERDATA))
				InvalidateRect(p->Label[i],NULL,TRUE);

	p->Focus = New;
	if (p->Focus)
	{
		POINT o;
		RECT r;
		int Min=MAX_INT;
		int Max=0;

		if (p->Focus->Control && GetParent(p->Focus->Control)==p->WndDialog)
		{
			GetWindowRect(p->Focus->Control,&r);
			if (r.bottom > Max)
				Max = r.bottom;
			if (r.top < Min)
				Min = r.top;
		}

		for (i=0;i<p->LabelCount;++i)
			if (p->Focus == (backup*)GetWindowLong(p->Label[i],GWL_USERDATA))
			{
				GetWindowRect(p->Label[i],&r);
				if (r.bottom > Max)
					Max = r.bottom;
				if (r.top < Min)
					Min = r.top;
				InvalidateRect(p->Label[i],NULL,TRUE);
			}

		if (Max > Min)
		{
			o.x = o.y = 0;
			ClientToScreen(p->WndDialog,&o);
			GetClientRect(p->WndDialog,&r);

			r.top += o.y;
			r.bottom += o.y;

			if (r.top > Min)
				VScroll(p,0,Min-r.top);
			else
			if (r.bottom < Max)
				VScroll(p,0,Max-r.bottom);
		}

		if (Set)
			if (p->Focus->Control)
				SetFocus(p->Focus->Control);
			else
				SetFocus(p->WndDialog);
	}
	else
	if (Set)
		SetFocus(p->WndDialog);
}

void WinNext(win* p,bool_t b)
{
	if (p->Focus)
	{
		backup* Last = p->Focus;
		backup* i;
		do
		{
			if (b)
			{
				for (i=p->Chain;i->Chain && i->Chain!=Last;i=i->Chain);
			}
			else
			{
				i = Last->Chain;
				if (!i) 
					i=p->Chain;
			}
			Last = i;
		}
		while (i && i!=p->Focus && i->Disabled);

		WinPropFocus(p,i,1);
	}
}

bool_t WinOpen(win* p,win* Parent,bool_t Modal )
{
	HWND Wnd;
	MSG Msg;
	int Result;
	int Style = WS_VISIBLE;
	int ExStyle = 0;
	int x,y;
	int Width,Height;
	
	if (!p) 
		return 0;

	p->Result = 0;
	p->Modal = Modal;
	p->Parent = Parent;
	p->BitmapNo = 0;

#if defined(_WIN32_WCE)
	if (AygShell && Modal)
		ExStyle |= WS_EX_CAPTIONOKBTN;

	if (Parent)
		Style |= WS_POPUP;

	{
		RECT r;
		SystemParametersInfo(SPI_GETWORKAREA,0,&r,0);
		x = r.left;
		y = r.top;
		Width = r.right - r.left;
		Height = r.bottom - r.top;
	}
#else
	Style |= WS_OVERLAPPEDWINDOW;
	y = x = CW_USEDEFAULT;
	Width = p->Width;
	Height = p->Height;
#endif

	Wnd = CreateWindowEx(ExStyle,WinClass.lpszClassName,p->Title,Style,x,y,Width,Height,
		Parent ? Parent->Wnd:NULL,NULL,WinClass.hInstance,p);

	if (Wnd)
	{
		if (p->Parent)
		{
			p->Parent->Child = p;
			EnableWindow(p->Parent->Wnd,0);
		}

		while (p->Wnd && GetMessage(&Msg, NULL, 0, 0)) 
		{
#ifdef MSGLOG
			DebugMessage(T("WND:%08x Cmd:%08x wP:%08x wL:%08x"),Msg.hwnd,Msg.message,Msg.wParam,Msg.lParam);
#endif

			if (Msg.message == MSG_NOTIFY)
			{
				// send directly to window
				win* Win=(win*)Msg.lParam;
				if (Win && Win->Proc(Win,Msg.message,Msg.wParam,Msg.lParam,&Result))
					continue;
			}

			if (Msg.message >= WM_APP + 0x200 && 
				Msg.message <= WM_APP + 0x220)
			{
				// send to application window
				win* Appl=p;
				while (Appl->Parent)
					Appl=Appl->Parent;

				if (Appl->Proc(Appl,Msg.message,Msg.wParam,Msg.lParam,&Result))
					continue;
			}

			if (Msg.message == WM_KEYDOWN)
			{
				if (p->CaptureKeys && p->Proc && p->Proc(p,Msg.message,Msg.wParam,Msg.lParam,&Result))
					continue;

				if (p->Focus)
				{
					if (p->Focus->DataDef.Type == TYPE_BOOL || p->Focus->DataDef.Type == TYPE_FUNC || p->Focus->DataDef.Type == TYPE_HOTKEY)
					{
						if (Msg.wParam == VK_RETURN)
						{
							SendMessage(p->Focus->Control,BM_CLICK,0,0);
							continue;
						}
						if (Msg.wParam == VK_LEFT || Msg.wParam == VK_RIGHT)
						{
							SendMessage(p->WndDialog,Msg.message,Msg.wParam,Msg.lParam);
							continue;
						}
					}
					if (Msg.wParam == VK_RETURN && p->Focus->ComboBox && !p->ComboOpen)
					{
						SendMessage(p->Focus->Control,CB_SHOWDROPDOWN,1,0);
						continue;
					}
					if (Msg.wParam == VK_DOWN && !p->ComboOpen && !p->Focus->ListView)
					{
						WinNext(p,FALSE);
						continue;
					}
					if (Msg.wParam == VK_UP && !p->ComboOpen && !p->Focus->ListView)
					{
						WinNext(p,TRUE);
						continue;
					}
					if (Msg.wParam == VK_TAB)
					{
						WinNext(p,GetKeyState(VK_SHIFT)<0);
						continue;
					}
				}
			}
			TranslateMessage(&Msg);
			DispatchMessage(&Msg);
		}

		DEBUG_MSG(DEBUG_WIN,T("ENDMESSAGELOOP"));
	}
	
	return p->Result;
}

static void VScroll(win* p,int Cmd,int Delta)
{
	int Pos;
	SCROLLINFO Scroll;

	Scroll.cbSize = sizeof(Scroll);
	Scroll.fMask  = SIF_ALL;
	GetScrollInfo(p->WndDialog, SB_VERT, &Scroll);

	Pos = Scroll.nPos;

	if (Delta)
		Scroll.nPos += Delta;
	else
	switch (Cmd)
	{
	case SB_TOP:
		Scroll.nPos = Scroll.nMin;
		break;
	case SB_BOTTOM:
		Scroll.nPos = Scroll.nMax;
		break;
	case SB_LINEUP:
		Scroll.nPos -= Scale(16,LogPixelSY,72);
		break;
	case SB_LINEDOWN:
		Scroll.nPos += Scale(16,LogPixelSY,72);
		break;
	case SB_PAGEUP:
		Scroll.nPos -= Scroll.nPage;
		break;
	case SB_PAGEDOWN:
		Scroll.nPos += Scroll.nPage;
		break;
	case SB_THUMBTRACK:
		Scroll.nPos = Scroll.nTrackPos;
		break;
	default:
		break;
	}

	Scroll.fMask = SIF_POS;
	SetScrollInfo(p->WndDialog, SB_VERT, &Scroll, TRUE);
	GetScrollInfo(p->WndDialog, SB_VERT, &Scroll);

	if (Scroll.nPos != Pos)
		WinScroll(p,0,Pos - Scroll.nPos);
}

static void HScroll(win* p,int Cmd,int Delta)
{
	int Pos;
	SCROLLINFO Scroll;

	Scroll.cbSize = sizeof(Scroll);
	Scroll.fMask  = SIF_ALL;
	GetScrollInfo(p->WndDialog, SB_HORZ, &Scroll);

	Pos = Scroll.nPos;

	if (Delta)
		Scroll.nPos += Delta;
	else
	switch (Cmd)
	{
	case SB_TOP:
		Scroll.nPos = Scroll.nMin;
		break;
	case SB_BOTTOM:
		Scroll.nPos = Scroll.nMax;
		break;
	case SB_LINEUP:
		Scroll.nPos -= Scale(16,LogPixelSX,72);
		break;
	case SB_LINEDOWN:
		Scroll.nPos += Scale(16,LogPixelSX,72);
		break;
	case SB_PAGEUP:
		Scroll.nPos -= Scroll.nPage;
		break;
	case SB_PAGEDOWN:
		Scroll.nPos += Scroll.nPage;
		break;
	case SB_THUMBTRACK:
		Scroll.nPos = Scroll.nTrackPos;
		break;
	default:
		break;
	}

	Scroll.fMask = SIF_POS;
	SetScrollInfo(p->WndDialog, SB_HORZ, &Scroll, TRUE);
	GetScrollInfo(p->WndDialog, SB_HORZ, &Scroll);

	if (Scroll.nPos != Pos)
		WinScroll(p,Pos - Scroll.nPos,0);
}

static void SetTrack(win* p,backup* i,int Value)
{
	if (i->DataDef.Format2 - i->DataDef.Format1 <= TRACKMAX)
		Value += i->DataDef.Format1;
	else
		Value = i->DataDef.Format1 + Scale(i->DataDef.Format2 - i->DataDef.Format1,Value,TRACKMAX);

	switch (i->DataDef.Type)
	{
	case TYPE_INT:
		*(int*)i->Data = Value;
		break;
	case TYPE_TICK:
		*(tick_t*)i->Data = Value;
		break;
	}
}

#define VALIDBACKUP(i) ((int)(i) & ~0xFFFF)

static LRESULT DialogProc(HWND Wnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	HWND Control;
	backup* i;
	CREATESTRUCT* Param;
	int Result;
	tchar_t s[32];
	win* p = (win*)GetWindowLong(Wnd,GWL_USERDATA);

#ifdef MSGLOG
	if (p)
		DebugMessage(T("DLG:%08x Cmd:%08x wP:%08x wL:%08x"),p->Id,Msg,wParam,lParam);
#endif

	if (p && p->DialogProc)
	{
		if (p->DialogProc(p,Msg,wParam,lParam,&Result))
			return Result;
	}

	switch (Msg)
	{
	case WM_CREATE:
		Param = (CREATESTRUCT*)lParam;
		p = (win*)Param->lpCreateParams;
		p->WndDialog = Wnd;
		p->LabelCount = 0;
		p->Chain = NULL;
		p->Focus = NULL;

		SetWindowLong(Wnd,GWL_USERDATA,(LONG)p);
		break;

	case WM_COMMAND:
		Control = (HWND)lParam;
		i = (backup*) GetWindowLong(Control,GWL_USERDATA);
		if (VALIDBACKUP(i))
			switch (HIWORD(wParam))
			{
			case CBN_CLOSEUP:
				if (i->ComboBox)
					p->ComboOpen = 0;
				break;

			case CBN_DROPDOWN:
				if (i->ComboBox)
					p->ComboOpen = 1;
				break;

			case LBN_SETFOCUS: //CBN_KILLFOCUS
				if (QueryPlatform(PLATFORM_TYPENO) == TYPE_SMARTPHONE)
					WinPropFocus(p,i,0);
				p->ComboOpen = 0;
				break;

			case EN_SETFOCUS:
				if (Control != i->Edit)
					WinShowSip(1);

				//no break
			case CBN_SETFOCUS:
			case BN_SETFOCUS:
				WinPropFocus(p,i,0);
				break;

			case EN_KILLFOCUS:
				if (Control != i->Edit)
					WinShowSip(0);
				break;

			case EN_CHANGE:
				if (i->DataDef.Type == TYPE_STRING)
					GetWindowText(Control,(tchar_t*)i->Data,MAXDATA/sizeof(tchar_t));
				else
				if (GetWindowText(Control,s,sizeof(s)/sizeof(tchar_t))>0)
				{
					int v = StringToInt(s,0);

					switch (i->DataDef.Type)
					{
					case TYPE_INT:
						if (i->DataDef.Flags & DF_PERCENT)
							v = (v*PERCENT_ONE)/100;
						*(int*)i->Data = v;
						break;

					case TYPE_TICK:
						if (i->DataDef.Flags & DF_MSEC)
							v = (v*TICKSPERSEC)/1000;
						else
							v *= TICKSPERSEC;
						*(tick_t*)i->Data = v;
						break;

					case TYPE_FRACTION:
						((fraction*)i->Data)->Num = v;
						((fraction*)i->Data)->Den = 100;
						break;
					}
				}
				break;

			case STN_CLICKED: //BN_CLICKED:
			case LBN_SELCHANGE: //STN_DBLCLK: CBN_SELCHANGE:
				if (Control != i->Control)
				{
					WinPropFocus(p,i,1);
				}
				else
				if (i->DataDef.Type == TYPE_FUNC)
				{
					int No; 

					((winbuttonfunc)i->DataDef.Format1)(p,p->Chain);

					for (No=0;No<p->LabelCount;++No)
						WinUpdateControl(p,(backup*)GetWindowLong(p->Label[No],GWL_USERDATA));
				}
				else
				if (i->DataDef.Type == TYPE_BOOL)
				{
					*(bool_t*)i->Data = !*(bool_t*)i->Data;
					SendMessage(Control,BM_SETCHECK,*(bool_t*)i->Data?BST_CHECKED:BST_UNCHECKED,0);
				}
				else
				if (i->DataDef.Type == TYPE_INT)
				{
					int No;
					if (QueryPlatform(PLATFORM_TYPENO)==TYPE_SMARTPHONE)
					{
						No = SendMessage(Control,LB_GETCURSEL,0,0);
						if (No>=0)
							No = SendMessage(Control,LB_GETITEMDATA,No,0);
						else
							No=0;
					}
					else
					{
						No = SendMessage(Control,CB_GETCURSEL,0,0);
						if (No>=0)
							No = SendMessage(Control,CB_GETITEMDATA,No,0);
						else
							No=0;
					}
					*(int*)i->Data = No;
				}
				else
				if (i->DataDef.Type == TYPE_HOTKEY)
				{
					if (!*(int*)i->Data)
					{
						node* QueryKey = NodeFind(QUERYKEY_ID);
						if (QueryKey)
						{
							int Key;
							int Key2;
							WinOpen((win*)QueryKey,p,1);
							QueryKey->Get(QueryKey,QUERYKEY_KEY,&Key,sizeof(Key));
							QueryKey->Get(QueryKey,QUERYKEY_KEY2,&Key2,sizeof(Key2));
							if (Key)
							{
								int No; 
								int KeyCode = Key & ~HOTKEY_KEEP;

								for (No=0;No<p->LabelCount;++No)
								{
									backup* j = (backup*)GetWindowLong(p->Label[No],GWL_USERDATA);
									if (j!=i && j->DataDef.Type == TYPE_HOTKEY && 
										((*(int*)j->Data & ~HOTKEY_KEEP) == (Key & ~HOTKEY_KEEP) || 
										 (*(int*)j->Data & ~HOTKEY_KEEP) == (Key2 & ~HOTKEY_KEEP)))
									{
										*(int*)j->Data = 0;
										WinUpdateControl(p,j);
									}
								}

								if (WinEssentialKey(KeyCode) && QueryPlatform(PLATFORM_TYPENO)==TYPE_SMARTPHONE)
									Key &= ~HOTKEY_KEEP;

								*(int*)i->Data = Key;
								WinUpdateControl(p,i);
							}
						}
					}
					else
					{
						*(int*)i->Data = 0;
						WinUpdateControl(p,i);
					}
				}
				break;
			}
		break;

	case WM_SIZE:
		if (p)
			UpdateDialogSize(p);
		break;

	case WM_SETFOCUS:
		if (p->Focus && p->Focus->Control)
			SetFocus(p->Focus->Control);
		break;

	case WM_DESTROY:
		p->WndDialog = NULL;
		break;

	case WM_HSCROLL:
		if (!lParam || !(GetWindowLong((HWND)lParam,GWL_STYLE) & (TBS_BOTH|TBS_NOTICKS|TBS_FIXEDLENGTH)))
			HScroll(p,LOWORD(wParam),0);
		else
		{
			Control = (HWND)lParam;
			i = (backup*) GetWindowLong(Control,GWL_USERDATA);
			if (VALIDBACKUP(i))
			{
				WinPropFocus(p,i,1);
				SetTrack(p,i,SendMessage(Control,TBM_GETPOS,0,0));
			}
		}
		break;
	case WM_VSCROLL:
		if (!lParam || !(GetWindowLong((HWND)lParam,GWL_STYLE) & (TBS_BOTH|TBS_NOTICKS|TBS_FIXEDLENGTH)))
			VScroll(p,LOWORD(wParam),0);
		break;

	case WM_KEYDOWN:
		if (wParam == VK_ESCAPE)
			PostMessage(p->Wnd,WM_CLOSE,0,0);
		if (p->ScrollV && wParam == VK_UP)
			VScroll(p,SB_LINEUP,0);
		if (p->ScrollV && wParam == VK_DOWN)
			VScroll(p,SB_LINEDOWN,0);
		if (p->ScrollH && wParam == VK_LEFT)
			HScroll(p,SB_LINEUP,0);
		if (p->ScrollH && wParam == VK_RIGHT)
			HScroll(p,SB_LINEDOWN,0);
		break;

#if defined(_WIN32_WCE)
	case WM_CTLCOLOREDIT:
		//i = (backup*)GetWindowLong((HWND)lParam,GWL_USERDATA);
		SetBkColor((HDC)wParam,GetSysColor(COLOR_BTNFACE));
		return (LRESULT) GetSysColorBrush(COLOR_BTNFACE);
#endif

	case WM_CTLCOLORSTATIC:
		i = (backup*)GetWindowLong((HWND)lParam,GWL_USERDATA);
		if (p->Focus && p->Focus->Control != (HWND)lParam && p->Focus->Edit != (HWND)lParam &&
			p->Focus == i)
		{
			if (QueryPlatform(PLATFORM_TYPENO)==TYPE_SMARTPHONE)
			{
				SetTextColor((HDC)wParam,GetSysColor(COLOR_HIGHLIGHTTEXT));
				SetBkColor((HDC)wParam,GetSysColor(COLOR_HIGHLIGHT));
				return (LRESULT) GetSysColorBrush(COLOR_HIGHLIGHT);
			}
			SetTextColor((HDC)wParam,0xC00000);
		}
		SetBkColor((HDC)wParam,GetSysColor(DIALOG_COLOR));
		return (LRESULT) GetSysColorBrush(DIALOG_COLOR);
	}

	return DefWindowProc(Wnd,Msg,wParam,lParam);
}

static LRESULT Proc(HWND Wnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	int Result = 0;
	CREATESTRUCT* Param;
	win* p = (win*)GetWindowLong(Wnd,GWL_USERDATA);

#ifdef MSGLOG
	if (p)
		DebugMessage(T("WIN:%08x Cmd:%08x wP:%08x wL:%08x"),p->Id,Msg,wParam,lParam);
#endif

	switch (Msg)
	{
#if defined(_WIN32_WCE)
	case WM_HOTKEY:
		if (p->Modal && HIWORD(lParam)==VK_ESCAPE)
		{
			HWND Focus = GetFocus();
			backup* i = (backup*) GetWindowLong(Focus,GWL_USERDATA);
			if (FuncSHSendBackToFocusWindow && ((VALIDBACKUP(i) && i->Editor) || (!VALIDBACKUP(i) && i)))
				FuncSHSendBackToFocusWindow(Msg,wParam,lParam);
			else
			{
				p->HotKeyEscape = !p->HotKeyEscape;
				if (!p->HotKeyEscape)
					PostMessage(p->Wnd,WM_CLOSE,0,0);
			}
			return 0;
		}
		if (LOWORD(lParam) & MOD_KEYUP)
			return 0;
		break;
#endif

	case WM_CREATE:

		Param = (CREATESTRUCT*)lParam;
		p = (win*)Param->lpCreateParams;
		p->Wnd = Wnd;
		p->Module = Param->hInstance;
		p->Result = 1;

#if defined(_WIN32_WCE)
		p->Activate = CAlloc(sizeof(SHACTIVATEINFO),1);
		if (p->Activate)
			((SHACTIVATEINFO*)p->Activate)->cbSize = sizeof(SHACTIVATEINFO);
#endif

		p->Proc(p,MSG_PREPARE,0,0,&Result);
		CreateToolBar(p);

		SetWindowLong(Wnd,GWL_USERDATA,(LONG)p);
		break;

	case WM_CLOSE:
		if (p->Modal)
		{
			if (p->Parent)
			{
				EnableWindow(p->Parent->Wnd,TRUE);
				p->Parent->Child = NULL;
				SetForegroundWindow(p->Parent->Wnd);
			}
			DestroyWindow(p->Wnd);
		}
		break;

	case WM_SETFOCUS:
		if (p->WndDialog)
			SetFocus(p->WndDialog);
		break;

	case WM_SIZE:
		if (p)
		{
			RECT r;
			GetClientRect(p->Wnd,&r);

			if (p->ToolBarHeight)
				MoveWindow(p->WndTB,0,0,r.right,p->ToolBarHeight,TRUE);

			if (p->WndDialog && !p->OwnDialogSize)
				MoveWindow(p->WndDialog,0,p->ToolBarHeight,r.right,r.bottom-p->ToolBarHeight,TRUE);
		}
		break;

	case WM_DESTROY:
		WinSetFullScreen(p,0);
		DestroyToolBar(p);
		if (p->Menu)
			DestroyMenu(p->Menu);
		p->Modal = 0;
		p->Wnd = NULL;
		p->WndTB = NULL;
		p->Menu = NULL;
		break;

	case WM_COMMAND:
		if (LOWORD(wParam) == WIN_A)
			wParam = p->WinCmd[0];
		if (LOWORD(wParam) == WIN_B)
			wParam = p->WinCmd[1];
		if (LOWORD(wParam) == IDOK)
			wParam = PLATFORM_OK;
		break;

	case WM_ACTIVATE:
		DEBUG_MSG3(DEBUG_WIN,T("ACTIVATE %d %08x %08x"),LOWORD(wParam)!=WA_INACTIVE,Wnd,lParam);

		if (LOWORD(wParam)!=WA_INACTIVE && p->Child)
		{
			SetForegroundWindow(p->Child->Wnd);
			return 0;
		}

#if defined(_WIN32_WCE)
		if (FuncSHHandleWMActivate && p->Activate)
			FuncSHHandleWMActivate(Wnd, wParam, lParam, p->Activate, FALSE);
#endif
		break;

	case WM_SETTINGCHANGE:
#if defined(_WIN32_WCE) 
		if (FuncSHHandleWMSettingChange && p->Activate)
			FuncSHHandleWMSettingChange(Wnd, wParam, lParam, p->Activate);
#endif
		UpdateDPI();
		break;
	}

	if (p)
	{
		if (p->Proc(p,Msg,wParam,lParam,&Result))
			return Result;

#if defined(_WIN32_WCE)
		if (Msg == WM_CREATE && !p->AygShellTB && (p->Dialog || GetSystemMetrics(SM_CXSCREEN)>240))
			CommandBar_AddAdornments(p->WndTB,p->NeedOK ? CMDBAR_OK:0,0);
#endif
		if (p->WndDialog && (Msg == WM_CREATE || Msg == WM_SIZE))
			WinUpdateScroll(p);
	}

	return DefWindowProc(Wnd,Msg,wParam,lParam);
}

win* WinGetObject(HWND Wnd)
{
	return (win*) GetWindowLong(Wnd,GWL_USERDATA);
}

void WinSetObject(HWND Wnd,win* p)
{
	SetWindowLong(Wnd,GWL_USERDATA,(LONG)p);
}

const tchar_t* WinClassName()
{
	stprintf(ClassName,T("%s_Win"),NodeProgramId());
	return ClassName;
}

void Win_Init(HMODULE Module,int IconId)
{
	InitCommonControls();

	memset(&WinClass,0,sizeof(WinClass));
    WinClass.style			= CS_HREDRAW | CS_VREDRAW;
    WinClass.lpfnWndProc	= (WNDPROC) Proc;
    WinClass.cbClsExtra		= 0;
    WinClass.cbWndExtra		= 0;
    WinClass.hInstance		= Module;
	if (IconId)
		WinClass.hIcon		= LoadIcon(Module, MAKEINTRESOURCE(IconId));
    WinClass.hCursor		= WinCursorArrow();
    WinClass.hbrBackground	= NULL;
    WinClass.lpszMenuName	= 0;
    WinClass.lpszClassName	= WinClassName();
	RegisterClass(&WinClass);

	memset(&DialogClass,0,sizeof(DialogClass));
    DialogClass.style			= CS_HREDRAW | CS_VREDRAW;
    DialogClass.lpfnWndProc	= (WNDPROC) DialogProc;
    DialogClass.cbClsExtra		= 0;
    DialogClass.cbWndExtra		= 0;
    DialogClass.hInstance		= Module;
    DialogClass.hCursor			= WinCursorArrow();
#if defined(_WIN32_WCE)
	DialogClass.hbrBackground	= GetSysColorBrush(COLOR_STATIC);
#else			
	DialogClass.hbrBackground	= GetSysColorBrush(COLOR_BTNFACE);
#endif
    DialogClass.lpszMenuName	= 0;
    DialogClass.lpszClassName	= T("DialogBase");
	RegisterClass(&DialogClass);

	memset(&FontCache,0,sizeof(FontCache));

#if defined(_WIN32_WCE)
	if (LOWORD(LinkVer) >= 3 && !QueryAdvPlatform(ADVPLATFORM_OLDSHELL))
	{
		AygShell = LoadLibrary(T("aygshell.dll"));
		*(FARPROC*)&FuncSHCreateMenuBar = GetProcAddress(AygShell,T("SHCreateMenuBar"));
		*(FARPROC*)&FuncSHInitDialog = GetProcAddress(AygShell,T("SHInitDialog"));
		*(FARPROC*)&FuncSHFullScreen = GetProcAddress(AygShell,T("SHFullScreen"));
		*(FARPROC*)&FuncSHHandleWMActivate = GetProcAddress(AygShell,MAKEINTRESOURCE(84));
		*(FARPROC*)&FuncSHHandleWMSettingChange = GetProcAddress(AygShell,MAKEINTRESOURCE(83));
		*(FARPROC*)&FuncSHSendBackToFocusWindow = GetProcAddress(AygShell,MAKEINTRESOURCE(97));
	}
	CoreDLL = LoadLibrary(T("coredll.dll"));
	*(FARPROC*)&FuncDeferWindowPos = GetProcAddress(CoreDLL,T("DeferWindowPos"));
	*(FARPROC*)&FuncBeginDeferWindowPos = GetProcAddress(CoreDLL,T("BeginDeferWindowPos"));
	*(FARPROC*)&FuncEndDeferWindowPos = GetProcAddress(CoreDLL,T("EndDeferWindowPos"));
	*(FARPROC*)&FuncUnregisterFunc1 = GetProcAddress(CoreDLL,T("UnregisterFunc1"));
	*(FARPROC*)&FuncAllKeys = GetProcAddress(CoreDLL,T("AllKeys")); //MAKEINTRESOURCE(1453));
	*(FARPROC*)&FuncSipShowIM = GetProcAddress(CoreDLL,T("SipShowIM"));

#else
	FuncDeferWindowPos		= DeferWindowPos;
	FuncBeginDeferWindowPos = BeginDeferWindowPos;
	FuncEndDeferWindowPos	= EndDeferWindowPos;
#endif

	UpdateDPI();
}

void Win_Done()
{
	ReleaseFonts();
#if defined(_WIN32_WCE)
	if (CoreDLL)
		FreeLibrary(CoreDLL);
	if (AygShell)
		FreeLibrary(AygShell);
#endif
}

void WinRegister(win* p,int Priority)
{
	p->VMT.Enum = WinEnum;
	p->VMT.Set = WinSet;
	p->VMT.Get = WinGet;

	NodeRegister(&p->VMT,Priority);
}

void WinUnRegister(win* p)
{
	NodeUnRegister(&p->VMT);
}

HFONT WinFont(int* FontSize,bool_t Bold)
{
	HFONT Font;
	if (*FontSize && *FontSize < MAXFONTSIZE)
	{
		if (!FontCache[*FontSize][Bold])
		{
			LOGFONT LogFont;

			memset(&LogFont,0,sizeof(LogFont));
			tcscpy(LogFont.lfFaceName,T("MS Sans Serif"));

			LogFont.lfHeight = Scale(*FontSize,LogPixelSY,72);
			LogFont.lfWidth = 0;

			switch (StringLang)
			{
			case FOURCC('C','H','T','_'):
				LogFont.lfCharSet = CHINESEBIG5_CHARSET;
				break;
			case FOURCC('C','H','S','_'):
				LogFont.lfCharSet = GB2312_CHARSET;
				break;
			case FOURCC('R','U','_','_'):
				LogFont.lfCharSet = RUSSIAN_CHARSET;
				break;
			case FOURCC('J','A','_','_'):
				LogFont.lfCharSet = SHIFTJIS_CHARSET;
				break;
			case FOURCC('T','R','_','_'):
				LogFont.lfCharSet = TURKISH_CHARSET;
				break;
			case FOURCC('K','O','_','_'):
				LogFont.lfCharSet = HANGEUL_CHARSET;
				break;
			default:
				LogFont.lfCharSet = OEM_CHARSET;
			}

			if (Bold)
				LogFont.lfWeight = FW_BOLD;

			FontCache[*FontSize][Bold] = CreateFontIndirect(&LogFont);
		}
		Font = FontCache[*FontSize][Bold];
	}
	else
	{
		LOGFONT LogFont;
		Font = (HFONT)GetStockObject(SYSTEM_FONT);
		GetObject(Font,sizeof(LogFont),&LogFont);
		*FontSize = LogFont.lfHeight;
	}
	return Font;
}

void WinProp( win* p, int* x0,int* x1)
{
	*x0 = 2;
	*x1 = p->LabelWidth + *x0;

	if (*x1 > (p->DlgWidth*3)/4)
		*x1 = (p->DlgWidth*3)/4;

	*x1++;
}

static void AddChain(win* p,backup* i)
{
	i->Chain = NULL;
	if (!p->Chain)
	{
		p->Chain = i;
		WinPropFocus(p,i,1);
	}
	else
	{
		backup* j=p->Chain;
		for (j=p->Chain;j->Chain;j=j->Chain);
		j->Chain = i;
	}
}

int WinPropLabel( win* p, int DlgTop, const tchar_t* Msg, const tchar_t* Value, backup* i )
{
	int x0,x1;
	int y2,y1;
	WinProp(p,&x0,&x1);
	y1 = WinLabel(p,DlgTop,x0,x1-x0-2,Msg,PROPSIZE,0,i);
	y2 = WinLabel(p,DlgTop,x1,-1,Value,PROPSIZE,LABEL_BOLD,0);
	if (i)
		AddChain(p,i);
	return y1>y2 ? y1:y2;
}

void WinUpdateControl(win* p,backup* i)
{
	tchar_t s[256];
	int Value;

	switch (i->DataDef.Type)
	{
	case TYPE_HOTKEY:
		HotKeyToString(s,*(int*)i->Data,sizeof(s));
		SetWindowText(i->Edit,s);
		SetWindowText(i->Control,String(PLATFORM_ID,*(int*)i->Data ? PLATFORM_CLEAR:PLATFORM_ASSIGN));
		break;

	case TYPE_BOOL:
		SendMessage(i->Control,BM_SETCHECK,*(bool_t*)i->Data?BST_CHECKED:BST_UNCHECKED,0);
		break;

	case TYPE_INT:
		if (i->DataDef.Flags & (DF_ENUMCLASS|DF_ENUMSTRING))
		{
			int GetCount;
			int GetItemData;
			int SetCurSel;
			int No,Count;

#if defined(_WIN32_WCE)
			if (QueryPlatform(PLATFORM_TYPENO) == TYPE_SMARTPHONE)
			{
				GetCount = LB_GETCOUNT;
				GetItemData = LB_GETITEMDATA;
				SetCurSel = LB_SETCURSEL;
			}
			else
#endif
			{
				GetCount = CB_GETCOUNT;
				GetItemData = CB_GETITEMDATA;
				SetCurSel = CB_SETCURSEL;
			}

			Count = SendMessage(i->Control,GetCount,0,0);
			for (No=0;No<Count;++No)
				if (SendMessage(i->Control,GetItemData,No,0) == *(int*)i->Data)
					break;
			if (No==Count)
				No=-1;
			SendMessage(i->Control,SetCurSel,No,0);
			break;
		}
		else
		if (i->DataDef.Flags & DF_MINMAX)
		{
			if (i->DataDef.Type == TYPE_TICK)
				Value = *(tick_t*)i->Data;
			else
				Value = *(int*)i->Data;

			if (i->DataDef.Format2 - i->DataDef.Format1 <= TRACKMAX)
				Value -= i->DataDef.Format1;
			else
				Value = Scale(Value - i->DataDef.Format1,TRACKMAX,i->DataDef.Format2 - i->DataDef.Format1);
			SendMessage(i->Control, TBM_SETPOS,1,Value);
			break;
		}
		else
		{
			if (i->DataDef.Flags & DF_PERCENT)
			{
				if (*(int*)i->Data > 0)
					IntToString(s,(*(int*)i->Data*100+(PERCENT_ONE>>1))/PERCENT_ONE,sizeof(s),0);
				else
					s[0]=0;
			}
			else
				IntToString(s,*(int*)i->Data,sizeof(s),0);
			SetWindowText(i->Control,s);
			SendMessage(i->Control,EM_SETSEL,tcslen(s),tcslen(s));
		}
		break;

	case TYPE_STRING:
		SetWindowText(i->Control,(tchar_t*)i->Data);
		SendMessage(i->Control,EM_SETSEL,tcslen(s),tcslen(s));
		break;

	case TYPE_TICK:
		if (i->DataDef.Flags & DF_MSEC)
			stprintf(s,T("%d"),(*(tick_t*)i->Data*1000+(TICKSPERSEC>>1)) / TICKSPERSEC);
		else
			stprintf(s,T("%d"),(*(tick_t*)i->Data+(TICKSPERSEC>>1)) / TICKSPERSEC);
		SetWindowText(i->Control,s);
		SendMessage(i->Control,EM_SETSEL,tcslen(s),tcslen(s));
		break;

	case TYPE_FRACTION:
		if (((fraction*)i->Data)->Num > 0)
			stprintf(s,T("%d"),(((fraction*)i->Data)->Num*100)/((fraction*)i->Data)->Den);
		else
			s[0] = 0;
		SetWindowText(i->Control,s);
		SendMessage(i->Control,EM_SETSEL,tcslen(s),tcslen(s));
		break;
	}
}

static int CreateEnum( win* p, backup* i, int Left, int Top, int Width, int Height )
{
	int No;
	int Pos;
	int AddString;
	int ItemData;
	int MaxWidth = 	(i->DataDef.Flags & DF_ENUMSTRING) ? 80:180;
	int Plus = 0;
	int FontSize;

	MaxWidth = Scale(MaxWidth,LogPixelSX,72);
	if (Width > MaxWidth)
		Width = MaxWidth;

	if (QueryPlatform(PLATFORM_TYPENO) == TYPE_SMARTPHONE)
	{
		i->Control = CreateWindow(T("listbox"), NULL, WS_TABSTOP|WS_VISIBLE|WS_CHILD|LBS_NOTIFY|LBS_NOINTEGRALHEIGHT|LBS_SORT, 
			Left, Top, Width, Height, p->WndDialog, NULL, p->Module, 0L);

		i->Edit = CreateWindow(UPDOWN_CLASS, NULL, , WS_CHILD|WS_VISIBLE | UDS_HORZ |UDS_ALIGNRIGHT | UDS_ARROWKEYS |  UDS_SETBUDDYINT | UDS_WRAP | UDS_EXPANDABLE, 
			0, 0, 0, 0, p->WndDialog, NULL, p->Module, 0L);

		SendMessage(i->Edit, UDM_SETBUDDY, (WPARAM)i->Control, 0);	

		AddString = LB_ADDSTRING;
		ItemData = LB_SETITEMDATA;

		FontSize = PROPSIZE;
		SendMessage(i->Edit,WM_SETFONT,(WPARAM)WinFont(&FontSize,1),0);
	}
	else
	{
		i->Control = CreateWindow(T("COMBOBOX"), NULL, WS_VSCROLL|WS_TABSTOP|WS_VISIBLE|WS_CHILD|CBS_DROPDOWNLIST|CBS_SORT, 
			Left, Top, Width, Height*(MenuSmall?10:15), p->WndDialog, NULL, p->Module, 0L);

		i->ComboBox = 1;
		AddString = CB_ADDSTRING;
		ItemData = CB_SETITEMDATA;
		Plus = 6;

		FontSize = PROPSIZE;
		SendMessage(i->Control,WM_SETFONT,(WPARAM)WinFont(&FontSize,1),0);
	}

	if (i->DataDef.Flags & DF_ENUMCLASS)
	{
		node* Node;
		int Id;
		for (No=0;(Node = NodeEnum(i->DataDef.Format1,No))!=NULL;++No)
		{
			Node->Get(Node,NODE_ID,&Id,sizeof(Id));
			Pos = SendMessage(i->Control,AddString,0,(LPARAM)String(Id,0));
			SendMessage(i->Control,ItemData,Pos,Id);
		}
	}
	else
	if (i->DataDef.Flags & DF_ENUMSTRING)
	{
		const stringdef* s;
		for (No=0;(s=StringEnum(i->DataDef.Format1,No))!=NULL;++No)
		{
			Pos = SendMessage(i->Control,AddString,0,(LPARAM)s->String);
			SendMessage(i->Control,ItemData,Pos,s->Id);
		}
	}

	return Plus;
}

static void CreateEditor( win* p, backup* i, int Left, int Top, int Width, int Height )
{
	tchar_t Post[32];
	int FontSize;
	int Style = ES_NUMBER;

	if (i->DataDef.Type==TYPE_STRING || (i->DataDef.Flags & DF_NEG))
		Style = 0;

	if ((Width > 30) && i->DataDef.Type!=TYPE_STRING)
		Width = 30;

	i->Control = CreateWindow(T("EDIT"), NULL, WS_TABSTOP|WS_VISIBLE|WS_CHILD|Style, 
		Scale(Left,LogPixelSX,72), Scale(Top,LogPixelSX,72), 
		Scale(Width,LogPixelSX,72), Scale(Height,LogPixelSX,72), 
		p->WndDialog, NULL, p->Module, 0L);

	FontSize = PROPSIZE;
	SendMessage(i->Control,WM_SETFONT,(WPARAM)WinFont(&FontSize,1),0);

	i->Editor = 1;
	Post[0] = 0;

	if (i->DataDef.Flags & DF_KBYTE)
		tcscpy(Post,T("KB"));
	else	
	if (i->DataDef.Flags & DF_MHZ)
		tcscpy(Post,T("Mhz"));
	else	
	if (i->DataDef.Flags & DF_PERCENT)
		tcscpy(Post,T("%"));
	else
	if (i->DataDef.Type == TYPE_TICK)
		tcscpy(Post,(i->DataDef.Flags & DF_MSEC) ? T("msec"):T("sec"));

	if (Post[0])
		WinLabel(p,Top,Left+Width+1,-1,Post,11,0,NULL);
}

int WinPropControl( win* p, int DlgTop, backup* i )
{
	int FontSize;
	int y2,y1;
	int Height,Width;
	int Top = Scale(DlgTop,LogPixelSY,72);
	int Left;
	int x0,x1;
	int x;

	if (i->DataDef.Flags & DF_SHORTLABEL)
	{
		if (QueryPlatform(PLATFORM_TYPENO) == TYPE_SMARTPHONE)
			x1 = 36;
		else
			x1 = 48;
		x0 = 2;
		Width = x1-x0-2; 
	}
	else
	if ((i->DataDef.Flags & DF_CHECKLIST) && i->DataDef.Type==TYPE_BOOL)
	{
		x0 = 4+PROPSIZE;
		x1 = 2;
		Width = -1;
	}
	else
	{
		WinProp(p,&x0,&x1);
		Width = x1-x0-2;
	}

	if (i->DataDef.Flags & DF_NOWRAP)
		Width = -1;

	if (i->DataDef.Type != TYPE_FUNC)
		y1 = WinLabel(p,DlgTop,x0,Width,i->DataDef.Name,PROPSIZE,i->Disabled ? LABEL_DISABLED:0,i);
	else
	{
		y1 = DlgTop;
		x1 = x0;
	}

	y2 = DlgTop + PROPSIZE+1;
	Width = Scale(PROPSIZE+1,LogPixelSX,72);
	Height = Scale(PROPSIZE+1,LogPixelSY,72);
	Left = Scale(x1,LogPixelSX,72);

	i->Control = NULL;
	i->Edit = NULL;
	i->Editor = 0;
	i->ComboBox = 0;
	i->ListView = 0;

	switch (i->DataDef.Type)
	{
	case TYPE_FUNC:
		i->Control = CreateWindow(T("BUTTON"),NULL,WS_TABSTOP|WS_VISIBLE|WS_CHILD|BS_PUSHBUTTON|BS_NOTIFY,
			Left+Scale(2,LogPixelSX,72),Top,Width*5,Scale(PROPSIZE+2,LogPixelSY,72), p->WndDialog, NULL, p->Module, NULL );

		y2 += 2;
		FontSize = PROPSIZE;
		SendMessage(i->Control,WM_SETFONT,(WPARAM)WinFont(&FontSize,0),0);
		SetWindowText(i->Control,i->DataDef.Name);
		break;

	case TYPE_HOTKEY:

		x = (Width*7)/2;
		i->Edit = CreateWindow(T("EDIT"),NULL,ES_READONLY|WS_VISIBLE|WS_CHILD,
			Left,Top,x,Height, p->WndDialog, NULL, p->Module, NULL );

		i->Control = CreateWindow(T("BUTTON"),NULL,WS_TABSTOP|WS_VISIBLE|WS_CHILD|BS_PUSHBUTTON|BS_NOTIFY,
			Left+x+Scale(2,LogPixelSX,72),Top,(Width*5)/2,Scale(PROPSIZE+2,LogPixelSY,72), p->WndDialog, NULL, p->Module, NULL );

		FontSize = PROPSIZE;
		SendMessage(i->Edit,WM_SETFONT,(WPARAM)WinFont(&FontSize,1),0);

		FontSize = PROPSIZE;
		SendMessage(i->Control,WM_SETFONT,(WPARAM)WinFont(&FontSize,0),0);

		y2 += 2;
		break;

	case TYPE_BOOL:
		i->Control = CreateWindow(T("BUTTON"),NULL,WS_TABSTOP|WS_VISIBLE|WS_CHILD|BS_CHECKBOX|BS_NOTIFY,
			Left,Top,Width,Height, p->WndDialog, NULL, p->Module, NULL );

		break;

	case TYPE_INT:
		if (i->DataDef.Flags & (DF_ENUMCLASS|DF_ENUMSTRING))
		{
			y2 += CreateEnum(p,i,Left,Top,Scale(p->DlgWidth-x1,LogPixelSX,72),Height);
			break;
		}
		else
		if (i->DataDef.Flags & DF_MINMAX)
		{
			int Interval;

			Width = p->DlgWidth-x1;
			if (Width > 100)
				Width = 100;
			if (Width < 65 && !(i->DataDef.Flags & DF_NOWRAP))
			{
				y2 = y1 + PROPSIZE+1;
				x1 = x0;
				Width = p->DlgWidth-x1;
				if (Width > 200)
					Width = 200;
				Left = Scale(x1,LogPixelSX,72);
				Top = Scale(y1,LogPixelSY,72);
			}

			i->Control = CreateWindowEx( 
				0,                    
				TRACKBAR_CLASS,       
				NULL,   
				WS_TABSTOP|WS_CHILD | WS_VISIBLE | TBS_HORZ|TBS_BOTH|TBS_NOTICKS|TBS_FIXEDLENGTH, 
				Left,Top,    
				Scale(Width,LogPixelSX,72), Height,   
				p->WndDialog,	
				NULL,						
				p->Module,                     
				NULL); 

			Interval = i->DataDef.Format2 - i->DataDef.Format1;
			if (Interval > TRACKMAX)
				Interval = TRACKMAX;

			SendMessage(i->Control, TBM_SETRANGE, FALSE,MAKELONG(0,Interval));
			SendMessage(i->Control, TBM_SETPAGESIZE,0,Interval/32);
			SendMessage(i->Control, TBM_SETLINESIZE,0,Interval/32);
			SendMessage(i->Control, TBM_SETTHUMBLENGTH, Scale(TRACKTHUMB,WinPixelSY(),72),0);
			break;
		}
		// no break

	case TYPE_STRING:
	case TYPE_TICK:
	case TYPE_FRACTION:
		CreateEditor(p,i,x1,DlgTop,p->DlgWidth-x1,PROPSIZE+1);
		y2 += 1;
		break;
	}

	WinUpdateControl(p,i);

	SetWindowLong(i->Control,GWL_USERDATA,(LONG)i);
	if (i->Disabled)
		EnableWindow(i->Control,FALSE);

	AddChain(p,i);
	return y1>y2?y1:y2;
}

int WinLabel(win* p,int DlgTop,int DlgLeft,int DlgWidth,const tchar_t* Msg,int FontSize,int Flags,void* User)
{
	tchar_t s[256];
	SIZE Size;
	HWND Wnd;
	HDC DC = GetDC(p->Wnd);
	HFONT OldFont;
	HFONT Font = WinFont(&FontSize,(Flags & LABEL_BOLD)!=0);
	int Style = WS_VISIBLE | WS_CHILD;
	int Left = Scale(DlgLeft,LogPixelSX,72);
	int Width = DlgWidth>0 ? Scale(DlgWidth,LogPixelSX,72) : -1;
	int Extra = Scale(2,LogPixelSX,72);

	if (User)
		Style |= SS_NOTIFY;

	if (Flags & LABEL_CENTER)
		Style |= SS_CENTER;
	else
		Style |= SS_LEFTNOWORDWRAP;

	OldFont = SelectObject(DC,Font);

	if (Msg)
		while (*Msg)
		{
			while (*Msg == 10)
			{
				++Msg;
				DlgTop += FontSize/3; // smaller gaps
			}

			if (*Msg)
			{
				tchar_t* i;
				TcsNCpy(s,Msg,256);

				i = tcschr(s,10);
				if (i) *i = 0;

				for (;;)
				{
					GetTextExtentPoint(DC,s,tcslen(s),&Size);
					Size.cx += Extra;

					if (!s[0] || !s[1] || Width<0 || Size.cx <= Width)
						break;

					i = tcsrchr(s,' ');
					if (i)
						*i = 0;
					else
						s[tcslen(s)-1] = 0;
				}

				Wnd = CreateWindow(T("STATIC"),s,Style,Left,Scale(DlgTop,LogPixelSY,72),
					Width>0 ? Width : Size.cx, Scale(FontSize,LogPixelSY,72), p->WndDialog, NULL, p->Module, NULL );

				p->Label[p->LabelCount++] = Wnd;

				SetWindowLong(Wnd,GWL_USERDATA,(LONG)User);
				SendMessage(Wnd,WM_SETFONT,(WPARAM)Font,0);

				if (Flags & LABEL_DISABLED)
					EnableWindow(Wnd,FALSE);

				DlgTop += FontSize;
				Msg += tcslen(s);

				while (*Msg == ' ') ++Msg;
				if (*Msg == 10) ++Msg;
			}
		}

	SelectObject(DC,OldFont);
	ReleaseDC(p->Wnd,DC);

	return DlgTop;
}

bool_t WinEssentialKey(int HotKey)
{
	HotKey &= ~HOTKEY_KEEP;
	return 
		HotKey == 0x86 || //action
		HotKey == VK_RETURN ||
		HotKey == VK_LEFT ||
		HotKey == VK_RIGHT ||
		HotKey == VK_UP ||
		HotKey == VK_DOWN; // should use normal WM_KEYDOWN event for them
}

void WinRegisterHotKey(win* p,int Id,int HotKey)
{
#if defined(NDEBUG) || defined(_WIN32_WCE)
	int Mod = MOD_KEYUP;
	HotKey &= ~HOTKEY_KEEP;
	if (HotKey & HOTKEY_ALT) Mod |= MOD_ALT;
	if (HotKey & HOTKEY_SHIFT) Mod |= MOD_SHIFT;
	if (HotKey & HOTKEY_CTRL) Mod |= MOD_CONTROL;
	if (HotKey & HOTKEY_WIN) Mod |= MOD_WIN;
	HotKey &= HOTKEY_MASK;
	if (FuncUnregisterFunc1)
		FuncUnregisterFunc1(Mod,HotKey);
	RegisterHotKey(p->Wnd,Id,Mod,HotKey);
#endif
}

void WinUnRegisterHotKey(win* p,int Id)
{
#if defined(NDEBUG) || defined(_WIN32_WCE)
	UnregisterHotKey(p->Wnd,Id);
#endif
}

int WinKeyState(int Key)
{
	if (GetKeyState(VK_CONTROL)<0) Key |= HOTKEY_CTRL;
	if (GetKeyState(VK_SHIFT)<0) Key |= HOTKEY_SHIFT;
	if (GetKeyState(VK_MENU)<0) Key |= HOTKEY_ALT;
	if (GetKeyState(VK_LWIN)<0 || GetKeyState(VK_RWIN)<0) Key |= HOTKEY_WIN;
	//if (Key >= 0xC1 && Key <= 0xCF)	Key |= HOTKEY_WIN; // h3800 ???
	return Key;
}

void WinAllKeys(bool_t State)
{
	if (FuncAllKeys)
		FuncAllKeys(State);
}
		
void WinShowSip(bool_t b)
{
	DEBUG_MSG1(DEBUG_WIN,T("SIP %d"),b);
#if defined(_WIN32_WCE)
	if (FuncSipShowIM)
		FuncSipShowIM(b ? SIPF_ON:SIPF_OFF);
#endif
}

HCURSOR WinCursorArrow()
{
#if defined(_WIN32_WCE)
	return NULL;
#else
	return LoadCursor(NULL,IDC_ARROW);
#endif
}