/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: win32.h 144 2004-12-12 15:28:39Z picard $
 *
 * BetaPlayer
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#ifndef __WIN32_H
#define __WIN32_H

#define WIN_CLASS			FOURCC('W','I','N','_')

#include "resource.h"

#define LABEL_BOLD		1
#define LABEL_CENTER	2
#define LABEL_DISABLED  4

#define PROPSIZE		11

#define MENU_SMALL		0x0100
#define MENU_GRAYED		0x0200
#define MENU_NOTSMALL	0x0400
#define MENU_ADV		0x0800

#define MSG_NOTIFY			(WM_APP+0x240)
#define MSG_PREPARE			(WM_APP+0x250)

#define SETTINGCHANGE_RESET  0x3002

typedef struct menudef
{
	int Level;
	int Class;
	int Id;
	int Flags;

} menudef;

#define MENUDEF_END { -1 }

typedef struct backup
{
	datadef DataDef;
	struct backup* Next;
	char Old[MAXDATA];
	char Data[MAXDATA];

	HWND Control;
	HWND Edit;
	bool_t Editor;
	bool_t ComboBox;
	bool_t ListView;
	bool_t Disabled;
	struct backup* Chain; // window control chain
} backup;

typedef	void (*winbuttonfunc)(void* Win, backup* Chain);

typedef struct win
{
	node VMT;
	bool_t (*Proc)(void*, int Msg, WPARAM wParam, LPARAM lParam,int* Result);
	bool_t (*DialogProc)(void*, int Msg, WPARAM wParam, LPARAM lParam,int* Result);

	// setup
	int Id;
	int Width;
	int Height;
	menudef* MenuDef;
	bool_t Dialog;
	const tchar_t* Title;
	nodeenum Enum;
	nodeget Get;
	nodeset	Set;
	int LabelWidth; // for prop dialog
	bool_t CaptureKeys;
	bool_t OwnDialogSize;

	// state
	int Result;
	int DlgWidth;
	int DlgHeight;
	HMODULE Module;
	HWND Wnd;
	HWND WndDialog;
	HWND WndTB;
	HMENU Menu;
	struct win* Parent;
	struct win* Child;
	int ToolBarHeight;
	bool_t FullScreen;
	bool_t Modal;
	int Style; 
	POINT SaveScreen;
	RECT SaveRect;
	bool_t AygShellTB;
	bool_t TitleOK;
	bool_t TitleClose;
	bool_t NeedOK;
	bool_t InUpdateScroll;
	bool_t ScrollV;
	bool_t ScrollH;
	bool_t HotKeyEscape;
	bool_t ComboOpen;
	void* Activate;
	int WinCmd[2];
	backup* Focus;
	backup* Chain;
	int LabelCount;
	int BitmapNo;
	HWND Label[256];

} win;

void Win_Init(HMODULE Module,int IconId);
void Win_Done();

void WinRegister(win*,int Priority);
void WinUnRegister(win*);

bool_t WinOpen(win*,win* Parent,bool_t Modal);
void WinCreateDialog(win*);
void WinUpdateScroll(win*);
void WinSetFullScreen(win*,bool_t b);
void WinBitmap(win*,int BitmapId16,int BitmapId32,int BitmapNum);
void WinDeleteButton(win*,int Id);
void WinAddButton(win*,int Id,int Bitmap,tchar_t* Name,bool_t Check); // -1 for separator
void WinButtonEnable(win*,int Id,bool_t);
void WinButtonBitmap(win*,int Id,int Bitmap);

int WinLabel(win*,int DlgTop,int DlgLeft,int DlgWidth,const tchar_t* Msg,int FontSize,int Flags,void* User);

const tchar_t* WinClassName();
void WinSetObject(HWND,win*);
win* WinGetObject(HWND);
int WinPixelSX();
int WinPixelSY();
HFONT WinFont(int* FontSize,bool_t Bold);
HCURSOR WinCursorArrow();
void WinShowSip(bool_t);

void WinPropClear( win* );
int WinPropLabel( win*, int DlgTop, const tchar_t* Msg, const tchar_t* Value, backup* );
int WinPropControl( win*, int DlgTop, backup* );
void WinPropFocus(win* p,backup*,bool_t SetFocus);
void WinUpdateControl(win* p,backup* i);

bool_t WinEssentialKey(int HotKey);
void WinRegisterHotKey(win* p,int Id,int HotKey);
void WinUnRegisterHotKey(win* p,int Id);
void WinAllKeys(bool_t State);
int WinKeyState(int Key);

#endif