/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: setup.c 111 2004-11-24 12:54:10Z picard $
 *
 ****************************************************************************/

#include "..\common\common.h"
#include <windows.h>
#include "benchmark.h"
#include "..\player\player.h"

#if defined(_WIN32_WCE)
#define ACCESS_LOAD		0
#define ACCESS_SAVE		0 
#else
#define ACCESS_LOAD		KEY_READ
#define ACCESS_SAVE		KEY_READ | KEY_WRITE | KEY_SET_VALUE 
#endif

DLLEXPORT int WINAPI Install_Init( HWND Parent, BOOL FirstCall, BOOL PrevInstalled, LPCTSTR InstallDir )
{
    return 0;
}

DLLEXPORT int WINAPI Install_Exit( HWND Parent, LPCTSTR InstallDir, WORD FailedDirs,WORD FailedFiles,WORD FailedRegKeys,WORD FailedRegVals,WORD FailedShortcuts )
{
	HMODULE GAPI;
	HMODULE CoreDLL;
	TCHAR LocalPath[MAXPATH];
	TCHAR WinPath[MAXPATH];
	int32_t Slow = -1;

	DWORD Disp;
	HKEY Key;

	tchar_t* AdvBase = T("SOFTWARE\\BetaPlayer\\ADVP");

#ifdef ARM
	tchar_t* Base = T("System\\GDI\\Drivers\\ATI");
	tchar_t* Name = T("DisableDeviceBitmap");
	HANDLE Handle;
#endif

	//-------------------------------------
	// Remove possible old files

	tcscpy(LocalPath,InstallDir);
	tcscat(LocalPath,T("\\lang_en.plg"));
	DeleteFile(LocalPath);

	//-------------------------------------
	// GAPI's gx.dll keep it or not?

	WinPath[0] = 0;
	CoreDLL = LoadLibrary(T("coredll.dll"));
	if (CoreDLL)
	{
		BOOL (WINAPI* FuncSHGetSpecialFolderPath)(HWND,LPTSTR,int,BOOL);

		*(FARPROC*)&FuncSHGetSpecialFolderPath = GetProcAddress(CoreDLL,T("SHGetSpecialFolderPath"));

		if (FuncSHGetSpecialFolderPath)
			FuncSHGetSpecialFolderPath(NULL,WinPath,0x24/*CSIDL_WINDOWS*/,FALSE);

		FreeLibrary(CoreDLL);
	}
	if (!WinPath[0])
		tcscpy(WinPath,T("\\Windows"));

	tcscat(WinPath,T("\\gx.dll"));
	tcscpy(LocalPath,InstallDir);
	tcscat(LocalPath,T("\\gx.dll"));

	GAPI = LoadLibrary(WinPath);
	if (GAPI)
	{
		DeleteFile(LocalPath);
	}
	else
	{
		GAPI = LoadLibrary(LocalPath); // can we load our gx.dll? aygshell.dll available?
		if (!GAPI)
		{
			DeleteFile(LocalPath);
			GAPI = LoadLibrary(T("gx.dll")); // try load on path
		}
	}

	//-------------------------------------
	// Benchmark video memory

	SlowVideoRAW(&Slow);

	if (GAPI)
	{
		if (Slow == -1)
			SlowVideoGAPI(GAPI,Parent,&Slow);
		FreeLibrary(GAPI);
	}

	if (Slow != -1)
	{
		if (RegCreateKeyEx( HKEY_LOCAL_MACHINE, AdvBase, 0, NULL, 0, 0, NULL, &Key, &Disp ) == ERROR_SUCCESS)
		{
			tchar_t Name[32];
			stprintf(Name,T("%d"),ADVPLATFORM_SLOW_VIDEO);
			RegSetValueEx( Key, Name, 0, REG_DWORD, (LPBYTE)&Slow, sizeof(Slow) );
			RegCloseKey( Key );
		}
	}

	//-------------------------------------
	// ATI Imageon 3200 registry settings

#ifdef ARM
	Handle = LoadLibrary(T("ACE_DDI.DLL"));
	if (Handle)
	{
		FreeLibrary(Handle);

		if (RegCreateKeyEx( HKEY_LOCAL_MACHINE, Base, 0, NULL, 0, ACCESS_SAVE, NULL, &Key, &Disp ) == ERROR_SUCCESS)
		{
			DWORD Value;
			DWORD RegType;
			DWORD RegSize;

			RegSize=sizeof(Value);
			if (RegQueryValueEx( Key, Name, 0, &RegType, (LPBYTE)&Value, &RegSize ) != ERROR_SUCCESS || !Value)
			{
				Value = 1;
				RegSetValueEx( Key, Name, 0, REG_DWORD, (LPBYTE)&Value, sizeof(Value) );

				MessageBox(Parent,
					T("ATI IMAGEON device bitmap caching disabled. ") 
					T("After setup please warm boot the device for changes to take effect! ")
					T("You can change this setting later in the player too."),T("Setup"),MB_OK|MB_SETFOREGROUND);
			}
			RegCloseKey( Key );
		}
	}
#endif

	return 0;
}

DLLEXPORT int WINAPI Uninstall_Init(HWND Parent, LPCTSTR InstallDir)
{
    return 0;
}

DLLEXPORT int WINAPI Uninstall_Exit(HWND hwndParent)
{
	HKEY Key;
	if (RegOpenKeyEx( HKEY_LOCAL_MACHINE, T("SOFTWARE"), 0, ACCESS_SAVE, &Key ) == ERROR_SUCCESS)
	{
		RegDeleteKey(Key,T("BetaPlayer"));
		RegCloseKey( Key );
	}
	return 0;
}
